=== modified file 'Makefile'
--- Makefile	2016-11-18 20:05:05 +0000
+++ Makefile	2017-01-16 20:10:58 +0000
@@ -1,8 +1,10 @@
 p=test*.py
+py3="assess_model_change_watcher.py"
 test:
-	TMPDIR=/tmp python -m unittest discover -vv ./tests -p "$(p)"
+	TMPDIR=/tmp python -m unittest discover -vv . -p "$(p)"
 lint:
-	flake8 $$(find -name '*.py') --builtins xrange,basestring
+	python3 -m flake8 --builtins xrange,basestring $(py3)
+	flake8 $$(find -name '*.py') --builtins xrange,basestring --exclude $(py3)
 cover:
 	python -m coverage run --source="./" --omit "./tests/*" -m unittest discover -vv ./tests
 	python -m coverage report
@@ -18,7 +20,7 @@
 install-deps: juju-ci-tools.common_0.1.4-0_all.deb apt-update
 	sudo dpkg -i juju-ci-tools.common_0.1.4-0_all.deb || true
 	sudo apt-get install -y -f
-	./pipdeps.py install
+	./pipdeps.py -v install
 name=NAMEHERE
 assess_file=assess_$(name).py
 test_assess_file=tests/test_assess_$(name).py

=== modified file 'assess_add_cloud.py'
--- assess_add_cloud.py	2016-11-17 20:16:39 +0000
+++ assess_add_cloud.py	2017-01-20 20:41:30 +0000
@@ -4,14 +4,16 @@
 from collections import namedtuple
 from copy import deepcopy
 import logging
+import re
 import sys
 
 import yaml
 
 from jujupy import (
     AuthNotAccepted,
-    EnvJujuClient,
+    ModelClient,
     get_client_class,
+    InvalidEndpoint,
     JujuData,
     NameNotAccepted,
     TypeNotAccepted,
@@ -23,10 +25,63 @@
     )
 
 
-CloudSpec = namedtuple('CloudSpec', ['label', 'name', 'config', 'exception'])
+class CloudMismatch(JujuAssertionError):
+    """The clouds did not match in some way."""
+
+    def __init__(self):
+        super(CloudMismatch, self).__init__('Cloud mismatch')
+
+
+class NameMismatch(JujuAssertionError):
+    """The cloud names did not match."""
+
+    def __init__(self):
+        super(NameMismatch, self).__init__('Name mismatch')
+
+
+class NotRaised(Exception):
+    """An expected exception was not raised."""
+
+    def __init__(self):
+        msg = 'Expected exception not raised: {}'.format(
+            cloud_spec.exception)
+        super(NotRaised, self).__init__(msg)
+
+
+CloudSpec = namedtuple('CloudSpec', [
+    'label', 'name', 'config', 'exception', 'xfail_bug'])
+
+
+def cloud_spec(label, name, config, exception=None, xfail_bug=None):
+    """Generate a CloudSpec, with defaults.
+
+    :param label: The label to display in test results.
+    :param name: The name to use for the cloud.
+    :param config: The cloud-config.
+    :param exception: The exception that is expected to be raised (if any).
+    :param xfail_bug: If this CloudSpec represents an expected failure, the
+        bug number.
+    """
+    return CloudSpec(label, name, config, exception, xfail_bug)
+
+
+def xfail(spec, bug, xfail_exception):
+    """Return a variant of a CloudSpec that is expected to fail.
+
+    Wrapping the original spec improves maintainability, because the xfail can
+    be removed to restore the original value.
+    """
+    return CloudSpec(spec.label, spec.name, spec.config, xfail_exception, bug)
 
 
 def assess_cloud(client, cloud_name, example_cloud):
+    """Assess interactively adding a cloud.
+
+    Will raise an exception
+    - If no clouds are present after interactive add-cloud.
+    - If the resulting cloud name doesn't match the supplied cloud-name.
+    - If the cloud data doesn't match the supplied cloud data.
+    """
     clouds = client.env.read_clouds()
     if len(clouds['clouds']) > 0:
         raise AssertionError('Clouds already present!')
@@ -35,33 +90,44 @@
     if len(clouds['clouds']) == 0:
         raise JujuAssertionError('Clouds missing!')
     if clouds['clouds'].keys() != [cloud_name]:
-        raise JujuAssertionError('Name mismatch')
+        raise NameMismatch()
     if clouds['clouds'][cloud_name] != example_cloud:
         sys.stderr.write('\nExpected:\n')
         yaml.dump(example_cloud, sys.stderr)
         sys.stderr.write('\nActual:\n')
         yaml.dump(clouds['clouds'][cloud_name], sys.stderr)
-        raise JujuAssertionError('Cloud mismatch')
-
-
-def iter_clouds(clouds):
-    yield CloudSpec('bogus-type', 'bogus-type', {'type': 'bogus'},
-                    exception=TypeNotAccepted)
-    for cloud_name, cloud in clouds.items():
-        yield CloudSpec(cloud_name, cloud_name, cloud, exception=None)
-
-    for cloud_name, cloud in clouds.items():
-        yield CloudSpec('long-name-{}'.format(cloud_name), 'A' * 4096, cloud,
-                        exception=None)
-        yield CloudSpec('invalid-name-{}'.format(cloud_name), 'invalid/name',
-                        cloud, exception=NameNotAccepted)
+        raise CloudMismatch()
+
+
+def iter_clouds(clouds, endpoint_validation):
+    """Iterate through CloudSpecs."""
+    yield cloud_spec('bogus-type', 'bogus-type', {'type': 'bogus'},
+                     exception=TypeNotAccepted)
+    for cloud_name, cloud in clouds.items():
+        spec = cloud_spec(cloud_name, cloud_name, cloud)
+        if cloud['type'] == 'manual' and endpoint_validation:
+            spec = xfail(spec, 1649721, InvalidEndpoint)
+        yield spec
+
+    for cloud_name, cloud in clouds.items():
+        spec = xfail(cloud_spec('long-name-{}'.format(cloud_name), 'A' * 4096,
+                                cloud, NameNotAccepted), 1641970, NameMismatch)
+        if cloud['type'] == 'manual' and endpoint_validation:
+            spec = xfail(spec, 1649721, InvalidEndpoint)
+        yield spec
+        spec = xfail(
+            cloud_spec('invalid-name-{}'.format(cloud_name), 'invalid/name',
+                       cloud, NameNotAccepted), 1641981, None)
+        if cloud['type'] == 'manual' and endpoint_validation:
+            spec = xfail(spec, 1649721, InvalidEndpoint)
+        yield spec
 
         if cloud['type'] not in ('maas', 'manual', 'vsphere'):
             variant = deepcopy(cloud)
             variant_name = 'bogus-auth-{}'.format(cloud_name)
             variant['auth-types'] = ['asdf']
-            yield CloudSpec(variant_name, cloud_name, variant,
-                            AuthNotAccepted)
+            yield cloud_spec(variant_name, cloud_name, variant,
+                             AuthNotAccepted)
 
         if 'endpoint' in cloud:
             variant = deepcopy(cloud)
@@ -70,7 +136,11 @@
                 for region in variant['regions'].values():
                     region['endpoint'] = variant['endpoint']
             variant_name = 'long-endpoint-{}'.format(cloud_name)
-            yield CloudSpec(variant_name, cloud_name, variant, exception=None)
+            spec = cloud_spec(variant_name, cloud_name, variant,
+                              InvalidEndpoint)
+            if variant['type'] == 'vsphere' or not endpoint_validation:
+                spec = xfail(spec, 1641970, CloudMismatch)
+            yield spec
 
         for region_name in cloud.get('regions', {}).keys():
             if cloud['type'] == 'vsphere':
@@ -80,36 +150,49 @@
             region['endpoint'] = 'A' * 4096
             variant_name = 'long-endpoint-{}-{}'.format(cloud_name,
                                                         region_name)
-            yield CloudSpec(variant_name, cloud_name, variant,
-                            exception=None)
-
-
-def assess_all_clouds(client, clouds):
+            spec = cloud_spec(variant_name, cloud_name, variant,
+                              InvalidEndpoint)
+            if not endpoint_validation:
+                spec = xfail(spec, 1641970, CloudMismatch)
+            yield spec
+
+
+def assess_all_clouds(client, cloud_specs):
+    """Test all the supplied cloud_specs and return the results.
+
+    Returns a tuple of succeeded, expected_failed, and failed.
+    succeeded and failed are sets of cloud labels.  expected_failed is a dict
+    linking a given bug to its associated failures.
+    """
     succeeded = set()
+    xfailed = {}
     failed = set()
     client.env.load_yaml()
-    for cloud_label, cloud_name, cloud, expected in iter_clouds(clouds):
-        sys.stdout.write('Testing {}.\n'.format(cloud_label))
+    for cloud_spec in cloud_specs:
+        sys.stdout.write('Testing {}.\n'.format(cloud_spec.label))
         try:
-            if expected is None:
-                assess_cloud(client, cloud_name, cloud)
+            if cloud_spec.exception is None:
+                assess_cloud(client, cloud_spec.name, cloud_spec.config)
             else:
                 try:
-                    assess_cloud(client, cloud_name, cloud)
-                except expected:
+                    assess_cloud(client, cloud_spec.name, cloud_spec.config)
+                except cloud_spec.exception:
                     pass
                 else:
-                    raise Exception(
-                        'Expected exception not raised: {}'.format(expected))
+                    raise NotRaised(cloud_spec.exception)
         except Exception as e:
             logging.exception(e)
-            failed.add(cloud_label)
+            failed.add(cloud_spec.label)
         else:
-            succeeded.add(cloud_label)
+            if cloud_spec.xfail_bug is not None:
+                xfailed.setdefault(
+                    cloud_spec.xfail_bug, set()).add(cloud_spec.label)
+            else:
+                succeeded.add(cloud_spec.label)
         finally:
             client.env.clouds = {'clouds': {}}
             client.env.dump_yaml(client.env.juju_home, {})
-    return succeeded, failed
+    return succeeded, xfailed, failed
 
 
 def write_status(status, tests):
@@ -131,17 +214,21 @@
 def main():
     args = parse_args()
     juju_bin = args.juju_bin
-    version = EnvJujuClient.get_version(juju_bin)
+    version = ModelClient.get_version(juju_bin)
     client_class = get_client_class(version)
     if client_class.config_class is not JujuData:
         logging.warn('This test does not support old jujus.')
     with open(args.example_clouds) as f:
         clouds = yaml.safe_load(f)['clouds']
+    endpoint_validation = bool(not re.match('2\.1[^\d]', version))
+    cloud_specs = iter_clouds(clouds, endpoint_validation)
     with temp_dir() as juju_home:
         env = JujuData('foo', config=None, juju_home=juju_home)
         client = client_class(env, version, juju_bin)
-        succeeded, failed = assess_all_clouds(client, clouds)
+        succeeded, xfailed, failed = assess_all_clouds(client, cloud_specs)
     write_status('Succeeded', succeeded)
+    for bug, failures in sorted(xfailed.items()):
+        write_status('Expected fail (bug #{})'.format(bug), failures)
     write_status('Failed', failed)
     if len(failed) > 0:
         return 1

=== modified file 'assess_autoload_credentials.py'
--- assess_autoload_credentials.py	2016-11-08 15:48:34 +0000
+++ assess_autoload_credentials.py	2017-01-20 20:25:32 +0000
@@ -146,9 +146,9 @@
 def ensure_autoload_credentials_stores_details(client_base, cloud_details_fn):
     """Test covering loading and storing credentials using autoload-credentials
 
-    :param client: EnvJujuClient object to use for the test run.
+    :param client: ModelClient object to use for the test run.
     :param cloud_details_fn: A callable that takes the 3 arguments `user`
-      string, `tmp_dir` path string and client EnvJujuClient and will returns a
+      string, `tmp_dir` path string and client ModelClient and will returns a
       `CloudDetails` object used to setup creation of credential details &
       comparison of the result.
 
@@ -173,9 +173,9 @@
                                                    cloud_details_fn):
     """Storing credentials using autoload-credentials must overwrite existing.
 
-    :param client: EnvJujuClient object to use for the test run.
+    :param client: ModelClient object to use for the test run.
     :param cloud_details_fn: A callable that takes the 3 arguments `user`
-      string, `tmp_dir` path string and client EnvJujuClient and will returns a
+      string, `tmp_dir` path string and client ModelClient and will returns a
       `CloudDetails` object used to setup creation of credential details &
       comparison of the result.
 
@@ -259,7 +259,7 @@
     Simple interaction, calls juju autoload-credentials selects the first
     option and then quits.
 
-    :param client: EnvJujuClient from which juju will be called.
+    :param client: ModelClient from which juju will be called.
     :param envvars: Dictionary containing environment variables to be used
       during execution.
     :param answers: ExpectAnswers object containing answers for the interactive

=== modified file 'assess_cloud.py'
--- assess_cloud.py	2016-11-23 17:33:54 +0000
+++ assess_cloud.py	2017-01-20 20:36:36 +0000
@@ -6,15 +6,13 @@
 from deploy_stack import (
     BootstrapManager,
     )
-from fakejuju import (
+from jujuconfig import get_juju_home
+from jujupy import (
+    ConditionList,
+    ModelClient,
     FakeBackend,
     FakeControllerState,
-    )
-from jujuconfig import get_juju_home
-from jujupy import (
-    EnvJujuClient,
     get_client_class,
-    WaitMachineNotPresent,
     )
 from utility import (
     add_basic_testing_arguments,
@@ -29,20 +27,25 @@
     client is determined based on the path and version.
     """
     if args.juju_bin == 'FAKE':
-        client_class = EnvJujuClient
+        client_class = ModelClient
         controller_state = FakeControllerState()
         version = '2.0.0'
         backend = FakeBackend(controller_state, full_path=args.juju_bin,
                               version=version)
     else:
-        version = EnvJujuClient.get_version(args.juju_bin)
+        version = ModelClient.get_version(args.juju_bin)
         client_class = get_client_class(version)
         backend = None
     juju_home = get_juju_home()
     with open(args.clouds_file) as f:
         clouds = yaml.safe_load(f)
+    if args.config is None:
+        config = {}
+    else:
+        with open(args.config) as f:
+            config = yaml.safe_load(f)
     juju_data = client_class.config_class.from_cloud_region(
-        args.cloud, args.region, {}, clouds, juju_home)
+        args.cloud, args.region, config, clouds, juju_home)
     return client_class(juju_data, version, args.juju_bin, debug=args.debug,
                         soft_deadline=args.deadline, _backend=backend)
 
@@ -59,8 +62,35 @@
         new_status = client.wait_for_started()
         new_machines = [k for k, v in new_status.iter_new_machines(old_status)]
         client.juju('remove-unit', 'ubuntu/0')
-        new_status = client.wait_for([WaitMachineNotPresent(n)
-                                      for n in new_machines])
+        new_status = client.wait_for(
+            ConditionList([client.make_remove_machine_condition(n)
+                           for n in new_machines]))
+
+
+def assess_cloud_provisioning(bs_manager):
+    """Assess provisioning operations on a cloud.
+
+    This was created for testing Azure streams.  It tests bootstrap,
+    add-machine, remove-machine and destroy-controller.  It does not test
+    charms.
+
+    It tests "trusty" and "win2012r2" as representative series on Azure.  It
+    does not test CentOS, because that is known-broken at the moment (bug
+    #1571982).  It tests "trusty" rather than "xenial" because "xenial" will
+    be used for bootstrap by default.
+    """
+    client = bs_manager.client
+    with bs_manager.booted_context(upload_tools=False):
+        old_status = client.get_status()
+        client.juju('add-machine', ('--series', 'win2012r2'))
+        client.juju('add-machine', ('--series', 'trusty'))
+        new_status = client.wait_for_started()
+        new_machines = [k for k, v in new_status.iter_new_machines(old_status)]
+        for machine in new_machines:
+            client.juju('remove-machine', (machine,))
+        new_status = client.wait_for(
+            ConditionList([client.make_remove_machine_condition(n)
+                           for n in new_machines]))
 
 
 def assess_cloud_kill_controller(bs_manager):
@@ -83,12 +113,13 @@
         cloud to test.
         """))
     subparsers = parser.add_subparsers(dest='test')
-    for test in ['combined', 'kill-controller']:
+    for test in ['combined', 'kill-controller', 'provisioning']:
         subparser = subparsers.add_parser(test)
         subparser.add_argument('clouds_file',
                                help='A clouds.yaml file to use for testing.')
         subparser.add_argument('cloud', help='Specific cloud to test.')
         add_basic_testing_arguments(subparser, env=False)
+        subparser.add_argument('--config')
     return parser.parse_args(args)
 
 
@@ -99,6 +130,8 @@
     bs_manager = BootstrapManager.from_client(args, client)
     if args.test == 'combined':
         assess_cloud_combined(bs_manager)
+    if args.test == 'provisioning':
+        assess_cloud_provisioning(bs_manager)
     else:
         assess_cloud_kill_controller(bs_manager)
 

=== modified file 'assess_constraints.py'
--- assess_constraints.py	2016-10-22 02:03:03 +0000
+++ assess_constraints.py	2016-11-29 22:51:05 +0000
@@ -219,7 +219,7 @@
     status = client.get_status()
     app_data = status.get_applications()[application]
     machines = [unit_data['machine'] for unit_data in
-                app_data['units'].itervalues()]
+                app_data['units'].values()]
     return machines
 
 

=== modified file 'assess_container_networking.py'
--- assess_container_networking.py	2016-12-06 11:19:11 +0000
+++ assess_container_networking.py	2017-01-20 20:58:41 +0000
@@ -90,7 +90,7 @@
        Two host machines.
        Two of each container_type on one host machine.
        One of each container_type on one host machine.
-    :param client: An EnvJujuClient
+    :param client: A ModelClient
     :param container_types: list of containers to create
     :return: hosts (list), containers {container_type}{host}[containers]
     """

=== modified file 'assess_heterogeneous_control.py'
--- assess_heterogeneous_control.py	2016-11-25 16:23:45 +0000
+++ assess_heterogeneous_control.py	2017-01-17 18:04:08 +0000
@@ -7,18 +7,15 @@
 from subprocess import CalledProcessError
 import sys
 
-from fakejuju import (
-    fake_juju_client,
-)
 from jujucharm import (
     local_charm_path,
 )
 from jujupy import (
     client_from_config,
     EnvJujuClient1X,
+    fake_juju_client,
     IncompatibleConfigClass,
     SimpleEnvironment,
-    until_timeout,
     )
 from deploy_stack import (
     BootstrapManager,
@@ -28,6 +25,7 @@
 from jujuci import add_credential_args
 from utility import (
     configure_logging,
+    until_timeout,
 )
 
 

=== modified file 'assess_log_forward.py'
--- assess_log_forward.py	2016-11-15 02:53:11 +0000
+++ assess_log_forward.py	2016-12-16 15:43:19 +0000
@@ -76,8 +76,7 @@
     """
     enable_log_forwarding(dummy)
 
-    model1 = dummy.add_model(
-        dummy.env.clone('{}-{}'.format(dummy.env.environment, 'model1')))
+    model1 = dummy.add_model('{}-{}'.format(dummy.env.environment, 'model1'))
 
     charm_path = local_charm_path(
         charm='dummy-source', juju_ver=model1.version)

=== added file 'assess_model_change_watcher.py'
--- assess_model_change_watcher.py	1970-01-01 00:00:00 +0000
+++ assess_model_change_watcher.py	2016-12-08 02:39:36 +0000
@@ -0,0 +1,122 @@
+#!/usr/bin/env python3
+
+import argparse
+import asyncio
+import logging
+import os
+import sys
+
+from juju.client.connection import Connection
+from juju.client import watcher
+
+from deploy_stack import (
+    BootstrapManager,
+    until_timeout,
+    )
+from jujucharm import local_charm_path
+from utility import (
+    add_basic_testing_arguments,
+    configure_logging,
+    JujuAssertionError,
+    )
+
+
+log = logging.getLogger("assess_model_change_watcher")
+TOKEN = "1234asdf"
+
+
+def is_config_change_in_event(event):
+    message = _get_message(event)
+    return all([
+        "application" in event,
+        "change" in event,
+        is_in_dict("config", {"token": TOKEN}, message),
+    ])
+
+
+def _get_message(event):
+    for message in event:
+        if isinstance(message, dict):
+            return message
+    return None
+
+
+def is_in_dict(key, value, items):
+    return items.get(key) == value
+
+
+async def listen_to_watcher(event_found, conn, future):
+    logging.info("Starting to listen for the watcher.")
+    all_watcher = watcher.AllWatcher()
+    all_watcher.connect(conn)
+    for _ in until_timeout(120):
+        logging.info("Listening for events...")
+        change = await all_watcher.Next()
+        for delta in change.deltas:
+            logging.info("Event received: {}".format(str(delta.deltas)))
+            if event_found(delta.deltas) is True:
+                await all_watcher.Stop()
+                await conn.close()
+                logging.info("Event found: {}".format(str(delta.deltas)))
+                future.set_result(True)
+                return
+
+    await all_watcher.Stop()
+    await conn.close()
+    logging.warning("Event not found.")
+    future.set_result(False)
+
+
+def run_listener(client, event, juju_bin):
+    logging.info("Running listener.")
+    loop = asyncio.get_event_loop()
+    future = asyncio.Future()
+
+    logging.info("Connect to the current model.")
+    os.environ['JUJU_DATA'] = client.env.juju_home
+    os.environ['PATH'] = "{}{}{}".format(
+        juju_bin, os.pathsep, os.environ.get('PATH', ''))
+    conn = loop.run_until_complete(Connection.connect_current())
+    logging.info("Connected to the current model.")
+
+    asyncio.ensure_future(listen_to_watcher(event, conn, future))
+    return loop, future
+
+
+def assess_model_change_watcher(client, charm_series, juju_bin):
+    charm = local_charm_path(
+        charm='dummy-source', juju_ver=client.version, series=charm_series,
+        platform='ubuntu')
+    client.deploy(charm)
+    client.wait_for_started()
+
+    loop, future = run_listener(client, is_config_change_in_event, juju_bin)
+
+    logging.info("Making config change.")
+    client.set_config('dummy-source', {'token': TOKEN})
+
+    loop.run_until_complete(future)
+    result = future.result()
+    if result is not True:
+        raise JujuAssertionError("Config change event was not sent.")
+    loop.close()
+
+
+def parse_args(argv):
+    parser = argparse.ArgumentParser(description="Assess config change.")
+    add_basic_testing_arguments(parser)
+    return parser.parse_args(argv)
+
+
+def main(argv=None):
+    args = parse_args(argv)
+    configure_logging(args.verbose)
+    bs_manager = BootstrapManager.from_args(args)
+    with bs_manager.booted_context(args.upload_tools):
+        assess_model_change_watcher(
+            bs_manager.client, bs_manager.series, args.juju_bin)
+    return 0
+
+
+if __name__ == '__main__':
+    sys.exit(main())

=== added file 'assess_model_defaults.py'
--- assess_model_defaults.py	1970-01-01 00:00:00 +0000
+++ assess_model_defaults.py	2016-12-14 20:03:27 +0000
@@ -0,0 +1,136 @@
+#!/usr/bin/env python
+"""Assess the model-defaults command."""
+
+from __future__ import print_function
+
+import argparse
+import logging
+import sys
+
+from deploy_stack import (
+    BootstrapManager,
+    )
+from utility import (
+    add_basic_testing_arguments,
+    configure_logging,
+    JujuAssertionError,
+    )
+
+
+__metaclass__ = type
+
+
+log = logging.getLogger('assess_model_defaults')
+
+
+def assemble_model_default(model_key, default, controller=None, regions=None):
+    """Create a dict that contrains the formatted model-defaults data."""
+    # Ordering in the regions argument is lost.
+    defaults = {'default': default}
+    if controller is not None:
+        defaults['controller'] = controller
+    if regions is not None:
+        defaults['regions'] = [
+            {'name': region, 'value': region_default}
+            for (region, region_default) in regions.items()]
+    return {model_key: defaults}
+
+
+def juju_assert_equal(lhs, rhs, msg):
+    if lhs != rhs:
+        raise JujuAssertionError(msg, lhs, rhs)
+
+
+def get_new_model_config(client, region=None, model_name=None):
+    """Create a new model, get it's config and then destroy it.
+
+    :param client: Client to use create the new model on.
+    :param region: If given and not None, create the new model in that region
+        otherwise create it in the client region.
+    :param model_name: Name of the new model.
+    """
+    if model_name is None:
+        model_name = 'temp-model'
+    new_env = client.env.clone(model_name)
+    if region is not None:
+        new_env.set_region(region)
+    new_model = client.add_model(new_env)
+    config_data = new_model.get_model_config()
+    new_model.destroy_model()
+    return config_data
+
+
+def assess_model_defaults_case(client, model_key, value, expected_default,
+                               cloud=None, region=None):
+    """Check the setting and unsetting of a region field."""
+    base_line = client.get_model_defaults(model_key)
+
+    client.set_model_defaults(model_key, value, cloud, region)
+    juju_assert_equal(expected_default, client.get_model_defaults(model_key),
+                      'Mismatch after setting model-default.')
+    config = get_new_model_config(client, region)
+    juju_assert_equal(value, config[model_key]['value'],
+                      'New model did not use the default.')
+
+    client.unset_model_defaults(model_key, cloud, region)
+    juju_assert_equal(base_line, client.get_model_defaults(model_key),
+                      'Mismatch after resetting model-default.')
+
+
+def assess_model_defaults_no_region(client, model_key, value):
+    """Check the setting and unsetting of the controller field."""
+    default = client.get_model_defaults(model_key)[model_key]['default']
+    assess_model_defaults_case(
+        client, model_key, value,
+        assemble_model_default(model_key, default, str(value)))
+
+
+def assess_model_defaults_region(client, model_key, value,
+                                 cloud=None, region=None):
+    """Check the setting and unsetting of a region field."""
+    default = client.get_model_defaults(model_key)[model_key]['default']
+    assess_model_defaults_case(
+        client, model_key, value,
+        assemble_model_default(model_key, default, None, {region: str(value)}),
+        cloud, region)
+
+
+def assess_model_defaults(client, other_region):
+    log.info('Checking controller model-defaults.')
+    assess_model_defaults_no_region(
+        client, 'automatically-retry-hooks', False)
+    region = client.env.get_region()
+    if region is not None:
+        log.info('Checking region model-defaults.')
+        assess_model_defaults_region(
+            client, 'default-series', 'trusty', region=region)
+    if other_region is not None:
+        log.info('Checking other region model-defaults.')
+        assess_model_defaults_region(
+            client, 'default-series', 'trusty', region=other_region)
+
+
+def parse_args(argv):
+    """Parse all arguments."""
+    parser = argparse.ArgumentParser(
+        description='Assess the model-defaults command.')
+    add_basic_testing_arguments(parser)
+    parser.add_argument('--other-region',
+                        help='Set the model default for a different region.')
+    return parser.parse_args(argv)
+
+
+def main(argv=None):
+    args = parse_args(argv)
+    configure_logging(args.verbose)
+    bs_manager = BootstrapManager.from_args(args)
+    if (args.other_region is not None and
+            args.other_region == bs_manager.client.env.get_region()):
+        raise ValueError('Other region is a repeat of region.')
+    with bs_manager.booted_context(args.upload_tools):
+        assess_model_defaults(bs_manager.client, args.other_region)
+    return 0
+
+
+if __name__ == '__main__':
+    sys.exit(main())

=== modified file 'assess_model_migration.py'
--- assess_model_migration.py	2016-12-02 18:07:25 +0000
+++ assess_model_migration.py	2017-01-20 20:30:41 +0000
@@ -17,7 +17,7 @@
 from deploy_stack import (
     BootstrapManager,
     get_random_string
-)
+    )
 from jujucharm import local_charm_path
 from remote import remote_from_address
 from utility import (
@@ -28,7 +28,7 @@
     get_unit_ipaddress,
     temp_dir,
     until_timeout,
-)
+    )
 
 
 __metaclass__ = type
@@ -45,8 +45,9 @@
         with bs2.existing_booted_context(args.upload_tools):
             source_client = bs1.client
             dest_client = bs2.client
-            ensure_able_to_migrate_model_between_controllers(
+            ensure_migration_including_resources_succeeds(
                 source_client, dest_client)
+            ensure_model_logs_are_migrated(source_client, dest_client)
             with temp_dir() as temp:
                 ensure_migrating_with_insufficient_user_permissions_fails(
                     source_client, dest_client, temp)
@@ -57,16 +58,11 @@
                 if args.use_develop:
                     ensure_superuser_can_migrate_other_user_models(
                         source_client, dest_client, temp)
-            # These too.
+
             if args.use_develop:
-                ensure_model_logs_are_migrated(source_client, dest_client)
                 ensure_migration_rolls_back_on_failure(
                     source_client, dest_client)
-                ensure_migration_of_resources_succeeds(
-                    source_client, dest_client)
                 ensure_api_login_redirects(source_client, dest_client)
-                ensure_migrating_to_target_and_back_to_source_succeeds(
-                    source_client, dest_client)
 
 
 def parse_args(argv):
@@ -145,10 +141,13 @@
                 ))
 
 
-def assert_deployed_charm_is_responding(client, expected_ouput):
+def assert_deployed_charm_is_responding(client, expected_output=None):
     """Ensure that the deployed simple-server charm is still responding."""
+    # Set default value if needed.
+    if expected_output is None:
+        expected_output = 'simple-server.'
     ipaddress = get_unit_ipaddress(client, 'simple-resource-http/0')
-    if expected_ouput != get_server_response(ipaddress):
+    if expected_output != get_server_response(ipaddress):
         raise JujuAssertionError('Server charm is not responding as expected.')
 
 
@@ -156,54 +155,6 @@
     return urlopen('http://{}'.format(ipaddress)).read().rstrip()
 
 
-def test_deployed_mongo_is_up(client):
-    """Ensure the mongo service is running as expected."""
-    try:
-        output = client.get_juju_output(
-            'run', '--unit', 'mongodb/0', 'mongo --eval "db.getMongo()"')
-        if 'connecting to: test' in output:
-            return
-    except CalledProcessError as e:
-        # Pass through to assertion error
-        log.error('Mongodb check command failed: {}'.format(e))
-    raise AssertionError('Mongo db is not in an expected state.')
-
-
-def ensure_able_to_migrate_model_between_controllers(
-        source_client, dest_client):
-    """Test simple migration of a model to another controller.
-
-    Ensure that migration a model that has an application deployed upon it is
-    able to continue it's operation after the migration process.
-
-    Given 2 bootstrapped environments:
-      - Deploy an application
-        - ensure it's operating as expected
-      - Migrate that model to the other environment
-        - Ensure it's operating as expected
-        - Add a new unit to the application to ensure the model is functional
-      - Migrate the model back to the original environment
-        - Note: Test for lp:1607457
-        - Ensure it's operating as expected
-        - Add a new unit to the application to ensure the model is functional
-    """
-    application = 'mongodb'
-    test_model = deploy_mongodb_to_new_model(
-        source_client, model_name='example-model')
-    log.info('Initiating migration process')
-    migration_target_client = migrate_model_to_controller(
-        test_model, dest_client)
-    assert_model_migrated_successfully(migration_target_client, application)
-    migration_target_client.remove_service(application)
-    log.info('SUCCESS: model migrated.')
-
-
-def assert_model_migrated_successfully(client, application):
-    client.wait_for_workloads()
-    test_deployed_mongo_is_up(client)
-    ensure_model_is_functional(client, application)
-
-
 def ensure_api_login_redirects(source_client, dest_client):
     """Login attempts must get transparently redirected to the new controller.
     """
@@ -233,12 +184,12 @@
     # Check file for details.
     assert_data_file_lists_correct_controller_for_model(
         migrated_model_client,
-        expected_controller=dest_client.controller.name)
+        expected_controller=dest_client.env.controller.name)
 
 
 def assert_data_file_lists_correct_controller_for_model(
         client, expected_controller):
-    models_path = os.path.join(client, 'models.yaml')
+    models_path = os.path.join(client.env.juju_home, 'models.yaml')
     with open(models_path, 'rt') as f:
         models_data = yaml.safe_load(f)
 
@@ -258,7 +209,7 @@
         raise JujuAssertionError()
 
 
-def ensure_migration_of_resources_succeeds(source_client, dest_client):
+def ensure_migration_including_resources_succeeds(source_client, dest_client):
     """Test simple migration of a model to another controller.
 
     Ensure that migration a model that has an application, that uses resources,
@@ -266,31 +217,33 @@
     process. This includes assertion that the resources are migrated correctly
     too.
 
-    Almost identical to ensure_able_to_migrate_model_between_controllers except
-    this test uses a charm with a resource.
-
-    Note: This test will supersede
-    ensure_able_to_migrate_model_between_controllers when the develop branch is
-    merged into master.
+    Given 2 bootstrapped environments:
+      - Deploy an application with a resource
+        - ensure it's operating as expected
+      - Migrate that model to the other environment
+        - Ensure it's operating as expected
+        - Add a new unit to the application to ensure the model is functional
 
     """
-    # Don't move the default model so we can reuse it in later tests.
-    test_model = source_client.add_model(
-        source_client.env.clone('example-model-resource'))
     resource_contents = get_random_string()
-    application = deploy_simple_resource_server(test_model, resource_contents)
-    assert_deployed_charm_is_responding(test_model, resource_contents)
-    log.info('Initiating migration process')
+    test_model, application = deploy_simple_server_to_new_model(
+        source_client, 'example-model-resource', resource_contents)
     migration_target_client = migrate_model_to_controller(
         test_model, dest_client)
-    migration_target_client.wait_for_workloads()
-    assert_deployed_charm_is_responding(
-        migration_target_client, resource_contents)
-    ensure_model_is_functional(migration_target_client, application)
+    assert_model_migrated_successfully(
+        migration_target_client, application, resource_contents)
+
     migration_target_client.remove_service(application)
     log.info('SUCCESS: resources migrated')
 
 
+def assert_model_migrated_successfully(
+        client, application, resource_contents=None):
+    client.wait_for_workloads()
+    assert_deployed_charm_is_responding(client, resource_contents)
+    ensure_model_is_functional(client, application)
+
+
 def ensure_superuser_can_migrate_other_user_models(
         source_client, dest_client, tmp_dir):
 
@@ -318,41 +271,13 @@
     migration_client.wait_for_started()
 
 
-def ensure_migrating_to_target_and_back_to_source_succeeds(
-        source_client, dest_client):
-    """Test migration from source to target and back again.
-
-    Almost a duplicate of 'ensure_able_to_migrate_model_between_controllers'
-    except adds the extra step of migrating the model back to the original
-    controller.
-
-    Note: Test for lp:1641824
-    """
-    application = 'mongodb'
-    test_model = deploy_mongodb_to_new_model(
-        source_client, model_name='example-model')
-    log.info('Initiating migration process')
-    migration_target_client = migrate_model_to_controller(
-        test_model, dest_client)
-    assert_model_migrated_successfully(migration_target_client, application)
-    # Ensure migration works back to the original controller as per lp:1641824
-    re_migrate_client = migrate_model_to_controller(
-        migration_target_client, source_client)
-    assert_model_migrated_successfully(re_migrate_client, application)
-    re_migrate_client.remove_service(application)
-    log.info('SUCCESS: model migrated back to source.')
-
-
-def deploy_mongodb_to_new_model(client, model_name):
-    bundle = 'mongodb'
-    log.info('Deploying charm')
-    # Don't move the default model so we can reuse it in later tests.
-    test_model = client.add_model(client.env.clone(model_name))
-    test_model.juju("deploy", (bundle))
-    test_model.wait_for_started()
-    test_model.wait_for_workloads()
-    test_deployed_mongo_is_up(test_model)
-    return test_model
+def deploy_simple_server_to_new_model(
+        client, model_name, resource_contents=None):
+    new_model = client.add_model(client.env.clone(model_name))
+    application = deploy_simple_resource_server(new_model, resource_contents)
+    assert_deployed_charm_is_responding(new_model, resource_contents)
+
+    return new_model, application
 
 
 def deploy_dummy_source_to_new_model(client, model_name):
@@ -366,24 +291,29 @@
     return new_model_client
 
 
-def deploy_simple_resource_server(client, resource_contents):
+def deploy_simple_resource_server(client, resource_contents=None):
     application_name = 'simple-resource-http'
     log.info('Deploying charm: '.format(application_name))
     charm_path = local_charm_path(
         charm=application_name, juju_ver=client.version)
     # Create a temp file which we'll use as the resource.
-    with temp_dir() as temp:
-        index_file = os.path.join(temp, 'index.html')
-        with open(index_file, 'wt') as f:
-            f.write(resource_contents)
-        client.deploy(charm_path, resource='index={}'.format(index_file))
-        client.wait_for_started()
-        client.wait_for_workloads()
-        return application_name
-
-
-def migrate_model_to_controller(source_client, dest_client,
-                                include_user_name=False):
+    if resource_contents is not None:
+        with temp_dir() as temp:
+            index_file = os.path.join(temp, 'index.html')
+            with open(index_file, 'wt') as f:
+                f.write(resource_contents)
+            client.deploy(charm_path, resource='index={}'.format(index_file))
+    else:
+        client.deploy(charm_path)
+
+    client.wait_for_started()
+    client.wait_for_workloads()
+    return application_name
+
+
+def migrate_model_to_controller(
+        source_client, dest_client, include_user_name=False):
+    log.info('Initiating migration process')
     if include_user_name:
         model_name = '{}/{}'.format(
             source_client.env.user_name, source_client.env.environment)
@@ -442,7 +372,7 @@
     log.info('Attempting migration process')
     migrated_model = migrate_model_to_controller(new_model_client, dest_client)
     after_migration_logs = migrated_model.get_juju_output(
-        'debug-log', '--no-tail', '-l', 'DEBUG')
+        'debug-log', '--no-tail', '--replay', '-l', 'DEBUG')
     if before_migration_logs not in after_migration_logs:
         raise JujuAssertionError('Logs failed to be migrated.')
     log.info('SUCCESS: logs migrated.')
@@ -455,9 +385,8 @@
     the migration must roll back and continue to be available on the source
     controller.
     """
-    application = 'mongodb'
-    test_model = deploy_mongodb_to_new_model(
-        source_client, model_name='rollmeback')
+    test_model, application = deploy_simple_server_to_new_model(
+        source_client, 'rollmeback')
     test_model.controller_juju(
         'migrate',
         (test_model.env.environment,
@@ -470,7 +399,7 @@
         log.info('Waiting for migration rollback to complete.')
         wait_for_model(test_model, test_model.env.environment)
         test_model.wait_for_started()
-        test_deployed_mongo_is_up(test_model)
+        assert_deployed_charm_is_responding(test_model)
         ensure_model_is_functional(test_model, application)
     test_model.remove_service(application)
     log.info('SUCCESS: migration rolled back.')
@@ -538,12 +467,12 @@
                                tmp_dir, username, permission):
     """Create a user on both supplied controller with the permissions supplied.
 
-    :param source_client: EnvJujuClient object to create user on.
-    :param dest_client: EnvJujuClient object to create user on.
+    :param source_client: ModelClient object to create user on.
+    :param dest_client: ModelClient object to create user on.
     :param tmp_dir: Path to base new users JUJU_DATA directory in.
     :param username: String of username to use.
     :param permission: String for permissions to grant user on both
-      controllers. Valid values are `EnvJujuClient.controller_permissions`.
+      controllers. Valid values are `ModelClient.controller_permissions`.
     """
     new_user_home = os.path.join(tmp_dir, username)
     os.makedirs(new_user_home)

=== modified file 'assess_multi_series_charms.py'
--- assess_multi_series_charms.py	2016-06-27 11:41:04 +0000
+++ assess_multi_series_charms.py	2017-01-20 20:39:27 +0000
@@ -67,7 +67,7 @@
     """Assess multi series charms.
 
     :param client: Juju client.
-    :type client: jujupy.EnvJujuClient
+    :type client: jujupy.ModelClient
     :return: None
     """
     tests = [
@@ -104,7 +104,7 @@
     """Deploy a charm and assert a success or fail.
 
     :param client: Juju client
-    :type client: jujupy.EnvJujuClient
+    :type client: jujupy.ModelClient
     :param test: Deploy test data.
     :type  test: Test
     :param charm_dir:

=== renamed file 'assess_jes_deploy.py' => 'assess_multimodel.py'
--- assess_jes_deploy.py	2016-09-20 14:23:52 +0000
+++ assess_multimodel.py	2016-12-16 16:03:27 +0000
@@ -1,4 +1,5 @@
 #!/usr/bin/env python
+"""Assess multimodel support."""
 
 from argparse import ArgumentParser
 from contextlib import contextmanager
@@ -22,10 +23,10 @@
     add_basic_testing_arguments,
     ensure_dir,
     print_now,
-)
-
-
-def test_jes_deploy(client, charm_series, log_dir, base_env):
+    )
+
+
+def assess_multimodel_deploy(client, charm_series, log_dir, base_env):
     """Deploy the dummy stack in two hosted environments."""
     # deploy into system env
     deploy_dummy_stack(client, charm_series)
@@ -41,8 +42,17 @@
             check_services(env2_client)
 
 
+def assess_destroy_current(client):
+    model_name = client.model_name
+    new_model = client.add_model('delete-me')
+    new_model.switch('delete-me')
+    new_model.destroy_model()
+    new_model.show_controller()
+    client.switch(model_name)
+
+
 @contextmanager
-def jes_setup(args):
+def multimodel_setup(args):
     """
     Sets up the juju client and its environment.
 
@@ -81,7 +91,7 @@
 @contextmanager
 def hosted_environment(system_client, log_dir, suffix):
     env_name = '{}-{}'.format(system_client.env.environment, suffix)
-    client = system_client.add_model(system_client.env.clone(env_name))
+    client = system_client.add_model(env_name)
     try:
         yield client
     except:
@@ -104,12 +114,18 @@
     check_token(client, token)
 
 
-def main():
+def parse_args(argv=None):
+    """Parse all arguments."""
     parser = ArgumentParser()
     add_basic_testing_arguments(parser, using_jes=True, deadline=True)
-    args = parser.parse_args()
-    with jes_setup(args) as (client, charm_series, base_env):
-        test_jes_deploy(client, charm_series, args.logs, base_env)
+    return parser.parse_args(argv)
+
+
+def main():
+    args = parse_args()
+    with multimodel_setup(args) as (client, charm_series, base_env):
+        assess_multimodel_deploy(client, charm_series, args.logs, base_env)
+        assess_destroy_current(client)
 
 
 if __name__ == '__main__':

=== modified file 'assess_public_clouds.py'
--- assess_public_clouds.py	2016-11-22 21:58:33 +0000
+++ assess_public_clouds.py	2016-12-18 17:35:13 +0000
@@ -1,4 +1,6 @@
 #!/usr/bin/env python
+"""Assess basic quality of public clouds."""
+
 from __future__ import print_function
 
 from argparse import ArgumentParser
@@ -36,14 +38,15 @@
     return log_dir
 
 
-def prepare_cloud(config, region, client, log_dir):
+def make_bootstrap_manager(config, region, client, log_dir):
     env_name = 'boot-cpc-{}-{}'.format(client.env.get_cloud(), region)[:30]
     logging_dir = make_logging_dir(log_dir, config, region)
     bs_manager = BootstrapManager(
         env_name, client, client, bootstrap_host=None, machines=[],
         series=None, agent_url=None, agent_stream=None, region=region,
-        log_dir=logging_dir, keep_env=False, permanent=True, jes_enabled=True)
-    assess_cloud_combined(bs_manager)
+        log_dir=logging_dir, keep_env=False, permanent=True, jes_enabled=True,
+        logged_exception_exit=False)
+    return bs_manager
 
 
 def iter_cloud_regions(public_clouds, credentials):
@@ -75,7 +78,9 @@
         try:
             client = client_from_config(
                 config, args.juju_bin, args.debug, args.deadline)
-            prepare_cloud(config, region, client, args.logs)
+            bs_manager = make_bootstrap_manager(config, region, client,
+                                                args.logs)
+            assess_cloud_combined(bs_manager)
         except LoggedException as error:
             yield config, region, error.exception
         except Exception as error:
@@ -85,7 +90,7 @@
 def parse_args(argv):
     """Parse all arguments."""
     parser = ArgumentParser(
-        description="This is a quick hack to test 0052b26.  HERE BE DRAGONS!")
+        description='Assess basic quality of public clouds.')
     parser.add_argument('juju_bin', nargs='?',
                         help='Full path to the Juju binary. By default, this'
                         ' will use $GOPATH/bin/juju or /usr/bin/juju in that'
@@ -117,7 +122,6 @@
     configure_logging(logging.INFO)
     args = parse_args(None)
     default_log_dir(args)
-    logging.warning('This is a quick hack to test 0052b26.  HERE BE DRAGONS!')
     public_clouds = yaml_file_load('public-clouds.yaml')['clouds']
     credentials = yaml_file_load('credentials.yaml')['credentials']
     failures = []

=== modified file 'assess_spaces_subnets.py'
--- assess_spaces_subnets.py	2016-02-11 15:22:35 +0000
+++ assess_spaces_subnets.py	2017-01-20 20:38:11 +0000
@@ -50,7 +50,7 @@
 
 def assess_spaces_subnets(client):
     """Check that space and subnet functionality works as expected
-    :param client: EnvJujuClient
+    :param client: ModelClient
     """
     network_config = {
         'default': ['subnet-0fb97566', 'subnet-d27d91a9'],
@@ -75,7 +75,7 @@
 
 def _assess_spaces_subnets(client, network_config, charms_to_space):
     """Check that space and subnet functionality works as expected
-    :param client: EnvJujuClient
+    :param client: ModelClient
     :param network_config: Map of 'space name' to ['subnet', 'list']
     :param charms_to_space: Map of 'unit name' to
            {'space': 'space name', 'charm': 'charm name (if not same as unit)}

=== modified file 'assess_unregister.py'
--- assess_unregister.py	2016-07-25 20:50:28 +0000
+++ assess_unregister.py	2017-01-20 20:48:06 +0000
@@ -79,7 +79,7 @@
 def assert_controller_list(client, controller_list):
     """Assert that clients controller list only contains names provided.
 
-    :param client: EnvJujuClient to retrieve controllers of.
+    :param client: ModelClient to retrieve controllers of.
     :param controller_list: list of strings for expected controller names.
 
     """

=== modified file 'assess_win_client.py'
--- assess_win_client.py	2016-11-23 13:47:35 +0000
+++ assess_win_client.py	2017-01-17 15:04:17 +0000
@@ -19,25 +19,21 @@
     installer = [u for u in urls if re.search('juju-setup-.*\.exe', u)][-1]
     s3_cmd(['sync', installer, '.'])
     install_file = installer.split('/')[-1]
+    deploy_job = (
+        'c:\\\\Users\\\\Administrator\\\\juju-ci-tools\\\\deploy_job.py')
     with open('run-file', 'w') as run_file:
         run_file.write(dedent("""
             ci/$1 /verysilent
             juju version
             juju destroy-environment --force -y win-client-deploy
             mkdir logs
-            python ci\\\\deploy_job.py parallel-win-client \
+            python {deploy_job} parallel-win-client \
                 'c:\\Program Files (x86)\\Juju\\juju.exe' \
-                logs win-client-deploy --series trusty \
+                logs win-client-deploy --series xenial --use-charmstore \
                 --agent-stream revision-build-{revision_build}
-            """.format(revision_build=revision_build)))
+            """.format(revision_build=revision_build, deploy_job=deploy_job)))
 
-    ci = [os.path.join(script_dir, f) for f in [
-        'deploy_stack.py', 'deploy_job.py', 'fakejuju.py', 'jujupy.py',
-        'jujuconfig.py', 'remote.py', 'substrate.py', 'utility.py',
-        'get_ami.py', 'chaos.py', 'timeout.py', 'jujucharm.py',
-        'winazurearm.py', 'gce.py',
-        ]]
-    ci.extend([install_file, 'run-file'])
+    ci = [install_file, 'run-file']
     with open('foo.yaml', 'w') as config:
         yaml.dump({
             'install': {'ci': ci},

=== modified file 'bootstrap'
--- bootstrap	2016-10-30 15:02:38 +0000
+++ bootstrap	2017-01-20 21:43:09 +0000
@@ -5,8 +5,6 @@
 from jujupy import (
     bootstrap_from_env,
     client_from_config,
-    EnvJujuClient,
-    SimpleEnvironment,
 )
 
 

=== added file 'clean_lxd.py'
--- clean_lxd.py	1970-01-01 00:00:00 +0000
+++ clean_lxd.py	2017-01-12 16:42:06 +0000
@@ -0,0 +1,52 @@
+#!/usr/bin/env python
+from __future__ import print_function
+
+from argparse import ArgumentParser
+from datetime import (
+    datetime,
+    timedelta,
+    )
+import json
+import os
+import subprocess
+import sys
+
+from dateutil import (
+    parser as date_parser,
+    tz,
+    )
+
+
+def list_old_juju_containers(hours):
+    env = dict(os.environ)
+    containers = json.loads(subprocess.check_output([
+        'lxc', 'list', '--format', 'json'], env=env))
+    now = datetime.now(tz.gettz('UTC'))
+    for container in containers:
+        name = container['name']
+        if not name.startswith('juju-'):
+            continue
+        created_at = date_parser.parse(container['created_at'])
+        age = now - created_at
+        if age <= timedelta(hours=hours):
+            continue
+        yield name, age
+
+
+def main():
+    parser = ArgumentParser('Delete old juju containers')
+    parser.add_argument('--dry-run', action='store_true',
+                        help='Do not actually delete.')
+    parser.add_argument('--hours', type=int, default=1,
+                        help='Number of hours a juju container may exist.')
+    args = parser.parse_args()
+    for container, age in list_old_juju_containers(args.hours):
+        print('deleting {} ({} old)'.format(container, age))
+        if args.dry_run:
+            continue
+        subprocess.check_call(('lxc', 'delete', '--verbose', '--force',
+                               container))
+
+
+if __name__ == '__main__':
+    sys.exit(main())

=== modified file 'clean_maas.py'
--- clean_maas.py	2016-04-29 00:19:30 +0000
+++ clean_maas.py	2017-01-18 16:57:54 +0000
@@ -4,25 +4,36 @@
 from __future__ import print_function
 
 import argparse
+from datetime import (
+    datetime,
+    timedelta,
+    )
 import sys
 
-import jujuconfig
+from jujupy import JujuData
 import substrate
 
 
 def main(argv):
-    parser = argparse.ArgumentParser(description="Delete all machines in MAAS")
+    parser = argparse.ArgumentParser(
+        description="Delete the machines in MAAS.")
     parser.add_argument("name", help="Name of the MAAS in juju config.")
+    parser.add_argument('--hours', help='Minimum age in hours.', type=float)
+    parser.add_argument('--dry-run', action='store_true',
+                        help="Show what would be deleted, but don't delete.")
     args = parser.parse_args(argv[1:])
-    environments = jujuconfig.get_environments()
-    if args.name not in environments:
-        parser.error("No maas '{}' found in juju config".format(args.name))
-    config = environments[args.name]
-    with substrate.maas_account_from_config(config) as manager:
+    boot_config = JujuData.from_config(args.name)
+    with substrate.maas_account_from_boot_config(boot_config) as manager:
         machines = manager.get_allocated_nodes()
+        if args.hours is not None:
+            threshold = datetime.now() - timedelta(hours=args.hours)
+            machines = dict(
+                (k, v) for k, v in machines.items()
+                if manager.get_acquire_date(v['system_id']) < threshold)
         print("Found {} machines: {}".format(len(machines), machines.keys()))
-        manager.terminate_instances(machine["resource_uri"]
-                                    for machine in machines.values())
+        if not args.dry_run:
+            manager.terminate_instances(machine["resource_uri"]
+                                        for machine in machines.values())
         print("Released.")
     return 0
 

=== modified file 'convert_config.py'
--- convert_config.py	2016-11-01 19:27:30 +0000
+++ convert_config.py	2017-01-17 18:04:08 +0000
@@ -8,9 +8,9 @@
 
 import yaml
 
-from fakejuju import fake_juju_client
 from jujuconfig import get_environments
 from jujupy import (
+    fake_juju_client,
     JujuData,
     )
 

=== modified file 'deploy_stack.py'
--- deploy_stack.py	2016-12-01 16:39:37 +0000
+++ deploy_stack.py	2017-01-20 20:23:55 +0000
@@ -3,10 +3,13 @@
 
 
 from argparse import ArgumentParser
-from contextlib import (
-    contextmanager,
-    nested,
-)
+from contextlib import contextmanager
+try:
+    from contextlib import nested
+except ImportError:
+    from contextlib import ExitStack as nested
+
+
 from datetime import (
     datetime,
 )
@@ -25,10 +28,6 @@
 import shutil
 
 from chaos import background_chaos
-from fakejuju import (
-    FakeBackend,
-    fake_juju_client,
-)
 from jujucharm import (
     local_charm_path,
 )
@@ -38,12 +37,16 @@
 )
 from jujupy import (
     client_from_config,
-    get_local_root,
+    FakeBackend,
+    fake_juju_client,
     get_machine_dns_name,
     jes_home_path,
     NoProvider,
     SimpleEnvironment,
     temp_bootstrap_env,
+    )
+from jujupy.client import (
+    get_local_root,
 )
 from remote import (
     remote_from_address,
@@ -83,7 +86,7 @@
         destroy_job_instances(instance_tag)
 
 
-def deploy_dummy_stack(client, charm_series):
+def deploy_dummy_stack(client, charm_series, use_charmstore=False):
     """"Deploy a dummy stack in the specified environment."""
     # Centos requires specific machine configuration (i.e. network device
     # order).
@@ -94,12 +97,18 @@
         platform = 'win'
     elif charm_series.startswith('centos'):
         platform = 'centos'
-    charm = local_charm_path(charm='dummy-source', juju_ver=client.version,
-                             series=charm_series, platform=platform)
-    client.deploy(charm, series=charm_series)
-    charm = local_charm_path(charm='dummy-sink', juju_ver=client.version,
-                             series=charm_series, platform=platform)
-    client.deploy(charm, series=charm_series)
+    if use_charmstore:
+        dummy_source = "cs:~juju-qa/dummy-source"
+        dummy_sink = "cs:~juju-qa/dummy-sink"
+    else:
+        dummy_source = local_charm_path(
+            charm='dummy-source', juju_ver=client.version, series=charm_series,
+            platform=platform)
+        dummy_sink = local_charm_path(
+            charm='dummy-sink', juju_ver=client.version, series=charm_series,
+            platform=platform)
+    client.deploy(dummy_source, series=charm_series)
+    client.deploy(dummy_sink, series=charm_series)
     client.juju('add-relation', ('dummy-source', 'dummy-sink'))
     client.juju('expose', ('dummy-sink',))
     if client.env.kvm or client.env.maas:
@@ -119,20 +128,6 @@
     check_token(client, token)
 
 
-GET_TOKEN_SCRIPT = """
-        for x in $(seq 120); do
-          if [ -f /var/run/dummy-sink/token ]; then
-            if [ "$(cat /var/run/dummy-sink/token)" != "" ]; then
-              break
-            fi
-          fi
-          sleep 1
-        done
-        cat /var/run/dummy-sink/token
-        sleep 2
-    """
-
-
 def get_token_from_status(client):
     """Return the token from the application status message or None."""
     status = client.get_status()
@@ -146,41 +141,48 @@
     return None
 
 
+token_pattern = re.compile(r'([^\n\r]*)\r?\n?')
+
+
+def _get_token(remote, token_path="/var/run/dummy-sink/token"):
+    """Check for token with remote, handling missing error if needed."""
+    try:
+        contents = remote.cat(token_path)
+    except subprocess.CalledProcessError as e:
+        if e.returncode != 1:
+            raise
+        return ""
+    return token_pattern.match(contents).group(1)
+
+
 def check_token(client, token, timeout=120):
     """Check the token found on dummy-sink/0 or raise ValueError."""
     logging.info('Waiting for applications to reach ready.')
     client.wait_for_workloads()
-    # Wait up to 120 seconds for token to be created.
     logging.info('Retrieving token.')
     remote = remote_from_unit(client, "dummy-sink/0")
     # Update remote with real address if needed.
     resolve_remote_dns_names(client.env, [remote])
+    # By this point the workloads should be ready and token will have been
+    # sent successfully, but fallback to timeout as previously for now.
     start = time.time()
     while True:
         if remote.is_windows():
             result = get_token_from_status(client)
             if not result:
-                try:
-                    result = remote.cat("%ProgramData%\\dummy-sink\\token")
-                except winrm.exceptions.WinRMTransportError as e:
-                    logging.warning(
-                        "Skipping token check because of: {}".format(str(e)))
-                    return
+                result = _get_token(remote, "%ProgramData%\\dummy-sink\\token")
         else:
-            result = remote.run(GET_TOKEN_SCRIPT)
-        token_pattern = re.compile(r'([^\n\r]*)\r?\n?')
-        result = token_pattern.match(result).group(1)
+            result = _get_token(remote)
         if result == token:
             logging.info("Token matches expected %r", result)
             return
         if time.time() - start > timeout:
-            if not remote.is_windows() and remote.use_juju_ssh:
+            if remote.use_juju_ssh and _can_run_ssh():
                 # 'juju ssh' didn't error, but try raw ssh to verify
                 # the result is the same.
                 remote.get_address()
                 remote.use_juju_ssh = False
-                result = remote.run(GET_TOKEN_SCRIPT)
-                result = token_pattern.match(result).group(1)
+                result = _get_token(remote)
                 if result == token:
                     logging.info("Token matches expected %r", result)
                     logging.error("juju ssh to unit is broken.")
@@ -274,7 +276,7 @@
         if machine_id not in machines:
             machines[machine_id] = remote_from_address(address)
     # Update remote machines in place with real addresses if substrate needs.
-    resolve_remote_dns_names(client.env, machines.itervalues())
+    resolve_remote_dns_names(client.env, machines.values())
     return machines
 
 
@@ -460,6 +462,8 @@
             'Host with a controller to use.  If supplied, SSO_EMAIL and'
             ' SSO_PASSWORD environment variables will be used for oauth'
             ' authentication.'))
+    parser.add_argument('--use-charmstore', action='store_true',
+                        help='Deploy dummy charms from the charmstore.')
     return parser.parse_args(argv)
 
 
@@ -603,8 +607,8 @@
 
     :ivar temp_env_name: a unique name for the juju env, such as a Jenkins
         job name.
-    :ivar client: an EnvJujuClient.
-    :ivar tear_down_client: an EnvJujuClient for tearing down the environment
+    :ivar client: a ModelClient.
+    :ivar tear_down_client: a ModelClient for tearing down the controller
         (may be more reliable/capable/compatible than client.)
     :ivar bootstrap_host: None, or the address of a manual or MAAS host to
         bootstrap on.
@@ -624,7 +628,8 @@
 
     def __init__(self, temp_env_name, client, tear_down_client, bootstrap_host,
                  machines, series, agent_url, agent_stream, region, log_dir,
-                 keep_env, permanent, jes_enabled, controller_strategy=None):
+                 keep_env, permanent, jes_enabled, controller_strategy=None,
+                 logged_exception_exit=True):
         """Constructor.
 
         Please see see `BootstrapManager` for argument descriptions.
@@ -649,6 +654,7 @@
         if controller_strategy is None:
             controller_strategy = CreateController(client, tear_down_client)
         self.controller_strategy = controller_strategy
+        self.logged_exception_exit = logged_exception_exit
         self.has_controller = False
 
     @property
@@ -825,6 +831,7 @@
             wait_for_port(machine, 22, timeout=120)
 
         with self.handle_bootstrap_exceptions():
+            self.has_controller = True
             yield
 
     @contextmanager
@@ -956,7 +963,7 @@
         environment will be destroyed when the test completes or there is an
         unrecoverable error.
 
-        The temporary environment is created by updating a EnvJujuClient's
+        The temporary environment is created by updating a ModelClient's
         config with series, agent_url, agent_stream.
 
         :param upload_tools: False or True to upload the local agent instead
@@ -977,7 +984,9 @@
                         m_client.show_status()
                     yield machines
         except LoggedException:
-            sys.exit(1)
+            if self.logged_exception_exit:
+                sys.exit(1)
+            raise
 
     @contextmanager
     def existing_booted_context(self, upload_tools, **kwargs):
@@ -1007,12 +1016,12 @@
     will be destroyed when the test completes or there is an unrecoverable
     error.
 
-    The temporary environment is created by updating a EnvJujuClient's config
+    The temporary environment is created by updating a ModelClient's config
     with series, agent_url, agent_stream.
 
     :param temp_env_name: a unique name for the juju env, such as a Jenkins
         job name.
-    :param client: an EnvJujuClient.
+    :param client: an ModelClient.
     :param bootstrap_host: None, or the address of a manual or MAAS host to
         bootstrap on.
     :param machine: [] or a list of machines to use add to a manual env
@@ -1057,10 +1066,6 @@
         args.keep_env, permanent=jes_enabled, jes_enabled=jes_enabled,
         controller_strategy=controller_strategy)
     with bs_manager.booted_context(args.upload_tools):
-        if sys.platform in ('win32', 'darwin'):
-            # The win and osx client tests only verify the client
-            # can bootstrap and call the state-server.
-            return
         if args.with_chaos > 0:
             manager = background_chaos(args.temp_env_name, client,
                                        args.logs, args.with_chaos)
@@ -1069,9 +1074,11 @@
             # deploy_dummy_stack(), as was the case prior to this revision.
             manager = nested()
         with manager:
-            deploy_dummy_stack(client, charm_series)
+            deploy_dummy_stack(client, charm_series, args.use_charmstore)
         assess_juju_relations(client)
-        skip_juju_run = charm_series.startswith(("centos", "win"))
+        skip_juju_run = (
+            (client.version < "2" and sys.platform in ("win32", "darwin")) or
+            charm_series.startswith(("centos", "win")))
         if not skip_juju_run:
             assess_juju_run(client)
         if args.upgrade:

=== modified file 'generate_perfscale_results.py'
--- generate_perfscale_results.py	2016-10-27 05:30:29 +0000
+++ generate_perfscale_results.py	2017-01-23 16:59:17 +0000
@@ -125,7 +125,7 @@
     for the controller during the test run.
 
     :param target_test: A callable that takes 2 arguments:
-        - EnvJujuClient client object  (bootstrapped)
+        - ModelClient client object  (bootstrapped)
         - argparse args object
       This callable must return a `DeployDetails` object.
 
@@ -220,16 +220,17 @@
 
         admin_client.juju(
             'scp',
-            ('--',
+            ('--proxy', '--',
              '-r',
              '{}:/var/lib/collectd/rrd/localhost/*'.format(machine_id),
              results_dir)
         )
         try:
             admin_client.juju(
-                'scp', ('{}:/tmp/mongodb-stats.log'.format(machine_id),
-                        results_dir)
-            )
+                'scp',
+                ('--proxy',
+                 '{}:/tmp/mongodb-stats.log'.format(machine_id),
+                 results_dir))
         except subprocess.CalledProcessError as e:
             log.error('Failed to copy mongodb stats for machine {}: {}'.format(
                 machine_id, e))
@@ -564,14 +565,14 @@
     """
     admin_client.juju(
         'scp',
-        (PATHS.collectd_config_path, '{}:{}'.format(
+        ('--proxy', PATHS.collectd_config_path, '{}:{}'.format(
             machine_id, PATHS.collectd_config_dest_file)))
 
     admin_client.juju(
         'scp',
-        (PATHS.installer_script_path, '{}:{}'.format(
+        ('--proxy', PATHS.installer_script_path, '{}:{}'.format(
             machine_id, PATHS.installer_script_dest_path)))
-    admin_client.juju('ssh', (machine_id, 'chmod +x {}'.format(
+    admin_client.juju('ssh', ('--proxy', machine_id, 'chmod +x {}'.format(
         PATHS.installer_script_dest_path)))
 
 
@@ -580,10 +581,14 @@
     # Respawn incase the initial execution fails for whatever reason.
     admin_client.juju(
         'ssh',
-        (machine_id, '{installer} {config_file} {output_file}'.format(
-            installer=PATHS.installer_script_dest_path,
-            config_file=PATHS.collectd_config_dest_file,
-            output_file=PATHS.runner_script_dest_path)))
+        ('--proxy',
+         machine_id,
+         '{installer} {config_file} {output_file}'.format(
+             installer=PATHS.installer_script_dest_path,
+             config_file=PATHS.collectd_config_dest_file,
+             output_file=PATHS.runner_script_dest_path)))
 
-    admin_client.juju('ssh', (machine_id, '--', 'daemon --respawn {}'.format(
-        PATHS.runner_script_dest_path)))
+    admin_client.juju(
+        'ssh',
+        ('--proxy', machine_id, '--', 'daemon --respawn {}'.format(
+            PATHS.runner_script_dest_path)))

=== modified file 'gotestwin.py'
--- gotestwin.py	2016-10-15 18:31:14 +0000
+++ gotestwin.py	2017-01-25 02:32:29 +0000
@@ -31,10 +31,10 @@
         s3_ci_path = join(SCRIPTS, 's3ci.py')
         downloaded = subprocess.check_output([
             s3_ci_path, 'get', revision, 'build-revision',
-            'juju_core.*.tar.gz', './'])
-        tarfile = basename(downloaded)
+            '.*.tar.gz', './'])
         job_name = os.environ.get('job_name', 'GoTestWin')
         subprocess.check_call([s3_ci_path, 'get-summary', revision, job_name])
+    tarfile = basename(downloaded)
     with open('temp-config.yaml', 'w') as temp_file:
         dump({
             'install': {'ci': [

=== modified file 'industrial_test.py'
--- industrial_test.py	2016-11-01 21:09:56 +0000
+++ industrial_test.py	2017-01-20 20:58:04 +0000
@@ -22,11 +22,11 @@
     Charm,
     local_charm_path,
 )
-from jujupy import (
+from jujupy import ModelClient
+from jujupy.client import (
     AgentsNotStarted,
     AGENTS_READY,
     coalesce_agent_status,
-    EnvJujuClient,
     get_machine_dns_name,
     LXC_MACHINE,
     LXD_MACHINE,
@@ -151,7 +151,7 @@
 
     def make_industrial_test(self):
         """Create an IndustrialTest for this MultiIndustrialTest."""
-        stable_path = EnvJujuClient.get_full_path()
+        stable_path = ModelClient.get_full_path()
         paths = [self.really_old_path, stable_path, self.new_juju_path]
         upgrade_sequence = [p for p in paths if p is not None]
         stage_attempts = [self.stages.factory(upgrade_sequence,
@@ -217,8 +217,8 @@
     def __init__(self, old_client, new_client, stage_attempts):
         """Constructor.
 
-        :param old_client: An EnvJ/maujuClient for the old juju.
-        :param new_client: An EnvJujuClient for the new juju.
+        :param old_client: An ModelClient for the old juju.
+        :param new_client: A ModelClient for the new juju.
         :param stage_attemps: List of stages to attempt.
         """
         self.old_client = old_client

=== modified file 'jujucharm.py'
--- jujucharm.py	2016-09-21 19:03:50 +0000
+++ jujucharm.py	2017-01-12 22:58:16 +0000
@@ -173,6 +173,7 @@
 
 
 def sane_charm_store_api_url(url):
+    """Ensure the store url includes the right parts."""
     if url is None:
         return CharmCommand.default_api_url
-    return '{}/charmstore'.format(re.sub('//(www\.)?', '//api.', url))
+    return '{}/charmstore'.format(url)

=== added directory 'jujupy'
=== added file 'jujupy/__init__.py'
--- jujupy/__init__.py	1970-01-01 00:00:00 +0000
+++ jujupy/__init__.py	2017-01-20 21:41:35 +0000
@@ -0,0 +1,83 @@
+from jujupy.client import (
+    AgentsNotStarted,
+    AuthNotAccepted,
+    ConditionList,
+    get_cache_path,
+    get_machine_dns_name,
+    get_timeout_prefix,
+    InvalidEndpoint,
+    jes_home_path,
+    JESNotSupported,
+    JujuData,
+    JUJU_DEV_FEATURE_FLAGS,
+    Juju2Backend,
+    KILL_CONTROLLER,
+    KVM_MACHINE,
+    LXC_MACHINE,
+    LXD_MACHINE,
+    Machine,
+    ModelClient,
+    NameNotAccepted,
+    NoProvider,
+    parse_new_state_server_from_error,
+    SimpleEnvironment,
+    SoftDeadlineExceeded,
+    Status,
+    temp_bootstrap_env,
+    _temp_env,
+    TypeNotAccepted,
+    uniquify_local,
+    )
+from jujupy.fake import (
+    FakeBackend,
+    FakeControllerState,
+    fake_juju_client,
+    fake_juju_client_optional_jes,
+    )
+from jujupy.version_client import (
+    client_from_config,
+    EnvJujuClient1X,
+    EnvJujuClient25,
+    get_client_class,
+    IncompatibleConfigClass,
+    )
+
+__all__ = [
+    'AgentsNotStarted',
+    'AuthNotAccepted',
+    'client_from_config',
+    'ConditionList',
+    'EnvJujuClient1X',
+    'EnvJujuClient25',
+    'FakeBackend',
+    'FakeControllerState',
+    'fake_juju_client',
+    'fake_juju_client_optional_jes',
+    'get_cache_path',
+    'get_client_class',
+    'get_machine_dns_name',
+    'get_timeout_prefix',
+    'IncompatibleConfigClass',
+    'InvalidEndpoint',
+    'jes_home_path',
+    'JESNotSupported',
+    'JujuData',
+    'JUJU_DEV_FEATURE_FLAGS',
+    'Juju2Backend',
+    'KILL_CONTROLLER',
+    'KVM_MACHINE',
+    'LXC_MACHINE',
+    'LXD_MACHINE',
+    'Machine',
+    'ModelClient',
+    'NameNotAccepted',
+    'NoProvider',
+    'parse_new_state_server_from_error',
+    'SimpleEnvironment',
+    'SoftDeadlineExceeded',
+    'Status',
+    'temp_bootstrap_env',
+    '_temp_env',
+    'TypeNotAccepted',
+    'uniquify_local',
+    ]

=== renamed file 'jujupy.py' => 'jujupy/client.py'
--- jujupy.py	2016-12-06 10:09:46 +0000
+++ jujupy/client.py	2017-01-20 20:03:12 +0000
@@ -21,7 +21,6 @@
 import shutil
 import subprocess
 import sys
-from tempfile import NamedTemporaryFile
 import time
 
 from dateutil.parser import parse as datetime_parse
@@ -37,8 +36,8 @@
     )
 from utility import (
     check_free_disk_space,
-    ensure_deleted,
     ensure_dir,
+    get_timeout_path,
     is_ipv6_address,
     JujuResourceTimeout,
     pause,
@@ -48,6 +47,7 @@
     skip_on_missing_file,
     split_address_port,
     temp_dir,
+    temp_yaml_file,
     unqualified_model_name,
     until_timeout,
     )
@@ -55,6 +55,12 @@
 
 __metaclass__ = type
 
+# Python 2 and 3 compatibility
+try:
+    argtype = basestring
+except NameError:
+    argtype = str
+
 AGENTS_READY = set(['started', 'idle'])
 WIN_JUJU_CMD = os.path.join('\\', 'Progra~2', 'Juju', 'juju.exe')
 
@@ -86,10 +92,6 @@
     """Raised when 'juju status' timed out."""
 
 
-class IncompatibleConfigClass(Exception):
-    """Raised when a client is initialised with the wrong config class."""
-
-
 class SoftDeadlineExceeded(Exception):
     """Raised when an overall client operation takes too long."""
 
@@ -110,15 +112,14 @@
     """Raised when the provided name was not accepted."""
 
 
+class InvalidEndpoint(Exception):
+    """Raised when the provided endpoint was deemed invalid."""
+
+
 class AuthNotAccepted(Exception):
     """Raised when the provided auth was not accepted."""
 
 
-def get_timeout_path():
-    import timeout
-    return os.path.abspath(timeout.__file__)
-
-
 def get_timeout_prefix(duration, timeout_path=None):
     """Return extra arguments to run a command with a timeout."""
     if timeout_path is None:
@@ -156,14 +157,6 @@
         self.state = state
 
 
-class BootstrapMismatch(Exception):
-
-    def __init__(self, arg_name, arg_val, env_name, env_val):
-        super(BootstrapMismatch, self).__init__(
-            '--{} {} does not match {}: {}'.format(
-                arg_name, arg_val, env_name, env_val))
-
-
 class UpgradeMongoNotSupported(Exception):
 
     def __init__(self):
@@ -185,13 +178,6 @@
             'This client does not need to enable JES')
 
 
-class VersionNotTestedError(Exception):
-
-    def __init__(self, version):
-        super(VersionNotTestedError, self).__init__(
-            'Tests for juju {} are no longer supported.'.format(version))
-
-
 Machine = namedtuple('Machine', ['machine_id', 'info'])
 
 
@@ -250,17 +236,6 @@
     _fmt = 'Timed out waiting for voting to be enabled in {env}.'
 
 
-@contextmanager
-def temp_yaml_file(yaml_dict):
-    temp_file = NamedTemporaryFile(suffix='.yaml', delete=False)
-    try:
-        with temp_file:
-            yaml.safe_dump(yaml_dict, temp_file)
-        yield temp_file.name
-    finally:
-        os.unlink(temp_file.name)
-
-
 class SimpleEnvironment:
     """Represents a model in a JUJU_HOME directory for juju 1."""
 
@@ -327,8 +302,27 @@
         except KeyError:
             raise NoProvider('No provider specified.')
 
+    def is_cloud_provider(self):
+        """Return True if the commandline cloud is a provider.
+
+        This is True when LXD or Manual would be specified on the commandline,
+        false otherwse.
+        """
+        return bool(self.provider in ('lxd', 'manual'))
+
     def get_region(self):
+        """Determine the region from a 1.x-style config.
+
+        This requires translating Azure's and Joyent's conventions for
+        specifying region.
+
+        It means that endpoint, rather than region, should be supplied if the
+        cloud (not the provider) is named "lxd" or "manual".
+
+        May return None for MAAS or LXD clouds.
+        """
         provider = self.provider
+        # In 1.x, providers define region differently.  Translate.
         if provider == 'azure':
             if 'tenant-id' not in self._config:
                 return self._config['location'].replace(' ', '').lower()
@@ -336,33 +330,60 @@
         elif provider == 'joyent':
             matcher = re.compile('https://(.*).api.joyentcloud.com')
             return matcher.match(self._config['sdc-url']).group(1)
-        elif provider == 'lxd':
+        elif provider == 'maas':
+            return None
+        # In 2.x, certain providers can be specified on the commandline in
+        # place of a cloud.  The "region" in these cases is the endpoint.
+        elif self.is_cloud_provider():
+            return self._get_config_endpoint()
+        else:
+            # The manual provider is typically used without a region.
+            if provider == 'manual':
+                return self._config.get('region')
+            return self._config['region']
+
+    def _get_config_endpoint(self):
+        if self.provider == 'lxd':
             return self._config.get('region', 'localhost')
-        elif provider == 'manual':
+        elif self.provider == 'manual':
             return self._config['bootstrap-host']
-        elif provider in ('maas', 'manual'):
-            return None
-        else:
-            return self._config['region']
 
     def set_region(self, region):
+        """Assign the region to a 1.x-style config.
+
+        This requires translating Azure's and Joyent's conventions for
+        specifying region.
+
+        It means that endpoint, rather than region, should be updated if the
+        cloud (not the provider) is named "lxd" or "manual".
+
+        Only None is acccepted for MAAS.
+        """
         try:
             provider = self.provider
+            cloud_is_provider = self.is_cloud_provider()
         except NoProvider:
             provider = None
+            cloud_is_provider = False
         if provider == 'azure':
             self._config['location'] = region
         elif provider == 'joyent':
             self._config['sdc-url'] = (
                 'https://{}.api.joyentcloud.com'.format(region))
-        elif provider == 'manual':
-            self._config['bootstrap-host'] = region
+        elif cloud_is_provider:
+            self._set_config_endpoint(region)
         elif provider == 'maas':
             if region is not None:
                 raise ValueError('Only None allowed for maas.')
         else:
             self._config['region'] = region
 
+    def _set_config_endpoint(self, endpoint):
+        if self.provider == 'lxd':
+            self._config['region'] = endpoint
+        elif self.provider == 'manual':
+            self._config['bootstrap-host'] = endpoint
+
     def clone(self, model_name=None):
         config = deepcopy(self._config)
         if model_name is None:
@@ -611,6 +632,17 @@
         """Return the credentials for this model's cloud."""
         return self.get_cloud_credentials_item()[1]
 
+    def is_cloud_provider(self):
+        """Return True if the commandline cloud is a provider.
+
+        Examples: lxd, manual
+        """
+        # if the commandline cloud is "lxd" or "manual", the provider type
+        # should match, and shortcutting get_cloud avoids pointless test
+        # breakage.
+        return bool(self.provider in ('lxd', 'manual') and
+                    self.get_cloud() in ('lxd', 'manual'))
+
 
 class StatusError(Exception):
     """Generic error for Status."""
@@ -673,8 +705,6 @@
     # This is the time limit set by IS for recovery from an agent error.
     a_reasonable_time = timedelta(minutes=5)
 
-    recoverable = False
-
 
 StatusError.ordering = [
     ProvisioningError, MachineError, InstallError, AgentUnresolvedError,
@@ -782,6 +812,10 @@
             status_yaml = yaml.safe_load(text)
         return cls(status_yaml, text)
 
+    @property
+    def model_name(self):
+        return self.status['model']['name']
+
     def get_applications(self):
         return self.status.get('applications', {})
 
@@ -952,46 +986,6 @@
             raise errors[0]
 
 
-class Status1X(Status):
-
-    def get_applications(self):
-        return self.status.get('services', {})
-
-    def condense_status(self, item_value):
-        """Condense the scattered agent-* fields into a status dict."""
-        def shift_field(dest_dict, dest_name, src_dict, src_name):
-            if src_name in src_dict:
-                dest_dict[dest_name] = src_dict[src_name]
-        condensed = {}
-        shift_field(condensed, 'current', item_value, 'agent-state')
-        shift_field(condensed, 'version', item_value, 'agent-version')
-        shift_field(condensed, 'message', item_value, 'agent-state-info')
-        return condensed
-
-    def iter_status(self):
-        SERVICE = 'service-status'
-        AGENT = 'agent-status'
-        for machine_name, machine_value in self.get_machines({}).items():
-            yield StatusItem(StatusItem.JUJU, machine_name,
-                             self.condense_status(machine_value))
-        for app_name, app_value in self.get_applications().items():
-            if SERVICE in app_value:
-                yield StatusItem(
-                    StatusItem.APPLICATION, app_name,
-                    {StatusItem.APPLICATION: app_value[SERVICE]})
-            for unit_name, unit_value in app_value.get('units', {}).items():
-                if StatusItem.WORKLOAD in unit_value:
-                    yield StatusItem(StatusItem.WORKLOAD,
-                                     unit_name, unit_value)
-                if AGENT in unit_value:
-                    yield StatusItem(
-                        StatusItem.JUJU, unit_name,
-                        {StatusItem.JUJU: unit_value[AGENT]})
-                else:
-                    yield StatusItem(StatusItem.JUJU, unit_name,
-                                     self.condense_status(unit_value))
-
-
 def describe_substrate(env):
     if env.provider == 'local':
         return {
@@ -1121,7 +1115,7 @@
 
         # If args is a string, make it a tuple. This makes writing commands
         # with one argument a bit nicer.
-        if isinstance(args, basestring):
+        if isinstance(args, argtype):
             args = (args,)
         # we split the command here so that the caller can control where the -m
         # model flag goes.  Everything in the command string is put before the
@@ -1214,92 +1208,77 @@
         pause(seconds)
 
 
-class Juju1XBackend(Juju2Backend):
-    """Backend for Juju 1.x versions.
-
-    Uses -e to specify models ("environments", uses JUJU_HOME to specify home
-    directory.
-    """
-
-    _model_flag = '-e'
-
-    def shell_environ(self, used_feature_flags, juju_home):
-        """Generate a suitable shell environment.
-
-        For 2.0-alpha1 and earlier set only JUJU_HOME and not JUJU_DATA.
-        """
-        env = super(Juju1XBackend, self).shell_environ(used_feature_flags,
-                                                       juju_home)
-        env['JUJU_HOME'] = juju_home
-        del env['JUJU_DATA']
-        return env
-
-
-def get_client_class(version):
-    if version.startswith('1.16'):
-        raise VersionNotTestedError(version)
-    elif re.match('^1\.22[.-]', version):
-        client_class = EnvJujuClient22
-    elif re.match('^1\.24[.-]', version):
-        client_class = EnvJujuClient24
-    elif re.match('^1\.25[.-]', version):
-        client_class = EnvJujuClient25
-    elif re.match('^1\.26[.-]', version):
-        raise VersionNotTestedError(version)
-    elif re.match('^1\.', version):
-        client_class = EnvJujuClient1X
-    elif re.match('^2\.0-(alpha|beta)', version):
-        raise VersionNotTestedError(version)
-    elif re.match('^2\.0-rc[1-3]', version):
-        client_class = EnvJujuClientRC
-    else:
-        client_class = EnvJujuClient
-    return client_class
-
-
-def client_from_config(config, juju_path, debug=False, soft_deadline=None):
-    """Create a client from an environment's configuration.
-
-    :param config: Name of the environment to use the config from.
-    :param juju_path: Path to juju binary the client should wrap.
-    :param debug=False: The debug flag for the client, False by default.
-    :param soft_deadline: A datetime representing the deadline by which
-        normal operations should complete.  If None, no deadline is
-        enforced.
-    """
-    version = EnvJujuClient.get_version(juju_path)
-    client_class = get_client_class(version)
-    if config is None:
-        env = client_class.config_class('', {})
-    else:
-        env = client_class.config_class.from_config(config)
-    if juju_path is None:
-        full_path = EnvJujuClient.get_full_path()
-    else:
-        full_path = os.path.abspath(juju_path)
-    return client_class(env, version, full_path, debug=debug,
-                        soft_deadline=soft_deadline)
-
-
-class WaitMachineNotPresent:
+class BaseCondition:
+    """Base class for conditions that support client.wait_for."""
+
+    def __init__(self, timeout=300, already_satisfied=False):
+        self.timeout = timeout
+        self.already_satisfied = already_satisfied
+
+
+class ConditionList(BaseCondition):
+    """A list of conditions that support client.wait_for.
+
+    This combines the supplied list of conditions.  It is only satisfied when
+    all conditions are met.  It times out when any member times out.  When
+    asked to raise, it causes the first condition to raise an exception.  An
+    improvement would be to raise the first condition whose timeout has been
+    exceeded.
+    """
+
+    def __init__(self, conditions):
+        if len(conditions) == 0:
+            timeout = 300
+        else:
+            timeout = max(c.timeout for c in conditions)
+        already_satisfied = all(c.already_satisfied for c in conditions)
+        super(ConditionList, self).__init__(timeout, already_satisfied)
+        self._conditions = conditions
+
+    def iter_blocking_state(self, status):
+        for condition in self._conditions:
+            for item, state in condition.iter_blocking_state(status):
+                yield item, state
+
+    def do_raise(self, model_name, status):
+        self._conditions[0].do_raise(model_name, status)
+
+
+class WaitMachineNotPresent(BaseCondition):
     """Condition satisfied when a given machine is not present."""
 
-    def __init__(self, machine):
+    def __init__(self, machine, timeout=300):
+        super(WaitMachineNotPresent, self).__init__(timeout)
         self.machine = machine
 
-    def is_satisfied(self, status):
+    def iter_blocking_state(self, status):
         for machine, info in status.iter_machines():
             if machine == self.machine:
-                return False
-        else:
-            return True
+                yield machine, 'still-present'
 
-    def do_raise(self):
+    def do_raise(self, model_name, status):
         raise Exception("Timed out waiting for machine removal %s" %
                         self.machine)
 
 
-class EnvJujuClient:
+class WaitVersion(BaseCondition):
+
+    def __init__(self, target_version, timeout=300):
+        super(WaitVersion, self).__init__(timeout)
+        self.target_version = target_version
+
+    def iter_blocking_state(self, status):
+        for version, agents in status.get_agent_versions().items():
+            if version == self.target_version:
+                continue
+            for agent in agents:
+                yield agent, version
+
+    def do_raise(self, model_name, status):
+        raise VersionsNotUpdated(model_name, status)
+
+
+class ModelClient:
     """Wraps calls to a juju instance, associated with a single model.
 
     Note: A model is often called an enviroment (Juju 1 legacy).
@@ -1347,6 +1326,9 @@
 
     command_set_all = 'all'
 
+    REGION_ENDPOINT_PROMPT = (
+        r'Enter the API endpoint url for the region \[use cloud api url\]:')
+
     @classmethod
     def preferred_container(cls):
         for container_type in [LXD_MACHINE, LXC_MACHINE]:
@@ -1363,7 +1345,8 @@
         """
         if juju_path is None:
             juju_path = 'juju'
-        return subprocess.check_output((juju_path, '--version')).strip()
+        version = subprocess.check_output((juju_path, '--version')).strip()
+        return version.decode("utf-8")
 
     def check_timeouts(self):
         return self._backend._check_timeouts()
@@ -1417,6 +1400,7 @@
 
     def clone_path_cls(self, juju_path):
         """Clone using the supplied path to determine the class."""
+        from jujupy.version_client import get_client_class
         version = self.get_version(juju_path)
         cls = get_client_class(version)
         if juju_path is None:
@@ -1427,7 +1411,7 @@
 
     def clone(self, env=None, version=None, full_path=None, debug=None,
               cls=None):
-        """Create a clone of this EnvJujuClient.
+        """Create a clone of this ModelClient.
 
         By default, the class, environment, version, full_path, and debug
         settings will match the original, but each can be overridden.
@@ -1561,6 +1545,18 @@
                 pause(30)
                 self.juju('add-machine', ('ssh:' + machine,))
 
+    def make_remove_machine_condition(self, machine):
+        """Return a condition object representing a machine removal.
+
+        The timeout varies depending on the provider.
+        See wait_for.
+        """
+        if self.env.provider == 'azure':
+            timeout = 1200
+        else:
+            timeout = 600
+        return WaitMachineNotPresent(machine, timeout)
+
     @staticmethod
     def get_cloud_region(cloud, region):
         if region is None:
@@ -1607,7 +1603,11 @@
     def add_model(self, env):
         """Add a model to this model's controller and return its client.
 
-        :param env: Class representing the new model/environment."""
+        :param env: Either a class representing the new model/environment
+            or the name of the new model/environment which will then be
+            otherwise identical to the current model/environment."""
+        if not isinstance(env, SimpleEnvironment):
+            env = self.env.clone(env)
         model_client = self.clone(env)
         with model_client._bootstrap_config() as config_file:
             self._add_model(env.environment, config_file)
@@ -1777,7 +1777,7 @@
                 return self.status_class.from_text(
                     self.get_juju_output(
                         self._show_status, '--format', 'yaml',
-                        controller=controller))
+                        controller=controller).decode('utf-8'))
             except subprocess.CalledProcessError:
                 pass
         raise StatusTimeout(
@@ -1834,6 +1834,47 @@
         """Unset the value of the option in the environment."""
         return self.juju('model-config', ('--reset', option,))
 
+    @staticmethod
+    def _format_cloud_region(cloud=None, region=None):
+        """Return the [[cloud/]region] in a tupple."""
+        if cloud and region:
+            return ('{}/{}'.format(cloud, region),)
+        elif region:
+            return (region,)
+        elif cloud:
+            raise ValueError('The cloud must be followed by a region.')
+        else:
+            return ()
+
+    def get_model_defaults(self, model_key, cloud=None, region=None):
+        """Return a dict with information on model-defaults for model-key.
+
+        Giving cloud/region acts as a filter."""
+        cloud_region = self._format_cloud_region(cloud, region)
+        gjo_args = ('--format', 'yaml') + cloud_region + (model_key,)
+        raw_yaml = self.get_juju_output('model-defaults', *gjo_args,
+                                        include_e=False)
+        return yaml.safe_load(raw_yaml)
+
+    def set_model_defaults(self, model_key, value, cloud=None, region=None):
+        """Set a model-defaults entry for model_key to value.
+
+        Giving cloud/region sets the default for that region, otherwise the
+        controller default is set."""
+        cloud_region = self._format_cloud_region(cloud, region)
+        self.juju('model-defaults',
+                  cloud_region + ('{}={}'.format(model_key, value),),
+                  include_e=False)
+
+    def unset_model_defaults(self, model_key, cloud=None, region=None):
+        """Unset a model-defaults entry for model_key.
+
+        Giving cloud/region unsets the default for that region, otherwise the
+        controller default is unset."""
+        cloud_region = self._format_cloud_region(cloud, region)
+        self.juju('model-defaults',
+                  cloud_region + ('--reset', model_key), include_e=False)
+
     def get_agent_metadata_url(self):
         return self.get_env_option(self.agent_metadata_url)
 
@@ -2076,15 +2117,8 @@
             reporter, status_to_subordinate_states, AgentsNotStarted,
             timeout=timeout, start=start)
 
-    def wait_for_version(self, version, timeout=300, start=None):
-        def status_to_version(status):
-            versions = status.get_agent_versions()
-            if versions.keys() == [version]:
-                return None
-            return versions
-        reporter = GroupReporter(sys.stdout, version)
-        self._wait_for_status(reporter, status_to_version, VersionsNotUpdated,
-                              timeout=timeout, start=start)
+    def wait_for_version(self, version, timeout=300):
+        self.wait_for(WaitVersion(version, timeout))
 
     def list_models(self):
         """List the models registered with the current controller."""
@@ -2173,14 +2207,13 @@
         self.juju('list-controllers', (), include_e=False)
 
     def get_controller_endpoint(self):
-        """Return the address of the controller leader."""
+        """Return the host and port of the controller leader."""
         controller = self.env.controller.name
         output = self.get_juju_output(
             'show-controller', controller, include_e=False)
         info = yaml.safe_load(output)
         endpoint = info[controller]['details']['api-endpoints'][0]
-        address, port = split_address_port(endpoint)
-        return address
+        return split_address_port(endpoint)
 
     def get_controller_members(self):
         """Return a list of Machines that are members of the controller.
@@ -2198,7 +2231,7 @@
         # Search for the leader and make it the first in the list.
         # If the endpoint address is not the same as the leader's dns_name,
         # the members are return in the order they were discovered.
-        endpoint = self.get_controller_endpoint()
+        endpoint = self.get_controller_endpoint()[0]
         log.debug('Controller endpoint is at {}'.format(endpoint))
         members.sort(key=lambda m: m.info.get('dns-name') != endpoint)
         return members
@@ -2278,29 +2311,33 @@
         self._wait_for_status(reporter, status_to_workloads, WorkloadsNotReady,
                               timeout=timeout, start=start)
 
-    def wait_for(self, conditions, timeout=300):
+    def wait_for(self, condition, quiet=False):
         """Wait until the supplied conditions are satisfied.
 
         The supplied conditions must be an iterable of objects like
         WaitMachineNotPresent.
         """
-        if len(conditions) == 0:
+        if condition.already_satisfied:
             return self.get_status()
+        reporter = GroupReporter(sys.stdout, 'started')
+        status = None
         try:
-            for status in self.status_until(timeout):
+            for status in self.status_until(condition.timeout):
                 status.raise_highest_error(ignore_recoverable=True)
-                pending = []
-                for condition in conditions:
-                    if not condition.is_satisfied(status):
-                        pending.append(condition)
-                if len(pending) == 0:
-                    return status
-                conditions = pending
+                states = {}
+                for item, state in condition.iter_blocking_state(status):
+                    states.setdefault(state, []).append(item)
+                if len(states) == 0:
+                    return
+                if not quiet:
+                    reporter.update(states)
             else:
                 status.raise_highest_error(ignore_recoverable=False)
         except StatusTimeout:
             pass
-        conditions[0].do_raise()
+        finally:
+            reporter.finish()
+        condition.do_raise(self.model_name, status)
 
     def get_matching_agent_version(self, no_build=False):
         # strip the series and srch from the built version.
@@ -2608,7 +2645,8 @@
             child.logfile = sys.stdout
             child.expect('Select cloud type:')
             child.sendline(cloud['type'])
-            child.expect('(Enter a name for the cloud:)|(Select cloud type:)')
+            child.expect('(Enter a name for your .* cloud:)|'
+                         '(Select cloud type:)')
             if child.match.group(2) is not None:
                 raise TypeNotAccepted('Cloud type not accepted.')
             child.sendline(cloud_name)
@@ -2618,7 +2656,7 @@
             if cloud['type'] == 'manual':
                 child.expect(
                     "(Enter the controller's hostname or IP address:)|"
-                    "(Enter a name for the cloud:)")
+                    "(Enter a name for your .* cloud:)")
                 if child.match.group(2) is not None:
                     raise NameNotAccepted('Cloud name not accepted.')
                 child.sendline(cloud['endpoint'])
@@ -2626,7 +2664,10 @@
                 child.expect('Enter the API endpoint url for the cloud:')
                 child.sendline(cloud['endpoint'])
                 child.expect(
-                    'Select one or more auth types separated by commas:')
+                    "(Select one or more auth types separated by commas:)|"
+                    "(Can't validate endpoint)")
+                if child.match.group(2) is not None:
+                    raise InvalidEndpoint()
                 child.sendline(','.join(cloud['auth-types']))
                 for num, (name, values) in enumerate(cloud['regions'].items()):
                     child.expect(
@@ -2635,9 +2676,12 @@
                     if child.match.group(2) is not None:
                         raise AuthNotAccepted('Auth was not compatible.')
                     child.sendline(name)
-                    child.expect('Enter the API endpoint url for the region:')
+                    child.expect(self.REGION_ENDPOINT_PROMPT)
                     child.sendline(values['endpoint'])
-                    child.expect('Enter another region\? \(Y/n\):')
+                    child.expect("(Enter another region\? \(Y/n\):)|"
+                                 "(Can't validate endpoint)")
+                    if child.match.group(2) is not None:
+                        raise InvalidEndpoint()
                     if num + 1 < len(cloud['regions']):
                         child.sendline('y')
                     else:
@@ -2654,7 +2698,10 @@
                     else:
                         child.sendline('n')
 
-            child.expect(pexpect.EOF)
+            child.expect([pexpect.EOF, "Can't validate endpoint"])
+            if child.match != pexpect.EOF:
+                if child.match.group(0) == "Can't validate endpoint":
+                    raise InvalidEndpoint()
         except pexpect.TIMEOUT:
             raise Exception(
                 'Adding cloud failed: pexpect session timed out')
@@ -2711,526 +2758,12 @@
             return self.juju('sync-tools', ('--local-dir', local_dir),
                              include_e=False)
 
-
-class EnvJujuClientRC(EnvJujuClient):
-
-    def get_bootstrap_args(
-            self, upload_tools, config_filename, bootstrap_series=None,
-            credential=None, auto_upgrade=False, metadata_source=None,
-            to=None, no_gui=False, agent_version=None):
-        """Return the bootstrap arguments for the substrate."""
-        if self.env.joyent:
-            # Only accept kvm packages by requiring >1 cpu core, see lp:1446264
-            constraints = 'mem=2G cpu-cores=1'
-        else:
-            constraints = 'mem=2G'
-        cloud_region = self.get_cloud_region(self.env.get_cloud(),
-                                             self.env.get_region())
-        # Note controller name before cloud_region
-        args = ['--constraints', constraints,
-                self.env.environment,
-                cloud_region,
-                '--config', config_filename,
-                '--default-model', self.env.environment]
-        if upload_tools:
-            if agent_version is not None:
-                raise ValueError(
-                    'agent-version may not be given with upload-tools.')
-            args.insert(0, '--upload-tools')
-        else:
-            if agent_version is None:
-                agent_version = self.get_matching_agent_version()
-            args.extend(['--agent-version', agent_version])
-        if bootstrap_series is not None:
-            args.extend(['--bootstrap-series', bootstrap_series])
-        if credential is not None:
-            args.extend(['--credential', credential])
-        if metadata_source is not None:
-            args.extend(['--metadata-source', metadata_source])
-        if auto_upgrade:
-            args.append('--auto-upgrade')
-        if to is not None:
-            args.extend(['--to', to])
-        if no_gui:
-            args.append('--no-gui')
-        return tuple(args)
-
-
-class EnvJujuClient1X(EnvJujuClientRC):
-    """Base for all 1.x client drivers."""
-
-    default_backend = Juju1XBackend
-
-    config_class = SimpleEnvironment
-
-    status_class = Status1X
-
-    # The environments.yaml options that are replaced by bootstrap options.
-    # For Juju 1.x, no bootstrap options are used.
-    bootstrap_replaces = frozenset()
-
-    destroy_model_command = 'destroy-environment'
-
-    supported_container_types = frozenset([KVM_MACHINE, LXC_MACHINE])
-
-    agent_metadata_url = 'tools-metadata-url'
-
-    _show_status = 'status'
-
-    command_set_destroy_model = 'destroy-environment'
-
-    command_set_all = 'all-changes'
-
-    @classmethod
-    def _get_env(cls, env):
-        if isinstance(env, JujuData):
-            raise IncompatibleConfigClass(
-                'JujuData cannot be used with {}'.format(cls.__name__))
-        return env
-
-    def get_cache_path(self):
-        return get_cache_path(self.env.juju_home, models=False)
-
-    def remove_service(self, service):
-        self.juju('destroy-service', (service,))
-
-    def backup(self):
-        environ = self._shell_environ()
-        # juju-backup does not support the -e flag.
-        environ['JUJU_ENV'] = self.env.environment
-        try:
-            # Mutate os.environ instead of supplying env parameter so Windows
-            # can search env['PATH']
-            with scoped_environ(environ):
-                args = ['juju', 'backup']
-                log.info(' '.join(args))
-                output = subprocess.check_output(args)
-        except subprocess.CalledProcessError as e:
-            log.info(e.output)
-            raise
-        log.info(output)
-        backup_file_pattern = re.compile('(juju-backup-[0-9-]+\.(t|tar.)gz)')
-        match = backup_file_pattern.search(output)
-        if match is None:
-            raise Exception("The backup file was not found in output: %s" %
-                            output)
-        backup_file_name = match.group(1)
-        backup_file_path = os.path.abspath(backup_file_name)
-        log.info("State-Server backup at %s", backup_file_path)
-        return backup_file_path
-
-    def restore_backup(self, backup_file):
-        return self.get_juju_output('restore', '--constraints', 'mem=2G',
-                                    backup_file)
-
-    def restore_backup_async(self, backup_file):
-        return self.juju_async('restore', ('--constraints', 'mem=2G',
-                                           backup_file))
-
-    def enable_ha(self):
-        self.juju('ensure-availability', ('-n', '3'))
-
-    def list_models(self):
-        """List the models registered with the current controller."""
-        log.info('The model is environment {}'.format(self.env.environment))
-
-    def list_clouds(self, format='json'):
-        """List all the available clouds."""
-        return {}
-
-    def show_controller(self, format='json'):
-        """Show controller's status."""
-        return {}
-
-    def get_models(self):
-        """return a models dict with a 'models': [] key-value pair."""
-        return {}
-
-    def list_controllers(self):
-        """List the controllers."""
-        log.info(
-            'The controller is environment {}'.format(self.env.environment))
-
-    @staticmethod
-    def get_controller_member_status(info_dict):
-        return info_dict.get('state-server-member-status')
-
-    def action_fetch(self, id, action=None, timeout="1m"):
-        """Fetches the results of the action with the given id.
-
-        Will wait for up to 1 minute for the action results.
-        The action name here is just used for an more informational error in
-        cases where it's available.
-        Returns the yaml output of the fetched action.
-        """
-        # the command has to be "action fetch" so that the -e <env> args are
-        # placed after "fetch", since that's where action requires them to be.
-        out = self.get_juju_output("action fetch", id, "--wait", timeout)
-        status = yaml.safe_load(out)["status"]
-        if status != "completed":
-            name = ""
-            if action is not None:
-                name = " " + action
-            raise Exception(
-                "timed out waiting for action%s to complete during fetch" %
-                name)
-        return out
-
-    def action_do(self, unit, action, *args):
-        """Performs the given action on the given unit.
-
-        Action params should be given as args in the form foo=bar.
-        Returns the id of the queued action.
-        """
-        args = (unit, action) + args
-
-        # the command has to be "action do" so that the -e <env> args are
-        # placed after "do", since that's where action requires them to be.
-        output = self.get_juju_output("action do", *args)
-        action_id_pattern = re.compile(
-            'Action queued with id: ([a-f0-9\-]{36})')
-        match = action_id_pattern.search(output)
-        if match is None:
-            raise Exception("Action id not found in output: %s" %
-                            output)
-        return match.group(1)
-
-    def run(self, commands, applications):
-        responses = self.get_juju_output(
-            'run', '--format', 'json', '--service', ','.join(applications),
-            *commands)
-        return json.loads(responses)
-
-    def list_space(self):
-        return yaml.safe_load(self.get_juju_output('space list'))
-
-    def add_space(self, space):
-        self.juju('space create', (space),)
-
-    def add_subnet(self, subnet, space):
-        self.juju('subnet add', (subnet, space))
-
-    def add_user_perms(self, username, models=None, permissions='read'):
-        raise JESNotSupported()
-
-    def grant(self, user_name, permission, model=None):
-        raise JESNotSupported()
-
-    def revoke(self, username, models=None, permissions='read'):
-        raise JESNotSupported()
-
-    def set_model_constraints(self, constraints):
-        constraint_strings = self._dict_as_option_strings(constraints)
-        return self.juju('set-constraints', constraint_strings)
-
-    def set_config(self, service, options):
-        option_strings = ['{}={}'.format(*item) for item in options.items()]
-        self.juju('set', (service,) + tuple(option_strings))
-
-    def get_config(self, service):
-        return yaml.safe_load(self.get_juju_output('get', service))
-
-    def get_model_config(self):
-        """Return the value of the environment's configured option."""
-        return yaml.safe_load(self.get_juju_output('get-env'))
-
-    def get_env_option(self, option):
-        """Return the value of the environment's configured option."""
-        return self.get_juju_output('get-env', option)
-
-    def set_env_option(self, option, value):
-        """Set the value of the option in the environment."""
-        option_value = "%s=%s" % (option, value)
-        return self.juju('set-env', (option_value,))
-
-    def unset_env_option(self, option):
-        """Unset the value of the option in the environment."""
-        return self.juju('set-env', ('{}='.format(option),))
-
-    def _cmd_model(self, include_e, controller):
-        if controller:
-            return self.get_controller_model_name()
-        elif self.env is None or not include_e:
-            return None
-        else:
-            return unqualified_model_name(self.model_name)
-
-    def update_user_name(self):
-        return
-
-    def _get_substrate_constraints(self):
-        if self.env.joyent:
-            # Only accept kvm packages by requiring >1 cpu core, see lp:1446264
-            return 'mem=2G cpu-cores=1'
-        else:
-            return 'mem=2G'
-
-    def get_bootstrap_args(self, upload_tools, bootstrap_series=None,
-                           credential=None):
-        """Return the bootstrap arguments for the substrate."""
-        if credential is not None:
-            raise ValueError(
-                '--credential is not supported by this juju version.')
-        constraints = self._get_substrate_constraints()
-        args = ('--constraints', constraints)
-        if upload_tools:
-            args = ('--upload-tools',) + args
-        if bootstrap_series is not None:
-            env_val = self.env.get_option('default-series')
-            if bootstrap_series != env_val:
-                raise BootstrapMismatch(
-                    'bootstrap-series', bootstrap_series, 'default-series',
-                    env_val)
-        return args
-
-    def bootstrap(self, upload_tools=False, bootstrap_series=None):
-        """Bootstrap a controller."""
-        self._check_bootstrap()
-        args = self.get_bootstrap_args(upload_tools, bootstrap_series)
-        self.juju('bootstrap', args)
-
-    @contextmanager
-    def bootstrap_async(self, upload_tools=False):
-        self._check_bootstrap()
-        args = self.get_bootstrap_args(upload_tools)
-        with self.juju_async('bootstrap', args):
-            yield
-            log.info('Waiting for bootstrap of {}.'.format(
-                self.env.environment))
-
-    def get_jes_command(self):
-        raise JESNotSupported()
-
-    def enable_jes(self):
-        raise JESNotSupported()
-
-    def upgrade_juju(self, force_version=True):
-        args = ()
-        if force_version:
-            version = self.get_matching_agent_version(no_build=True)
-            args += ('--version', version)
-        if self.env.local:
-            args += ('--upload-tools',)
-        self._upgrade_juju(args)
-
-    def make_model_config(self):
-        config_dict = make_safe_config(self)
-        # Strip unneeded variables.
-        return config_dict
-
-    def _add_model(self, model_name, config_file):
-        seen_cmd = self.get_jes_command()
-        if seen_cmd == SYSTEM:
-            controller_option = ('-s', self.env.environment)
-        else:
-            controller_option = ('-c', self.env.environment)
-        self.juju(_jes_cmds[seen_cmd]['create'], controller_option + (
-            model_name, '--config', config_file), include_e=False)
-
-    def destroy_model(self):
-        """With JES enabled, destroy-environment destroys the model."""
-        return self.destroy_environment(force=False)
-
-    def kill_controller(self, check=False):
-        """Destroy the environment, with force. Hard kill option.
-
-        :return: Subprocess's exit code."""
-        return self.juju(
-            'destroy-environment', (self.env.environment, '--force', '-y'),
-            check=check, include_e=False, timeout=get_teardown_timeout(self))
-
-    def destroy_controller(self, all_models=False):
-        """Destroy the environment, with force. Soft kill option.
-
-        :param all_models: Ignored.
-        :raises: subprocess.CalledProcessError if the operation fails."""
-        return self.juju(
-            'destroy-environment', (self.env.environment, '-y'),
-            include_e=False, timeout=get_teardown_timeout(self))
-
-    def destroy_environment(self, force=True, delete_jenv=False):
-        if force:
-            force_arg = ('--force',)
-        else:
-            force_arg = ()
-        exit_status = self.juju(
-            'destroy-environment',
-            (self.env.environment,) + force_arg + ('-y',),
-            check=False, include_e=False,
-            timeout=get_teardown_timeout(self))
-        if delete_jenv:
-            jenv_path = get_jenv_path(self.env.juju_home, self.env.environment)
-            ensure_deleted(jenv_path)
-        return exit_status
-
-    def _get_models(self):
-        """return a list of model dicts."""
-        try:
-            return yaml.safe_load(self.get_juju_output(
-                'environments', '-s', self.env.environment, '--format', 'yaml',
-                include_e=False))
-        except subprocess.CalledProcessError:
-            # This *private* method attempts to use a 1.25 JES feature.
-            # The JES design is dead. The private method is not used to
-            # directly test juju cli; the failure is not a contract violation.
-            log.info('Call to JES juju environments failed, falling back.')
-            return []
-
-    def get_model_uuid(self):
-        raise JESNotSupported()
-
-    def deploy_bundle(self, bundle, timeout=_DEFAULT_BUNDLE_TIMEOUT):
-        """Deploy bundle using deployer for Juju 1.X version."""
-        self.deployer(bundle, timeout=timeout)
-
-    def deployer(self, bundle_template, name=None, deploy_delay=10,
-                 timeout=3600):
-        """Deploy a bundle using deployer."""
-        bundle = self.format_bundle(bundle_template)
-        args = (
-            '--debug',
-            '--deploy-delay', str(deploy_delay),
-            '--timeout', str(timeout),
-            '--config', bundle,
-        )
-        if name:
-            args += (name,)
-        self.juju('deployer', args)
-
-    def deploy(self, charm, repository=None, to=None, series=None,
-               service=None, force=False, storage=None, constraints=None):
-        args = [charm]
-        if repository is not None:
-            args.extend(['--repository', repository])
-        if to is not None:
-            args.extend(['--to', to])
-        if service is not None:
-            args.extend([service])
-        if storage is not None:
-            args.extend(['--storage', storage])
-        if constraints is not None:
-            args.extend(['--constraints', constraints])
-        return self.juju('deploy', tuple(args))
-
-    def upgrade_charm(self, service, charm_path=None):
-        args = (service,)
-        if charm_path is not None:
-            repository = os.path.dirname(os.path.dirname(charm_path))
-            args = args + ('--repository', repository)
-        self.juju('upgrade-charm', args)
-
-    def get_controller_client(self):
-        """Return a client for the controller model.  May return self."""
-        return self
-
-    def get_controller_model_name(self):
-        """Return the name of the 'controller' model.
-
-        Return the name of the 1.x environment."""
-        return self.env.environment
-
-    def get_controller_endpoint(self):
-        """Return the address of the state-server leader."""
-        endpoint = self.get_juju_output('api-endpoints')
-        address, port = split_address_port(endpoint)
-        return address
-
-    def upgrade_mongo(self):
-        raise UpgradeMongoNotSupported()
-
-    def create_cloned_environment(
-            self, cloned_juju_home, controller_name, user_name=None):
-        """Create a cloned environment.
-
-        `user_name` is unused in this version of juju.
-        """
-        user_client = self.clone(env=self.env.clone())
-        user_client.env.juju_home = cloned_juju_home
-        # New user names the controller.
-        user_client.env.controller = Controller(controller_name)
-        return user_client
-
-    def add_storage(self, unit, storage_type, amount="1"):
-        """Add storage instances to service.
-
-        Only type 'disk' is able to add instances.
-        """
-        self.juju('storage add', (unit, storage_type + "=" + amount))
-
-    def list_storage(self):
-        """Return the storage list."""
-        return self.get_juju_output('storage list', '--format', 'json')
-
-    def list_storage_pool(self):
-        """Return the list of storage pool."""
-        return self.get_juju_output('storage pool list', '--format', 'json')
-
-    def create_storage_pool(self, name, provider, size):
-        """Create storage pool."""
-        self.juju('storage pool create',
-                  (name, provider,
-                   'size={}'.format(size)))
-
-    def ssh_keys(self, full=False):
-        """Give the ssh keys registered for the current model."""
-        args = []
-        if full:
-            args.append('--full')
-        return self.get_juju_output('authorized-keys list', *args)
-
-    def add_ssh_key(self, *keys):
-        """Add one or more ssh keys to the current model."""
-        return self.get_juju_output('authorized-keys add', *keys,
-                                    merge_stderr=True)
-
-    def remove_ssh_key(self, *keys):
-        """Remove one or more ssh keys from the current model."""
-        return self.get_juju_output('authorized-keys delete', *keys,
-                                    merge_stderr=True)
-
-    def import_ssh_key(self, *keys):
-        """Import ssh keys from one or more identities to the current model."""
-        return self.get_juju_output('authorized-keys import', *keys,
-                                    merge_stderr=True)
-
-    def list_disabled_commands(self):
-        """List all the commands disabled on the model."""
-        raw = self.get_juju_output('block list', '--format', 'yaml')
-        return yaml.safe_load(raw)
-
-    def disable_command(self, command_set, message=''):
-        """Disable a command-set."""
-        return self.juju('block {}'.format(command_set), (message, ))
-
-    def enable_command(self, args):
-        """Enable a command-set."""
-        return self.juju('unblock', args)
-
-
-class EnvJujuClient22(EnvJujuClient1X):
-
-    used_feature_flags = frozenset(['actions'])
-
-    def __init__(self, *args, **kwargs):
-        super(EnvJujuClient22, self).__init__(*args, **kwargs)
-        self.feature_flags.add('actions')
-
-
-class EnvJujuClient25(EnvJujuClient1X):
-    """Drives Juju 2.5-series clients."""
-
-    used_feature_flags = frozenset()
-
-    def disable_jes(self):
-        self.feature_flags.discard('jes')
-
-
-class EnvJujuClient24(EnvJujuClient25):
-    """Similar to EnvJujuClient25."""
-
-    def add_ssh_machines(self, machines):
-        for machine in machines:
-            self.juju('add-machine', ('ssh:' + machine,))
+    def switch(self, model=None, controller=None):
+        """Switch between models."""
+        args = [x for x in [controller, model] if x]
+        if not args:
+            raise ValueError('No target to switch to has been given.')
+        self.juju('switch', (':'.join(args),), include_e=False)
 
 
 def get_local_root(juju_home, env):

=== renamed file 'fakejuju.py' => 'jujupy/fake.py'
--- fakejuju.py	2016-11-22 19:17:42 +0000
+++ jujupy/fake.py	2017-01-20 21:41:35 +0000
@@ -13,7 +13,7 @@
 import yaml
 
 from jujupy import (
-    EnvJujuClient,
+    ModelClient,
     JESNotSupported,
     JujuData,
     SoftDeadlineExceeded,
@@ -21,14 +21,11 @@
 
 __metaclass__ = type
 
-
-def check_juju_output(func):
-    def wrapper(*args, **kwargs):
-        result = func(*args, **kwargs)
-        if 'service' in result:
-            raise AssertionError('Result contained service')
-        return result
-    return wrapper
+# Python 2 and 3 compatibility
+try:
+    argtype = basestring
+except NameError:
+    argtype = str
 
 
 class ControllerOperation(Exception):
@@ -106,7 +103,7 @@
         self.state = 'registered'
 
     def destroy(self, kill=False):
-        for model in self.models.values():
+        for model in list(self.models.values()):
             model.destroy_model()
         self.models.clear()
         if kill:
@@ -146,7 +143,7 @@
 
     def add_machine(self, host_name=None, machine_id=None):
         if machine_id is None:
-            machine_id = str(self.machine_id_iter.next())
+            machine_id = str(next(self.machine_id_iter))
         self.machines.add(machine_id)
         if host_name is None:
             host_name = '{}.example.com'.format(machine_id)
@@ -278,7 +275,11 @@
                 'relations': self.relations.get(service, {}),
                 'exposed': service in self.exposed,
                 }
-        return {'machines': machines, 'applications': services}
+        return {
+            'machines': machines,
+            'applications': services,
+            'model': {'name': self.name},
+            }
 
     def add_ssh_key(self, keys_to_add):
         errors = []
@@ -382,23 +383,31 @@
         self.values = {}
         self.lines = []
 
-    def expect(self, regex):
+    def expect(self, pattern):
+        if type(pattern) is not list:
+            pattern = [pattern]
         try:
             prompt = self.prompts.next()
         except StopIteration:
-            if regex is not pexpect.EOF:
+            if pexpect.EOF not in pattern:
                 raise
             self.close()
             return
-        if regex is pexpect.EOF:
-            raise ValueError('Expected EOF. got "{}"'.format(prompt))
+        for regex in pattern:
+            if regex is pexpect.EOF:
+                continue
+            regex_match = re.search(regex, prompt)
+            if regex_match is not None:
+                self.match = regex_match
+                break
+        else:
+            if pexpect.EOF in pattern:
+                raise ValueError('Expected EOF. got "{}"'.format(prompt))
+            else:
+                raise ValueError(
+                    'Regular expression did not match prompt.  Regex: "{}",'
+                    ' prompt "{}"'.format(pattern, prompt))
         super(PromptingExpectChild, self).expect(regex)
-        regex_match = re.search(regex, prompt)
-        if regex_match is None:
-            raise ValueError(
-                'Regular expression did not match prompt.  Regex: "{}",'
-                ' prompt "{}"'.format(regex, prompt))
-        self.match = regex_match
 
     def sendline(self, line=''):
         full_match = self.match.group(0)
@@ -428,7 +437,13 @@
 
 class AddCloud(PromptingExpectChild):
 
-    NAME = 'Enter a name for the cloud:'
+    @property
+    def provider(self):
+        return self.values[self.TYPE]
+
+    @property
+    def name_prompt(self):
+        return 'Enter a name for your {} cloud:'.format(self.provider)
 
     REGION_NAME = 'Enter region name:'
 
@@ -440,12 +455,28 @@
 
     CLOUD_ENDPOINT = 'Enter the API endpoint url for the cloud:'
 
-    REGION_ENDPOINT = 'Enter the API endpoint url for the region:'
+    REGION_ENDPOINT = (
+        'Enter the API endpoint url for the region [use cloud api url]:')
 
     HOST = "Enter the controller's hostname or IP address:"
 
     ANOTHER_REGION = 'Enter another region? (Y/n):'
 
+    def cant_validate(self, endpoint):
+        if self.provider in ('openstack', 'maas'):
+            if self.provider == 'openstack':
+                server_type = 'Openstack'
+                reprompt = self.CLOUD_ENDPOINT
+            else:
+                server_type = 'MAAS'
+                reprompt = self.API_ENDPOINT
+            msg = 'No {} server running at {}'.format(server_type, endpoint)
+        elif self.provider == 'manual':
+            msg = 'ssh: Could not resolve hostname {}'.format(endpoint)
+            reprompt = self.HOST
+        return "Can't validate endpoint: {}\n{}".format(
+            msg, reprompt)
+
     def __init__(self, backend, juju_home, extra_env):
         super(AddCloud, self).__init__(
             backend, juju_home, extra_env, self.iter_prompts())
@@ -456,15 +487,24 @@
             if self.values[self.TYPE] != 'bogus':
                 break
         while True:
-            yield self.NAME
-            if '/' not in self.values[self.NAME]:
+            yield self.name_prompt
+            if '/' not in self.values[self.name_prompt]:
                 break
         if self.values[self.TYPE] == 'maas':
             yield self.API_ENDPOINT
+            endpoint = self.values[self.API_ENDPOINT]
+            while len(endpoint) > 1000:
+                yield self.cant_validate(endpoint)
         elif self.values[self.TYPE] == 'manual':
             yield self.HOST
+            endpoint = self.values[self.HOST]
+            while len(endpoint) > 1000:
+                yield self.cant_validate(endpoint)
         elif self.values[self.TYPE] == 'openstack':
             yield self.CLOUD_ENDPOINT
+            endpoint = self.values[self.CLOUD_ENDPOINT]
+            while len(endpoint) > 1000:
+                yield self.cant_validate(endpoint)
             while True:
                 yield self.AUTH
                 if 'invalid' not in self.values[self.AUTH]:
@@ -472,6 +512,9 @@
             while self.values.get(self.ANOTHER_REGION) != 'n':
                 yield self.REGION_NAME
                 yield self.REGION_ENDPOINT
+                endpoint = self.values[self.REGION_ENDPOINT]
+                if len(endpoint) > 1000:
+                    yield self.cant_validate(endpoint)
                 yield self.ANOTHER_REGION
         if self.values['Select cloud type:'] == 'vsphere':
             yield self.CLOUD_ENDPOINT
@@ -510,8 +553,12 @@
                 'endpoint': self.values[self.CLOUD_ENDPOINT],
                 'regions': regions,
                 })
-        self.backend.clouds[self.values[self.NAME]] = cloud
-        self.backend.clouds[self.values[self.NAME]] = cloud
+        self.backend.clouds[self.values[self.name_prompt]] = cloud
+
+
+class AddCloud2_1(AddCloud):
+
+    REGION_ENDPOINT = 'Enter the API endpoint url for the region:'
 
 
 class FakeBackend:
@@ -638,6 +685,7 @@
         parser = ArgumentParser()
         parser.add_argument('host_placement', nargs='*')
         parser.add_argument('-n', type=int, dest='count', default='1')
+        parser.add_argument('--series')
         parsed = parser.parse_args(args)
         if len(parsed.host_placement) == 1:
             split = parsed.host_placement[0].split(':')
@@ -728,13 +776,14 @@
         if model is not None:
             full_args.extend(['-m', model])
         full_args.extend(args)
-        self.log.log(level, ' '.join(full_args))
+        self.log.log(level, u' '.join(full_args))
 
     def juju(self, command, args, used_feature_flags,
              juju_home, model=None, check=True, timeout=None, extra_env=None):
         if 'service' in command:
             raise Exception('Command names must not contain "service".')
-        if isinstance(args, basestring):
+
+        if isinstance(args, argtype):
             args = (args,)
         self._log_command(command, args, model)
         if model is not None:
@@ -863,7 +912,6 @@
         self.juju(command, args, used_feature_flags,
                   juju_home, model, timeout=timeout)
 
-    @check_juju_output
     def get_juju_output(self, command, args, used_feature_flags, juju_home,
                         model=None, timeout=None, user_name=None,
                         merge_stderr=False):
@@ -875,8 +923,8 @@
                 if ':' in model:
                     model = model.split(':')[1]
                 model_state = self.controller_state.models[model]
-            from deploy_stack import GET_TOKEN_SCRIPT
-            if (command, args) == ('ssh', ('dummy-sink/0', GET_TOKEN_SCRIPT)):
+            sink_cat = ('dummy-sink/0', 'cat', '/var/run/dummy-sink/token')
+            if (command, args) == ('ssh', sink_cat):
                 return model_state.token
             if (command, args) == ('ssh', ('0', 'lsb_release', '-c')):
                 return 'Codename:\t{}\n'.format(
@@ -906,7 +954,7 @@
                 status_dict = model_state.get_status_dict()
                 # Parsing JSON is much faster than parsing YAML, and JSON is a
                 # subset of YAML, so emit JSON.
-                return json.dumps(status_dict)
+                return json.dumps(status_dict).encode('utf-8')
             if command == 'create-backup':
                 self.controller_state.require_controller('backup', model)
                 return 'juju-backup-0.tar.gz'
@@ -951,7 +999,18 @@
     return b64encode(sha512(username).digest())
 
 
-class FakeBackend2B7(FakeBackend):
+class FakeBackend2_1(FakeBackend):
+    """Backend for 2.1 and earlier."""
+    def expect(self, command, args, used_feature_flags, juju_home, model=None,
+               timeout=None, extra_env=None):
+        if command == 'add-cloud':
+            return AddCloud2_1(self, juju_home, extra_env)
+        return super(FakeBackend2_1, self).expect(
+            command, args, used_feature_flags, juju_home, model, timeout,
+            extra_env)
+
+
+class FakeBackend2B7(FakeBackend2_1):
 
     def juju(self, command, args, used_feature_flags,
              juju_home, model=None, check=True, timeout=None, extra_env=None):
@@ -973,17 +1032,19 @@
 
 
 def fake_juju_client(env=None, full_path=None, debug=False, version='2.0.0',
-                     _backend=None, cls=EnvJujuClient):
+                     _backend=None, cls=ModelClient, juju_home=None):
+    if juju_home is None:
+        if env is None or env.juju_home is None:
+            juju_home = 'foo'
+        else:
+            juju_home = env.juju_home
     if env is None:
         env = JujuData('name', {
             'type': 'foo',
             'default-series': 'angsty',
             'region': 'bar',
-            }, juju_home='foo')
+            }, juju_home=juju_home)
         env.credentials = {'credentials': {'foo': {'creds': {}}}}
-    juju_home = env.juju_home
-    if juju_home is None:
-        juju_home = 'foo'
     if _backend is None:
         backend_state = FakeControllerState()
         _backend = FakeBackend(
@@ -1011,7 +1072,7 @@
     return client
 
 
-class FakeJujuClientOptionalJES(EnvJujuClient):
+class FakeJujuClientOptionalJES(ModelClient):
 
     def get_controller_model_name(self):
         return self._backend.controller_state.controller_model.name

=== added directory 'jujupy/tests'
=== added file 'jujupy/tests/__init__.py'
=== renamed file 'tests/test_jujupy.py' => 'jujupy/tests/test_client.py'
--- tests/test_jujupy.py	2016-12-06 10:09:46 +0000
+++ jujupy/tests/test_client.py	2017-01-20 20:03:12 +0000
@@ -23,32 +23,26 @@
     )
 import yaml
 
-from fakejuju import (
-    fake_juju_client,
-    get_user_register_command_info,
-    get_user_register_token,
-    )
 from jujuconfig import (
     get_environments_path,
     get_jenv_path,
     NoSuchEnvironment,
     )
 from jujupy import (
+    FakeControllerState,
+    fake_juju_client,
+    get_client_class,
+    )
+from jujupy.client import (
     AuthNotAccepted,
     AgentError,
     AgentUnresolvedError,
     AppError,
-    BootstrapMismatch,
+    BaseCondition,
     CannotConnectEnv,
-    client_from_config,
+    ConditionList,
     Controller,
     describe_substrate,
-    EnvJujuClient,
-    EnvJujuClient1X,
-    EnvJujuClient22,
-    EnvJujuClient24,
-    EnvJujuClient25,
-    EnvJujuClientRC,
     ErroredUnit,
     GroupReporter,
     get_cache_path,
@@ -57,11 +51,9 @@
     get_timeout_path,
     get_timeout_prefix,
     HookFailedError,
-    IncompatibleConfigClass,
     InstallError,
+    InvalidEndpoint,
     jes_home_path,
-    JESNotSupported,
-    Juju1XBackend,
     Juju2Backend,
     JujuData,
     JUJU_DEV_FEATURE_FLAGS,
@@ -69,6 +61,7 @@
     Machine,
     MachineError,
     make_safe_config,
+    ModelClient,
     NameNotAccepted,
     NoProvider,
     parse_new_state_server_from_error,
@@ -76,22 +69,26 @@
     SimpleEnvironment,
     SoftDeadlineExceeded,
     Status,
-    Status1X,
     StatusError,
     StatusItem,
     StatusNotMet,
     StatusTimeout,
     SYSTEM,
     temp_bootstrap_env,
-    _temp_env as temp_env,
     temp_yaml_file,
     TypeNotAccepted,
     uniquify_local,
     UnitError,
-    UpgradeMongoNotSupported,
-    VersionNotTestedError,
     WaitMachineNotPresent,
     )
+from jujupy.fake import (
+    get_user_register_command_info,
+    get_user_register_token,
+)
+from jujupy.version_client import (
+    EnvJujuClient1X,
+    EnvJujuClient24,
+    )
 from tests import (
     assert_juju_call,
     client_past_deadline,
@@ -122,10 +119,30 @@
 
     def setUp(self):
         super(ClientTest, self).setUp()
-        patcher = patch('jujupy.pause')
+        patcher = patch('jujupy.client.pause')
         self.addCleanup(patcher.stop)
         self.pause_mock = patcher.start()
 
+    def fake_juju_client(self):
+        """Provide a faked juju client for the version under test.
+
+        Requires client_version, client_class, fake_backend_class to be
+        defined.
+        """
+        backend_state = FakeControllerState()
+        backend = self.fake_backend_class(backend_state,
+                                          version=self.client_version)
+        return fake_juju_client(version=self.client_version,
+                                cls=self.client_class, _backend=backend)
+
+    def check_basics(self):
+        """Basic tests for a recent ClientTest.
+
+        These are opt-in because not every ClientTest declares client_class
+        and client_version.
+        """
+        self.assertIs(self.client_class, get_client_class(self.client_version))
+
 
 class TestTempYamlFile(TestCase):
 
@@ -278,279 +295,69 @@
                     backend.get_juju_output('cmd', ('args',), [], 'home')
 
 
-class TestJuju1XBackend(TestCase):
-
-    def test_full_args_model(self):
-        backend = Juju1XBackend('/bin/path/juju', '1.25', set(), False, None)
-        full = backend.full_args('help', ('commands',), 'test', None)
-        self.assertEqual(('juju', '--show-log', 'help', '-e', 'test',
-                          'commands'), full)
-
-
-class TestEnvJujuClient25(ClientTest):
-
-    client_class = EnvJujuClient25
-
-    def test_enable_jes(self):
-        client = self.client_class(
-            SimpleEnvironment('baz', {}),
-            '1.25-foobar', 'path')
-        with self.assertRaises(JESNotSupported):
-            client.enable_jes()
-
-    def test_disable_jes(self):
-        client = self.client_class(
-            SimpleEnvironment('baz', {}),
-            '1.25-foobar', 'path')
-        client.feature_flags.add('jes')
-        client.disable_jes()
-        self.assertNotIn('jes', client.feature_flags)
-
-    def test_clone_unchanged(self):
-        client1 = self.client_class(
-            SimpleEnvironment('foo'), '1.27', 'full/path', debug=True)
-        client2 = client1.clone()
-        self.assertIsNot(client1, client2)
-        self.assertIs(type(client1), type(client2))
-        self.assertIs(client1.env, client2.env)
-        self.assertEqual(client1.version, client2.version)
-        self.assertEqual(client1.full_path, client2.full_path)
-        self.assertIs(client1.debug, client2.debug)
-        self.assertEqual(client1._backend, client2._backend)
-
-    def test_clone_changed(self):
-        client1 = self.client_class(
-            SimpleEnvironment('foo'), '1.27', 'full/path', debug=True)
-        env2 = SimpleEnvironment('bar')
-        client2 = client1.clone(env2, '1.28', 'other/path', debug=False,
-                                cls=EnvJujuClient1X)
-        self.assertIs(EnvJujuClient1X, type(client2))
-        self.assertIs(env2, client2.env)
-        self.assertEqual('1.28', client2.version)
-        self.assertEqual('other/path', client2.full_path)
-        self.assertIs(False, client2.debug)
-
-    def test_clone_defaults(self):
-        client1 = self.client_class(
-            SimpleEnvironment('foo'), '1.27', 'full/path', debug=True)
-        client2 = client1.clone()
-        self.assertIsNot(client1, client2)
-        self.assertIs(self.client_class, type(client2))
-        self.assertEqual(set(), client2.feature_flags)
-
-
-class TestEnvJujuClient22(ClientTest):
-
-    client_class = EnvJujuClient22
-
-    def test__shell_environ(self):
-        client = self.client_class(
-            SimpleEnvironment('baz', {'type': 'ec2'}), '1.22-foobar', 'path')
-        env = client._shell_environ()
-        self.assertEqual(env.get(JUJU_DEV_FEATURE_FLAGS), 'actions')
-
-    def test__shell_environ_juju_home(self):
-        client = self.client_class(
-            SimpleEnvironment('baz', {'type': 'ec2'}), '1.22-foobar', 'path',
-            'asdf')
-        env = client._shell_environ()
-        self.assertEqual(env['JUJU_HOME'], 'asdf')
-
-
-class TestEnvJujuClient24(ClientTest):
-
-    client_class = EnvJujuClient24
-
-    def test_no_jes(self):
-        client = self.client_class(
-            SimpleEnvironment('baz', {}),
-            '1.25-foobar', 'path')
-        with self.assertRaises(JESNotSupported):
-            client.enable_jes()
-        client._use_jes = True
-        env = client._shell_environ()
-        self.assertNotIn('jes', env.get(JUJU_DEV_FEATURE_FLAGS, '').split(","))
-
-    def test_add_ssh_machines(self):
-        client = self.client_class(SimpleEnvironment('foo', {}), None, 'juju')
-        with patch('subprocess.check_call', autospec=True) as cc_mock:
-            client.add_ssh_machines(['m-foo', 'm-bar', 'm-baz'])
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 0)
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-bar'), 1)
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-baz'), 2)
-        self.assertEqual(cc_mock.call_count, 3)
-
-    def test_add_ssh_machines_no_retry(self):
-        client = self.client_class(SimpleEnvironment('foo', {}), None, 'juju')
-        with patch('subprocess.check_call', autospec=True,
-                   side_effect=[subprocess.CalledProcessError(None, None),
-                                None, None, None]) as cc_mock:
-            with self.assertRaises(subprocess.CalledProcessError):
-                client.add_ssh_machines(['m-foo', 'm-bar', 'm-baz'])
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'))
-
-
-class TestClientFromConfig(ClientTest):
-
-    @contextmanager
-    def assertRaisesVersionNotTested(self, version):
-        with self.assertRaisesRegexp(
-                VersionNotTestedError, 'juju ' + version):
-            yield
-
-    @patch.object(JujuData, 'from_config', return_value=JujuData('', {}))
-    @patch.object(SimpleEnvironment, 'from_config',
-                  return_value=SimpleEnvironment('', {}))
-    @patch.object(EnvJujuClient, 'get_full_path', return_value='fake-path')
-    def test_from_config(self, gfp_mock, se_fc_mock, jd_fc_mock):
-        def juju_cmd_iterator():
-            yield '1.17'
-            yield '1.16'
-            yield '1.16.1'
-            yield '1.15'
-            yield '1.22.1'
-            yield '1.24-alpha1'
-            yield '1.24.7'
-            yield '1.25.1'
-            yield '1.26.1'
-            yield '1.27.1'
-            yield '2.0-alpha1'
-            yield '2.0-alpha2'
-            yield '2.0-alpha3'
-            yield '2.0-beta1'
-            yield '2.0-beta2'
-            yield '2.0-beta3'
-            yield '2.0-beta4'
-            yield '2.0-beta5'
-            yield '2.0-beta6'
-            yield '2.0-beta7'
-            yield '2.0-beta8'
-            yield '2.0-beta9'
-            yield '2.0-beta10'
-            yield '2.0-beta11'
-            yield '2.0-beta12'
-            yield '2.0-beta13'
-            yield '2.0-beta14'
-            yield '2.0-beta15'
-            yield '2.0-rc1'
-            yield '2.0-rc2'
-            yield '2.0-rc3'
-            yield '2.0-delta1'
-
-        context = patch.object(
-            EnvJujuClient, 'get_version',
-            side_effect=juju_cmd_iterator().send)
-        with context:
-            self.assertIs(EnvJujuClient1X,
-                          type(client_from_config('foo', None)))
-
-            def test_fc(version, cls):
-                if cls is not None:
-                    client = client_from_config('foo', None)
-                    if isinstance(client, EnvJujuClient1X):
-                        self.assertEqual(se_fc_mock.return_value, client.env)
-                    else:
-                        self.assertEqual(jd_fc_mock.return_value, client.env)
-                    self.assertIs(cls, type(client))
-                    self.assertEqual(version, client.version)
-                else:
-                    with self.assertRaisesVersionNotTested(version):
-                        client_from_config('foo', None)
-
-            test_fc('1.16', None)
-            test_fc('1.16.1', None)
-            test_fc('1.15', EnvJujuClient1X)
-            test_fc('1.22.1', EnvJujuClient22)
-            test_fc('1.24-alpha1', EnvJujuClient24)
-            test_fc('1.24.7', EnvJujuClient24)
-            test_fc('1.25.1', EnvJujuClient25)
-            test_fc('1.26.1', None)
-            test_fc('1.27.1', EnvJujuClient1X)
-            test_fc('2.0-alpha1', None)
-            test_fc('2.0-alpha2', None)
-            test_fc('2.0-alpha3', None)
-            test_fc('2.0-beta1', None)
-            test_fc('2.0-beta2', None)
-            test_fc('2.0-beta3', None)
-            test_fc('2.0-beta4', None)
-            test_fc('2.0-beta5', None)
-            test_fc('2.0-beta6', None)
-            test_fc('2.0-beta7', None)
-            test_fc('2.0-beta8', None)
-            test_fc('2.0-beta9', None)
-            test_fc('2.0-beta10', None)
-            test_fc('2.0-beta11', None)
-            test_fc('2.0-beta12', None)
-            test_fc('2.0-beta13', None)
-            test_fc('2.0-beta14', None)
-            test_fc('2.0-beta15', None)
-            test_fc('2.0-rc1', EnvJujuClientRC)
-            test_fc('2.0-rc2', EnvJujuClientRC)
-            test_fc('2.0-rc3', EnvJujuClientRC)
-            test_fc('2.0-delta1', EnvJujuClient)
-            with self.assertRaises(StopIteration):
-                client_from_config('foo', None)
-
-    def test_client_from_config_path(self):
-        with patch('subprocess.check_output', return_value=' 4.3') as vsn:
-            with patch.object(JujuData, 'from_config'):
-                client = client_from_config('foo', 'foo/bar/qux')
-        vsn.assert_called_once_with(('foo/bar/qux', '--version'))
-        self.assertNotEqual(client.full_path, 'foo/bar/qux')
-        self.assertEqual(client.full_path, os.path.abspath('foo/bar/qux'))
-
-    def test_client_from_config_keep_home(self):
-        env = JujuData({}, juju_home='/foo/bar')
-        with patch('subprocess.check_output', return_value='2.0.0'):
-            with patch.object(JujuData, 'from_config',
-                              side_effect=lambda x: JujuData(x, {})):
-                client_from_config('foo', 'foo/bar/qux')
-        self.assertEqual('/foo/bar', env.juju_home)
-
-    def test_client_from_config_deadline(self):
-        deadline = datetime(2012, 11, 10, 9, 8, 7)
-        with patch('subprocess.check_output', return_value='2.0.0'):
-            with patch.object(JujuData, 'from_config',
-                              side_effect=lambda x: JujuData(x, {})):
-                client = client_from_config(
-                    'foo', 'foo/bar/qux', soft_deadline=deadline)
-        self.assertEqual(client._backend.soft_deadline, deadline)
+class TestBaseCondition(ClientTest):
+
+    def test_timeout(self):
+        condition = BaseCondition()
+        self.assertEqual(300, condition.timeout)
+        condition = BaseCondition(600)
+        self.assertEqual(600, condition.timeout)
+
+
+class TestConditionList(ClientTest):
+
+    def test_uses_max_timeout(self):
+        conditions = ConditionList([Mock(timeout=300), Mock(timeout=400)])
+        self.assertEqual(400, conditions.timeout)
+
+    def test_empty_timeout(self):
+        conditions = ConditionList([])
+        self.assertEqual(300, conditions.timeout)
+
+    def test_iter_blocking_state(self):
+        mock_ab = Mock()
+        mock_ab.iter_blocking_state.return_value = [('a', 'b')]
+        mock_cd = Mock()
+        mock_cd.iter_blocking_state.return_value = [('c', 'd')]
+        conditions = ConditionList([mock_ab, mock_cd])
+        self.assertEqual([('a', 'b'), ('c', 'd')],
+                         list(conditions.iter_blocking_state(None)))
 
 
 class TestWaitMachineNotPresent(ClientTest):
 
-    def test_is_satisfied(self):
+    def test_iter_blocking_state(self):
         not_present = WaitMachineNotPresent('0')
         client = fake_juju_client()
         client.bootstrap()
-        self.assertIs(not_present.is_satisfied(client.get_status()), True)
+        self.assertItemsEqual(
+            [], not_present.iter_blocking_state(client.get_status()))
         client.juju('add-machine', ())
-        self.assertIs(not_present.is_satisfied(client.get_status()), False)
+        self.assertItemsEqual(
+            [('0', 'still-present')],
+            not_present.iter_blocking_state(client.get_status()))
         client.juju('remove-machine', ('0'))
-        self.assertIs(not_present.is_satisfied(client.get_status()), True)
+        self.assertItemsEqual(
+            [], not_present.iter_blocking_state(client.get_status()))
 
     def test_do_raise(self):
         not_present = WaitMachineNotPresent('0')
         with self.assertRaisesRegexp(
                 Exception, 'Timed out waiting for machine removal 0'):
-            not_present.do_raise()
-
-
-class TestEnvJujuClient(ClientTest):
+            not_present.do_raise('', None)
+
+
+class TestModelClient(ClientTest):
 
     def test_no_duplicate_env(self):
         env = JujuData('foo', {})
-        client = EnvJujuClient(env, '1.25', 'full_path')
+        client = ModelClient(env, '1.25', 'full_path')
         self.assertIs(env, client.env)
 
     def test_convert_to_juju_data(self):
         env = SimpleEnvironment('foo', {'type': 'bar'}, 'baz')
         with patch.object(JujuData, 'load_yaml'):
-            client = EnvJujuClient(env, '1.25', 'full_path')
+            client = ModelClient(env, '1.25', 'full_path')
             client.env.load_yaml.assert_called_once_with()
         self.assertIsInstance(client.env, JujuData)
         self.assertEqual(client.env.environment, 'foo')
@@ -560,17 +367,17 @@
     def test_get_version(self):
         value = ' 5.6 \n'
         with patch('subprocess.check_output', return_value=value) as vsn:
-            version = EnvJujuClient.get_version()
+            version = ModelClient.get_version()
         self.assertEqual('5.6', version)
         vsn.assert_called_with(('juju', '--version'))
 
     def test_get_version_path(self):
         with patch('subprocess.check_output', return_value=' 4.3') as vsn:
-            EnvJujuClient.get_version('foo/bar/baz')
+            ModelClient.get_version('foo/bar/baz')
         vsn.assert_called_once_with(('foo/bar/baz', '--version'))
 
     def test_get_matching_agent_version(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData(None, {'type': 'local'}, juju_home='foo'),
             '1.23-series-arch', None)
         self.assertEqual('1.23.1', client.get_matching_agent_version())
@@ -580,29 +387,29 @@
         self.assertEqual('1.20-beta1.1', client.get_matching_agent_version())
 
     def test_upgrade_juju_nonlocal(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('foo', {'type': 'nonlocal'}), '2.0-betaX', None)
         with patch.object(client, '_upgrade_juju') as juju_mock:
             client.upgrade_juju()
         juju_mock.assert_called_with(('--agent-version', '2.0'))
 
     def test_upgrade_juju_local(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('foo', {'type': 'local'}), '2.0-betaX', None)
         with patch.object(client, '_upgrade_juju') as juju_mock:
             client.upgrade_juju()
         juju_mock.assert_called_with(('--agent-version', '2.0',))
 
     def test_upgrade_juju_no_force_version(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('foo', {'type': 'local'}), '2.0-betaX', None)
         with patch.object(client, '_upgrade_juju') as juju_mock:
             client.upgrade_juju(force_version=False)
         juju_mock.assert_called_with(())
 
     def test_clone_unchanged(self):
-        client1 = EnvJujuClient(JujuData('foo'), '1.27', 'full/path',
-                                debug=True)
+        client1 = ModelClient(JujuData('foo'), '1.27', 'full/path',
+                              debug=True)
         client2 = client1.clone()
         self.assertIsNot(client1, client2)
         self.assertIs(type(client1), type(client2))
@@ -614,8 +421,8 @@
         self.assertEqual(client1._backend, client2._backend)
 
     def test_clone_changed(self):
-        client1 = EnvJujuClient(JujuData('foo'), '1.27', 'full/path',
-                                debug=True)
+        client1 = ModelClient(JujuData('foo'), '1.27', 'full/path',
+                              debug=True)
         env2 = SimpleEnvironment('bar')
         client2 = client1.clone(env2, '1.28', 'other/path', debug=False,
                                 cls=EnvJujuClient1X)
@@ -627,8 +434,8 @@
         self.assertEqual(client1.feature_flags, client2.feature_flags)
 
     def test_get_cache_path(self):
-        client = EnvJujuClient(JujuData('foo', juju_home='/foo/'),
-                               '1.27', 'full/path', debug=True)
+        client = ModelClient(JujuData('foo', juju_home='/foo/'),
+                             '1.27', 'full/path', debug=True)
         self.assertEqual('/foo/models/cache.yaml',
                          client.get_cache_path())
 
@@ -638,7 +445,7 @@
             'tools-metadata-url': 'bar',
             'type': 'baz',
             })
-        client = EnvJujuClient(env, None, 'my/juju/bin')
+        client = ModelClient(env, None, 'my/juju/bin')
         self.assertEqual({
             'agent-metadata-url': 'foo',
             'test-mode': True,
@@ -691,7 +498,7 @@
             'type': 'foo',
             'username': 'foo',
             }, 'home')
-        client = EnvJujuClient(env, None, 'my/juju/bin')
+        client = ModelClient(env, None, 'my/juju/bin')
         with client._bootstrap_config() as config_filename:
             with open(config_filename) as f:
                 self.assertEqual({
@@ -712,14 +519,14 @@
 
     def test_get_cloud_region(self):
         self.assertEqual(
-            'foo/bar', EnvJujuClient.get_cloud_region('foo', 'bar'))
+            'foo/bar', ModelClient.get_cloud_region('foo', 'bar'))
         self.assertEqual(
-            'foo', EnvJujuClient.get_cloud_region('foo', None))
+            'foo', ModelClient.get_cloud_region('foo', None))
 
     def test_bootstrap_maas(self):
         env = JujuData('maas', {'type': 'foo', 'region': 'asdf'})
-        with patch.object(EnvJujuClient, 'juju') as mock:
-            client = EnvJujuClient(env, '2.0-zeta1', None)
+        with patch.object(ModelClient, 'juju') as mock:
+            client = ModelClient(env, '2.0-zeta1', None)
             with patch.object(client.env, 'maas', lambda: True):
                 with observable_temp_file() as config_file:
                     client.bootstrap()
@@ -736,8 +543,8 @@
         # Disable space constraint with environment variable
         os.environ['JUJU_CI_SPACELESSNESS'] = "1"
         env = JujuData('maas', {'type': 'foo', 'region': 'asdf'})
-        with patch.object(EnvJujuClient, 'juju') as mock:
-            client = EnvJujuClient(env, '2.0-zeta1', None)
+        with patch.object(ModelClient, 'juju') as mock:
+            client = ModelClient(env, '2.0-zeta1', None)
             with patch.object(client.env, 'maas', lambda: True):
                 with observable_temp_file() as config_file:
                     client.bootstrap()
@@ -752,8 +559,8 @@
     def test_bootstrap_joyent(self):
         env = JujuData('joyent', {
             'type': 'joyent', 'sdc-url': 'https://foo.api.joyentcloud.com'})
-        with patch.object(EnvJujuClient, 'juju', autospec=True) as mock:
-            client = EnvJujuClient(env, '2.0-zeta1', None)
+        with patch.object(ModelClient, 'juju', autospec=True) as mock:
+            client = ModelClient(env, '2.0-zeta1', None)
             with patch.object(client.env, 'joyent', lambda: True):
                 with observable_temp_file() as config_file:
                     client.bootstrap()
@@ -768,8 +575,8 @@
     def test_bootstrap(self):
         env = JujuData('foo', {'type': 'bar', 'region': 'baz'})
         with observable_temp_file() as config_file:
-            with patch.object(EnvJujuClient, 'juju') as mock:
-                client = EnvJujuClient(env, '2.0-zeta1', None)
+            with patch.object(ModelClient, 'juju') as mock:
+                client = ModelClient(env, '2.0-zeta1', None)
                 client.bootstrap()
                 mock.assert_called_with(
                     'bootstrap', ('--constraints', 'mem=2G',
@@ -783,7 +590,7 @@
 
     def test_bootstrap_upload_tools(self):
         env = JujuData('foo', {'type': 'foo', 'region': 'baz'})
-        client = EnvJujuClient(env, '2.0-zeta1', None)
+        client = ModelClient(env, '2.0-zeta1', None)
         with observable_temp_file() as config_file:
             with patch.object(client, 'juju') as mock:
                 client.bootstrap(upload_tools=True)
@@ -796,7 +603,7 @@
 
     def test_bootstrap_credential(self):
         env = JujuData('foo', {'type': 'foo', 'region': 'baz'})
-        client = EnvJujuClient(env, '2.0-zeta1', None)
+        client = ModelClient(env, '2.0-zeta1', None)
         with observable_temp_file() as config_file:
             with patch.object(client, 'juju') as mock:
                 client.bootstrap(credential='credential_name')
@@ -810,7 +617,7 @@
 
     def test_bootstrap_bootstrap_series(self):
         env = JujuData('foo', {'type': 'bar', 'region': 'baz'})
-        client = EnvJujuClient(env, '2.0-zeta1', None)
+        client = ModelClient(env, '2.0-zeta1', None)
         with patch.object(client, 'juju') as mock:
             with observable_temp_file() as config_file:
                 client.bootstrap(bootstrap_series='angsty')
@@ -824,7 +631,7 @@
 
     def test_bootstrap_auto_upgrade(self):
         env = JujuData('foo', {'type': 'bar', 'region': 'baz'})
-        client = EnvJujuClient(env, '2.0-zeta1', None)
+        client = ModelClient(env, '2.0-zeta1', None)
         with patch.object(client, 'juju') as mock:
             with observable_temp_file() as config_file:
                 client.bootstrap(auto_upgrade=True)
@@ -837,7 +644,7 @@
 
     def test_bootstrap_no_gui(self):
         env = JujuData('foo', {'type': 'bar', 'region': 'baz'})
-        client = EnvJujuClient(env, '2.0-zeta1', None)
+        client = ModelClient(env, '2.0-zeta1', None)
         with patch.object(client, 'juju') as mock:
             with observable_temp_file() as config_file:
                 client.bootstrap(no_gui=True)
@@ -850,7 +657,7 @@
 
     def test_bootstrap_metadata(self):
         env = JujuData('foo', {'type': 'bar', 'region': 'baz'})
-        client = EnvJujuClient(env, '2.0-zeta1', None)
+        client = ModelClient(env, '2.0-zeta1', None)
         with patch.object(client, 'juju') as mock:
             with observable_temp_file() as config_file:
                 client.bootstrap(metadata_source='/var/test-source')
@@ -864,7 +671,7 @@
 
     def test_bootstrap_to(self):
         env = JujuData('foo', {'type': 'bar', 'region': 'baz'})
-        client = EnvJujuClient(env, '2.0-zeta1', None)
+        client = ModelClient(env, '2.0-zeta1', None)
         with patch.object(client, 'juju') as mock:
             with observable_temp_file() as config_file:
                 client.bootstrap(to='target')
@@ -877,8 +684,8 @@
 
     def test_bootstrap_async(self):
         env = JujuData('foo', {'type': 'bar', 'region': 'baz'})
-        with patch.object(EnvJujuClient, 'juju_async', autospec=True) as mock:
-            client = EnvJujuClient(env, '2.0-zeta1', None)
+        with patch.object(ModelClient, 'juju_async', autospec=True) as mock:
+            client = ModelClient(env, '2.0-zeta1', None)
             client.env.juju_home = 'foo'
             with observable_temp_file() as config_file:
                 with client.bootstrap_async():
@@ -892,8 +699,8 @@
 
     def test_bootstrap_async_upload_tools(self):
         env = JujuData('foo', {'type': 'bar', 'region': 'baz'})
-        with patch.object(EnvJujuClient, 'juju_async', autospec=True) as mock:
-            client = EnvJujuClient(env, '2.0-zeta1', None)
+        with patch.object(ModelClient, 'juju_async', autospec=True) as mock:
+            client = ModelClient(env, '2.0-zeta1', None)
             with observable_temp_file() as config_file:
                 with client.bootstrap_async(upload_tools=True):
                     mock.assert_called_with(
@@ -907,7 +714,7 @@
 
     def test_get_bootstrap_args_bootstrap_series(self):
         env = JujuData('foo', {'type': 'bar', 'region': 'baz'})
-        client = EnvJujuClient(env, '2.0-zeta1', None)
+        client = ModelClient(env, '2.0-zeta1', None)
         args = client.get_bootstrap_args(upload_tools=True,
                                          config_filename='config',
                                          bootstrap_series='angsty')
@@ -919,7 +726,7 @@
 
     def test_get_bootstrap_args_agent_version(self):
         env = JujuData('foo', {'type': 'bar', 'region': 'baz'})
-        client = EnvJujuClient(env, '2.0-zeta1', None)
+        client = ModelClient(env, '2.0-zeta1', None)
         args = client.get_bootstrap_args(upload_tools=False,
                                          config_filename='config',
                                          agent_version='2.0-lambda1')
@@ -930,7 +737,7 @@
 
     def test_get_bootstrap_args_upload_tools_and_agent_version(self):
         env = JujuData('foo', {'type': 'bar', 'region': 'baz'})
-        client = EnvJujuClient(env, '2.0-zeta1', None)
+        client = ModelClient(env, '2.0-zeta1', None)
         with self.assertRaises(ValueError):
             client.get_bootstrap_args(upload_tools=True,
                                       config_filename='config',
@@ -941,9 +748,9 @@
             'kill-controller', 'add-model', ('-c', 'foo'))
 
     def do_add_model(self, jes_command, create_cmd, controller_option):
-        controller_client = EnvJujuClient(JujuData('foo'), None, None)
+        controller_client = ModelClient(JujuData('foo'), None, None)
         model_data = JujuData('bar', {'type': 'foo'})
-        client = EnvJujuClient(model_data, None, None)
+        client = ModelClient(model_data, None, None)
         with patch.object(client, 'get_jes_command',
                           return_value=jes_command):
                 with patch.object(controller_client, 'juju') as ccj_mock:
@@ -966,14 +773,24 @@
             '--config', config_file.name),
             frozenset({'migration'}), 'foo', None, True, None, None)
 
+    def test_add_model_by_name(self):
+        client = fake_juju_client()
+        client.bootstrap()
+        with patch.object(client._backend, 'juju') as juju_mock:
+            with observable_temp_file() as config_file:
+                client.add_model('new-model')
+        juju_mock.assert_called_once_with('add-model', (
+            '-c', 'name', 'new-model', '--config', config_file.name),
+            frozenset({'migration'}), 'foo', None, True, None, None)
+
     def test_destroy_environment(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         self.assertIs(False, hasattr(client, 'destroy_environment'))
 
     def test_destroy_model(self):
         env = JujuData('foo', {'type': 'ec2'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         with patch.object(client, 'juju') as mock:
             client.destroy_model()
         mock.assert_called_with(
@@ -982,7 +799,7 @@
 
     def test_destroy_model_azure(self):
         env = JujuData('foo', {'type': 'azure'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         with patch.object(client, 'juju') as mock:
             client.destroy_model()
         mock.assert_called_with(
@@ -991,7 +808,7 @@
 
     def test_destroy_model_gce(self):
         env = JujuData('foo', {'type': 'gce'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         with patch.object(client, 'juju') as mock:
             client.destroy_model()
         mock.assert_called_with(
@@ -999,7 +816,7 @@
             include_e=False, timeout=1200)
 
     def test_kill_controller(self):
-        client = EnvJujuClient(JujuData('foo', {'type': 'ec2'}), None, None)
+        client = ModelClient(JujuData('foo', {'type': 'ec2'}), None, None)
         with patch.object(client, 'juju') as juju_mock:
             client.kill_controller()
         juju_mock.assert_called_once_with(
@@ -1007,7 +824,7 @@
             timeout=600)
 
     def test_kill_controller_check(self):
-        client = EnvJujuClient(JujuData('foo', {'type': 'ec2'}), None, None)
+        client = ModelClient(JujuData('foo', {'type': 'ec2'}), None, None)
         with patch.object(client, 'juju') as juju_mock:
             client.kill_controller(check=True)
         juju_mock.assert_called_once_with(
@@ -1015,7 +832,7 @@
             timeout=600)
 
     def do_kill_controller_azure(self, jes_command, kill_command):
-        client = EnvJujuClient(JujuData('foo', {'type': 'azure'}), None, None)
+        client = ModelClient(JujuData('foo', {'type': 'azure'}), None, None)
         with patch.object(client, 'get_jes_command',
                           return_value=jes_command):
             with patch.object(client, 'juju') as juju_mock:
@@ -1025,7 +842,7 @@
             timeout=1800)
 
     def test_kill_controller_gce(self):
-        client = EnvJujuClient(JujuData('foo', {'type': 'gce'}), None, None)
+        client = ModelClient(JujuData('foo', {'type': 'gce'}), None, None)
         with patch.object(client, 'juju') as juju_mock:
             client.kill_controller()
         juju_mock.assert_called_once_with(
@@ -1033,7 +850,7 @@
             timeout=1200)
 
     def test_destroy_controller(self):
-        client = EnvJujuClient(JujuData('foo', {'type': 'ec2'}), None, None)
+        client = ModelClient(JujuData('foo', {'type': 'ec2'}), None, None)
         with patch.object(client, 'juju') as juju_mock:
             client.destroy_controller()
         juju_mock.assert_called_once_with(
@@ -1041,7 +858,7 @@
             timeout=600)
 
     def test_destroy_controller_all_models(self):
-        client = EnvJujuClient(JujuData('foo', {'type': 'ec2'}), None, None)
+        client = ModelClient(JujuData('foo', {'type': 'ec2'}), None, None)
         with patch.object(client, 'juju') as juju_mock:
             client.destroy_controller(all_models=True)
         juju_mock.assert_called_once_with(
@@ -1071,7 +888,7 @@
 
     def test_tear_down(self):
         """Check that a successful tear_down calls destroy."""
-        client = EnvJujuClient(JujuData('foo', {'type': 'gce'}), None, None)
+        client = ModelClient(JujuData('foo', {'type': 'gce'}), None, None)
         with self.mock_tear_down(client) as (mock_destroy, mock_kill):
             client.tear_down()
         mock_destroy.assert_called_once_with(all_models=True)
@@ -1079,7 +896,7 @@
 
     def test_tear_down_fall_back(self):
         """Check that tear_down uses kill_controller if destroy fails."""
-        client = EnvJujuClient(JujuData('foo', {'type': 'gce'}), None, None)
+        client = ModelClient(JujuData('foo', {'type': 'gce'}), None, None)
         with self.mock_tear_down(client, True) as (mock_destroy, mock_kill):
             with self.assertRaises(subprocess.CalledProcessError) as err:
                 client.tear_down()
@@ -1089,7 +906,7 @@
 
     def test_tear_down_double_fail(self):
         """Check tear_down when both destroy and kill fail."""
-        client = EnvJujuClient(JujuData('foo', {'type': 'gce'}), None, None)
+        client = ModelClient(JujuData('foo', {'type': 'gce'}), None, None)
         with self.mock_tear_down(client, True, True) as (
                 mock_destroy, mock_kill):
             with self.assertRaises(subprocess.CalledProcessError) as err:
@@ -1100,7 +917,7 @@
 
     def test_get_juju_output(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         fake_popen = FakePopen('asdf', None, 0)
         with patch('subprocess.Popen', return_value=fake_popen) as mock:
             result = client.get_juju_output('bar')
@@ -1111,7 +928,7 @@
     def test_get_juju_output_accepts_varargs(self):
         env = JujuData('foo')
         fake_popen = FakePopen('asdf', None, 0)
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         with patch('subprocess.Popen', return_value=fake_popen) as mock:
             result = client.get_juju_output('bar', 'baz', '--qux')
         self.assertEqual('asdf', result)
@@ -1121,7 +938,7 @@
     def test_get_juju_output_stderr(self):
         env = JujuData('foo')
         fake_popen = FakePopen(None, 'Hello!', 1)
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         with self.assertRaises(subprocess.CalledProcessError) as exc:
             with patch('subprocess.Popen', return_value=fake_popen):
                 client.get_juju_output('bar')
@@ -1130,7 +947,7 @@
     def test_get_juju_output_merge_stderr(self):
         env = JujuData('foo')
         fake_popen = FakePopen('Err on out', None, 0)
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         with patch('subprocess.Popen', return_value=fake_popen) as mock_popen:
             result = client.get_juju_output('bar', merge_stderr=True)
         self.assertEqual(result, 'Err on out')
@@ -1142,7 +959,7 @@
     def test_get_juju_output_full_cmd(self):
         env = JujuData('foo')
         fake_popen = FakePopen(None, 'Hello!', 1)
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         with self.assertRaises(subprocess.CalledProcessError) as exc:
             with patch('subprocess.Popen', return_value=fake_popen):
                 client.get_juju_output('bar', '--baz', 'qux')
@@ -1153,7 +970,7 @@
     def test_get_juju_output_accepts_timeout(self):
         env = JujuData('foo')
         fake_popen = FakePopen('asdf', None, 0)
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         with patch('subprocess.Popen', return_value=fake_popen) as po_mock:
             client.get_juju_output('bar', timeout=5)
         self.assertEqual(
@@ -1162,7 +979,7 @@
              '--show-log', 'bar', '-m', 'foo:foo'))
 
     def test__shell_environ_juju_data(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('baz', {'type': 'ec2'}), '1.25-foobar', 'path', 'asdf')
         env = client._shell_environ()
         self.assertEqual(env['JUJU_DATA'], 'asdf')
@@ -1170,7 +987,7 @@
 
     def test_juju_output_supplies_path(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, '/foobar/bar')
+        client = ModelClient(env, None, '/foobar/bar')
 
         def check_path(*args, **kwargs):
             self.assertRegexpMatches(os.environ['PATH'], r'/foobar\:')
@@ -1186,7 +1003,7 @@
                 - c
                 """)
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         with patch.object(client, 'get_juju_output',
                           return_value=output_text) as gjo_mock:
             result = client.get_status()
@@ -1197,7 +1014,7 @@
 
     def test_get_status_retries_on_error(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         client.attempt = 0
 
         def get_juju_output(command, *args, **kwargs):
@@ -1211,22 +1028,24 @@
 
     def test_get_status_raises_on_timeout_1(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
 
         def get_juju_output(command, *args, **kwargs):
             raise subprocess.CalledProcessError(1, command)
 
         with patch.object(client, 'get_juju_output',
                           side_effect=get_juju_output):
-            with patch('jujupy.until_timeout', lambda x: iter([None, None])):
+            with patch('jujupy.client.until_timeout',
+                       lambda x: iter([None, None])):
                 with self.assertRaisesRegexp(
                         Exception, 'Timed out waiting for juju status'):
                     client.get_status()
 
     def test_get_status_raises_on_timeout_2(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, None)
-        with patch('jujupy.until_timeout', return_value=iter([1])) as mock_ut:
+        client = ModelClient(env, None, None)
+        with patch('jujupy.client.until_timeout',
+                   return_value=iter([1])) as mock_ut:
             with patch.object(client, 'get_juju_output',
                               side_effect=StopIteration):
                 with self.assertRaises(StopIteration):
@@ -1235,7 +1054,7 @@
 
     def test_show_model_uses_provided_model_name(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         show_model_output = dedent("""\
             bar:
                 status:
@@ -1254,7 +1073,7 @@
 
     def test_show_model_defaults_to_own_model_name(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         show_model_output = dedent("""\
             foo:
                 status:
@@ -1274,6 +1093,8 @@
     @staticmethod
     def make_status_yaml(key, machine_value, unit_value):
         return dedent("""\
+            model:
+              name: foo
             machines:
               "0":
                 {0}: {1}
@@ -1285,21 +1106,21 @@
         """.format(key, machine_value, unit_value))
 
     def test_deploy_non_joyent(self):
-        env = EnvJujuClient(
+        env = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         with patch.object(env, 'juju') as mock_juju:
             env.deploy('mondogb')
         mock_juju.assert_called_with('deploy', ('mondogb',))
 
     def test_deploy_joyent(self):
-        env = EnvJujuClient(
+        env = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         with patch.object(env, 'juju') as mock_juju:
             env.deploy('mondogb')
         mock_juju.assert_called_with('deploy', ('mondogb',))
 
     def test_deploy_repository(self):
-        env = EnvJujuClient(
+        env = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         with patch.object(env, 'juju') as mock_juju:
             env.deploy('/home/jrandom/repo/mongodb')
@@ -1307,7 +1128,7 @@
             'deploy', ('/home/jrandom/repo/mongodb',))
 
     def test_deploy_to(self):
-        env = EnvJujuClient(
+        env = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         with patch.object(env, 'juju') as mock_juju:
             env.deploy('mondogb', to='0')
@@ -1315,7 +1136,7 @@
             'deploy', ('mondogb', '--to', '0'))
 
     def test_deploy_service(self):
-        env = EnvJujuClient(
+        env = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         with patch.object(env, 'juju') as mock_juju:
             env.deploy('local:mondogb', service='my-mondogb')
@@ -1323,14 +1144,14 @@
             'deploy', ('local:mondogb', 'my-mondogb',))
 
     def test_deploy_force(self):
-        env = EnvJujuClient(
+        env = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         with patch.object(env, 'juju') as mock_juju:
             env.deploy('local:mondogb', force=True)
         mock_juju.assert_called_with('deploy', ('local:mondogb', '--force',))
 
     def test_deploy_series(self):
-        env = EnvJujuClient(
+        env = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         with patch.object(env, 'juju') as mock_juju:
             env.deploy('local:blah', series='xenial')
@@ -1338,7 +1159,7 @@
             'deploy', ('local:blah', '--series', 'xenial'))
 
     def test_deploy_resource(self):
-        env = EnvJujuClient(JujuData('foo', {'type': 'local'}), None, None)
+        env = ModelClient(JujuData('foo', {'type': 'local'}), None, None)
         with patch.object(env, 'juju') as mock_juju:
             env.deploy('local:blah', resource='foo=/path/dir')
         mock_juju.assert_called_with(
@@ -1361,14 +1182,14 @@
             'deploy', ('mondogb', '--constraints', 'virt-type=kvm'))
 
     def test_attach(self):
-        env = EnvJujuClient(JujuData('foo', {'type': 'local'}), None, None)
+        env = ModelClient(JujuData('foo', {'type': 'local'}), None, None)
         with patch.object(env, 'juju') as mock_juju:
             env.attach('foo', resource='foo=/path/dir')
         mock_juju.assert_called_with('attach', ('foo', 'foo=/path/dir'))
 
     def test_list_resources(self):
         data = 'resourceid: resource/foo'
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         with patch.object(
                 client, 'get_juju_output', return_value=data) as mock_gjo:
             status = client.list_resources('foo')
@@ -1377,7 +1198,7 @@
             'list-resources', '--format', 'yaml', 'foo', '--details')
 
     def test_wait_for_resource(self):
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         with patch.object(
                 client, 'list_resources',
                 return_value=make_resource_list()) as mock_lr:
@@ -1385,13 +1206,13 @@
         mock_lr.assert_called_once_with('foo')
 
     def test_wait_for_resource_timeout(self):
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         resource_list = make_resource_list()
         resource_list['resources'][0]['expected']['resourceid'] = 'bad_id'
         with patch.object(
                 client, 'list_resources',
                 return_value=resource_list) as mock_lr:
-            with patch('jujupy.until_timeout', autospec=True,
+            with patch('jujupy.client.until_timeout', autospec=True,
                        return_value=[0, 1]) as mock_ju:
                 with patch('time.sleep', autospec=True) as mock_ts:
                     with self.assertRaisesRegexp(
@@ -1404,7 +1225,7 @@
         self.assertEqual(mock_ju.mock_calls, [call(60)])
 
     def test_wait_for_resource_suppresses_deadline(self):
-        client = EnvJujuClient(JujuData('local', juju_home=''), None, None)
+        client = ModelClient(JujuData('local', juju_home=''), None, None)
         with client_past_deadline(client):
             real_check_timeouts = client.check_timeouts
 
@@ -1420,7 +1241,7 @@
 
     def test_wait_for_resource_checks_deadline(self):
         resource_list = make_resource_list()
-        client = EnvJujuClient(JujuData('local', juju_home=''), None, None)
+        client = ModelClient(JujuData('local', juju_home=''), None, None)
         with client_past_deadline(client):
             with patch.object(client, 'list_resources', autospec=True,
                               return_value=resource_list):
@@ -1428,23 +1249,23 @@
                     client.wait_for_resource('dummy-resource/foo', 'app_unit')
 
     def test_deploy_bundle_2x(self):
-        client = EnvJujuClient(JujuData('an_env', None),
-                               '1.23-series-arch', None)
+        client = ModelClient(JujuData('an_env', None),
+                             '1.23-series-arch', None)
         with patch.object(client, 'juju') as mock_juju:
             client.deploy_bundle('bundle:~juju-qa/some-bundle')
         mock_juju.assert_called_with(
             'deploy', ('bundle:~juju-qa/some-bundle'), timeout=3600)
 
     def test_deploy_bundle_template(self):
-        client = EnvJujuClient(JujuData('an_env', None),
-                               '1.23-series-arch', None)
+        client = ModelClient(JujuData('an_env', None),
+                             '1.23-series-arch', None)
         with patch.object(client, 'juju') as mock_juju:
             client.deploy_bundle('bundle:~juju-qa/some-{container}-bundle')
         mock_juju.assert_called_with(
             'deploy', ('bundle:~juju-qa/some-lxd-bundle'), timeout=3600)
 
     def test_upgrade_charm(self):
-        env = EnvJujuClient(
+        env = ModelClient(
             JujuData('foo', {'type': 'local'}), '2.34-74', None)
         with patch.object(env, 'juju') as mock_juju:
             env.upgrade_charm('foo-service',
@@ -1454,14 +1275,14 @@
                               '/bar/repository/angsty/mongodb',))
 
     def test_remove_service(self):
-        env = EnvJujuClient(
+        env = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         with patch.object(env, 'juju') as mock_juju:
             env.remove_service('mondogb')
         mock_juju.assert_called_with('remove-application', ('mondogb',))
 
     def test_status_until_always_runs_once(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         status_txt = self.make_status_yaml('agent-state', 'started', 'started')
         with patch.object(client, 'get_juju_output', return_value=status_txt):
@@ -1470,7 +1291,7 @@
             [r.status for r in result], [Status.from_text(status_txt).status])
 
     def test_status_until_timeout(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         status_txt = self.make_status_yaml('agent-state', 'started', 'started')
         status_yaml = yaml.safe_load(status_txt)
@@ -1479,7 +1300,7 @@
             return iter([None, None])
 
         with patch.object(client, 'get_juju_output', return_value=status_txt):
-            with patch('jujupy.until_timeout',
+            with patch('jujupy.client.until_timeout',
                        side_effect=until_timeout_stub) as ut_mock:
                 result = list(client.status_until(30, 70))
         self.assertEqual(
@@ -1498,7 +1319,7 @@
                 list(client.status_until(0))
 
     def test_add_ssh_machines(self):
-        client = EnvJujuClient(JujuData('foo'), None, 'juju')
+        client = ModelClient(JujuData('foo'), None, 'juju')
         with patch('subprocess.check_call', autospec=True) as cc_mock:
             client.add_ssh_machines(['m-foo', 'm-bar', 'm-baz'])
         assert_juju_call(
@@ -1524,8 +1345,23 @@
             2)
         self.assertEqual(cc_mock.call_count, 3)
 
+    def test_make_remove_machine_condition(self):
+        client = fake_juju_client()
+        condition = client.make_remove_machine_condition('0')
+        self.assertIs(WaitMachineNotPresent, type(condition))
+        self.assertEqual('0', condition.machine)
+        self.assertEqual(600, condition.timeout)
+
+    def test_make_remove_machine_condition_azure(self):
+        client = fake_juju_client()
+        client.env._config['type'] = 'azure'
+        condition = client.make_remove_machine_condition('0')
+        self.assertIs(WaitMachineNotPresent, type(condition))
+        self.assertEqual('0', condition.machine)
+        self.assertEqual(1200, condition.timeout)
+
     def test_add_ssh_machines_retry(self):
-        client = EnvJujuClient(JujuData('foo'), None, 'juju')
+        client = ModelClient(JujuData('foo'), None, 'juju')
         with patch('subprocess.check_call', autospec=True,
                    side_effect=[subprocess.CalledProcessError(None, None),
                                 None, None, None]) as cc_mock:
@@ -1562,7 +1398,7 @@
         self.assertEqual(cc_mock.call_count, 4)
 
     def test_add_ssh_machines_fail_on_second_machine(self):
-        client = EnvJujuClient(JujuData('foo'), None, 'juju')
+        client = ModelClient(JujuData('foo'), None, 'juju')
         with patch('subprocess.check_call', autospec=True, side_effect=[
                 None, subprocess.CalledProcessError(None, None), None, None
                 ]) as cc_mock:
@@ -1585,7 +1421,7 @@
         self.assertEqual(cc_mock.call_count, 2)
 
     def test_add_ssh_machines_fail_on_second_attempt(self):
-        client = EnvJujuClient(JujuData('foo'), None, 'juju')
+        client = ModelClient(JujuData('foo'), None, 'juju')
         with patch('subprocess.check_call', autospec=True, side_effect=[
                 subprocess.CalledProcessError(None, None),
                 subprocess.CalledProcessError(None, None)]) as cc_mock:
@@ -1609,14 +1445,15 @@
 
     def test_wait_for_started(self):
         value = self.make_status_yaml('agent-state', 'started', 'started')
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         with patch.object(client, 'get_juju_output', return_value=value):
             client.wait_for_started()
 
     def test_wait_for_started_timeout(self):
         value = self.make_status_yaml('agent-state', 'pending', 'started')
-        client = EnvJujuClient(JujuData('local'), None, None)
-        with patch('jujupy.until_timeout', lambda x, start=None: range(1)):
+        client = ModelClient(JujuData('local'), None, None)
+        with patch('jujupy.client.until_timeout',
+                   lambda x, start=None: range(1)):
             with patch.object(client, 'get_juju_output', return_value=value):
                 writes = []
                 with patch.object(GroupReporter, '_write', autospec=True,
@@ -1629,7 +1466,7 @@
 
     def test_wait_for_started_start(self):
         value = self.make_status_yaml('agent-state', 'started', 'pending')
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         now = datetime.now() + timedelta(days=1)
         with patch('utility.until_timeout.now', return_value=now):
             with patch.object(client, 'get_juju_output', return_value=value):
@@ -1658,7 +1495,7 @@
         Also, the client is patched so that the soft_deadline has been hit.
         """
         if client is None:
-            client = EnvJujuClient(JujuData('local', juju_home=''), None, None)
+            client = ModelClient(JujuData('local', juju_home=''), None, None)
         with client_past_deadline(client):
             # This will work even after we patch check_timeouts below.
             real_check_timeouts = client.check_timeouts
@@ -1687,7 +1524,7 @@
         Also, the client is patched so that the soft_deadline has been hit.
         """
         if client is None:
-            client = EnvJujuClient(JujuData('local', juju_home=''), None, None)
+            client = ModelClient(JujuData('local', juju_home=''), None, None)
         with client_past_deadline(client):
             status_obj = client.status_class(status, '')
             with patch.object(client, 'get_status', autospec=True,
@@ -1754,7 +1591,7 @@
 
     def test_wait_for_started_logs_status(self):
         value = self.make_status_yaml('agent-state', 'pending', 'started')
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         with patch.object(client, 'get_juju_output', return_value=value):
             writes = []
             with patch.object(GroupReporter, '_write', autospec=True,
@@ -1787,12 +1624,15 @@
                       sub3/0:
                         agent-state: started
         """)
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         now = datetime.now() + timedelta(days=1)
         with patch('utility.until_timeout.now', return_value=now):
             with patch.object(client, 'get_juju_output', return_value=value):
-                with patch('jujupy.GroupReporter.update') as update_mock:
-                    with patch('jujupy.GroupReporter.finish') as finish_mock:
+                with patch(
+                        'jujupy.client.GroupReporter.update') as update_mock:
+                    with patch(
+                            'jujupy.client.GroupReporter.finish'
+                            ) as finish_mock:
                         client.wait_for_subordinate_units(
                             'jenkins', 'sub1', start=now - timedelta(1200))
         self.assertEqual([], update_mock.call_args_list)
@@ -1822,12 +1662,15 @@
                         agent-status:
                           current: idle
         """)
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         now = datetime.now() + timedelta(days=1)
         with patch('utility.until_timeout.now', return_value=now):
             with patch.object(client, 'get_juju_output', return_value=value):
-                with patch('jujupy.GroupReporter.update') as update_mock:
-                    with patch('jujupy.GroupReporter.finish') as finish_mock:
+                with patch(
+                        'jujupy.client.GroupReporter.update') as update_mock:
+                    with patch(
+                            'jujupy.client.GroupReporter.finish'
+                            ) as finish_mock:
                         client.wait_for_subordinate_units(
                             'jenkins', 'sub1', start=now - timedelta(1200))
         self.assertEqual([], update_mock.call_args_list)
@@ -1850,12 +1693,15 @@
                       sub/1:
                         agent-state: started
         """)
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         now = datetime.now() + timedelta(days=1)
         with patch('utility.until_timeout.now', return_value=now):
             with patch.object(client, 'get_juju_output', return_value=value):
-                with patch('jujupy.GroupReporter.update') as update_mock:
-                    with patch('jujupy.GroupReporter.finish') as finish_mock:
+                with patch(
+                        'jujupy.client.GroupReporter.update') as update_mock:
+                    with patch(
+                            'jujupy.client.GroupReporter.finish'
+                            ) as finish_mock:
                         client.wait_for_subordinate_units(
                             'ubuntu', 'sub', start=now - timedelta(1200))
         self.assertEqual([], update_mock.call_args_list)
@@ -1874,7 +1720,7 @@
                       sub1:
                         agent-state: started
         """)
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         now = datetime.now() + timedelta(days=1)
         with patch('utility.until_timeout.now', return_value=now):
             with patch.object(client, 'get_juju_output', return_value=value):
@@ -1895,7 +1741,7 @@
                   jenkins/0:
                     agent-state: started
         """)
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         now = datetime.now() + timedelta(days=1)
         with patch('utility.until_timeout.now', return_value=now):
             with patch.object(client, 'get_juju_output', return_value=value):
@@ -1921,7 +1767,7 @@
         final_status = Status(copy.deepcopy(initial_status.status), None)
         final_status.status['applications']['jenkins']['units']['jenkins/0'][
             'workload-status']['current'] = 'active'
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         writes = []
         with patch('utility.until_timeout', autospec=True, return_value=[1]):
             with patch.object(client, 'get_status', autospec=True,
@@ -1944,7 +1790,7 @@
                       workload-status:
                         current: unknown
         """)
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         writes = []
         with patch('utility.until_timeout', autospec=True, return_value=[]):
             with patch.object(client, 'get_status', autospec=True,
@@ -1962,7 +1808,7 @@
                   jenkins/0:
                     agent-state: active
         """)
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         writes = []
         with patch('utility.until_timeout', autospec=True, return_value=[]):
             with patch.object(client, 'get_status', autospec=True,
@@ -1973,7 +1819,7 @@
         self.assertEqual(writes, [])
 
     def test_list_models(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'juju') as j_mock:
             client.list_models()
         j_mock.assert_called_once_with(
@@ -1990,7 +1836,7 @@
               owner: admin
             current-model: foo
         """
-        client = EnvJujuClient(JujuData('baz'), None, None)
+        client = ModelClient(JujuData('baz'), None, None)
         with patch.object(client, 'get_juju_output',
                           return_value=data) as gjo_mock:
             models = client.get_models()
@@ -2016,7 +1862,7 @@
               owner: admin
             current-model: foo
         """
-        client = EnvJujuClient(JujuData('foo', {}), None, None)
+        client = ModelClient(JujuData('foo', {}), None, None)
         with patch.object(client, 'get_juju_output', return_value=data):
             model_clients = list(client.iter_model_clients())
         self.assertEqual(2, len(model_clients))
@@ -2030,7 +1876,7 @@
                 {'name': 'bar', 'model-uuid': 'bbbb'}],
             'current-model': 'bar'
         }
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_models',
                           return_value=models) as gm_mock:
             controller_name = client.get_controller_model_name()
@@ -2044,13 +1890,13 @@
                 {'name': 'baz', 'model-uuid': 'bbbb'}],
             'current-model': 'bar'
         }
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_models', return_value=models):
             controller_name = client.get_controller_model_name()
         self.assertEqual('controller', controller_name)
 
     def test_get_controller_model_name_no_models(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_models', return_value={}):
             controller_name = client.get_controller_model_name()
         self.assertEqual('controller', controller_name)
@@ -2076,7 +1922,7 @@
               access: admin
               last-connection: just now
             """.format(uuid=model_uuid))
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_juju_output') as m_get_juju_output:
             m_get_juju_output.return_value = yaml_string
             self.assertEqual(
@@ -2109,7 +1955,7 @@
               access: admin
               last-connection: just now""".format(model=controller_model_uuid,
                                                   controller=controller_uuid))
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_juju_output') as m_get_juju_output:
             m_get_juju_output.return_value = yaml_string
             self.assertEqual(
@@ -2142,7 +1988,7 @@
             cloud: lxd
             cloud-type: lxd
             region: localhost""".format(uuid=controller_uuid))
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_juju_output') as m_get_juju_output:
             m_get_juju_output.return_value = yaml_string
             self.assertEqual(
@@ -2153,7 +1999,7 @@
                 'show-controller', '--format', 'yaml', 'foo', include_e=False)
 
     def test_get_controller_client(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('foo', {'bar': 'baz'}, 'myhome'), None, None)
         controller_client = client.get_controller_client()
         controller_env = controller_client.env
@@ -2162,7 +2008,7 @@
             {'bar': 'baz', 'name': 'controller'}, controller_env._config)
 
     def test_list_controllers(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'juju') as j_mock:
             client.list_controllers()
         j_mock.assert_called_once_with('list-controllers', (), include_e=False)
@@ -2173,11 +2019,11 @@
             details:
               api-endpoints: ['10.0.0.1:17070', '10.0.0.2:17070']
         """
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_juju_output',
                           return_value=data) as gjo_mock:
             endpoint = client.get_controller_endpoint()
-        self.assertEqual('10.0.0.1', endpoint)
+        self.assertEqual(('10.0.0.1', '17070'), endpoint)
         gjo_mock.assert_called_once_with(
             'show-controller', 'foo', include_e=False)
 
@@ -2187,11 +2033,11 @@
             details:
               api-endpoints: ['[::1]:17070', '[fe80::216:3eff:0:9dc7]:17070']
         """
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_juju_output',
                           return_value=data) as gjo_mock:
             endpoint = client.get_controller_endpoint()
-        self.assertEqual('::1', endpoint)
+        self.assertEqual(('::1', '17070'), endpoint)
         gjo_mock.assert_called_once_with(
             'show-controller', 'foo', include_e=False)
 
@@ -2201,7 +2047,7 @@
             details:
               api-endpoints: ['[::1]:17070', '[fe80::216:3eff:0:9dc7]:17070']
         """
-        client = EnvJujuClient(JujuData('foo', {}), None, None)
+        client = ModelClient(JujuData('foo', {}), None, None)
         controller_client = client.get_controller_client()
         client.env.controller.name = 'bar'
         with patch.object(controller_client, 'get_juju_output',
@@ -2209,7 +2055,7 @@
             endpoint = controller_client.get_controller_endpoint()
         gjo.assert_called_once_with('show-controller', 'bar',
                                     include_e=False)
-        self.assertEqual('::1', endpoint)
+        self.assertEqual(('::1', '17070'), endpoint)
 
     def test_get_controller_members(self):
         status = Status.from_text("""\
@@ -2231,11 +2077,11 @@
                 instance-id: juju-dddd-machine-3
                 controller-member-status: has-vote
         """)
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_status', autospec=True,
                           return_value=status):
             with patch.object(client, 'get_controller_endpoint', autospec=True,
-                              return_value='10.0.0.3') as gce_mock:
+                              return_value=('10.0.0.3', '17070')) as gce_mock:
                 with patch.object(client, 'get_controller_member_status',
                                   wraps=client.get_controller_member_status,
                                   ) as gcms_mock:
@@ -2270,7 +2116,7 @@
                 instance-id: juju-aaaa-machine-0
                 controller-member-status: has-vote
         """)
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_status', autospec=True,
                           return_value=status):
             with patch.object(client, 'get_controller_endpoint') as gce_mock:
@@ -2291,14 +2137,14 @@
             Machine('0', {}),
             Machine('2', {}),
         ]
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_controller_members', autospec=True,
                           return_value=members):
             leader = client.get_controller_leader()
         self.assertEqual(Machine('3', {}), leader)
 
     def make_controller_client(self):
-        client = EnvJujuClient(JujuData('local', {'name': 'test'}), None, None)
+        client = ModelClient(JujuData('local', {'name': 'test'}), None, None)
         return client.get_controller_client()
 
     def test_wait_for_ha(self):
@@ -2320,7 +2166,7 @@
         client = self.make_controller_client()
         with patch.object(client, 'get_juju_output', return_value=value):
             writes = []
-            with patch('jujupy.until_timeout', autospec=True,
+            with patch('jujupy.client.until_timeout', autospec=True,
                        return_value=[2, 1]):
                 with patch.object(GroupReporter, '_write', autospec=True,
                                   side_effect=lambda _, s: writes.append(s)):
@@ -2342,7 +2188,8 @@
         })
         client = self.make_controller_client()
         status = client.status_class.from_text(value)
-        with patch('jujupy.until_timeout', lambda x, start=None: range(0)):
+        with patch('jujupy.client.until_timeout',
+                   lambda x, start=None: range(0)):
             with patch.object(client, 'get_status', return_value=status
                               ) as get_status_mock:
                 with self.assertRaisesRegexp(
@@ -2360,7 +2207,8 @@
             'services': {},
         })
         client = self.make_controller_client()
-        with patch('jujupy.until_timeout', autospec=True, return_value=[2, 1]):
+        with patch('jujupy.client.until_timeout', autospec=True,
+                   return_value=[2, 1]):
             with patch.object(client, 'get_juju_output', return_value=value):
                 with self.assertRaisesRegexp(
                         ErroredUnit, '1 is in state error: foo'):
@@ -2390,7 +2238,7 @@
                 }
             }
         })
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         with patch.object(client, 'get_juju_output', return_value=value):
             client.wait_for_deploy_started()
 
@@ -2401,8 +2249,8 @@
             },
             'applications': {},
         })
-        client = EnvJujuClient(JujuData('local'), None, None)
-        with patch('jujupy.until_timeout', lambda x: range(0)):
+        client = ModelClient(JujuData('local'), None, None)
+        with patch('jujupy.client.until_timeout', lambda x: range(0)):
             with patch.object(client, 'get_juju_output', return_value=value):
                 with self.assertRaisesRegexp(
                         StatusNotMet,
@@ -2436,15 +2284,16 @@
 
     def test_wait_for_version(self):
         value = self.make_status_yaml('agent-version', '1.17.2', '1.17.2')
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         with patch.object(client, 'get_juju_output', return_value=value):
             client.wait_for_version('1.17.2')
 
     def test_wait_for_version_timeout(self):
         value = self.make_status_yaml('agent-version', '1.17.2', '1.17.1')
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         writes = []
-        with patch('jujupy.until_timeout', lambda x, start=None: [x]):
+        with patch('jujupy.client.until_timeout',
+                   lambda x, start=None: [x]):
             with patch.object(client, 'get_juju_output', return_value=value):
                 with patch.object(GroupReporter, '_write', autospec=True,
                                   side_effect=lambda _, s: writes.append(s)):
@@ -2467,7 +2316,7 @@
             else:
                 return action
 
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         with patch.object(client, 'get_juju_output', get_juju_output_fake):
             client.wait_for_version('1.17.2')
 
@@ -2483,7 +2332,7 @@
             else:
                 return action
 
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         with patch.object(client, 'get_juju_output', get_juju_output_fake):
             with self.assertRaisesRegexp(Exception, 'foo'):
                 client.wait_for_version('1.17.2')
@@ -2494,9 +2343,9 @@
                 '0': {'agent-state': 'started'},
             },
         })
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         with patch.object(client, 'get_juju_output', return_value=value):
-            client.wait_for([WaitMachineNotPresent('1')])
+            client.wait_for(WaitMachineNotPresent('1'), quiet=True)
 
     def test_wait_just_machine_0_timeout(self):
         value = yaml.safe_dump({
@@ -2505,34 +2354,71 @@
                 '1': {'agent-state': 'started'},
             },
         })
-        client = EnvJujuClient(JujuData('local'), None, None)
+        client = ModelClient(JujuData('local'), None, None)
         with patch.object(client, 'get_juju_output', return_value=value), \
-            patch('jujupy.until_timeout', lambda x: range(0)), \
+            patch('jujupy.client.until_timeout',
+                  lambda x, start=None: range(1)), \
             self.assertRaisesRegexp(
                 Exception,
                 'Timed out waiting for machine removal 1'):
-            client.wait_for([WaitMachineNotPresent('1')])
+            client.wait_for(WaitMachineNotPresent('1'), quiet=True)
 
     class NeverSatisfied:
 
+        timeout = 1234
+
+        already_satisfied = False
+
         class NeverSatisfiedException(Exception):
             pass
 
-        def is_satisfied(self, ignored):
-            return False
+        def iter_blocking_state(self, ignored):
+            yield ('global state', 'unsatisfied')
 
-        def do_raise(self):
+        def do_raise(self, model_name, status):
             raise self.NeverSatisfiedException()
 
     def test_wait_timeout(self):
         client = fake_juju_client()
         client.bootstrap()
-
         never_satisfied = self.NeverSatisfied()
         with self.assertRaises(never_satisfied.NeverSatisfiedException):
-            with patch.object(client, 'status_until', lambda timeout: iter(
-                    [Status({}, '')])):
-                client.wait_for([never_satisfied])
+            with patch.object(client, 'status_until', return_value=iter(
+                    [Status({}, '')])) as mock_su:
+                client.wait_for(never_satisfied, quiet=True)
+        mock_su.assert_called_once_with(1234)
+
+    def test_wait_for_emits_output(self):
+        client = fake_juju_client()
+        client.bootstrap()
+        mock_wait = Mock(timeout=300, already_satisfied=False)
+        mock_wait.iter_blocking_state.side_effect = [
+            [('0', 'still-present')],
+            [('0', 'still-present')],
+            [('0', 'still-present')],
+            [],
+            ]
+        writes = []
+        with patch.object(GroupReporter, '_write', autospec=True,
+                          side_effect=lambda _, s: writes.append(s)):
+            client.wait_for(mock_wait)
+        self.assertEqual('still-present: 0 ..\n', ''.join(writes))
+
+    def test_wait_for_quiet(self):
+        client = fake_juju_client()
+        client.bootstrap()
+        mock_wait = Mock(timeout=300)
+        mock_wait.iter_blocking_state.side_effect = [
+            [('0', 'still-present')],
+            [('0', 'still-present')],
+            [('0', 'still-present')],
+            [],
+            ]
+        writes = []
+        with patch.object(GroupReporter, '_write', autospec=True,
+                          side_effect=lambda _, s: writes.append(s)):
+            client.wait_for(mock_wait, quiet=True)
+        self.assertEqual('', ''.join(writes))
 
     def test_wait_bad_status(self):
         client = fake_juju_client()
@@ -2545,7 +2431,7 @@
         with self.assertRaises(MachineError):
             with patch.object(client, 'status_until', lambda timeout: iter(
                     [bad_status])):
-                client.wait_for([never_satisfied])
+                client.wait_for(never_satisfied, quiet=True)
 
     def test_wait_bad_status_recoverable_recovered(self):
         client = fake_juju_client()
@@ -2559,7 +2445,7 @@
         with self.assertRaises(never_satisfied.NeverSatisfiedException):
             with patch.object(client, 'status_until', lambda timeout: iter(
                     [bad_status, good_status])):
-                client.wait_for([never_satisfied])
+                client.wait_for(never_satisfied, quiet=True)
 
     def test_wait_bad_status_recoverable_timed_out(self):
         client = fake_juju_client()
@@ -2572,17 +2458,18 @@
         with self.assertRaises(AppError):
             with patch.object(client, 'status_until', lambda timeout: iter(
                     [bad_status])):
-                client.wait_for([never_satisfied])
+                client.wait_for(never_satisfied, quiet=True)
 
     def test_wait_empty_list(self):
         client = fake_juju_client()
         client.bootstrap()
         with patch.object(client, 'status_until', side_effect=StatusTimeout):
-            self.assertEqual(client.wait_for([]).status,
-                             client.get_status().status)
+            self.assertEqual(client.wait_for(
+                ConditionList([]), quiet=True).status,
+                client.get_status().status)
 
     def test_set_model_constraints(self):
-        client = EnvJujuClient(JujuData('bar', {}), None, '/foo')
+        client = ModelClient(JujuData('bar', {}), None, '/foo')
         with patch.object(client, 'juju') as juju_mock:
             client.set_model_constraints({'bar': 'baz'})
         juju_mock.assert_called_once_with('set-model-constraints',
@@ -2591,7 +2478,7 @@
     def test_get_model_config(self):
         env = JujuData('foo', None)
         fake_popen = FakePopen(yaml.safe_dump({'bar': 'baz'}), None, 0)
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         with patch('subprocess.Popen', return_value=fake_popen) as po_mock:
             result = client.get_model_config()
         assert_juju_call(
@@ -2603,7 +2490,7 @@
     def test_get_env_option(self):
         env = JujuData('foo', None)
         fake_popen = FakePopen('https://example.org/juju/tools', None, 0)
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         with patch('subprocess.Popen', return_value=fake_popen) as mock:
             result = client.get_env_option('tools-metadata-url')
         self.assertEqual(
@@ -2614,7 +2501,7 @@
 
     def test_set_env_option(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         with patch('subprocess.check_call') as mock:
             client.set_env_option(
                 'tools-metadata-url', 'https://example.org/juju/tools')
@@ -2626,7 +2513,7 @@
 
     def test_unset_env_option(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         with patch('subprocess.check_call') as mock:
             client.unset_env_option('tools-metadata-url')
         environ = dict(os.environ)
@@ -2635,9 +2522,67 @@
             ('juju', '--show-log', 'model-config', '-m', 'foo:foo',
              '--reset', 'tools-metadata-url'))
 
+    def test__format_cloud_region(self):
+        fcr = ModelClient._format_cloud_region
+        self.assertEqual(('aws/us-east-1',), fcr('aws', 'us-east-1'))
+        self.assertEqual(('us-east-1',), fcr(region='us-east-1'))
+        self.assertRaises(ValueError, fcr, cloud='aws')
+        self.assertEqual((), fcr())
+
+    def test_get_model_defaults(self):
+        data = {'some-key': {'default': 'black'}}
+        raw_yaml = yaml.safe_dump(data)
+        client = fake_juju_client()
+        with patch.object(client, 'get_juju_output', autospec=True,
+                          return_value=raw_yaml) as output_mock:
+            retval = client.get_model_defaults('some-key')
+        self.assertEqual(data, retval)
+        output_mock.assert_called_once_with(
+            'model-defaults', '--format', 'yaml', 'some-key', include_e=False)
+
+    def test_get_model_defaults_cloud_region(self):
+        raw_yaml = yaml.safe_dump({'some-key': {'default': 'red'}})
+        client = fake_juju_client()
+        with patch.object(client, 'get_juju_output', autospec=True,
+                          return_value=raw_yaml) as output_mock:
+            client.get_model_defaults('some-key', region='us-east-1')
+        output_mock.assert_called_once_with(
+            'model-defaults', '--format', 'yaml', 'us-east-1', 'some-key',
+            include_e=False)
+
+    def test_set_model_defaults(self):
+        client = fake_juju_client()
+        with patch.object(client, 'juju', autospec=True) as juju_mock:
+            client.set_model_defaults('some-key', 'white')
+        juju_mock.assert_called_once_with(
+            'model-defaults', ('some-key=white',), include_e=False)
+
+    def test_set_model_defaults_cloud_region(self):
+        client = fake_juju_client()
+        with patch.object(client, 'juju', autospec=True) as juju_mock:
+            client.set_model_defaults('some-key', 'white', region='us-east-1')
+        juju_mock.assert_called_once_with(
+            'model-defaults', ('us-east-1', 'some-key=white',),
+            include_e=False)
+
+    def test_unset_model_defaults(self):
+        client = fake_juju_client()
+        with patch.object(client, 'juju', autospec=True) as juju_mock:
+            client.unset_model_defaults('some-key')
+        juju_mock.assert_called_once_with(
+            'model-defaults', ('--reset', 'some-key'), include_e=False)
+
+    def test_unset_model_defaults_cloud_region(self):
+        client = fake_juju_client()
+        with patch.object(client, 'juju', autospec=True) as juju_mock:
+            client.unset_model_defaults('some-key', region='us-east-1')
+        juju_mock.assert_called_once_with(
+            'model-defaults', ('us-east-1', '--reset', 'some-key'),
+            include_e=False)
+
     def test_set_testing_agent_metadata_url(self):
         env = JujuData(None, {'type': 'foo'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         with patch.object(client, 'get_env_option') as mock_get:
             mock_get.return_value = 'https://example.org/juju/tools'
             with patch.object(client, 'set_env_option') as mock_set:
@@ -2649,7 +2594,7 @@
 
     def test_set_testing_agent_metadata_url_noop(self):
         env = JujuData(None, {'type': 'foo'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         with patch.object(client, 'get_env_option') as mock_get:
             mock_get.return_value = 'https://example.org/juju/testing/tools'
             with patch.object(client, 'set_env_option') as mock_set:
@@ -2659,7 +2604,7 @@
 
     def test_juju(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         with patch('subprocess.check_call') as mock:
             client.juju('foo', ('bar', 'baz'))
         environ = dict(os.environ)
@@ -2669,7 +2614,7 @@
 
     def test_expect_returns_pexpect_spawn_object(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         with patch('pexpect.spawn') as mock:
             process = client.expect('foo', ('bar', 'baz'))
 
@@ -2679,7 +2624,7 @@
     def test_expect_uses_provided_envvar_path(self):
         from pexpect import ExceptionPexpect
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
 
         with temp_dir() as empty_path:
             broken_envvars = dict(PATH=empty_path)
@@ -2691,7 +2636,7 @@
 
     def test_juju_env(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
 
         def check_path(*args, **kwargs):
             self.assertRegexpMatches(os.environ['PATH'], r'/foobar\:')
@@ -2700,7 +2645,7 @@
 
     def test_juju_no_check(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         environ = dict(os.environ)
         environ['JUJU_HOME'] = client.env.juju_home
         with patch('subprocess.call') as mock:
@@ -2710,7 +2655,7 @@
 
     def test_juju_no_check_env(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
 
         def check_path(*args, **kwargs):
             self.assertRegexpMatches(os.environ['PATH'], r'/foobar\:')
@@ -2719,7 +2664,7 @@
 
     def test_juju_timeout(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
         with patch('subprocess.check_call') as cc_mock:
             client.juju('foo', ('bar', 'baz'), timeout=58)
         self.assertEqual(cc_mock.call_args[0][0], (
@@ -2729,7 +2674,7 @@
     def test_juju_juju_home(self):
         env = JujuData('qux')
         os.environ['JUJU_HOME'] = 'foo'
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
 
         def check_home(*args, **kwargs):
             self.assertEqual(os.environ['JUJU_HOME'], 'foo')
@@ -2744,7 +2689,7 @@
 
     def test_juju_extra_env(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, 'juju')
+        client = ModelClient(env, None, 'juju')
         extra_env = {'JUJU': '/juju', 'JUJU_HOME': client.env.juju_home}
 
         def check_env(*args, **kwargs):
@@ -2758,7 +2703,7 @@
 
     def test_juju_backup_with_tgz(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
 
         with patch(
                 'subprocess.Popen',
@@ -2771,7 +2716,7 @@
 
     def test_juju_backup_with_tar_gz(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
         with patch('subprocess.Popen',
                    return_value=FakePopen(
                        'foojuju-backup-123-456.tar.gzbar', '', 0)):
@@ -2781,7 +2726,7 @@
 
     def test_juju_backup_no_file(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
         with patch('subprocess.Popen', return_value=FakePopen('', '', 0)):
             with self.assertRaisesRegexp(
                     Exception, 'The backup file was not found in output'):
@@ -2789,7 +2734,7 @@
 
     def test_juju_backup_wrong_file(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
         with patch('subprocess.Popen',
                    return_value=FakePopen('mumu-backup-24.tgz', '', 0)):
             with self.assertRaisesRegexp(
@@ -2798,7 +2743,7 @@
 
     def test_juju_backup_environ(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
         environ = client._shell_environ()
 
         def side_effect(*args, **kwargs):
@@ -2810,7 +2755,7 @@
 
     def test_restore_backup(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
         with patch.object(client, 'juju') as gjo_mock:
             client.restore_backup('quxx')
         gjo_mock.assert_called_once_with(
@@ -2819,7 +2764,7 @@
 
     def test_restore_backup_async(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
         with patch.object(client, 'juju_async') as gjo_mock:
             result = client.restore_backup_async('quxx')
         gjo_mock.assert_called_once_with('restore-backup', (
@@ -2828,7 +2773,7 @@
 
     def test_enable_ha(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
         with patch.object(client, 'juju', autospec=True) as eha_mock:
             client.enable_ha()
         eha_mock.assert_called_once_with(
@@ -2836,7 +2781,7 @@
 
     def test_juju_async(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
         with patch('subprocess.Popen') as popen_class_mock:
             with client.juju_async('foo', ('bar', 'baz')) as proc:
                 assert_juju_call(
@@ -2852,7 +2797,7 @@
 
     def test_juju_async_failure(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
         with patch('subprocess.Popen') as popen_class_mock:
             with self.assertRaises(subprocess.CalledProcessError) as err_cxt:
                 with client.juju_async('foo', ('bar', 'baz')):
@@ -2864,7 +2809,7 @@
 
     def test_juju_async_environ(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
         environ = client._shell_environ()
         proc_mock = Mock()
         with patch('subprocess.Popen') as popen_class_mock:
@@ -2879,10 +2824,10 @@
             self.assertNotEqual(environ, os.environ)
 
     def test_is_jes_enabled(self):
-        # EnvJujuClient knows that JES is always enabled, and doesn't need to
+        # ModelClient knows that JES is always enabled, and doesn't need to
         # shell out.
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
         fake_popen = FakePopen(' %s' % SYSTEM, None, 0)
         with patch('subprocess.Popen',
                    return_value=fake_popen) as po_mock:
@@ -2891,7 +2836,7 @@
 
     def test_get_jes_command(self):
         env = JujuData('qux')
-        client = EnvJujuClient(env, None, '/foobar/baz')
+        client = ModelClient(env, None, '/foobar/baz')
         # Juju 1.24 and older do not have a JES command. It is an error
         # to call get_jes_command when is_jes_enabled is False
         fake_popen = FakePopen(' %s' % SYSTEM, None, 0)
@@ -2902,7 +2847,7 @@
 
     def test_get_juju_timings(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, 'my/juju/bin')
+        client = ModelClient(env, None, 'my/juju/bin')
         client._backend.juju_timings = {("juju", "op1"): [1],
                                         ("juju", "op2"): [2]}
         flattened_timings = client.get_juju_timings()
@@ -2910,9 +2855,9 @@
         self.assertEqual(flattened_timings, expected)
 
     def test_deployer(self):
-        client = EnvJujuClient(JujuData('foo', {'type': 'local'}),
-                               '1.23-series-arch', None)
-        with patch.object(EnvJujuClient, 'juju') as mock:
+        client = ModelClient(JujuData('foo', {'type': 'local'}),
+                             '1.23-series-arch', None)
+        with patch.object(ModelClient, 'juju') as mock:
             client.deployer('bundle:~juju-qa/some-bundle')
         mock.assert_called_with(
             'deployer', ('-e', 'foo:foo', '--debug', '--deploy-delay',
@@ -2921,9 +2866,9 @@
             include_e=False)
 
     def test_deployer_with_bundle_name(self):
-        client = EnvJujuClient(JujuData('foo', {'type': 'local'}),
-                               '2.0.0-series-arch', None)
-        with patch.object(EnvJujuClient, 'juju') as mock:
+        client = ModelClient(JujuData('foo', {'type': 'local'}),
+                             '2.0.0-series-arch', None)
+        with patch.object(ModelClient, 'juju') as mock:
             client.deployer('bundle:~juju-qa/some-bundle', 'name')
         mock.assert_called_with(
             'deployer', ('-e', 'foo:foo', '--debug', '--deploy-delay',
@@ -2932,9 +2877,9 @@
             include_e=False)
 
     def test_quickstart_maas(self):
-        client = EnvJujuClient(JujuData(None, {'type': 'maas'}),
-                               '1.23-series-arch', '/juju')
-        with patch.object(EnvJujuClient, 'juju') as mock:
+        client = ModelClient(JujuData(None, {'type': 'maas'}),
+                             '1.23-series-arch', '/juju')
+        with patch.object(ModelClient, 'juju') as mock:
             client.quickstart('bundle:~juju-qa/some-bundle')
         mock.assert_called_with(
             'quickstart', ('--constraints', 'mem=2G', '--no-browser',
@@ -2942,9 +2887,9 @@
             extra_env={'JUJU': '/juju'})
 
     def test_quickstart_local(self):
-        client = EnvJujuClient(JujuData(None, {'type': 'local'}),
-                               '1.23-series-arch', '/juju')
-        with patch.object(EnvJujuClient, 'juju') as mock:
+        client = ModelClient(JujuData(None, {'type': 'local'}),
+                             '1.23-series-arch', '/juju')
+        with patch.object(ModelClient, 'juju') as mock:
             client.quickstart('bundle:~juju-qa/some-bundle')
         mock.assert_called_with(
             'quickstart', ('--constraints', 'mem=2G', '--no-browser',
@@ -2952,9 +2897,9 @@
             extra_env={'JUJU': '/juju'})
 
     def test_quickstart_template(self):
-        client = EnvJujuClient(JujuData(None, {'type': 'local'}),
-                               '1.23-series-arch', '/juju')
-        with patch.object(EnvJujuClient, 'juju') as mock:
+        client = ModelClient(JujuData(None, {'type': 'local'}),
+                             '1.23-series-arch', '/juju')
+        with patch.object(ModelClient, 'juju') as mock:
             client.quickstart('bundle:~juju-qa/some-{container}-bundle')
         mock.assert_called_with(
             'quickstart', ('--constraints', 'mem=2G', '--no-browser',
@@ -2962,9 +2907,9 @@
             extra_env={'JUJU': '/juju'})
 
     def test_action_do(self):
-        client = EnvJujuClient(JujuData(None, {'type': 'local'}),
-                               '1.23-series-arch', None)
-        with patch.object(EnvJujuClient, 'get_juju_output') as mock:
+        client = ModelClient(JujuData(None, {'type': 'local'}),
+                             '1.23-series-arch', None)
+        with patch.object(ModelClient, 'get_juju_output') as mock:
             mock.return_value = \
                 "Action queued with id: 5a92ec93-d4be-4399-82dc-7431dbfd08f9"
             id = client.action_do("foo/0", "myaction", "param=5")
@@ -2974,18 +2919,18 @@
         )
 
     def test_action_do_error(self):
-        client = EnvJujuClient(JujuData(None, {'type': 'local'}),
-                               '1.23-series-arch', None)
-        with patch.object(EnvJujuClient, 'get_juju_output') as mock:
+        client = ModelClient(JujuData(None, {'type': 'local'}),
+                             '1.23-series-arch', None)
+        with patch.object(ModelClient, 'get_juju_output') as mock:
             mock.return_value = "some bad text"
             with self.assertRaisesRegexp(Exception,
                                          "Action id not found in output"):
                 client.action_do("foo/0", "myaction", "param=5")
 
     def test_action_fetch(self):
-        client = EnvJujuClient(JujuData(None, {'type': 'local'}),
-                               '1.23-series-arch', None)
-        with patch.object(EnvJujuClient, 'get_juju_output') as mock:
+        client = ModelClient(JujuData(None, {'type': 'local'}),
+                             '1.23-series-arch', None)
+        with patch.object(ModelClient, 'get_juju_output') as mock:
             ret = "status: completed\nfoo: bar"
             mock.return_value = ret
             out = client.action_fetch("123")
@@ -2995,19 +2940,19 @@
         )
 
     def test_action_fetch_timeout(self):
-        client = EnvJujuClient(JujuData(None, {'type': 'local'}),
-                               '1.23-series-arch', None)
+        client = ModelClient(JujuData(None, {'type': 'local'}),
+                             '1.23-series-arch', None)
         ret = "status: pending\nfoo: bar"
-        with patch.object(EnvJujuClient,
+        with patch.object(ModelClient,
                           'get_juju_output', return_value=ret):
             with self.assertRaisesRegexp(Exception,
                                          "timed out waiting for action"):
                 client.action_fetch("123")
 
     def test_action_do_fetch(self):
-        client = EnvJujuClient(JujuData(None, {'type': 'local'}),
-                               '1.23-series-arch', None)
-        with patch.object(EnvJujuClient, 'get_juju_output') as mock:
+        client = ModelClient(JujuData(None, {'type': 'local'}),
+                             '1.23-series-arch', None)
+        with patch.object(ModelClient, 'get_juju_output') as mock:
             ret = "status: completed\nfoo: bar"
             # setting side_effect to an iterable will return the next value
             # from the list each time the function is called.
@@ -3018,7 +2963,7 @@
             self.assertEqual(out, ret)
 
     def test_run(self):
-        client = fake_juju_client(cls=EnvJujuClient)
+        client = fake_juju_client(cls=ModelClient)
         run_list = [
             {"MachineId": "1",
              "Stdout": "Linux\n",
@@ -3035,7 +2980,7 @@
             'name:name', user_name=None)
 
     def test_run_machines(self):
-        client = fake_juju_client(cls=EnvJujuClient)
+        client = fake_juju_client(cls=ModelClient)
         output = json.dumps({"ReturnCode": 255})
         with patch.object(client, 'get_juju_output',
                           return_value=output) as output_mock:
@@ -3044,15 +2989,15 @@
             'run', '--format', 'json', '--machine', '0,1,2', 'true')
 
     def test_run_use_json_false(self):
-        client = fake_juju_client(cls=EnvJujuClient)
+        client = fake_juju_client(cls=ModelClient)
         output = json.dumps({"ReturnCode": 255})
         with patch.object(client, 'get_juju_output', return_value=output):
             result = client.run(['true'], use_json=False)
         self.assertEqual(output, result)
 
     def test_list_space(self):
-        client = EnvJujuClient(JujuData(None, {'type': 'local'}),
-                               '1.23-series-arch', None)
+        client = ModelClient(JujuData(None, {'type': 'local'}),
+                             '1.23-series-arch', None)
         yaml_dict = {'foo': 'bar'}
         output = yaml.safe_dump(yaml_dict)
         with patch.object(client, 'get_juju_output', return_value=output,
@@ -3062,28 +3007,28 @@
         gjo_mock.assert_called_once_with('list-space')
 
     def test_add_space(self):
-        client = EnvJujuClient(JujuData(None, {'type': 'local'}),
-                               '1.23-series-arch', None)
+        client = ModelClient(JujuData(None, {'type': 'local'}),
+                             '1.23-series-arch', None)
         with patch.object(client, 'juju', autospec=True) as juju_mock:
             client.add_space('foo-space')
         juju_mock.assert_called_once_with('add-space', ('foo-space'))
 
     def test_add_subnet(self):
-        client = EnvJujuClient(JujuData(None, {'type': 'local'}),
-                               '1.23-series-arch', None)
+        client = ModelClient(JujuData(None, {'type': 'local'}),
+                             '1.23-series-arch', None)
         with patch.object(client, 'juju', autospec=True) as juju_mock:
             client.add_subnet('bar-subnet', 'foo-space')
         juju_mock.assert_called_once_with('add-subnet',
                                           ('bar-subnet', 'foo-space'))
 
     def test__shell_environ_uses_pathsep(self):
-        client = EnvJujuClient(JujuData('foo'), None, 'foo/bar/juju')
+        client = ModelClient(JujuData('foo'), None, 'foo/bar/juju')
         with patch('os.pathsep', '!'):
             environ = client._shell_environ()
         self.assertRegexpMatches(environ['PATH'], r'foo/bar\!')
 
     def test_set_config(self):
-        client = EnvJujuClient(JujuData('bar', {}), None, '/foo')
+        client = ModelClient(JujuData('bar', {}), None, '/foo')
         with patch.object(client, 'juju') as juju_mock:
             client.set_config('foo', {'bar': 'baz'})
         juju_mock.assert_called_once_with('config', ('foo', 'bar=baz'))
@@ -3103,7 +3048,7 @@
                 }
             })
         expected = yaml.safe_load(output())
-        client = EnvJujuClient(JujuData('bar', {}), None, '/foo')
+        client = ModelClient(JujuData('bar', {}), None, '/foo')
         with patch.object(client, 'get_juju_output',
                           side_effect=output) as gjo_mock:
             results = client.get_config('foo')
@@ -3125,43 +3070,43 @@
                 }
             })
         expected = yaml.safe_load(output())
-        client = EnvJujuClient(JujuData('bar', {}), None, '/foo')
+        client = ModelClient(JujuData('bar', {}), None, '/foo')
         with patch.object(client, 'get_juju_output', side_effect=output):
             results = client.get_service_config('foo')
         self.assertEqual(expected, results)
 
     def test_get_service_config_timesout(self):
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo')
-        with patch('jujupy.until_timeout', return_value=range(0)):
+        client = ModelClient(JujuData('foo', {}), None, '/foo')
+        with patch('jujupy.client.until_timeout', return_value=range(0)):
             with self.assertRaisesRegexp(
                     Exception, 'Timed out waiting for juju get'):
                 client.get_service_config('foo')
 
     def test_upgrade_mongo(self):
-        client = EnvJujuClient(JujuData('bar', {}), None, '/foo')
+        client = ModelClient(JujuData('bar', {}), None, '/foo')
         with patch.object(client, 'juju') as juju_mock:
             client.upgrade_mongo()
         juju_mock.assert_called_once_with('upgrade-mongo', ())
 
     def test_enable_feature(self):
-        client = EnvJujuClient(JujuData('bar', {}), None, '/foo')
+        client = ModelClient(JujuData('bar', {}), None, '/foo')
         self.assertEqual(set(), client.feature_flags)
         client.enable_feature('actions')
         self.assertEqual(set(['actions']), client.feature_flags)
 
     def test_enable_feature_invalid(self):
-        client = EnvJujuClient(JujuData('bar', {}), None, '/foo')
+        client = ModelClient(JujuData('bar', {}), None, '/foo')
         self.assertEqual(set(), client.feature_flags)
         with self.assertRaises(ValueError) as ctx:
             client.enable_feature('nomongo')
         self.assertEqual(str(ctx.exception), "Unknown feature flag: 'nomongo'")
 
     def test_is_juju1x(self):
-        client = EnvJujuClient(None, '1.25.5', None)
+        client = ModelClient(None, '1.25.5', None)
         self.assertTrue(client.is_juju1x())
 
     def test_is_juju1x_false(self):
-        client = EnvJujuClient(None, '2.0.0', None)
+        client = ModelClient(None, '2.0.0', None)
         self.assertFalse(client.is_juju1x())
 
     def test__get_register_command_returns_register_token(self):
@@ -3263,7 +3208,7 @@
     def test_disable_user(self):
         env = JujuData('foo')
         username = 'fakeuser'
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         with patch.object(client, 'juju') as mock:
             client.disable_user(username)
         mock.assert_called_with(
@@ -3272,7 +3217,7 @@
     def test_enable_user(self):
         env = JujuData('foo')
         username = 'fakeuser'
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         with patch.object(client, 'juju') as mock:
             client.enable_user(username)
         mock.assert_called_with(
@@ -3280,7 +3225,7 @@
 
     def test_logout(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         with patch.object(client, 'juju') as mock:
             client.logout()
         mock.assert_called_with(
@@ -3315,7 +3260,7 @@
 
     def test_list_clouds(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         with patch.object(client, 'get_juju_output') as mock:
             client.list_clouds()
         mock.assert_called_with(
@@ -3330,12 +3275,27 @@
         client.add_cloud_interactive('foo', clouds['foo'])
         self.assertEqual(client._backend.clouds, clouds)
 
+    def test_add_cloud_interactive_maas_invalid_endpoint(self):
+        client = fake_juju_client()
+        clouds = {'foo': {
+            'type': 'maas',
+            'endpoint': 'B' * 4000,
+            }}
+        with self.assertRaises(InvalidEndpoint):
+            client.add_cloud_interactive('foo', clouds['foo'])
+
     def test_add_cloud_interactive_manual(self):
         client = fake_juju_client()
         clouds = {'foo': {'type': 'manual', 'endpoint': '127.100.100.1'}}
         client.add_cloud_interactive('foo', clouds['foo'])
         self.assertEqual(client._backend.clouds, clouds)
 
+    def test_add_cloud_interactive_manual_invalid_endpoint(self):
+        client = fake_juju_client()
+        clouds = {'foo': {'type': 'manual', 'endpoint': 'B' * 4000}}
+        with self.assertRaises(InvalidEndpoint):
+            client.add_cloud_interactive('foo', clouds['foo'])
+
     def get_openstack_clouds(self):
         return {'foo': {
             'type': 'openstack',
@@ -3353,6 +3313,20 @@
         client.add_cloud_interactive('foo', clouds['foo'])
         self.assertEqual(client._backend.clouds, clouds)
 
+    def test_add_cloud_interactive_openstack_invalid_endpoint(self):
+        client = fake_juju_client()
+        clouds = self.get_openstack_clouds()
+        clouds['foo']['endpoint'] = 'B' * 4000
+        with self.assertRaises(InvalidEndpoint):
+            client.add_cloud_interactive('foo', clouds['foo'])
+
+    def test_add_cloud_interactive_openstack_invalid_region_endpoint(self):
+        client = fake_juju_client()
+        clouds = self.get_openstack_clouds()
+        clouds['foo']['regions']['harvey']['endpoint'] = 'B' * 4000
+        with self.assertRaises(InvalidEndpoint):
+            client.add_cloud_interactive('foo', clouds['foo'])
+
     def test_add_cloud_interactive_openstack_invalid_auth(self):
         client = fake_juju_client()
         clouds = self.get_openstack_clouds()
@@ -3387,7 +3361,7 @@
 
     def test_show_controller(self):
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         with patch.object(client, 'get_juju_output') as mock:
             client.show_controller()
         mock.assert_called_with(
@@ -3400,7 +3374,7 @@
             series: xenial
         """
         env = JujuData('foo')
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         with patch.object(client, 'get_juju_output', autospec=True,
                           return_value=output) as mock:
             data = client.show_machine('0')
@@ -3408,7 +3382,7 @@
         self.assertEqual({'machines': {'0': {'series': 'xenial'}}}, data)
 
     def test_ssh_keys(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         given_output = 'ssh keys output'
         with patch.object(client, 'get_juju_output', autospec=True,
                           return_value=given_output) as mock:
@@ -3417,7 +3391,7 @@
         mock.assert_called_once_with('ssh-keys')
 
     def test_ssh_keys_full(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         given_output = 'ssh keys full output'
         with patch.object(client, 'get_juju_output', autospec=True,
                           return_value=given_output) as mock:
@@ -3426,7 +3400,7 @@
         mock.assert_called_once_with('ssh-keys', '--full')
 
     def test_add_ssh_key(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_juju_output', autospec=True,
                           return_value='') as mock:
             output = client.add_ssh_key('ak', 'bk')
@@ -3435,7 +3409,7 @@
             'add-ssh-key', 'ak', 'bk', merge_stderr=True)
 
     def test_remove_ssh_key(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_juju_output', autospec=True,
                           return_value='') as mock:
             output = client.remove_ssh_key('ak', 'bk')
@@ -3444,7 +3418,7 @@
             'remove-ssh-key', 'ak', 'bk', merge_stderr=True)
 
     def test_import_ssh_key(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_juju_output', autospec=True,
                           return_value='') as mock:
             output = client.import_ssh_key('gh:au', 'lp:bu')
@@ -3453,13 +3427,13 @@
             'import-ssh-key', 'gh:au', 'lp:bu', merge_stderr=True)
 
     def test_disable_commands_properties(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         self.assertEqual('destroy-model', client.command_set_destroy_model)
         self.assertEqual('remove-object', client.command_set_remove_object)
         self.assertEqual('all', client.command_set_all)
 
     def test_list_disabled_commands(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'get_juju_output', autospec=True,
                           return_value=dedent("""\
              - command-set: destroy-model
@@ -3473,1703 +3447,43 @@
                                      '--format', 'yaml')
 
     def test_disable_command(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'juju', autospec=True) as mock:
             client.disable_command('all', 'message')
         mock.assert_called_once_with('disable-command', ('all', 'message'))
 
     def test_enable_command(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'juju', autospec=True) as mock:
             client.enable_command('all')
         mock.assert_called_once_with('enable-command', 'all')
 
     def test_sync_tools(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'juju', autospec=True) as mock:
             client.sync_tools()
         mock.assert_called_once_with('sync-tools', ())
 
     def test_sync_tools_local_dir(self):
-        client = EnvJujuClient(JujuData('foo'), None, None)
+        client = ModelClient(JujuData('foo'), None, None)
         with patch.object(client, 'juju', autospec=True) as mock:
             client.sync_tools('/agents')
         mock.assert_called_once_with('sync-tools', ('--local-dir', '/agents'),
                                      include_e=False)
 
-
-class TestEnvJujuClientRC(ClientTest):
-
-    def test_bootstrap(self):
-        env = JujuData('foo', {'type': 'bar', 'region': 'baz'})
-        with observable_temp_file() as config_file:
-            with patch.object(EnvJujuClientRC, 'juju') as mock:
-                client = EnvJujuClientRC(env, '2.0-zeta1', None)
-                client.bootstrap()
-                mock.assert_called_with(
-                    'bootstrap', ('--constraints', 'mem=2G',
-                                  'foo', 'bar/baz',
-                                  '--config', config_file.name,
-                                  '--default-model', 'foo',
-                                  '--agent-version', '2.0'), include_e=False)
-                config_file.seek(0)
-                config = yaml.safe_load(config_file)
-        self.assertEqual({'test-mode': True}, config)
-
-
-class TestEnvJujuClient1X(ClientTest):
-
-    def test_raise_on_juju_data(self):
-        env = JujuData('foo', {'type': 'bar'}, 'baz')
-        with self.assertRaisesRegexp(
-                IncompatibleConfigClass,
-                'JujuData cannot be used with EnvJujuClient1X'):
-            EnvJujuClient1X(env, '1.25', 'full_path')
-
-    def test_no_duplicate_env(self):
-        env = SimpleEnvironment('foo', {})
-        client = EnvJujuClient1X(env, '1.25', 'full_path')
-        self.assertIs(env, client.env)
-
-    def test_not_supported(self):
-        client = EnvJujuClient1X(
-            SimpleEnvironment('foo', {}), '1.25', 'full_path')
-        with self.assertRaises(JESNotSupported):
-            client.add_user_perms('test-user')
-        with self.assertRaises(JESNotSupported):
-            client.grant('test-user', 'read')
-        with self.assertRaises(JESNotSupported):
-            client.revoke('test-user', 'read')
-        with self.assertRaises(JESNotSupported):
-            client.get_model_uuid()
-
-    def test_get_version(self):
-        value = ' 5.6 \n'
-        with patch('subprocess.check_output', return_value=value) as vsn:
-            version = EnvJujuClient1X.get_version()
-        self.assertEqual('5.6', version)
-        vsn.assert_called_with(('juju', '--version'))
-
-    def test_get_version_path(self):
-        with patch('subprocess.check_output', return_value=' 4.3') as vsn:
-            EnvJujuClient1X.get_version('foo/bar/baz')
-        vsn.assert_called_once_with(('foo/bar/baz', '--version'))
-
-    def test_get_matching_agent_version(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', None)
-        self.assertEqual('1.23.1', client.get_matching_agent_version())
-        self.assertEqual('1.23', client.get_matching_agent_version(
-                         no_build=True))
-        client = client.clone(version='1.20-beta1-series-arch')
-        self.assertEqual('1.20-beta1.1', client.get_matching_agent_version())
-
-    def test_upgrade_juju_nonlocal(self):
-        client = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'nonlocal'}), '1.234-76', None)
-        with patch.object(client, 'juju') as juju_mock:
-            client.upgrade_juju()
-        juju_mock.assert_called_with(
-            'upgrade-juju', ('--version', '1.234'))
-
-    def test_upgrade_juju_local(self):
-        client = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
-        with patch.object(client, 'juju') as juju_mock:
-            client.upgrade_juju()
-        juju_mock.assert_called_with(
-            'upgrade-juju', ('--version', '1.234', '--upload-tools',))
-
-    def test_upgrade_juju_no_force_version(self):
-        client = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
-        with patch.object(client, 'juju') as juju_mock:
-            client.upgrade_juju(force_version=False)
-        juju_mock.assert_called_with(
-            'upgrade-juju', ('--upload-tools',))
-
-    def test_upgrade_mongo_exception(self):
-        client = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
-        with self.assertRaises(UpgradeMongoNotSupported):
-            client.upgrade_mongo()
-
-    def test_get_cache_path(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo', juju_home='/foo/'),
-                                 '1.27', 'full/path', debug=True)
-        self.assertEqual('/foo/environments/cache.yaml',
-                         client.get_cache_path())
-
-    def test_bootstrap_maas(self):
-        env = SimpleEnvironment('maas')
-        with patch.object(EnvJujuClient1X, 'juju') as mock:
-            client = EnvJujuClient1X(env, None, None)
-            with patch.object(client.env, 'maas', lambda: True):
-                client.bootstrap()
-        mock.assert_called_once_with('bootstrap', ('--constraints', 'mem=2G'))
-
-    def test_bootstrap_joyent(self):
-        env = SimpleEnvironment('joyent')
-        with patch.object(EnvJujuClient1X, 'juju', autospec=True) as mock:
-            client = EnvJujuClient1X(env, None, None)
-            with patch.object(client.env, 'joyent', lambda: True):
-                client.bootstrap()
-        mock.assert_called_once_with(
-            client, 'bootstrap', ('--constraints', 'mem=2G cpu-cores=1'))
-
-    def test_bootstrap(self):
-        env = SimpleEnvironment('foo')
-        client = EnvJujuClient1X(env, None, None)
-        with patch.object(client, 'juju') as mock:
-            client.bootstrap()
-        mock.assert_called_with('bootstrap', ('--constraints', 'mem=2G'))
-
-    def test_bootstrap_upload_tools(self):
-        env = SimpleEnvironment('foo')
-        client = EnvJujuClient1X(env, None, None)
-        with patch.object(client, 'juju') as mock:
-            client.bootstrap(upload_tools=True)
-        mock.assert_called_with(
-            'bootstrap', ('--upload-tools', '--constraints', 'mem=2G'))
-
-    def test_bootstrap_args(self):
-        env = SimpleEnvironment('foo', {})
-        client = EnvJujuClient1X(env, None, None)
-        with self.assertRaisesRegexp(
-                BootstrapMismatch,
-                '--bootstrap-series angsty does not match default-series:'
-                ' None'):
-            client.bootstrap(bootstrap_series='angsty')
-        env.update_config({
-            'default-series': 'angsty',
-            })
-        with patch.object(client, 'juju') as mock:
-            client.bootstrap(bootstrap_series='angsty')
-        mock.assert_called_with('bootstrap', ('--constraints', 'mem=2G'))
-
-    def test_bootstrap_async(self):
-        env = SimpleEnvironment('foo')
-        with patch.object(EnvJujuClient, 'juju_async', autospec=True) as mock:
-            client = EnvJujuClient1X(env, None, None)
-            client.env.juju_home = 'foo'
-            with client.bootstrap_async():
-                mock.assert_called_once_with(
-                    client, 'bootstrap', ('--constraints', 'mem=2G'))
-
-    def test_bootstrap_async_upload_tools(self):
-        env = SimpleEnvironment('foo')
-        with patch.object(EnvJujuClient, 'juju_async', autospec=True) as mock:
-            client = EnvJujuClient1X(env, None, None)
-            with client.bootstrap_async(upload_tools=True):
-                mock.assert_called_with(
-                    client, 'bootstrap', ('--upload-tools', '--constraints',
-                                          'mem=2G'))
-
-    def test_get_bootstrap_args_bootstrap_series(self):
-        env = SimpleEnvironment('foo', {})
-        client = EnvJujuClient1X(env, None, None)
-        with self.assertRaisesRegexp(
-                BootstrapMismatch,
-                '--bootstrap-series angsty does not match default-series:'
-                ' None'):
-            client.get_bootstrap_args(upload_tools=True,
-                                      bootstrap_series='angsty')
-        env.update_config({'default-series': 'angsty'})
-        args = client.get_bootstrap_args(upload_tools=True,
-                                         bootstrap_series='angsty')
-        self.assertEqual(args, ('--upload-tools', '--constraints', 'mem=2G'))
-
-    def test_create_environment_system(self):
-        self.do_create_environment(
-            'system', 'system create-environment', ('-s', 'foo'))
-
-    def test_create_environment_controller(self):
-        self.do_create_environment(
-            'controller', 'controller create-environment', ('-c', 'foo'))
-
-    def test_create_environment_hypenated_controller(self):
-        self.do_create_environment(
-            'kill-controller', 'create-environment', ('-c', 'foo'))
-
-    def do_create_environment(self, jes_command, create_cmd,
-                              controller_option):
-        controller_client = EnvJujuClient1X(SimpleEnvironment('foo'), '1.26.1',
-                                            None)
-        model_env = SimpleEnvironment('bar', {'type': 'foo'})
-        with patch.object(controller_client, 'get_jes_command',
-                          return_value=jes_command):
-            with patch.object(controller_client, 'juju') as juju_mock:
-                with observable_temp_file() as config_file:
-                    controller_client.add_model(model_env)
-        juju_mock.assert_called_once_with(
-            create_cmd, controller_option + (
-                'bar', '--config', config_file.name), include_e=False)
-
-    def test_destroy_environment(self):
-        env = SimpleEnvironment('foo', {'type': 'ec2'})
-        client = EnvJujuClient1X(env, None, None)
-        with patch.object(client, 'juju') as mock:
-            client.destroy_environment()
-        mock.assert_called_with(
-            'destroy-environment', ('foo', '--force', '-y'),
-            check=False, include_e=False, timeout=600)
-
-    def test_destroy_environment_no_force(self):
-        env = SimpleEnvironment('foo', {'type': 'ec2'})
-        client = EnvJujuClient1X(env, None, None)
-        with patch.object(client, 'juju') as mock:
-            client.destroy_environment(force=False)
-            mock.assert_called_with(
-                'destroy-environment', ('foo', '-y'),
-                check=False, include_e=False, timeout=600)
-
-    def test_destroy_environment_azure(self):
-        env = SimpleEnvironment('foo', {'type': 'azure'})
-        client = EnvJujuClient1X(env, None, None)
-        with patch.object(client, 'juju') as mock:
-            client.destroy_environment(force=False)
-            mock.assert_called_with(
-                'destroy-environment', ('foo', '-y'),
-                check=False, include_e=False, timeout=1800)
-
-    def test_destroy_environment_gce(self):
-        env = SimpleEnvironment('foo', {'type': 'gce'})
-        client = EnvJujuClient1X(env, None, None)
-        with patch.object(client, 'juju') as mock:
-            client.destroy_environment(force=False)
-            mock.assert_called_with(
-                'destroy-environment', ('foo', '-y'),
-                check=False, include_e=False, timeout=1200)
-
-    def test_destroy_environment_delete_jenv(self):
-        env = SimpleEnvironment('foo', {'type': 'ec2'})
-        client = EnvJujuClient1X(env, None, None)
-        with patch.object(client, 'juju'):
-            with temp_env({}) as juju_home:
-                client.env.juju_home = juju_home
-                jenv_path = get_jenv_path(juju_home, 'foo')
-                os.makedirs(os.path.dirname(jenv_path))
-                open(jenv_path, 'w')
-                self.assertTrue(os.path.exists(jenv_path))
-                client.destroy_environment(delete_jenv=True)
-                self.assertFalse(os.path.exists(jenv_path))
-
-    def test_destroy_model(self):
-        env = SimpleEnvironment('foo', {'type': 'ec2'})
-        client = EnvJujuClient1X(env, None, None)
-        with patch.object(client, 'juju') as mock:
-            client.destroy_model()
-        mock.assert_called_with(
-            'destroy-environment', ('foo', '-y'),
-            check=False, include_e=False, timeout=600)
-
-    def test_kill_controller(self):
-        client = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'ec2'}), None, None)
-        with patch.object(client, 'juju') as juju_mock:
-            client.kill_controller()
-        juju_mock.assert_called_once_with(
-            'destroy-environment', ('foo', '--force', '-y'), check=False,
-            include_e=False, timeout=600)
-
-    def test_kill_controller_check(self):
-        client = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'ec2'}), None, None)
-        with patch.object(client, 'juju') as juju_mock:
-            client.kill_controller(check=True)
-        juju_mock.assert_called_once_with(
-            'destroy-environment', ('foo', '--force', '-y'), check=True,
-            include_e=False, timeout=600)
-
-    def test_destroy_controller(self):
-        client = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'ec2'}), None, None)
-        with patch.object(client, 'juju') as juju_mock:
-            client.destroy_controller()
-        juju_mock.assert_called_once_with(
-            'destroy-environment', ('foo', '-y'),
-            include_e=False, timeout=600)
-
-    def test_get_juju_output(self):
-        env = SimpleEnvironment('foo')
-        client = EnvJujuClient1X(env, None, 'juju')
-        fake_popen = FakePopen('asdf', None, 0)
-        with patch('subprocess.Popen', return_value=fake_popen) as mock:
-            result = client.get_juju_output('bar')
-        self.assertEqual('asdf', result)
-        self.assertEqual((('juju', '--show-log', 'bar', '-e', 'foo'),),
-                         mock.call_args[0])
-
-    def test_get_juju_output_accepts_varargs(self):
-        env = SimpleEnvironment('foo')
-        fake_popen = FakePopen('asdf', None, 0)
-        client = EnvJujuClient1X(env, None, 'juju')
-        with patch('subprocess.Popen', return_value=fake_popen) as mock:
-            result = client.get_juju_output('bar', 'baz', '--qux')
-        self.assertEqual('asdf', result)
-        self.assertEqual((('juju', '--show-log', 'bar', '-e', 'foo', 'baz',
-                           '--qux'),), mock.call_args[0])
-
-    def test_get_juju_output_stderr(self):
-        env = SimpleEnvironment('foo')
-        fake_popen = FakePopen('Hello', 'Error!', 1)
-        client = EnvJujuClient1X(env, None, 'juju')
-        with self.assertRaises(subprocess.CalledProcessError) as exc:
-            with patch('subprocess.Popen', return_value=fake_popen):
-                client.get_juju_output('bar')
-        self.assertEqual(exc.exception.output, 'Hello')
-        self.assertEqual(exc.exception.stderr, 'Error!')
-
-    def test_get_juju_output_full_cmd(self):
-        env = SimpleEnvironment('foo')
-        fake_popen = FakePopen(None, 'Hello!', 1)
-        client = EnvJujuClient1X(env, None, 'juju')
-        with self.assertRaises(subprocess.CalledProcessError) as exc:
-            with patch('subprocess.Popen', return_value=fake_popen):
-                client.get_juju_output('bar', '--baz', 'qux')
-        self.assertEqual(
-            ('juju', '--show-log', 'bar', '-e', 'foo', '--baz', 'qux'),
-            exc.exception.cmd)
-
-    def test_get_juju_output_accepts_timeout(self):
-        env = SimpleEnvironment('foo')
-        fake_popen = FakePopen('asdf', None, 0)
-        client = EnvJujuClient1X(env, None, 'juju')
-        with patch('subprocess.Popen', return_value=fake_popen) as po_mock:
-            client.get_juju_output('bar', timeout=5)
-        self.assertEqual(
-            po_mock.call_args[0][0],
-            (sys.executable, get_timeout_path(), '5.00', '--', 'juju',
-             '--show-log', 'bar', '-e', 'foo'))
-
-    def test__shell_environ_juju_home(self):
-        client = EnvJujuClient1X(
-            SimpleEnvironment('baz', {'type': 'ec2'}), '1.25-foobar', 'path',
-            'asdf')
-        env = client._shell_environ()
-        self.assertEqual(env['JUJU_HOME'], 'asdf')
-        self.assertNotIn('JUJU_DATA', env)
-
-    def test_juju_output_supplies_path(self):
-        env = SimpleEnvironment('foo')
-        client = EnvJujuClient1X(env, None, '/foobar/bar')
-
-        def check_path(*args, **kwargs):
-            self.assertRegexpMatches(os.environ['PATH'], r'/foobar\:')
-            return FakePopen(None, None, 0)
-        with patch('subprocess.Popen', autospec=True,
-                   side_effect=check_path):
-            client.get_juju_output('cmd', 'baz')
-
-    def test_get_status(self):
-        output_text = dedent("""\
-                - a
-                - b
-                - c
-                """)
-        env = SimpleEnvironment('foo')
-        client = EnvJujuClient1X(env, None, None)
-        with patch.object(client, 'get_juju_output',
-                          return_value=output_text) as gjo_mock:
-            result = client.get_status()
-        gjo_mock.assert_called_once_with(
-            'status', '--format', 'yaml', controller=False)
-        self.assertEqual(Status1X, type(result))
-        self.assertEqual(['a', 'b', 'c'], result.status)
-
-    def test_get_status_retries_on_error(self):
-        env = SimpleEnvironment('foo')
-        client = EnvJujuClient1X(env, None, None)
-        client.attempt = 0
-
-        def get_juju_output(command, *args, **kwargs):
-            if client.attempt == 1:
-                return '"hello"'
-            client.attempt += 1
-            raise subprocess.CalledProcessError(1, command)
-
-        with patch.object(client, 'get_juju_output', get_juju_output):
-            client.get_status()
-
-    def test_get_status_raises_on_timeout_1(self):
-        env = SimpleEnvironment('foo')
-        client = EnvJujuClient1X(env, None, None)
-
-        def get_juju_output(command, *args, **kwargs):
-            raise subprocess.CalledProcessError(1, command)
-
-        with patch.object(client, 'get_juju_output',
-                          side_effect=get_juju_output):
-            with patch('jujupy.until_timeout', lambda x: iter([None, None])):
-                with self.assertRaisesRegexp(
-                        Exception, 'Timed out waiting for juju status'):
-                    client.get_status()
-
-    def test_get_status_raises_on_timeout_2(self):
-        env = SimpleEnvironment('foo')
-        client = EnvJujuClient1X(env, None, None)
-        with patch('jujupy.until_timeout', return_value=iter([1])) as mock_ut:
-            with patch.object(client, 'get_juju_output',
-                              side_effect=StopIteration):
-                with self.assertRaises(StopIteration):
-                    client.get_status(500)
-        mock_ut.assert_called_with(500)
-
-    def test_get_status_controller(self):
-        output_text = """\
-            - a
-            - b
-            - c
-        """
-        env = SimpleEnvironment('foo')
-        client = EnvJujuClient1X(env, None, None)
-        with patch.object(client, 'get_juju_output',
-                          return_value=output_text) as gjo_mock:
-            client.get_status(controller=True)
-        gjo_mock.assert_called_once_with(
-            'status', '--format', 'yaml', controller=True)
-
-    @staticmethod
-    def make_status_yaml(key, machine_value, unit_value):
-        return dedent("""\
-            machines:
-              "0":
-                {0}: {1}
-            services:
-              jenkins:
-                units:
-                  jenkins/0:
-                    {0}: {2}
-        """.format(key, machine_value, unit_value))
-
-    def test_deploy_non_joyent(self):
-        env = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
-        with patch.object(env, 'juju') as mock_juju:
-            env.deploy('mondogb')
-        mock_juju.assert_called_with('deploy', ('mondogb',))
-
-    def test_deploy_joyent(self):
-        env = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
-        with patch.object(env, 'juju') as mock_juju:
-            env.deploy('mondogb')
-        mock_juju.assert_called_with('deploy', ('mondogb',))
-
-    def test_deploy_repository(self):
-        env = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
-        with patch.object(env, 'juju') as mock_juju:
-            env.deploy('mondogb', '/home/jrandom/repo')
-        mock_juju.assert_called_with(
-            'deploy', ('mondogb', '--repository', '/home/jrandom/repo'))
-
-    def test_deploy_to(self):
-        env = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
-        with patch.object(env, 'juju') as mock_juju:
-            env.deploy('mondogb', to='0')
-        mock_juju.assert_called_with(
-            'deploy', ('mondogb', '--to', '0'))
-
-    def test_deploy_service(self):
-        env = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
-        with patch.object(env, 'juju') as mock_juju:
-            env.deploy('local:mondogb', service='my-mondogb')
-        mock_juju.assert_called_with(
-            'deploy', ('local:mondogb', 'my-mondogb',))
-
-    def test_upgrade_charm(self):
-        client = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
-        with patch.object(client, 'juju') as mock_juju:
-            client.upgrade_charm('foo-service',
-                                 '/bar/repository/angsty/mongodb')
-        mock_juju.assert_called_once_with(
-            'upgrade-charm', ('foo-service', '--repository',
-                              '/bar/repository',))
-
-    def test_remove_service(self):
-        client = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
-        with patch.object(client, 'juju') as mock_juju:
-            client.remove_service('mondogb')
-        mock_juju.assert_called_with('destroy-service', ('mondogb',))
-
-    def test_status_until_always_runs_once(self):
-        client = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
-        status_txt = self.make_status_yaml('agent-state', 'started', 'started')
-        with patch.object(client, 'get_juju_output', return_value=status_txt):
-            result = list(client.status_until(-1))
-        self.assertEqual(
-            [r.status for r in result], [Status.from_text(status_txt).status])
-
-    def test_status_until_timeout(self):
-        client = EnvJujuClient1X(
-            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
-        status_txt = self.make_status_yaml('agent-state', 'started', 'started')
-        status_yaml = yaml.safe_load(status_txt)
-
-        def until_timeout_stub(timeout, start=None):
-            return iter([None, None])
-
-        with patch.object(client, 'get_juju_output', return_value=status_txt):
-            with patch('jujupy.until_timeout',
-                       side_effect=until_timeout_stub) as ut_mock:
-                result = list(client.status_until(30, 70))
-        self.assertEqual(
-            [r.status for r in result], [status_yaml] * 3)
-        # until_timeout is called by status as well as status_until.
-        self.assertEqual(ut_mock.mock_calls,
-                         [call(60), call(30, start=70), call(60), call(60)])
-
-    def test_add_ssh_machines(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, 'juju')
-        with patch('subprocess.check_call', autospec=True) as cc_mock:
-            client.add_ssh_machines(['m-foo', 'm-bar', 'm-baz'])
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 0)
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-bar'), 1)
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-baz'), 2)
-        self.assertEqual(cc_mock.call_count, 3)
-
-    def test_add_ssh_machines_retry(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, 'juju')
-        with patch('subprocess.check_call', autospec=True,
-                   side_effect=[subprocess.CalledProcessError(None, None),
-                                None, None, None]) as cc_mock:
-            client.add_ssh_machines(['m-foo', 'm-bar', 'm-baz'])
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 0)
-        self.pause_mock.assert_called_once_with(30)
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 1)
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-bar'), 2)
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-baz'), 3)
-        self.assertEqual(cc_mock.call_count, 4)
-
-    def test_add_ssh_machines_fail_on_second_machine(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, 'juju')
-        with patch('subprocess.check_call', autospec=True, side_effect=[
-                None, subprocess.CalledProcessError(None, None), None, None
-                ]) as cc_mock:
-            with self.assertRaises(subprocess.CalledProcessError):
-                client.add_ssh_machines(['m-foo', 'm-bar', 'm-baz'])
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 0)
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-bar'), 1)
-        self.assertEqual(cc_mock.call_count, 2)
-
-    def test_add_ssh_machines_fail_on_second_attempt(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, 'juju')
-        with patch('subprocess.check_call', autospec=True, side_effect=[
-                subprocess.CalledProcessError(None, None),
-                subprocess.CalledProcessError(None, None)]) as cc_mock:
-            with self.assertRaises(subprocess.CalledProcessError):
-                client.add_ssh_machines(['m-foo', 'm-bar', 'm-baz'])
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 0)
-        assert_juju_call(self, cc_mock, client, (
-            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 1)
-        self.assertEqual(cc_mock.call_count, 2)
-
-    def test_wait_for_started(self):
-        value = self.make_status_yaml('agent-state', 'started', 'started')
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        with patch.object(client, 'get_juju_output', return_value=value):
-            client.wait_for_started()
-
-    def test_wait_for_started_timeout(self):
-        value = self.make_status_yaml('agent-state', 'pending', 'started')
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        with patch('jujupy.until_timeout', lambda x, start=None: range(1)):
-            with patch.object(client, 'get_juju_output', return_value=value):
-                writes = []
-                with patch.object(GroupReporter, '_write', autospec=True,
-                                  side_effect=lambda _, s: writes.append(s)):
-                    with self.assertRaisesRegexp(
-                            StatusNotMet,
-                            'Timed out waiting for agents to start in local'):
-                        client.wait_for_started()
-                self.assertEqual(writes, ['pending: 0', ' .', '\n'])
-
-    def test_wait_for_started_start(self):
-        value = self.make_status_yaml('agent-state', 'started', 'pending')
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        now = datetime.now() + timedelta(days=1)
-        with patch('utility.until_timeout.now', return_value=now):
-            with patch.object(client, 'get_juju_output', return_value=value):
-                writes = []
-                with patch.object(GroupReporter, '_write', autospec=True,
-                                  side_effect=lambda _, s: writes.append(s)):
-                    with self.assertRaisesRegexp(
-                            StatusNotMet,
-                            'Timed out waiting for agents to start in local'):
-                        client.wait_for_started(start=now - timedelta(1200))
-                self.assertEqual(writes, ['pending: jenkins/0', '\n'])
-
-    def test_wait_for_started_logs_status(self):
-        value = self.make_status_yaml('agent-state', 'pending', 'started')
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        with patch.object(client, 'get_juju_output', return_value=value):
-            writes = []
-            with patch.object(GroupReporter, '_write', autospec=True,
-                              side_effect=lambda _, s: writes.append(s)):
-                with self.assertRaisesRegexp(
-                        StatusNotMet,
-                        'Timed out waiting for agents to start in local'):
-                    client.wait_for_started(0)
-            self.assertEqual(writes, ['pending: 0', '\n'])
-        self.assertEqual(self.log_stream.getvalue(), 'ERROR %s\n' % value)
-
-    def test_wait_for_subordinate_units(self):
-        value = dedent("""\
-            machines:
-              "0":
-                agent-state: started
-            services:
-              jenkins:
-                units:
-                  jenkins/0:
-                    subordinates:
-                      sub1/0:
-                        agent-state: started
-              ubuntu:
-                units:
-                  ubuntu/0:
-                    subordinates:
-                      sub2/0:
-                        agent-state: started
-                      sub3/0:
-                        agent-state: started
-        """)
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        now = datetime.now() + timedelta(days=1)
-        with patch('utility.until_timeout.now', return_value=now):
-            with patch.object(client, 'get_juju_output', return_value=value):
-                with patch('jujupy.GroupReporter.update') as update_mock:
-                    with patch('jujupy.GroupReporter.finish') as finish_mock:
-                        client.wait_for_subordinate_units(
-                            'jenkins', 'sub1', start=now - timedelta(1200))
-        self.assertEqual([], update_mock.call_args_list)
-        finish_mock.assert_called_once_with()
-
-    def test_wait_for_multiple_subordinate_units(self):
-        value = dedent("""\
-            machines:
-              "0":
-                agent-state: started
-            services:
-              ubuntu:
-                units:
-                  ubuntu/0:
-                    subordinates:
-                      sub/0:
-                        agent-state: started
-                  ubuntu/1:
-                    subordinates:
-                      sub/1:
-                        agent-state: started
-        """)
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        now = datetime.now() + timedelta(days=1)
-        with patch('utility.until_timeout.now', return_value=now):
-            with patch.object(client, 'get_juju_output', return_value=value):
-                with patch('jujupy.GroupReporter.update') as update_mock:
-                    with patch('jujupy.GroupReporter.finish') as finish_mock:
-                        client.wait_for_subordinate_units(
-                            'ubuntu', 'sub', start=now - timedelta(1200))
-        self.assertEqual([], update_mock.call_args_list)
-        finish_mock.assert_called_once_with()
-
-    def test_wait_for_subordinate_units_checks_slash_in_unit_name(self):
-        value = dedent("""\
-            machines:
-              "0":
-                agent-state: started
-            services:
-              jenkins:
-                units:
-                  jenkins/0:
-                    subordinates:
-                      sub1:
-                        agent-state: started
-        """)
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        now = datetime.now() + timedelta(days=1)
-        with patch('utility.until_timeout.now', return_value=now):
-            with patch.object(client, 'get_juju_output', return_value=value):
-                with self.assertRaisesRegexp(
-                        StatusNotMet,
-                        'Timed out waiting for agents to start in local'):
-                    client.wait_for_subordinate_units(
-                        'jenkins', 'sub1', start=now - timedelta(1200))
-
-    def test_wait_for_subordinate_units_no_subordinate(self):
-        value = dedent("""\
-            machines:
-              "0":
-                agent-state: started
-            services:
-              jenkins:
-                units:
-                  jenkins/0:
-                    agent-state: started
-        """)
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        now = datetime.now() + timedelta(days=1)
-        with patch('utility.until_timeout.now', return_value=now):
-            with patch.object(client, 'get_juju_output', return_value=value):
-                with self.assertRaisesRegexp(
-                        StatusNotMet,
-                        'Timed out waiting for agents to start in local'):
-                    client.wait_for_subordinate_units(
-                        'jenkins', 'sub1', start=now - timedelta(1200))
-
-    def test_wait_for_workload(self):
-        initial_status = Status1X.from_text("""\
-            services:
-              jenkins:
-                units:
-                  jenkins/0:
-                    workload-status:
-                      current: waiting
-                  subordinates:
-                    ntp/0:
-                      workload-status:
-                        current: unknown
-        """)
-        final_status = Status(copy.deepcopy(initial_status.status), None)
-        final_status.status['services']['jenkins']['units']['jenkins/0'][
-            'workload-status']['current'] = 'active'
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        writes = []
-        with patch('utility.until_timeout', autospec=True, return_value=[1]):
-            with patch.object(client, 'get_status', autospec=True,
-                              side_effect=[initial_status, final_status]):
-                with patch.object(GroupReporter, '_write', autospec=True,
-                                  side_effect=lambda _, s: writes.append(s)):
-                    client.wait_for_workloads()
-        self.assertEqual(writes, ['waiting: jenkins/0', '\n'])
-
-    def test_wait_for_workload_all_unknown(self):
-        status = Status.from_text("""\
-            services:
-              jenkins:
-                units:
-                  jenkins/0:
-                    workload-status:
-                      current: unknown
-                  subordinates:
-                    ntp/0:
-                      workload-status:
-                        current: unknown
-        """)
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        writes = []
-        with patch('utility.until_timeout', autospec=True, return_value=[]):
-            with patch.object(client, 'get_status', autospec=True,
-                              return_value=status):
-                with patch.object(GroupReporter, '_write', autospec=True,
-                                  side_effect=lambda _, s: writes.append(s)):
-                    client.wait_for_workloads(timeout=1)
-        self.assertEqual(writes, [])
-
-    def test_wait_for_workload_no_workload_status(self):
-        status = Status.from_text("""\
-            services:
-              jenkins:
-                units:
-                  jenkins/0:
-                    agent-state: active
-        """)
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        writes = []
-        with patch('utility.until_timeout', autospec=True, return_value=[]):
-            with patch.object(client, 'get_status', autospec=True,
-                              return_value=status):
-                with patch.object(GroupReporter, '_write', autospec=True,
-                                  side_effect=lambda _, s: writes.append(s)):
-                    client.wait_for_workloads(timeout=1)
-        self.assertEqual(writes, [])
-
-    def test_wait_for_ha(self):
-        value = yaml.safe_dump({
-            'machines': {
-                '0': {'state-server-member-status': 'has-vote'},
-                '1': {'state-server-member-status': 'has-vote'},
-                '2': {'state-server-member-status': 'has-vote'},
-            },
-            'services': {},
-        })
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        with patch.object(client, 'get_juju_output', return_value=value):
-            client.wait_for_ha()
-
-    def test_wait_for_ha_no_has_vote(self):
-        value = yaml.safe_dump({
-            'machines': {
-                '0': {'state-server-member-status': 'no-vote'},
-                '1': {'state-server-member-status': 'no-vote'},
-                '2': {'state-server-member-status': 'no-vote'},
-            },
-            'services': {},
-        })
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        with patch.object(client, 'get_juju_output', return_value=value):
-            writes = []
-            with patch('jujupy.until_timeout', autospec=True,
-                       return_value=[2, 1]):
-                with patch.object(GroupReporter, '_write', autospec=True,
-                                  side_effect=lambda _, s: writes.append(s)):
-                    with self.assertRaisesRegexp(
-                            StatusNotMet,
-                            'Timed out waiting for voting to be enabled.'):
-                        client.wait_for_ha()
-            self.assertEqual(writes[:2], ['no-vote: 0, 1, 2', ' .'])
-            self.assertEqual(writes[2:-1], ['.'] * (len(writes) - 3))
-            self.assertEqual(writes[-1:], ['\n'])
-
-    def test_wait_for_ha_timeout(self):
-        value = yaml.safe_dump({
-            'machines': {
-                '0': {'state-server-member-status': 'has-vote'},
-                '1': {'state-server-member-status': 'has-vote'},
-            },
-            'services': {},
-        })
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        status = client.status_class.from_text(value)
-        with patch('jujupy.until_timeout', lambda x, start=None: range(0)):
-            with patch.object(client, 'get_status', return_value=status):
-                with self.assertRaisesRegexp(
-                        StatusNotMet,
-                        'Timed out waiting for voting to be enabled.'):
-                    client.wait_for_ha()
-
-    def test_wait_for_deploy_started(self):
-        value = yaml.safe_dump({
-            'machines': {
-                '0': {'agent-state': 'started'},
-            },
-            'services': {
-                'jenkins': {
-                    'units': {
-                        'jenkins/1': {'baz': 'qux'}
-                    }
-                }
-            }
-        })
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        with patch.object(client, 'get_juju_output', return_value=value):
-            client.wait_for_deploy_started()
-
-    def test_wait_for_deploy_started_timeout(self):
-        value = yaml.safe_dump({
-            'machines': {
-                '0': {'agent-state': 'started'},
-            },
-            'services': {},
-        })
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        with patch('jujupy.until_timeout', lambda x: range(0)):
-            with patch.object(client, 'get_juju_output', return_value=value):
-                with self.assertRaisesRegexp(
-                        StatusNotMet,
-                        'Timed out waiting for applications to start.'):
-                    client.wait_for_deploy_started()
-
-    def test_wait_for_version(self):
-        value = self.make_status_yaml('agent-version', '1.17.2', '1.17.2')
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        with patch.object(client, 'get_juju_output', return_value=value):
-            client.wait_for_version('1.17.2')
-
-    def test_wait_for_version_timeout(self):
-        value = self.make_status_yaml('agent-version', '1.17.2', '1.17.1')
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        writes = []
-        with patch('jujupy.until_timeout', lambda x, start=None: [x]):
-            with patch.object(client, 'get_juju_output', return_value=value):
-                with patch.object(GroupReporter, '_write', autospec=True,
-                                  side_effect=lambda _, s: writes.append(s)):
-                    with self.assertRaisesRegexp(
-                            StatusNotMet, 'Some versions did not update'):
-                        client.wait_for_version('1.17.2')
-        self.assertEqual(writes, ['1.17.1: jenkins/0', ' .', '\n'])
-
-    def test_wait_for_version_handles_connection_error(self):
-        err = subprocess.CalledProcessError(2, 'foo')
-        err.stderr = 'Unable to connect to environment'
-        err = CannotConnectEnv(err)
-        status = self.make_status_yaml('agent-version', '1.17.2', '1.17.2')
-        actions = [err, status]
-
-        def get_juju_output_fake(*args, **kwargs):
-            action = actions.pop(0)
-            if isinstance(action, Exception):
-                raise action
-            else:
-                return action
-
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        with patch.object(client, 'get_juju_output', get_juju_output_fake):
-            client.wait_for_version('1.17.2')
-
-    def test_wait_for_version_raises_non_connection_error(self):
-        err = Exception('foo')
-        status = self.make_status_yaml('agent-version', '1.17.2', '1.17.2')
-        actions = [err, status]
-
-        def get_juju_output_fake(*args, **kwargs):
-            action = actions.pop(0)
-            if isinstance(action, Exception):
-                raise action
-            else:
-                return action
-
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        with patch.object(client, 'get_juju_output', get_juju_output_fake):
-            with self.assertRaisesRegexp(Exception, 'foo'):
-                client.wait_for_version('1.17.2')
-
-    def test_wait_just_machine_0(self):
-        value = yaml.safe_dump({
-            'machines': {
-                '0': {'agent-state': 'started'},
-            },
-        })
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        with patch.object(client, 'get_juju_output', return_value=value):
-            client.wait_for([WaitMachineNotPresent('1')])
-
-    def test_wait_just_machine_0_timeout(self):
-        value = yaml.safe_dump({
-            'machines': {
-                '0': {'agent-state': 'started'},
-                '1': {'agent-state': 'started'},
-            },
-        })
-        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
-        with patch.object(client, 'get_juju_output', return_value=value), \
-            patch('jujupy.until_timeout', lambda x: range(0)), \
-            self.assertRaisesRegexp(
-                Exception,
-                'Timed out waiting for machine removal 1'):
-            client.wait_for([WaitMachineNotPresent('1')])
-
-    def test_set_model_constraints(self):
-        client = EnvJujuClient1X(SimpleEnvironment('bar', {}), None, '/foo')
-        with patch.object(client, 'juju') as juju_mock:
-            client.set_model_constraints({'bar': 'baz'})
-        juju_mock.assert_called_once_with('set-constraints', ('bar=baz',))
-
-    def test_get_model_config(self):
-        env = SimpleEnvironment('foo', None)
-        fake_popen = FakePopen(yaml.safe_dump({'bar': 'baz'}), None, 0)
-        client = EnvJujuClient1X(env, None, 'juju')
-        with patch('subprocess.Popen', return_value=fake_popen) as po_mock:
-            result = client.get_model_config()
-        assert_juju_call(
-            self, po_mock, client, (
-                'juju', '--show-log', 'get-env', '-e', 'foo'))
-        self.assertEqual({'bar': 'baz'}, result)
-
-    def test_get_env_option(self):
-        env = SimpleEnvironment('foo', None)
-        fake_popen = FakePopen('https://example.org/juju/tools', None, 0)
-        client = EnvJujuClient1X(env, None, 'juju')
-        with patch('subprocess.Popen', return_value=fake_popen) as mock:
-            result = client.get_env_option('tools-metadata-url')
-        self.assertEqual(
-            mock.call_args[0][0],
-            ('juju', '--show-log', 'get-env', '-e', 'foo',
-             'tools-metadata-url'))
-        self.assertEqual('https://example.org/juju/tools', result)
-
-    def test_set_env_option(self):
-        env = SimpleEnvironment('foo')
-        client = EnvJujuClient1X(env, None, 'juju')
-        with patch('subprocess.check_call') as mock:
-            client.set_env_option(
-                'tools-metadata-url', 'https://example.org/juju/tools')
-        environ = dict(os.environ)
-        environ['JUJU_HOME'] = client.env.juju_home
-        mock.assert_called_with(
-            ('juju', '--show-log', 'set-env', '-e', 'foo',
-             'tools-metadata-url=https://example.org/juju/tools'))
-
-    def test_unset_env_option(self):
-        env = SimpleEnvironment('foo')
-        client = EnvJujuClient1X(env, None, 'juju')
-        with patch('subprocess.check_call') as mock:
-            client.unset_env_option('tools-metadata-url')
-        environ = dict(os.environ)
-        environ['JUJU_HOME'] = client.env.juju_home
-        mock.assert_called_with(
-            ('juju', '--show-log', 'set-env', '-e', 'foo',
-             'tools-metadata-url='))
-
-    def test_set_testing_agent_metadata_url(self):
-        env = SimpleEnvironment(None, {'type': 'foo'})
-        client = EnvJujuClient1X(env, None, None)
-        with patch.object(client, 'get_env_option') as mock_get:
-            mock_get.return_value = 'https://example.org/juju/tools'
-            with patch.object(client, 'set_env_option') as mock_set:
-                client.set_testing_agent_metadata_url()
-        mock_get.assert_called_with('tools-metadata-url')
-        mock_set.assert_called_with(
-            'tools-metadata-url',
-            'https://example.org/juju/testing/tools')
-
-    def test_set_testing_agent_metadata_url_noop(self):
-        env = SimpleEnvironment(None, {'type': 'foo'})
-        client = EnvJujuClient1X(env, None, None)
-        with patch.object(client, 'get_env_option') as mock_get:
-            mock_get.return_value = 'https://example.org/juju/testing/tools'
-            with patch.object(client, 'set_env_option') as mock_set:
-                client.set_testing_agent_metadata_url()
-        mock_get.assert_called_with('tools-metadata-url')
-        self.assertEqual(0, mock_set.call_count)
-
-    def test_juju(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, 'juju')
-        with patch('subprocess.check_call') as mock:
-            client.juju('foo', ('bar', 'baz'))
-        environ = dict(os.environ)
-        environ['JUJU_HOME'] = client.env.juju_home
-        mock.assert_called_with(('juju', '--show-log', 'foo', '-e', 'qux',
-                                 'bar', 'baz'))
-
-    def test_juju_env(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-
-        def check_path(*args, **kwargs):
-            self.assertRegexpMatches(os.environ['PATH'], r'/foobar\:')
-        with patch('subprocess.check_call', side_effect=check_path):
-            client.juju('foo', ('bar', 'baz'))
-
-    def test_juju_no_check(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, 'juju')
-        environ = dict(os.environ)
-        environ['JUJU_HOME'] = client.env.juju_home
-        with patch('subprocess.call') as mock:
-            client.juju('foo', ('bar', 'baz'), check=False)
-        mock.assert_called_with(('juju', '--show-log', 'foo', '-e', 'qux',
-                                 'bar', 'baz'))
-
-    def test_juju_no_check_env(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-
-        def check_path(*args, **kwargs):
-            self.assertRegexpMatches(os.environ['PATH'], r'/foobar\:')
-        with patch('subprocess.call', side_effect=check_path):
-            client.juju('foo', ('bar', 'baz'), check=False)
-
-    def test_juju_timeout(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-        with patch('subprocess.check_call') as cc_mock:
-            client.juju('foo', ('bar', 'baz'), timeout=58)
-        self.assertEqual(cc_mock.call_args[0][0], (
-            sys.executable, get_timeout_path(), '58.00', '--', 'baz',
-            '--show-log', 'foo', '-e', 'qux', 'bar', 'baz'))
-
-    def test_juju_juju_home(self):
-        env = SimpleEnvironment('qux')
-        os.environ['JUJU_HOME'] = 'foo'
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-
-        def check_home(*args, **kwargs):
-            self.assertEqual(os.environ['JUJU_HOME'], 'foo')
-            yield
-            self.assertEqual(os.environ['JUJU_HOME'], 'asdf')
-            yield
-
-        with patch('subprocess.check_call', side_effect=check_home):
-            client.juju('foo', ('bar', 'baz'))
-            client.env.juju_home = 'asdf'
-            client.juju('foo', ('bar', 'baz'))
-
-    def test_juju_extra_env(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, 'juju')
-        extra_env = {'JUJU': '/juju', 'JUJU_HOME': client.env.juju_home}
-
-        def check_env(*args, **kwargs):
-            self.assertEqual('/juju', os.environ['JUJU'])
-
-        with patch('subprocess.check_call', side_effect=check_env) as mock:
-            client.juju('quickstart', ('bar', 'baz'), extra_env=extra_env)
-        mock.assert_called_with(
-            ('juju', '--show-log', 'quickstart', '-e', 'qux', 'bar', 'baz'))
-
-    def test_juju_backup_with_tgz(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-
-        def check_env(*args, **kwargs):
-            self.assertEqual(os.environ['JUJU_ENV'], 'qux')
-            return 'foojuju-backup-24.tgzz'
-        with patch('subprocess.check_output',
-                   side_effect=check_env) as co_mock:
-            backup_file = client.backup()
-        self.assertEqual(backup_file, os.path.abspath('juju-backup-24.tgz'))
-        assert_juju_call(self, co_mock, client, ['juju', 'backup'])
-
-    def test_juju_backup_with_tar_gz(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-        with patch('subprocess.check_output',
-                   return_value='foojuju-backup-123-456.tar.gzbar'):
-            backup_file = client.backup()
-        self.assertEqual(
-            backup_file, os.path.abspath('juju-backup-123-456.tar.gz'))
-
-    def test_juju_backup_no_file(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-        with patch('subprocess.check_output', return_value=''):
-            with self.assertRaisesRegexp(
-                    Exception, 'The backup file was not found in output'):
-                client.backup()
-
-    def test_juju_backup_wrong_file(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-        with patch('subprocess.check_output',
-                   return_value='mumu-backup-24.tgz'):
-            with self.assertRaisesRegexp(
-                    Exception, 'The backup file was not found in output'):
-                client.backup()
-
-    def test_juju_backup_environ(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-        environ = client._shell_environ()
-        environ['JUJU_ENV'] = client.env.environment
-
-        def side_effect(*args, **kwargs):
-            self.assertEqual(environ, os.environ)
-            return 'foojuju-backup-123-456.tar.gzbar'
-        with patch('subprocess.check_output', side_effect=side_effect):
-            client.backup()
-            self.assertNotEqual(environ, os.environ)
-
-    def test_restore_backup(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-        with patch.object(client, 'get_juju_output') as gjo_mock:
-            result = client.restore_backup('quxx')
-        gjo_mock.assert_called_once_with('restore', '--constraints',
-                                         'mem=2G', 'quxx')
-        self.assertIs(gjo_mock.return_value, result)
-
-    def test_restore_backup_async(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-        with patch.object(client, 'juju_async') as gjo_mock:
-            result = client.restore_backup_async('quxx')
-        gjo_mock.assert_called_once_with(
-            'restore', ('--constraints', 'mem=2G', 'quxx'))
-        self.assertIs(gjo_mock.return_value, result)
-
-    def test_enable_ha(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-        with patch.object(client, 'juju', autospec=True) as eha_mock:
-            client.enable_ha()
-        eha_mock.assert_called_once_with('ensure-availability', ('-n', '3'))
-
-    def test_juju_async(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-        with patch('subprocess.Popen') as popen_class_mock:
-            with client.juju_async('foo', ('bar', 'baz')) as proc:
-                assert_juju_call(self, popen_class_mock, client, (
-                    'baz', '--show-log', 'foo', '-e', 'qux', 'bar', 'baz'))
-                self.assertIs(proc, popen_class_mock.return_value)
-                self.assertEqual(proc.wait.call_count, 0)
-                proc.wait.return_value = 0
-        proc.wait.assert_called_once_with()
-
-    def test_juju_async_failure(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-        with patch('subprocess.Popen') as popen_class_mock:
-            with self.assertRaises(subprocess.CalledProcessError) as err_cxt:
-                with client.juju_async('foo', ('bar', 'baz')):
-                    proc_mock = popen_class_mock.return_value
-                    proc_mock.wait.return_value = 23
-        self.assertEqual(err_cxt.exception.returncode, 23)
-        self.assertEqual(err_cxt.exception.cmd, (
-            'baz', '--show-log', 'foo', '-e', 'qux', 'bar', 'baz'))
-
-    def test_juju_async_environ(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-        environ = client._shell_environ()
-        proc_mock = Mock()
-        with patch('subprocess.Popen') as popen_class_mock:
-
-            def check_environ(*args, **kwargs):
-                self.assertEqual(environ, os.environ)
-                return proc_mock
-            popen_class_mock.side_effect = check_environ
-            proc_mock.wait.return_value = 0
-            with client.juju_async('foo', ('bar', 'baz')):
-                pass
-            self.assertNotEqual(environ, os.environ)
-
-    def test_is_jes_enabled(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-        fake_popen = FakePopen(' %s' % SYSTEM, None, 0)
-        with patch('subprocess.Popen',
-                   return_value=fake_popen) as po_mock:
-            self.assertIsFalse(client.is_jes_enabled())
-        self.assertEqual(0, po_mock.call_count)
-
-    def test_get_jes_command(self):
-        env = SimpleEnvironment('qux')
-        client = EnvJujuClient1X(env, None, '/foobar/baz')
-        # Juju 1.24 and older do not have a JES command. It is an error
-        # to call get_jes_command when is_jes_enabled is False
-        fake_popen = FakePopen(' %s' % SYSTEM, None, 0)
-        with patch('subprocess.Popen',
-                   return_value=fake_popen) as po_mock:
-            with self.assertRaises(JESNotSupported):
-                client.get_jes_command()
-        self.assertEqual(0, po_mock.call_count)
-
-    def test_get_juju_timings(self):
-        env = SimpleEnvironment('foo')
-        client = EnvJujuClient1X(env, None, 'my/juju/bin')
-        client._backend.juju_timings = {("juju", "op1"): [1],
-                                        ("juju", "op2"): [2]}
-        flattened_timings = client.get_juju_timings()
-        expected = {"juju op1": [1], "juju op2": [2]}
-        self.assertEqual(flattened_timings, expected)
-
-    def test_deploy_bundle_1x(self):
-        client = EnvJujuClient1X(SimpleEnvironment('an_env', None),
-                                 '1.23-series-arch', None)
-        with patch.object(client, 'juju') as mock_juju:
-            client.deploy_bundle('bundle:~juju-qa/some-bundle')
-        mock_juju.assert_called_with(
-            'deployer', ('--debug', '--deploy-delay', '10', '--timeout',
-                         '3600', '--config', 'bundle:~juju-qa/some-bundle'))
-
-    def test_deploy_bundle_template(self):
-        client = EnvJujuClient1X(SimpleEnvironment('an_env', None),
-                                 '1.23-series-arch', None)
-        with patch.object(client, 'juju') as mock_juju:
-            client.deploy_bundle('bundle:~juju-qa/some-{container}-bundle')
-        mock_juju.assert_called_with(
-            'deployer', (
-                '--debug', '--deploy-delay', '10', '--timeout', '3600',
-                '--config', 'bundle:~juju-qa/some-lxc-bundle'))
-
-    def test_deployer(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', None)
-        with patch.object(EnvJujuClient1X, 'juju') as mock:
-            client.deployer('bundle:~juju-qa/some-bundle')
-        mock.assert_called_with(
-            'deployer', ('--debug', '--deploy-delay', '10', '--timeout',
-                         '3600', '--config', 'bundle:~juju-qa/some-bundle'))
-
-    def test_deployer_template(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', None)
-        with patch.object(EnvJujuClient1X, 'juju') as mock:
-            client.deployer('bundle:~juju-qa/some-{container}-bundle')
-        mock.assert_called_with(
-            'deployer', (
-                '--debug', '--deploy-delay', '10', '--timeout', '3600',
-                '--config', 'bundle:~juju-qa/some-lxc-bundle'))
-
-    def test_deployer_with_bundle_name(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', None)
-        with patch.object(EnvJujuClient1X, 'juju') as mock:
-            client.deployer('bundle:~juju-qa/some-bundle', 'name')
-        mock.assert_called_with(
-            'deployer', ('--debug', '--deploy-delay', '10', '--timeout',
-                         '3600', '--config', 'bundle:~juju-qa/some-bundle',
-                         'name'))
-
-    def test_quickstart_maas(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'maas'}),
-                                 '1.23-series-arch', '/juju')
-        with patch.object(EnvJujuClient1X, 'juju') as mock:
-            client.quickstart('bundle:~juju-qa/some-bundle')
-        mock.assert_called_with(
-            'quickstart', ('--constraints', 'mem=2G', '--no-browser',
-                           'bundle:~juju-qa/some-bundle'),
-            extra_env={'JUJU': '/juju'})
-
-    def test_quickstart_local(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', '/juju')
-        with patch.object(EnvJujuClient1X, 'juju') as mock:
-            client.quickstart('bundle:~juju-qa/some-bundle')
-        mock.assert_called_with(
-            'quickstart', ('--constraints', 'mem=2G', '--no-browser',
-                           'bundle:~juju-qa/some-bundle'),
-            extra_env={'JUJU': '/juju'})
-
-    def test_quickstart_template(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', '/juju')
-        with patch.object(EnvJujuClient1X, 'juju') as mock:
-            client.quickstart('bundle:~juju-qa/some-{container}-bundle')
-        mock.assert_called_with(
-            'quickstart', (
-                '--constraints', 'mem=2G', '--no-browser',
-                'bundle:~juju-qa/some-lxc-bundle'),
-            extra_env={'JUJU': '/juju'})
-
-    def test_list_models(self):
-        env = SimpleEnvironment('foo', {'type': 'local'})
-        client = EnvJujuClient1X(env, '1.23-series-arch', None)
-        client.list_models()
-        self.assertEqual(
-            'INFO The model is environment foo\n',
-            self.log_stream.getvalue())
-
-    def test__get_models(self):
-        data = """\
-            - name: foo
-              model-uuid: aaaa
-            - name: bar
-              model-uuid: bbbb
-        """
-        env = SimpleEnvironment('foo', {'type': 'local'})
-        client = fake_juju_client(cls=EnvJujuClient1X, env=env)
-        with patch.object(client, 'get_juju_output', return_value=data):
-            models = client._get_models()
-            self.assertEqual(
-                [{'name': 'foo', 'model-uuid': 'aaaa'},
-                 {'name': 'bar', 'model-uuid': 'bbbb'}],
-                models)
-
-    def test__get_models_exception(self):
-        env = SimpleEnvironment('foo', {'type': 'local'})
-        client = fake_juju_client(cls=EnvJujuClient1X, env=env)
-        with patch.object(client, 'get_juju_output',
-                          side_effect=subprocess.CalledProcessError('a', 'b')):
-            self.assertEqual([], client._get_models())
-
-    def test_get_models(self):
-        env = SimpleEnvironment('foo', {'type': 'local'})
-        client = EnvJujuClient1X(env, '1.23-series-arch', None)
-        self.assertEqual({}, client.get_models())
-
-    def test_iter_model_clients(self):
-        data = """\
-            - name: foo
-              model-uuid: aaaa
-              owner: admin@local
-            - name: bar
-              model-uuid: bbbb
-              owner: admin@local
-        """
-        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, None)
-        with patch.object(client, 'get_juju_output', return_value=data):
-            model_clients = list(client.iter_model_clients())
-        self.assertEqual(2, len(model_clients))
-        self.assertIs(client, model_clients[0])
-        self.assertEqual('bar', model_clients[1].env.environment)
-
-    def test_get_controller_client(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo'), {'bar': 'baz'},
-                                 'myhome')
-        controller_client = client.get_controller_client()
-        self.assertIs(client, controller_client)
-
-    def test_list_controllers(self):
-        env = SimpleEnvironment('foo', {'type': 'local'})
-        client = EnvJujuClient1X(env, '1.23-series-arch', None)
-        client.list_controllers()
-        self.assertEqual(
-            'INFO The controller is environment foo\n',
-            self.log_stream.getvalue())
-
-    def test_get_controller_model_name(self):
-        env = SimpleEnvironment('foo', {'type': 'local'})
-        client = EnvJujuClient1X(env, '1.23-series-arch', None)
-        controller_name = client.get_controller_model_name()
-        self.assertEqual('foo', controller_name)
-
-    def test_get_controller_endpoint_ipv4(self):
-        env = SimpleEnvironment('foo', {'type': 'local'})
-        client = EnvJujuClient1X(env, '1.23-series-arch', None)
-        with patch.object(client, 'get_juju_output',
-                          return_value='10.0.0.1:17070') as gjo_mock:
-            endpoint = client.get_controller_endpoint()
-        self.assertEqual('10.0.0.1', endpoint)
-        gjo_mock.assert_called_once_with('api-endpoints')
-
-    def test_get_controller_endpoint_ipv6(self):
-        env = SimpleEnvironment('foo', {'type': 'local'})
-        client = EnvJujuClient1X(env, '1.23-series-arch', None)
-        with patch.object(client, 'get_juju_output',
-                          return_value='[::1]:17070') as gjo_mock:
-            endpoint = client.get_controller_endpoint()
-        self.assertEqual('::1', endpoint)
-        gjo_mock.assert_called_once_with('api-endpoints')
-
-    def test_action_do(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', None)
-        with patch.object(EnvJujuClient1X, 'get_juju_output') as mock:
-            mock.return_value = \
-                "Action queued with id: 5a92ec93-d4be-4399-82dc-7431dbfd08f9"
-            id = client.action_do("foo/0", "myaction", "param=5")
-            self.assertEqual(id, "5a92ec93-d4be-4399-82dc-7431dbfd08f9")
-        mock.assert_called_once_with(
-            'action do', 'foo/0', 'myaction', "param=5"
-        )
-
-    def test_action_do_error(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', None)
-        with patch.object(EnvJujuClient1X, 'get_juju_output') as mock:
-            mock.return_value = "some bad text"
-            with self.assertRaisesRegexp(Exception,
-                                         "Action id not found in output"):
-                client.action_do("foo/0", "myaction", "param=5")
-
-    def test_action_fetch(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', None)
-        with patch.object(EnvJujuClient1X, 'get_juju_output') as mock:
-            ret = "status: completed\nfoo: bar"
-            mock.return_value = ret
-            out = client.action_fetch("123")
-            self.assertEqual(out, ret)
-        mock.assert_called_once_with(
-            'action fetch', '123', "--wait", "1m"
-        )
-
-    def test_action_fetch_timeout(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', None)
-        ret = "status: pending\nfoo: bar"
-        with patch.object(EnvJujuClient1X,
-                          'get_juju_output', return_value=ret):
-            with self.assertRaisesRegexp(Exception,
-                                         "timed out waiting for action"):
-                client.action_fetch("123")
-
-    def test_action_do_fetch(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', None)
-        with patch.object(EnvJujuClient1X, 'get_juju_output') as mock:
-            ret = "status: completed\nfoo: bar"
-            # setting side_effect to an iterable will return the next value
-            # from the list each time the function is called.
-            mock.side_effect = [
-                "Action queued with id: 5a92ec93-d4be-4399-82dc-7431dbfd08f9",
-                ret]
-            out = client.action_do_fetch("foo/0", "myaction", "param=5")
-            self.assertEqual(out, ret)
-
-    def test_run(self):
-        env = SimpleEnvironment('name', {}, 'foo')
-        client = fake_juju_client(cls=EnvJujuClient1X, env=env)
-        run_list = [
-            {"MachineId": "1",
-             "Stdout": "Linux\n",
-             "ReturnCode": 255,
-             "Stderr": "Permission denied (publickey,password)"}]
-        run_output = json.dumps(run_list)
-        with patch.object(client._backend, 'get_juju_output',
-                          return_value=run_output) as gjo_mock:
-            result = client.run(('wname',), applications=['foo', 'bar'])
-        self.assertEqual(run_list, result)
-        gjo_mock.assert_called_once_with(
-            'run', ('--format', 'json', '--service', 'foo,bar', 'wname'),
-            frozenset(['migration']),
-            'foo', 'name', user_name=None)
-
-    def test_list_space(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', None)
-        yaml_dict = {'foo': 'bar'}
-        output = yaml.safe_dump(yaml_dict)
-        with patch.object(client, 'get_juju_output', return_value=output,
-                          autospec=True) as gjo_mock:
-            result = client.list_space()
-        self.assertEqual(result, yaml_dict)
-        gjo_mock.assert_called_once_with('space list')
-
-    def test_add_space(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', None)
-        with patch.object(client, 'juju', autospec=True) as juju_mock:
-            client.add_space('foo-space')
-        juju_mock.assert_called_once_with('space create', ('foo-space'))
-
-    def test_add_subnet(self):
-        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
-                                 '1.23-series-arch', None)
-        with patch.object(client, 'juju', autospec=True) as juju_mock:
-            client.add_subnet('bar-subnet', 'foo-space')
-        juju_mock.assert_called_once_with('subnet add',
-                                          ('bar-subnet', 'foo-space'))
-
-    def test__shell_environ_uses_pathsep(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo'), None,
-                                 'foo/bar/juju')
-        with patch('os.pathsep', '!'):
-            environ = client._shell_environ()
-        self.assertRegexpMatches(environ['PATH'], r'foo/bar\!')
-
-    def test_set_config(self):
-        client = EnvJujuClient1X(SimpleEnvironment('bar', {}), None, '/foo')
-        with patch.object(client, 'juju') as juju_mock:
-            client.set_config('foo', {'bar': 'baz'})
-        juju_mock.assert_called_once_with('set', ('foo', 'bar=baz'))
-
-    def test_get_config(self):
-        def output(*args, **kwargs):
-            return yaml.safe_dump({
-                'charm': 'foo',
-                'service': 'foo',
-                'settings': {
-                    'dir': {
-                        'default': 'true',
-                        'description': 'bla bla',
-                        'type': 'string',
-                        'value': '/tmp/charm-dir',
-                    }
-                }
-            })
-        expected = yaml.safe_load(output())
-        client = EnvJujuClient1X(SimpleEnvironment('bar', {}), None, '/foo')
-        with patch.object(client, 'get_juju_output',
-                          side_effect=output) as gjo_mock:
-            results = client.get_config('foo')
-        self.assertEqual(expected, results)
-        gjo_mock.assert_called_once_with('get', 'foo')
-
-    def test_get_service_config(self):
-        def output(*args, **kwargs):
-            return yaml.safe_dump({
-                'charm': 'foo',
-                'service': 'foo',
-                'settings': {
-                    'dir': {
-                        'default': 'true',
-                        'description': 'bla bla',
-                        'type': 'string',
-                        'value': '/tmp/charm-dir',
-                    }
-                }
-            })
-        expected = yaml.safe_load(output())
-        client = EnvJujuClient1X(SimpleEnvironment('bar', {}), None, '/foo')
-        with patch.object(client, 'get_juju_output', side_effect=output):
-            results = client.get_service_config('foo')
-        self.assertEqual(expected, results)
-
-    def test_get_service_config_timesout(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, '/foo')
-        with patch('jujupy.until_timeout', return_value=range(0)):
-            with self.assertRaisesRegexp(
-                    Exception, 'Timed out waiting for juju get'):
-                client.get_service_config('foo')
-
-    def test_ssh_keys(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, None)
-        given_output = 'ssh keys output'
-        with patch.object(client, 'get_juju_output', autospec=True,
-                          return_value=given_output) as mock:
-            output = client.ssh_keys()
-        self.assertEqual(output, given_output)
-        mock.assert_called_once_with('authorized-keys list')
-
-    def test_ssh_keys_full(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, None)
-        given_output = 'ssh keys full output'
-        with patch.object(client, 'get_juju_output', autospec=True,
-                          return_value=given_output) as mock:
-            output = client.ssh_keys(full=True)
-        self.assertEqual(output, given_output)
-        mock.assert_called_once_with('authorized-keys list', '--full')
-
-    def test_add_ssh_key(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, None)
-        with patch.object(client, 'get_juju_output', autospec=True,
-                          return_value='') as mock:
-            output = client.add_ssh_key('ak', 'bk')
-        self.assertEqual(output, '')
-        mock.assert_called_once_with(
-            'authorized-keys add', 'ak', 'bk', merge_stderr=True)
-
-    def test_remove_ssh_key(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, None)
-        with patch.object(client, 'get_juju_output', autospec=True,
-                          return_value='') as mock:
-            output = client.remove_ssh_key('ak', 'bk')
-        self.assertEqual(output, '')
-        mock.assert_called_once_with(
-            'authorized-keys delete', 'ak', 'bk', merge_stderr=True)
-
-    def test_import_ssh_key(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, None)
-        with patch.object(client, 'get_juju_output', autospec=True,
-                          return_value='') as mock:
-            output = client.import_ssh_key('gh:au', 'lp:bu')
-        self.assertEqual(output, '')
-        mock.assert_called_once_with(
-            'authorized-keys import', 'gh:au', 'lp:bu', merge_stderr=True)
-
-    def test_disable_commands_properties(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, None)
-        self.assertEqual(
-            'destroy-environment', client.command_set_destroy_model)
-        self.assertEqual('remove-object', client.command_set_remove_object)
-        self.assertEqual('all-changes', client.command_set_all)
-
-    def test_list_disabled_commands(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, None)
-        with patch.object(client, 'get_juju_output', autospec=True,
-                          return_value=dedent("""\
-             - command-set: destroy-model
-               message: Lock Models
-             - command-set: remove-object""")) as mock:
-            output = client.list_disabled_commands()
-        self.assertEqual([{'command-set': 'destroy-model',
-                           'message': 'Lock Models'},
-                          {'command-set': 'remove-object'}], output)
-        mock.assert_called_once_with('block list',
-                                     '--format', 'yaml')
-
-    def test_disable_command(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, None)
-        with patch.object(client, 'juju', autospec=True) as mock:
-            client.disable_command('all', 'message')
-        mock.assert_called_once_with('block all', ('message', ))
-
-    def test_enable_command(self):
-        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, None)
-        with patch.object(client, 'juju', autospec=True) as mock:
-            client.enable_command('all')
-        mock.assert_called_once_with('unblock', 'all')
+    def test_switch(self):
+        def run_switch_test(expect, model=None, controller=None):
+            client = ModelClient(JujuData('foo'), None, None)
+            with patch.object(client, 'juju', autospec=True) as mock:
+                client.switch(model=model, controller=controller)
+            mock.assert_called_once_with('switch', (expect,), include_e=False)
+        run_switch_test('default', 'default')
+        run_switch_test('other', controller='other')
+        run_switch_test('other:default', 'default', 'other')
+
+    def test_switch_no_target(self):
+        client = ModelClient(JujuData('foo'), None, None)
+        self.assertRaises(ValueError, client.switch)
 
 
 class TestUniquifyLocal(TestCase):
@@ -5223,7 +3537,7 @@
 @contextmanager
 def bootstrap_context(client=None):
     # Avoid unnecessary syscalls.
-    with patch('jujupy.check_free_disk_space'):
+    with patch('jujupy.client.check_free_disk_space'):
         with scoped_environ():
             with temp_dir() as fake_home:
                 os.environ['JUJU_HOME'] = fake_home
@@ -5273,7 +3587,7 @@
         with temp_dir() as juju_home:
             env = JujuData('foo', {'type': 'local'}, juju_home=juju_home)
             client = fake_juju_client(env)
-            with patch('jujupy.check_free_disk_space'):
+            with patch('jujupy.client.check_free_disk_space'):
                 config = make_safe_config(client)
         self.assertEqual(get_local_root(client.env.juju_home, client.env),
                          config['root-dir'])
@@ -5336,7 +3650,7 @@
             return juju_home
 
         with patch('utility.mkdtemp', side_effect=side_effect):
-            with patch('jujupy.check_free_disk_space', autospec=True):
+            with patch('jujupy.client.check_free_disk_space', autospec=True):
                 with temp_bootstrap_env(self.home_dir, client) as temp_home:
                     pass
         self.assertEqual(temp_home, juju_home)
@@ -5346,7 +3660,7 @@
         client = self.get_client(env)
         os.environ['JUJU_HOME'] = 'foo'
         os.environ['JUJU_DATA'] = 'bar'
-        with patch('jujupy.check_free_disk_space', autospec=True):
+        with patch('jujupy.client.check_free_disk_space', autospec=True):
             with temp_bootstrap_env(self.home_dir, client, set_home=False):
                 self.assertEqual(os.environ['JUJU_HOME'], 'foo')
                 self.assertEqual(os.environ['JUJU_DATA'], 'bar')
@@ -5389,7 +3703,7 @@
         env = SimpleEnvironment('qux', {'type': 'local'})
         with bootstrap_context() as fake_home:
             client = self.get_client(env)
-            with patch('jujupy.check_free_disk_space') as mock_cfds:
+            with patch('jujupy.client.check_free_disk_space') as mock_cfds:
                 with temp_bootstrap_env(fake_home, client):
                     stub_bootstrap(client)
         self.assertEqual(mock_cfds.mock_calls, [
@@ -5401,7 +3715,7 @@
         env = SimpleEnvironment('qux', {'type': 'local', 'container': 'kvm'})
         with bootstrap_context() as fake_home:
             client = self.get_client(env)
-            with patch('jujupy.check_free_disk_space') as mock_cfds:
+            with patch('jujupy.client.check_free_disk_space') as mock_cfds:
                 with temp_bootstrap_env(fake_home, client):
                     stub_bootstrap(client)
         self.assertEqual(mock_cfds.mock_calls, [
@@ -5561,6 +3875,10 @@
 
 class TestStatus(FakeHomeTestCase):
 
+    def test_model_name(self):
+        status = Status({'model': {'name': 'bar'}}, '')
+        self.assertEqual('bar', status.model_name)
+
     def test_iter_machines_no_containers(self):
         status = Status({
             'machines': {
@@ -6181,11 +4499,12 @@
             status.get_machine_dns_name('2')
 
     def test_from_text(self):
-        text = TestEnvJujuClient.make_status_yaml(
+        text = TestModelClient.make_status_yaml(
             'agent-state', 'pending', 'horsefeathers')
         status = Status.from_text(text)
         self.assertEqual(status.status_text, text)
         self.assertEqual(status.status, {
+            'model': {'name': 'foo'},
             'machines': {'0': {'agent-state': 'pending'}},
             'applications': {'jenkins': {'units': {'jenkins/0': {
                 'agent-state': 'horsefeathers'}}}}
@@ -6400,103 +4719,6 @@
         self.assertEqual({'application': {}}, status.get_applications())
 
 
-class TestStatus1X(FakeHomeTestCase):
-
-    def test_get_applications_gets_services(self):
-        status = Status1X({
-            'services': {'service': {}},
-            'applications': {'application': {}},
-            }, '')
-        self.assertEqual({'service': {}}, status.get_applications())
-
-    def test_condense_status(self):
-        status = Status1X({}, '')
-        self.assertEqual(status.condense_status(
-                             {'agent-state': 'started',
-                              'agent-state-info': 'all good',
-                              'agent-version': '1.25.1'}),
-                         {'current': 'started', 'message': 'all good',
-                          'version': '1.25.1'})
-
-    def test_condense_status_no_info(self):
-        status = Status1X({}, '')
-        self.assertEqual(status.condense_status(
-                             {'agent-state': 'started',
-                              'agent-version': '1.25.1'}),
-                         {'current': 'started', 'version': '1.25.1'})
-
-    @staticmethod
-    def run_iter_status():
-        status = Status1X({
-            'environment': 'fake-unit-test',
-            'machines': {
-                '0': {
-                    'agent-state': 'started',
-                    'agent-state-info': 'all good',
-                    'agent-version': '1.25.1',
-                    },
-                },
-            'services': {
-                'dummy-sink': {
-                    'units': {
-                        'dummy-sink/0': {
-                            'agent-state': 'started',
-                            'agent-version': '1.25.1',
-                            },
-                        'dummy-sink/1': {
-                            'workload-status': {
-                                'current': 'active',
-                                },
-                            'agent-status': {
-                                'current': 'executing',
-                                },
-                            'agent-state': 'started',
-                            'agent-version': '1.25.1',
-                            },
-                        }
-                    },
-                'dummy-source': {
-                    'service-status': {
-                        'current': 'active',
-                        },
-                    'units': {
-                        'dummy-source/0': {
-                            'agent-state': 'started',
-                            'agent-version': '1.25.1',
-                            }
-                        }
-                    },
-                },
-            }, '')
-        for sub_status in status.iter_status():
-            yield sub_status
-
-    def test_iter_status_range(self):
-        status_set = set([(status_item.item_name, status_item.status_name,
-                           status_item.current)
-                          for status_item in self.run_iter_status()])
-        APP = StatusItem.APPLICATION
-        WORK = StatusItem.WORKLOAD
-        JUJU = StatusItem.JUJU
-        self.assertEqual({
-            ('0', JUJU, 'started'), ('dummy-sink/0', JUJU, 'started'),
-            ('dummy-sink/1', JUJU, 'executing'),
-            ('dummy-sink/1', WORK, 'active'), ('dummy-source', APP, 'active'),
-            ('dummy-source/0', JUJU, 'started'),
-            }, status_set)
-
-    def test_iter_status_data(self):
-        iterator = self.run_iter_status()
-        self.assertEqual(iterator.next().status,
-                         dict(current='started', message='all good',
-                              version='1.25.1'))
-
-
-def fast_timeout(count):
-    if False:
-        yield
-
-
 @contextmanager
 def temp_config():
     with temp_dir() as home:
@@ -6956,6 +5178,108 @@
         juju_data = JujuData('foo', {'type': 'bar'}, 'home', cloud_name='baz')
         self.assertEqual('baz', juju_data.get_cloud())
 
+    def test_set_region(self):
+        env = JujuData('foo', {'type': 'bar'}, 'home')
+        env.set_region('baz')
+        self.assertEqual(env.get_option('region'), 'baz')
+        self.assertEqual(env.get_region(), 'baz')
+
+    def test_set_region_no_provider(self):
+        env = JujuData('foo', {}, 'home')
+        env.set_region('baz')
+        self.assertEqual(env.get_option('region'), 'baz')
+
+    def test_set_region_joyent(self):
+        env = JujuData('foo', {'type': 'joyent'}, 'home')
+        env.set_region('baz')
+        self.assertEqual(env.get_option('sdc-url'),
+                         'https://baz.api.joyentcloud.com')
+        self.assertEqual(env.get_region(), 'baz')
+
+    def test_set_region_azure(self):
+        env = JujuData('foo', {'type': 'azure'}, 'home')
+        env.set_region('baz')
+        self.assertEqual(env.get_option('location'), 'baz')
+        self.assertEqual(env.get_region(), 'baz')
+
+    def test_set_region_lxd(self):
+        env = JujuData('foo', {'type': 'lxd'}, 'home')
+        env.set_region('baz')
+        self.assertEqual(env.get_option('region'), 'baz')
+
+    def test_set_region_manual(self):
+        env = JujuData('foo', {'type': 'manual'}, 'home')
+        env.set_region('baz')
+        self.assertEqual(env.get_option('bootstrap-host'), 'baz')
+        self.assertEqual(env.get_region(), 'baz')
+
+    def test_set_region_manual_named_cloud(self):
+        env = JujuData('foo', {'type': 'manual'}, 'home', cloud_name='qux')
+        env.set_region('baz')
+        self.assertEqual(env.get_option('region'), 'baz')
+        self.assertEqual(env.get_region(), 'baz')
+
+    def test_set_region_maas(self):
+        env = JujuData('foo', {'type': 'maas'}, 'home')
+        with self.assertRaisesRegexp(ValueError,
+                                     'Only None allowed for maas.'):
+            env.set_region('baz')
+        env.set_region(None)
+        self.assertIs(env.get_region(), None)
+
+    def test_get_region(self):
+        self.assertEqual(
+            'bar', JujuData(
+                'foo', {'type': 'foo', 'region': 'bar'}, 'home').get_region())
+
+    def test_get_region_old_azure(self):
+        self.assertEqual('northeu', JujuData('foo', {
+            'type': 'azure', 'location': 'North EU'}, 'home').get_region())
+
+    def test_get_region_azure_arm(self):
+        self.assertEqual('bar', JujuData('foo', {
+            'type': 'azure', 'location': 'bar', 'tenant-id': 'baz'},
+            'home').get_region())
+
+    def test_get_region_joyent(self):
+        self.assertEqual('bar', JujuData('foo', {
+            'type': 'joyent', 'sdc-url': 'https://bar.api.joyentcloud.com'},
+            'home').get_region())
+
+    def test_get_region_lxd(self):
+        self.assertEqual('localhost', JujuData(
+            'foo', {'type': 'lxd'}, 'home').get_region())
+
+    def test_get_region_lxd_specified(self):
+        self.assertEqual('foo', JujuData(
+            'foo', {'type': 'lxd', 'region': 'foo'}, 'home').get_region())
+
+    def test_get_region_maas(self):
+        self.assertIs(None, JujuData('foo', {
+            'type': 'maas', 'region': 'bar',
+        }, 'home').get_region())
+
+    def test_get_region_manual(self):
+        self.assertEqual('baz', JujuData('foo', {
+            'type': 'manual', 'region': 'bar',
+            'bootstrap-host': 'baz'}, 'home').get_region())
+
+    def test_get_region_manual_named_cloud(self):
+        self.assertEqual('bar', JujuData('foo', {
+            'type': 'manual', 'region': 'bar',
+            'bootstrap-host': 'baz'}, 'home', cloud_name='qux').get_region())
+
+    def test_get_region_manual_cloud_name_manual(self):
+        self.assertEqual('baz', JujuData('foo', {
+            'type': 'manual', 'region': 'bar',
+            'bootstrap-host': 'baz'}, 'home',
+            cloud_name='manual').get_region())
+
+    def test_get_region_manual_named_cloud_no_region(self):
+        self.assertIs(None, JujuData('foo', {
+            'type': 'manual', 'bootstrap-host': 'baz',
+            }, 'home', cloud_name='qux').get_region())
+
     def test_dump_yaml(self):
         cloud_dict = {'clouds': {'foo': {}}}
         credential_dict = {'credential': {'bar': {}}}

=== added file 'jujupy/tests/test_version_client.py'
--- jujupy/tests/test_version_client.py	1970-01-01 00:00:00 +0000
+++ jujupy/tests/test_version_client.py	2017-01-20 20:03:12 +0000
@@ -0,0 +1,2234 @@
+from contextlib import contextmanager
+import copy
+from datetime import (
+    datetime,
+    timedelta,
+    )
+import json
+import os
+import subprocess
+import sys
+from textwrap import dedent
+
+from mock import (
+    call,
+    Mock,
+    patch,
+    )
+import yaml
+
+from jujuconfig import (
+    get_jenv_path,
+    )
+from jujupy import (
+    AuthNotAccepted,
+    fake_juju_client,
+    InvalidEndpoint,
+    JujuData,
+    JUJU_DEV_FEATURE_FLAGS,
+    JESNotSupported,
+    ModelClient,
+    NameNotAccepted,
+    SimpleEnvironment,
+    Status,
+    _temp_env as temp_env,
+    TypeNotAccepted,
+    )
+from jujupy.client import (
+    CannotConnectEnv,
+    GroupReporter,
+    StatusItem,
+    StatusNotMet,
+    SYSTEM,
+    UpgradeMongoNotSupported,
+    WaitMachineNotPresent,
+    )
+from jujupy.fake import (
+    FakeBackend2_1,
+    )
+from jujupy.version_client import (
+    BootstrapMismatch,
+    client_from_config,
+    EnvJujuClient1X,
+    EnvJujuClient22,
+    EnvJujuClient24,
+    EnvJujuClient25,
+    ModelClientRC,
+    IncompatibleConfigClass,
+    Juju1XBackend,
+    ModelClient2_1,
+    VersionNotTestedError,
+    Status1X,
+    )
+from jujupy.tests.test_client import (
+    ClientTest,
+    )
+from tests import (
+    assert_juju_call,
+    FakeHomeTestCase,
+    FakePopen,
+    observable_temp_file,
+    TestCase
+    )
+from utility import (
+    get_timeout_path,
+    )
+
+
+class TestJuju1XBackend(TestCase):
+
+    def test_full_args_model(self):
+        backend = Juju1XBackend('/bin/path/juju', '1.25', set(), False, None)
+        full = backend.full_args('help', ('commands',), 'test', None)
+        self.assertEqual(('juju', '--show-log', 'help', '-e', 'test',
+                          'commands'), full)
+
+
+class TestClientFromConfig(ClientTest):
+
+    @contextmanager
+    def assertRaisesVersionNotTested(self, version):
+        with self.assertRaisesRegexp(
+                VersionNotTestedError, 'juju ' + version):
+            yield
+
+    @patch.object(JujuData, 'from_config', return_value=JujuData('', {}))
+    @patch.object(SimpleEnvironment, 'from_config',
+                  return_value=SimpleEnvironment('', {}))
+    @patch.object(ModelClient, 'get_full_path', return_value='fake-path')
+    def test_from_config(self, gfp_mock, se_fc_mock, jd_fc_mock):
+        def juju_cmd_iterator():
+            yield '1.17'
+            yield '1.16'
+            yield '1.16.1'
+            yield '1.15'
+            yield '1.22.1'
+            yield '1.24-alpha1'
+            yield '1.24.7'
+            yield '1.25.1'
+            yield '1.26.1'
+            yield '1.27.1'
+            yield '2.0-alpha1'
+            yield '2.0-alpha2'
+            yield '2.0-alpha3'
+            yield '2.0-beta1'
+            yield '2.0-beta2'
+            yield '2.0-beta3'
+            yield '2.0-beta4'
+            yield '2.0-beta5'
+            yield '2.0-beta6'
+            yield '2.0-beta7'
+            yield '2.0-beta8'
+            yield '2.0-beta9'
+            yield '2.0-beta10'
+            yield '2.0-beta11'
+            yield '2.0-beta12'
+            yield '2.0-beta13'
+            yield '2.0-beta14'
+            yield '2.0-beta15'
+            yield '2.0-rc1'
+            yield '2.0-rc2'
+            yield '2.0-rc3'
+            yield '2.0-delta1'
+            yield '2.2.0'
+
+        context = patch.object(
+            ModelClient, 'get_version',
+            side_effect=juju_cmd_iterator().send)
+        with context:
+            self.assertIs(EnvJujuClient1X,
+                          type(client_from_config('foo', None)))
+
+            def test_fc(version, cls):
+                if cls is not None:
+                    client = client_from_config('foo', None)
+                    if isinstance(client, EnvJujuClient1X):
+                        self.assertEqual(se_fc_mock.return_value, client.env)
+                    else:
+                        self.assertEqual(jd_fc_mock.return_value, client.env)
+                    self.assertIs(cls, type(client))
+                    self.assertEqual(version, client.version)
+                else:
+                    with self.assertRaisesVersionNotTested(version):
+                        client_from_config('foo', None)
+
+            test_fc('1.16', None)
+            test_fc('1.16.1', None)
+            test_fc('1.15', EnvJujuClient1X)
+            test_fc('1.22.1', EnvJujuClient22)
+            test_fc('1.24-alpha1', EnvJujuClient24)
+            test_fc('1.24.7', EnvJujuClient24)
+            test_fc('1.25.1', EnvJujuClient25)
+            test_fc('1.26.1', None)
+            test_fc('1.27.1', EnvJujuClient1X)
+            test_fc('2.0-alpha1', None)
+            test_fc('2.0-alpha2', None)
+            test_fc('2.0-alpha3', None)
+            test_fc('2.0-beta1', None)
+            test_fc('2.0-beta2', None)
+            test_fc('2.0-beta3', None)
+            test_fc('2.0-beta4', None)
+            test_fc('2.0-beta5', None)
+            test_fc('2.0-beta6', None)
+            test_fc('2.0-beta7', None)
+            test_fc('2.0-beta8', None)
+            test_fc('2.0-beta9', None)
+            test_fc('2.0-beta10', None)
+            test_fc('2.0-beta11', None)
+            test_fc('2.0-beta12', None)
+            test_fc('2.0-beta13', None)
+            test_fc('2.0-beta14', None)
+            test_fc('2.0-beta15', None)
+            test_fc('2.0-rc1', ModelClientRC)
+            test_fc('2.0-rc2', ModelClientRC)
+            test_fc('2.0-rc3', ModelClientRC)
+            test_fc('2.0-delta1', ModelClient2_1)
+            test_fc('2.2.0', ModelClient)
+            with self.assertRaises(StopIteration):
+                client_from_config('foo', None)
+
+    def test_client_from_config_path(self):
+        with patch('subprocess.check_output', return_value=' 4.3') as vsn:
+            with patch.object(JujuData, 'from_config'):
+                client = client_from_config('foo', 'foo/bar/qux')
+        vsn.assert_called_once_with(('foo/bar/qux', '--version'))
+        self.assertNotEqual(client.full_path, 'foo/bar/qux')
+        self.assertEqual(client.full_path, os.path.abspath('foo/bar/qux'))
+
+    def test_client_from_config_keep_home(self):
+        env = JujuData({}, juju_home='/foo/bar')
+        with patch('subprocess.check_output', return_value='2.0.0'):
+            with patch.object(JujuData, 'from_config',
+                              side_effect=lambda x: JujuData(x, {})):
+                client_from_config('foo', 'foo/bar/qux')
+        self.assertEqual('/foo/bar', env.juju_home)
+
+    def test_client_from_config_deadline(self):
+        deadline = datetime(2012, 11, 10, 9, 8, 7)
+        with patch('subprocess.check_output', return_value='2.0.0'):
+            with patch.object(JujuData, 'from_config',
+                              side_effect=lambda x: JujuData(x, {})):
+                client = client_from_config(
+                    'foo', 'foo/bar/qux', soft_deadline=deadline)
+        self.assertEqual(client._backend.soft_deadline, deadline)
+
+
+class TestModelClient2_1(ClientTest):
+
+    client_version = '2.1.0'
+
+    client_class = ModelClient2_1
+
+    fake_backend_class = FakeBackend2_1
+
+    def test_basics(self):
+        self.check_basics()
+
+    def test_add_cloud_interactive_maas(self):
+        client = self.fake_juju_client()
+        clouds = {'foo': {
+            'type': 'maas',
+            'endpoint': 'http://bar.example.com',
+            }}
+        client.add_cloud_interactive('foo', clouds['foo'])
+        self.assertEqual(client._backend.clouds, clouds)
+
+    def test_add_cloud_interactive_maas_invalid_endpoint(self):
+        client = self.fake_juju_client()
+        clouds = {'foo': {
+            'type': 'maas',
+            'endpoint': 'B' * 4000,
+            }}
+        with self.assertRaises(InvalidEndpoint):
+            client.add_cloud_interactive('foo', clouds['foo'])
+
+    def test_add_cloud_interactive_manual(self):
+        client = self.fake_juju_client()
+        clouds = {'foo': {'type': 'manual', 'endpoint': '127.100.100.1'}}
+        client.add_cloud_interactive('foo', clouds['foo'])
+        self.assertEqual(client._backend.clouds, clouds)
+
+    def test_add_cloud_interactive_manual_invalid_endpoint(self):
+        client = self.fake_juju_client()
+        clouds = {'foo': {'type': 'manual', 'endpoint': 'B' * 4000}}
+        with self.assertRaises(InvalidEndpoint):
+            client.add_cloud_interactive('foo', clouds['foo'])
+
+    def get_openstack_clouds(self):
+        return {'foo': {
+            'type': 'openstack',
+            'endpoint': 'http://bar.example.com',
+            'auth-types': ['oauth1', 'oauth12'],
+            'regions': {
+                'harvey': {'endpoint': 'http://harvey.example.com'},
+                'steve': {'endpoint': 'http://steve.example.com'},
+                }
+            }}
+
+    def test_add_cloud_interactive_openstack(self):
+        client = self.fake_juju_client()
+        clouds = self.get_openstack_clouds()
+        client.add_cloud_interactive('foo', clouds['foo'])
+        self.assertEqual(client._backend.clouds, clouds)
+
+    def test_add_cloud_interactive_openstack_invalid_endpoint(self):
+        client = self.fake_juju_client()
+        clouds = self.get_openstack_clouds()
+        clouds['foo']['endpoint'] = 'B' * 4000
+        with self.assertRaises(InvalidEndpoint):
+            client.add_cloud_interactive('foo', clouds['foo'])
+
+    def test_add_cloud_interactive_openstack_invalid_region_endpoint(self):
+        client = self.fake_juju_client()
+        clouds = self.get_openstack_clouds()
+        clouds['foo']['regions']['harvey']['endpoint'] = 'B' * 4000
+        with self.assertRaises(InvalidEndpoint):
+            client.add_cloud_interactive('foo', clouds['foo'])
+
+    def test_add_cloud_interactive_openstack_invalid_auth(self):
+        client = self.fake_juju_client()
+        clouds = self.get_openstack_clouds()
+        clouds['foo']['auth-types'] = ['invalid', 'oauth12']
+        with self.assertRaises(AuthNotAccepted):
+            client.add_cloud_interactive('foo', clouds['foo'])
+
+    def test_add_cloud_interactive_vsphere(self):
+        client = self.fake_juju_client()
+        clouds = {'foo': {
+            'type': 'vsphere',
+            'endpoint': 'http://bar.example.com',
+            'regions': {
+                'harvey': {},
+                'steve': {},
+                }
+            }}
+        client.add_cloud_interactive('foo', clouds['foo'])
+        self.assertEqual(client._backend.clouds, clouds)
+
+    def test_add_cloud_interactive_bogus(self):
+        client = self.fake_juju_client()
+        clouds = {'foo': {'type': 'bogus'}}
+        with self.assertRaises(TypeNotAccepted):
+            client.add_cloud_interactive('foo', clouds['foo'])
+
+    def test_add_cloud_interactive_invalid_name(self):
+        client = self.fake_juju_client()
+        cloud = {'type': 'manual', 'endpoint': 'example.com'}
+        with self.assertRaises(NameNotAccepted):
+            client.add_cloud_interactive('invalid/name', cloud)
+
+
+class TestModelClientRC(ClientTest):
+
+    def test_bootstrap(self):
+        env = JujuData('foo', {'type': 'bar', 'region': 'baz'})
+        with observable_temp_file() as config_file:
+            with patch.object(ModelClientRC, 'juju') as mock:
+                client = ModelClientRC(env, '2.0-zeta1', None)
+                client.bootstrap()
+                mock.assert_called_with(
+                    'bootstrap', ('--constraints', 'mem=2G',
+                                  'foo', 'bar/baz',
+                                  '--config', config_file.name,
+                                  '--default-model', 'foo',
+                                  '--agent-version', '2.0'), include_e=False)
+                config_file.seek(0)
+                config = yaml.safe_load(config_file)
+        self.assertEqual({'test-mode': True}, config)
+
+
+class TestEnvJujuClient1X(ClientTest):
+
+    def test_raise_on_juju_data(self):
+        env = JujuData('foo', {'type': 'bar'}, 'baz')
+        with self.assertRaisesRegexp(
+                IncompatibleConfigClass,
+                'JujuData cannot be used with EnvJujuClient1X'):
+            EnvJujuClient1X(env, '1.25', 'full_path')
+
+    def test_no_duplicate_env(self):
+        env = SimpleEnvironment('foo', {})
+        client = EnvJujuClient1X(env, '1.25', 'full_path')
+        self.assertIs(env, client.env)
+
+    def test_not_supported(self):
+        client = EnvJujuClient1X(
+            SimpleEnvironment('foo', {}), '1.25', 'full_path')
+        with self.assertRaises(JESNotSupported):
+            client.add_user_perms('test-user')
+        with self.assertRaises(JESNotSupported):
+            client.grant('test-user', 'read')
+        with self.assertRaises(JESNotSupported):
+            client.revoke('test-user', 'read')
+        with self.assertRaises(JESNotSupported):
+            client.get_model_uuid()
+
+    def test_get_version(self):
+        value = ' 5.6 \n'
+        with patch('subprocess.check_output', return_value=value) as vsn:
+            version = EnvJujuClient1X.get_version()
+        self.assertEqual('5.6', version)
+        vsn.assert_called_with(('juju', '--version'))
+
+    def test_get_version_path(self):
+        with patch('subprocess.check_output', return_value=' 4.3') as vsn:
+            EnvJujuClient1X.get_version('foo/bar/baz')
+        vsn.assert_called_once_with(('foo/bar/baz', '--version'))
+
+    def test_get_matching_agent_version(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', None)
+        self.assertEqual('1.23.1', client.get_matching_agent_version())
+        self.assertEqual('1.23', client.get_matching_agent_version(
+                         no_build=True))
+        client = client.clone(version='1.20-beta1-series-arch')
+        self.assertEqual('1.20-beta1.1', client.get_matching_agent_version())
+
+    def test_upgrade_juju_nonlocal(self):
+        client = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'nonlocal'}), '1.234-76', None)
+        with patch.object(client, 'juju') as juju_mock:
+            client.upgrade_juju()
+        juju_mock.assert_called_with(
+            'upgrade-juju', ('--version', '1.234'))
+
+    def test_upgrade_juju_local(self):
+        client = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
+        with patch.object(client, 'juju') as juju_mock:
+            client.upgrade_juju()
+        juju_mock.assert_called_with(
+            'upgrade-juju', ('--version', '1.234', '--upload-tools',))
+
+    def test_upgrade_juju_no_force_version(self):
+        client = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
+        with patch.object(client, 'juju') as juju_mock:
+            client.upgrade_juju(force_version=False)
+        juju_mock.assert_called_with(
+            'upgrade-juju', ('--upload-tools',))
+
+    def test_upgrade_mongo_exception(self):
+        client = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
+        with self.assertRaises(UpgradeMongoNotSupported):
+            client.upgrade_mongo()
+
+    def test_get_cache_path(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo', juju_home='/foo/'),
+                                 '1.27', 'full/path', debug=True)
+        self.assertEqual('/foo/environments/cache.yaml',
+                         client.get_cache_path())
+
+    def test_bootstrap_maas(self):
+        env = SimpleEnvironment('maas')
+        with patch.object(EnvJujuClient1X, 'juju') as mock:
+            client = EnvJujuClient1X(env, None, None)
+            with patch.object(client.env, 'maas', lambda: True):
+                client.bootstrap()
+        mock.assert_called_once_with('bootstrap', ('--constraints', 'mem=2G'))
+
+    def test_bootstrap_joyent(self):
+        env = SimpleEnvironment('joyent')
+        with patch.object(EnvJujuClient1X, 'juju', autospec=True) as mock:
+            client = EnvJujuClient1X(env, None, None)
+            with patch.object(client.env, 'joyent', lambda: True):
+                client.bootstrap()
+        mock.assert_called_once_with(
+            client, 'bootstrap', ('--constraints', 'mem=2G cpu-cores=1'))
+
+    def test_bootstrap(self):
+        env = SimpleEnvironment('foo')
+        client = EnvJujuClient1X(env, None, None)
+        with patch.object(client, 'juju') as mock:
+            client.bootstrap()
+        mock.assert_called_with('bootstrap', ('--constraints', 'mem=2G'))
+
+    def test_bootstrap_upload_tools(self):
+        env = SimpleEnvironment('foo')
+        client = EnvJujuClient1X(env, None, None)
+        with patch.object(client, 'juju') as mock:
+            client.bootstrap(upload_tools=True)
+        mock.assert_called_with(
+            'bootstrap', ('--upload-tools', '--constraints', 'mem=2G'))
+
+    def test_bootstrap_args(self):
+        env = SimpleEnvironment('foo', {})
+        client = EnvJujuClient1X(env, None, None)
+        with self.assertRaisesRegexp(
+                BootstrapMismatch,
+                '--bootstrap-series angsty does not match default-series:'
+                ' None'):
+            client.bootstrap(bootstrap_series='angsty')
+        env.update_config({
+            'default-series': 'angsty',
+            })
+        with patch.object(client, 'juju') as mock:
+            client.bootstrap(bootstrap_series='angsty')
+        mock.assert_called_with('bootstrap', ('--constraints', 'mem=2G'))
+
+    def test_bootstrap_async(self):
+        env = SimpleEnvironment('foo')
+        with patch.object(ModelClient, 'juju_async', autospec=True) as mock:
+            client = EnvJujuClient1X(env, None, None)
+            client.env.juju_home = 'foo'
+            with client.bootstrap_async():
+                mock.assert_called_once_with(
+                    client, 'bootstrap', ('--constraints', 'mem=2G'))
+
+    def test_bootstrap_async_upload_tools(self):
+        env = SimpleEnvironment('foo')
+        with patch.object(ModelClient, 'juju_async', autospec=True) as mock:
+            client = EnvJujuClient1X(env, None, None)
+            with client.bootstrap_async(upload_tools=True):
+                mock.assert_called_with(
+                    client, 'bootstrap', ('--upload-tools', '--constraints',
+                                          'mem=2G'))
+
+    def test_get_bootstrap_args_bootstrap_series(self):
+        env = SimpleEnvironment('foo', {})
+        client = EnvJujuClient1X(env, None, None)
+        with self.assertRaisesRegexp(
+                BootstrapMismatch,
+                '--bootstrap-series angsty does not match default-series:'
+                ' None'):
+            client.get_bootstrap_args(upload_tools=True,
+                                      bootstrap_series='angsty')
+        env.update_config({'default-series': 'angsty'})
+        args = client.get_bootstrap_args(upload_tools=True,
+                                         bootstrap_series='angsty')
+        self.assertEqual(args, ('--upload-tools', '--constraints', 'mem=2G'))
+
+    def test_create_environment_system(self):
+        self.do_create_environment(
+            'system', 'system create-environment', ('-s', 'foo'))
+
+    def test_create_environment_controller(self):
+        self.do_create_environment(
+            'controller', 'controller create-environment', ('-c', 'foo'))
+
+    def test_create_environment_hypenated_controller(self):
+        self.do_create_environment(
+            'kill-controller', 'create-environment', ('-c', 'foo'))
+
+    def do_create_environment(self, jes_command, create_cmd,
+                              controller_option):
+        controller_client = EnvJujuClient1X(SimpleEnvironment('foo'), '1.26.1',
+                                            None)
+        model_env = SimpleEnvironment('bar', {'type': 'foo'})
+        with patch.object(controller_client, 'get_jes_command',
+                          return_value=jes_command):
+            with patch.object(controller_client, 'juju') as juju_mock:
+                with observable_temp_file() as config_file:
+                    controller_client.add_model(model_env)
+        juju_mock.assert_called_once_with(
+            create_cmd, controller_option + (
+                'bar', '--config', config_file.name), include_e=False)
+
+    def test_destroy_environment(self):
+        env = SimpleEnvironment('foo', {'type': 'ec2'})
+        client = EnvJujuClient1X(env, None, None)
+        with patch.object(client, 'juju') as mock:
+            client.destroy_environment()
+        mock.assert_called_with(
+            'destroy-environment', ('foo', '--force', '-y'),
+            check=False, include_e=False, timeout=600)
+
+    def test_destroy_environment_no_force(self):
+        env = SimpleEnvironment('foo', {'type': 'ec2'})
+        client = EnvJujuClient1X(env, None, None)
+        with patch.object(client, 'juju') as mock:
+            client.destroy_environment(force=False)
+            mock.assert_called_with(
+                'destroy-environment', ('foo', '-y'),
+                check=False, include_e=False, timeout=600)
+
+    def test_destroy_environment_azure(self):
+        env = SimpleEnvironment('foo', {'type': 'azure'})
+        client = EnvJujuClient1X(env, None, None)
+        with patch.object(client, 'juju') as mock:
+            client.destroy_environment(force=False)
+            mock.assert_called_with(
+                'destroy-environment', ('foo', '-y'),
+                check=False, include_e=False, timeout=1800)
+
+    def test_destroy_environment_gce(self):
+        env = SimpleEnvironment('foo', {'type': 'gce'})
+        client = EnvJujuClient1X(env, None, None)
+        with patch.object(client, 'juju') as mock:
+            client.destroy_environment(force=False)
+            mock.assert_called_with(
+                'destroy-environment', ('foo', '-y'),
+                check=False, include_e=False, timeout=1200)
+
+    def test_destroy_environment_delete_jenv(self):
+        env = SimpleEnvironment('foo', {'type': 'ec2'})
+        client = EnvJujuClient1X(env, None, None)
+        with patch.object(client, 'juju'):
+            with temp_env({}) as juju_home:
+                client.env.juju_home = juju_home
+                jenv_path = get_jenv_path(juju_home, 'foo')
+                os.makedirs(os.path.dirname(jenv_path))
+                open(jenv_path, 'w')
+                self.assertTrue(os.path.exists(jenv_path))
+                client.destroy_environment(delete_jenv=True)
+                self.assertFalse(os.path.exists(jenv_path))
+
+    def test_destroy_model(self):
+        env = SimpleEnvironment('foo', {'type': 'ec2'})
+        client = EnvJujuClient1X(env, None, None)
+        with patch.object(client, 'juju') as mock:
+            client.destroy_model()
+        mock.assert_called_with(
+            'destroy-environment', ('foo', '-y'),
+            check=False, include_e=False, timeout=600)
+
+    def test_kill_controller(self):
+        client = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'ec2'}), None, None)
+        with patch.object(client, 'juju') as juju_mock:
+            client.kill_controller()
+        juju_mock.assert_called_once_with(
+            'destroy-environment', ('foo', '--force', '-y'), check=False,
+            include_e=False, timeout=600)
+
+    def test_kill_controller_check(self):
+        client = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'ec2'}), None, None)
+        with patch.object(client, 'juju') as juju_mock:
+            client.kill_controller(check=True)
+        juju_mock.assert_called_once_with(
+            'destroy-environment', ('foo', '--force', '-y'), check=True,
+            include_e=False, timeout=600)
+
+    def test_destroy_controller(self):
+        client = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'ec2'}), None, None)
+        with patch.object(client, 'juju') as juju_mock:
+            client.destroy_controller()
+        juju_mock.assert_called_once_with(
+            'destroy-environment', ('foo', '-y'),
+            include_e=False, timeout=600)
+
+    def test_get_juju_output(self):
+        env = SimpleEnvironment('foo')
+        client = EnvJujuClient1X(env, None, 'juju')
+        fake_popen = FakePopen('asdf', None, 0)
+        with patch('subprocess.Popen', return_value=fake_popen) as mock:
+            result = client.get_juju_output('bar')
+        self.assertEqual('asdf', result)
+        self.assertEqual((('juju', '--show-log', 'bar', '-e', 'foo'),),
+                         mock.call_args[0])
+
+    def test_get_juju_output_accepts_varargs(self):
+        env = SimpleEnvironment('foo')
+        fake_popen = FakePopen('asdf', None, 0)
+        client = EnvJujuClient1X(env, None, 'juju')
+        with patch('subprocess.Popen', return_value=fake_popen) as mock:
+            result = client.get_juju_output('bar', 'baz', '--qux')
+        self.assertEqual('asdf', result)
+        self.assertEqual((('juju', '--show-log', 'bar', '-e', 'foo', 'baz',
+                           '--qux'),), mock.call_args[0])
+
+    def test_get_juju_output_stderr(self):
+        env = SimpleEnvironment('foo')
+        fake_popen = FakePopen('Hello', 'Error!', 1)
+        client = EnvJujuClient1X(env, None, 'juju')
+        with self.assertRaises(subprocess.CalledProcessError) as exc:
+            with patch('subprocess.Popen', return_value=fake_popen):
+                client.get_juju_output('bar')
+        self.assertEqual(exc.exception.output, 'Hello')
+        self.assertEqual(exc.exception.stderr, 'Error!')
+
+    def test_get_juju_output_full_cmd(self):
+        env = SimpleEnvironment('foo')
+        fake_popen = FakePopen(None, 'Hello!', 1)
+        client = EnvJujuClient1X(env, None, 'juju')
+        with self.assertRaises(subprocess.CalledProcessError) as exc:
+            with patch('subprocess.Popen', return_value=fake_popen):
+                client.get_juju_output('bar', '--baz', 'qux')
+        self.assertEqual(
+            ('juju', '--show-log', 'bar', '-e', 'foo', '--baz', 'qux'),
+            exc.exception.cmd)
+
+    def test_get_juju_output_accepts_timeout(self):
+        env = SimpleEnvironment('foo')
+        fake_popen = FakePopen('asdf', None, 0)
+        client = EnvJujuClient1X(env, None, 'juju')
+        with patch('subprocess.Popen', return_value=fake_popen) as po_mock:
+            client.get_juju_output('bar', timeout=5)
+        self.assertEqual(
+            po_mock.call_args[0][0],
+            (sys.executable, get_timeout_path(), '5.00', '--', 'juju',
+             '--show-log', 'bar', '-e', 'foo'))
+
+    def test__shell_environ_juju_home(self):
+        client = EnvJujuClient1X(
+            SimpleEnvironment('baz', {'type': 'ec2'}), '1.25-foobar', 'path',
+            'asdf')
+        env = client._shell_environ()
+        self.assertEqual(env['JUJU_HOME'], 'asdf')
+        self.assertNotIn('JUJU_DATA', env)
+
+    def test_juju_output_supplies_path(self):
+        env = SimpleEnvironment('foo')
+        client = EnvJujuClient1X(env, None, '/foobar/bar')
+
+        def check_path(*args, **kwargs):
+            self.assertRegexpMatches(os.environ['PATH'], r'/foobar\:')
+            return FakePopen(None, None, 0)
+        with patch('subprocess.Popen', autospec=True,
+                   side_effect=check_path):
+            client.get_juju_output('cmd', 'baz')
+
+    def test_get_status(self):
+        output_text = dedent("""\
+                - a
+                - b
+                - c
+                """)
+        env = SimpleEnvironment('foo')
+        client = EnvJujuClient1X(env, None, None)
+        with patch.object(client, 'get_juju_output',
+                          return_value=output_text) as gjo_mock:
+            result = client.get_status()
+        gjo_mock.assert_called_once_with(
+            'status', '--format', 'yaml', controller=False)
+        self.assertEqual(Status1X, type(result))
+        self.assertEqual(['a', 'b', 'c'], result.status)
+
+    def test_get_status_retries_on_error(self):
+        env = SimpleEnvironment('foo')
+        client = EnvJujuClient1X(env, None, None)
+        client.attempt = 0
+
+        def get_juju_output(command, *args, **kwargs):
+            if client.attempt == 1:
+                return '"hello"'
+            client.attempt += 1
+            raise subprocess.CalledProcessError(1, command)
+
+        with patch.object(client, 'get_juju_output', get_juju_output):
+            client.get_status()
+
+    def test_get_status_raises_on_timeout_1(self):
+        env = SimpleEnvironment('foo')
+        client = EnvJujuClient1X(env, None, None)
+
+        def get_juju_output(command, *args, **kwargs):
+            raise subprocess.CalledProcessError(1, command)
+
+        with patch.object(client, 'get_juju_output',
+                          side_effect=get_juju_output):
+            with patch('jujupy.client.until_timeout',
+                       lambda x: iter([None, None])):
+                with self.assertRaisesRegexp(
+                        Exception, 'Timed out waiting for juju status'):
+                    client.get_status()
+
+    def test_get_status_raises_on_timeout_2(self):
+        env = SimpleEnvironment('foo')
+        client = EnvJujuClient1X(env, None, None)
+        with patch('jujupy.client.until_timeout',
+                   return_value=iter([1])) as mock_ut:
+            with patch.object(client, 'get_juju_output',
+                              side_effect=StopIteration):
+                with self.assertRaises(StopIteration):
+                    client.get_status(500)
+        mock_ut.assert_called_with(500)
+
+    def test_get_status_controller(self):
+        output_text = """\
+            - a
+            - b
+            - c
+        """
+        env = SimpleEnvironment('foo')
+        client = EnvJujuClient1X(env, None, None)
+        with patch.object(client, 'get_juju_output',
+                          return_value=output_text) as gjo_mock:
+            client.get_status(controller=True)
+        gjo_mock.assert_called_once_with(
+            'status', '--format', 'yaml', controller=True)
+
+    @staticmethod
+    def make_status_yaml(key, machine_value, unit_value):
+        return dedent("""\
+            environment: foo
+            machines:
+              "0":
+                {0}: {1}
+            services:
+              jenkins:
+                units:
+                  jenkins/0:
+                    {0}: {2}
+        """.format(key, machine_value, unit_value))
+
+    def test_deploy_non_joyent(self):
+        env = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
+        with patch.object(env, 'juju') as mock_juju:
+            env.deploy('mondogb')
+        mock_juju.assert_called_with('deploy', ('mondogb',))
+
+    def test_deploy_joyent(self):
+        env = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
+        with patch.object(env, 'juju') as mock_juju:
+            env.deploy('mondogb')
+        mock_juju.assert_called_with('deploy', ('mondogb',))
+
+    def test_deploy_repository(self):
+        env = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
+        with patch.object(env, 'juju') as mock_juju:
+            env.deploy('mondogb', '/home/jrandom/repo')
+        mock_juju.assert_called_with(
+            'deploy', ('mondogb', '--repository', '/home/jrandom/repo'))
+
+    def test_deploy_to(self):
+        env = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
+        with patch.object(env, 'juju') as mock_juju:
+            env.deploy('mondogb', to='0')
+        mock_juju.assert_called_with(
+            'deploy', ('mondogb', '--to', '0'))
+
+    def test_deploy_service(self):
+        env = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
+        with patch.object(env, 'juju') as mock_juju:
+            env.deploy('local:mondogb', service='my-mondogb')
+        mock_juju.assert_called_with(
+            'deploy', ('local:mondogb', 'my-mondogb',))
+
+    def test_upgrade_charm(self):
+        client = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
+        with patch.object(client, 'juju') as mock_juju:
+            client.upgrade_charm('foo-service',
+                                 '/bar/repository/angsty/mongodb')
+        mock_juju.assert_called_once_with(
+            'upgrade-charm', ('foo-service', '--repository',
+                              '/bar/repository',))
+
+    def test_remove_service(self):
+        client = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
+        with patch.object(client, 'juju') as mock_juju:
+            client.remove_service('mondogb')
+        mock_juju.assert_called_with('destroy-service', ('mondogb',))
+
+    def test_status_until_always_runs_once(self):
+        client = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
+        status_txt = self.make_status_yaml('agent-state', 'started', 'started')
+        with patch.object(client, 'get_juju_output', return_value=status_txt):
+            result = list(client.status_until(-1))
+        self.assertEqual(
+            [r.status for r in result], [Status.from_text(status_txt).status])
+
+    def test_status_until_timeout(self):
+        client = EnvJujuClient1X(
+            SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
+        status_txt = self.make_status_yaml('agent-state', 'started', 'started')
+        status_yaml = yaml.safe_load(status_txt)
+
+        def until_timeout_stub(timeout, start=None):
+            return iter([None, None])
+
+        with patch.object(client, 'get_juju_output', return_value=status_txt):
+            with patch('jujupy.client.until_timeout',
+                       side_effect=until_timeout_stub) as ut_mock:
+                result = list(client.status_until(30, 70))
+        self.assertEqual(
+            [r.status for r in result], [status_yaml] * 3)
+        # until_timeout is called by status as well as status_until.
+        self.assertEqual(ut_mock.mock_calls,
+                         [call(60), call(30, start=70), call(60), call(60)])
+
+    def test_add_ssh_machines(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, 'juju')
+        with patch('subprocess.check_call', autospec=True) as cc_mock:
+            client.add_ssh_machines(['m-foo', 'm-bar', 'm-baz'])
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 0)
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-bar'), 1)
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-baz'), 2)
+        self.assertEqual(cc_mock.call_count, 3)
+
+    def test_add_ssh_machines_retry(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, 'juju')
+        with patch('subprocess.check_call', autospec=True,
+                   side_effect=[subprocess.CalledProcessError(None, None),
+                                None, None, None]) as cc_mock:
+            client.add_ssh_machines(['m-foo', 'm-bar', 'm-baz'])
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 0)
+        self.pause_mock.assert_called_once_with(30)
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 1)
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-bar'), 2)
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-baz'), 3)
+        self.assertEqual(cc_mock.call_count, 4)
+
+    def test_add_ssh_machines_fail_on_second_machine(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, 'juju')
+        with patch('subprocess.check_call', autospec=True, side_effect=[
+                None, subprocess.CalledProcessError(None, None), None, None
+                ]) as cc_mock:
+            with self.assertRaises(subprocess.CalledProcessError):
+                client.add_ssh_machines(['m-foo', 'm-bar', 'm-baz'])
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 0)
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-bar'), 1)
+        self.assertEqual(cc_mock.call_count, 2)
+
+    def test_add_ssh_machines_fail_on_second_attempt(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, 'juju')
+        with patch('subprocess.check_call', autospec=True, side_effect=[
+                subprocess.CalledProcessError(None, None),
+                subprocess.CalledProcessError(None, None)]) as cc_mock:
+            with self.assertRaises(subprocess.CalledProcessError):
+                client.add_ssh_machines(['m-foo', 'm-bar', 'm-baz'])
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 0)
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 1)
+        self.assertEqual(cc_mock.call_count, 2)
+
+    def test_wait_for_started(self):
+        value = self.make_status_yaml('agent-state', 'started', 'started')
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        with patch.object(client, 'get_juju_output', return_value=value):
+            client.wait_for_started()
+
+    def test_wait_for_started_timeout(self):
+        value = self.make_status_yaml('agent-state', 'pending', 'started')
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        with patch('jujupy.client.until_timeout',
+                   lambda x, start=None: range(1)):
+            with patch.object(client, 'get_juju_output', return_value=value):
+                writes = []
+                with patch.object(GroupReporter, '_write', autospec=True,
+                                  side_effect=lambda _, s: writes.append(s)):
+                    with self.assertRaisesRegexp(
+                            StatusNotMet,
+                            'Timed out waiting for agents to start in local'):
+                        client.wait_for_started()
+                self.assertEqual(writes, ['pending: 0', ' .', '\n'])
+
+    def test_wait_for_started_start(self):
+        value = self.make_status_yaml('agent-state', 'started', 'pending')
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        now = datetime.now() + timedelta(days=1)
+        with patch('utility.until_timeout.now', return_value=now):
+            with patch.object(client, 'get_juju_output', return_value=value):
+                writes = []
+                with patch.object(GroupReporter, '_write', autospec=True,
+                                  side_effect=lambda _, s: writes.append(s)):
+                    with self.assertRaisesRegexp(
+                            StatusNotMet,
+                            'Timed out waiting for agents to start in local'):
+                        client.wait_for_started(start=now - timedelta(1200))
+                self.assertEqual(writes, ['pending: jenkins/0', '\n'])
+
+    def test_wait_for_started_logs_status(self):
+        value = self.make_status_yaml('agent-state', 'pending', 'started')
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        with patch.object(client, 'get_juju_output', return_value=value):
+            writes = []
+            with patch.object(GroupReporter, '_write', autospec=True,
+                              side_effect=lambda _, s: writes.append(s)):
+                with self.assertRaisesRegexp(
+                        StatusNotMet,
+                        'Timed out waiting for agents to start in local'):
+                    client.wait_for_started(0)
+            self.assertEqual(writes, ['pending: 0', '\n'])
+        self.assertEqual(self.log_stream.getvalue(), 'ERROR %s\n' % value)
+
+    def test_wait_for_subordinate_units(self):
+        value = dedent("""\
+            machines:
+              "0":
+                agent-state: started
+            services:
+              jenkins:
+                units:
+                  jenkins/0:
+                    subordinates:
+                      sub1/0:
+                        agent-state: started
+              ubuntu:
+                units:
+                  ubuntu/0:
+                    subordinates:
+                      sub2/0:
+                        agent-state: started
+                      sub3/0:
+                        agent-state: started
+        """)
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        now = datetime.now() + timedelta(days=1)
+        with patch('utility.until_timeout.now', return_value=now):
+            with patch.object(client, 'get_juju_output', return_value=value):
+                with patch(
+                        'jujupy.client.GroupReporter.update') as update_mock:
+                    with patch(
+                            'jujupy.client.GroupReporter.finish'
+                            ) as finish_mock:
+                        client.wait_for_subordinate_units(
+                            'jenkins', 'sub1', start=now - timedelta(1200))
+        self.assertEqual([], update_mock.call_args_list)
+        finish_mock.assert_called_once_with()
+
+    def test_wait_for_multiple_subordinate_units(self):
+        value = dedent("""\
+            machines:
+              "0":
+                agent-state: started
+            services:
+              ubuntu:
+                units:
+                  ubuntu/0:
+                    subordinates:
+                      sub/0:
+                        agent-state: started
+                  ubuntu/1:
+                    subordinates:
+                      sub/1:
+                        agent-state: started
+        """)
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        now = datetime.now() + timedelta(days=1)
+        with patch('utility.until_timeout.now', return_value=now):
+            with patch.object(client, 'get_juju_output', return_value=value):
+                with patch(
+                        'jujupy.client.GroupReporter.update') as update_mock:
+                    with patch(
+                            'jujupy.client.GroupReporter.finish'
+                            ) as finish_mock:
+                        client.wait_for_subordinate_units(
+                            'ubuntu', 'sub', start=now - timedelta(1200))
+        self.assertEqual([], update_mock.call_args_list)
+        finish_mock.assert_called_once_with()
+
+    def test_wait_for_subordinate_units_checks_slash_in_unit_name(self):
+        value = dedent("""\
+            machines:
+              "0":
+                agent-state: started
+            services:
+              jenkins:
+                units:
+                  jenkins/0:
+                    subordinates:
+                      sub1:
+                        agent-state: started
+        """)
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        now = datetime.now() + timedelta(days=1)
+        with patch('utility.until_timeout.now', return_value=now):
+            with patch.object(client, 'get_juju_output', return_value=value):
+                with self.assertRaisesRegexp(
+                        StatusNotMet,
+                        'Timed out waiting for agents to start in local'):
+                    client.wait_for_subordinate_units(
+                        'jenkins', 'sub1', start=now - timedelta(1200))
+
+    def test_wait_for_subordinate_units_no_subordinate(self):
+        value = dedent("""\
+            machines:
+              "0":
+                agent-state: started
+            services:
+              jenkins:
+                units:
+                  jenkins/0:
+                    agent-state: started
+        """)
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        now = datetime.now() + timedelta(days=1)
+        with patch('utility.until_timeout.now', return_value=now):
+            with patch.object(client, 'get_juju_output', return_value=value):
+                with self.assertRaisesRegexp(
+                        StatusNotMet,
+                        'Timed out waiting for agents to start in local'):
+                    client.wait_for_subordinate_units(
+                        'jenkins', 'sub1', start=now - timedelta(1200))
+
+    def test_wait_for_workload(self):
+        initial_status = Status1X.from_text("""\
+            services:
+              jenkins:
+                units:
+                  jenkins/0:
+                    workload-status:
+                      current: waiting
+                  subordinates:
+                    ntp/0:
+                      workload-status:
+                        current: unknown
+        """)
+        final_status = Status(copy.deepcopy(initial_status.status), None)
+        final_status.status['services']['jenkins']['units']['jenkins/0'][
+            'workload-status']['current'] = 'active'
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        writes = []
+        with patch('utility.until_timeout', autospec=True, return_value=[1]):
+            with patch.object(client, 'get_status', autospec=True,
+                              side_effect=[initial_status, final_status]):
+                with patch.object(GroupReporter, '_write', autospec=True,
+                                  side_effect=lambda _, s: writes.append(s)):
+                    client.wait_for_workloads()
+        self.assertEqual(writes, ['waiting: jenkins/0', '\n'])
+
+    def test_wait_for_workload_all_unknown(self):
+        status = Status.from_text("""\
+            services:
+              jenkins:
+                units:
+                  jenkins/0:
+                    workload-status:
+                      current: unknown
+                  subordinates:
+                    ntp/0:
+                      workload-status:
+                        current: unknown
+        """)
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        writes = []
+        with patch('utility.until_timeout', autospec=True, return_value=[]):
+            with patch.object(client, 'get_status', autospec=True,
+                              return_value=status):
+                with patch.object(GroupReporter, '_write', autospec=True,
+                                  side_effect=lambda _, s: writes.append(s)):
+                    client.wait_for_workloads(timeout=1)
+        self.assertEqual(writes, [])
+
+    def test_wait_for_workload_no_workload_status(self):
+        status = Status.from_text("""\
+            services:
+              jenkins:
+                units:
+                  jenkins/0:
+                    agent-state: active
+        """)
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        writes = []
+        with patch('utility.until_timeout', autospec=True, return_value=[]):
+            with patch.object(client, 'get_status', autospec=True,
+                              return_value=status):
+                with patch.object(GroupReporter, '_write', autospec=True,
+                                  side_effect=lambda _, s: writes.append(s)):
+                    client.wait_for_workloads(timeout=1)
+        self.assertEqual(writes, [])
+
+    def test_wait_for_ha(self):
+        value = yaml.safe_dump({
+            'machines': {
+                '0': {'state-server-member-status': 'has-vote'},
+                '1': {'state-server-member-status': 'has-vote'},
+                '2': {'state-server-member-status': 'has-vote'},
+            },
+            'services': {},
+        })
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        with patch.object(client, 'get_juju_output', return_value=value):
+            client.wait_for_ha()
+
+    def test_wait_for_ha_no_has_vote(self):
+        value = yaml.safe_dump({
+            'machines': {
+                '0': {'state-server-member-status': 'no-vote'},
+                '1': {'state-server-member-status': 'no-vote'},
+                '2': {'state-server-member-status': 'no-vote'},
+            },
+            'services': {},
+        })
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        with patch.object(client, 'get_juju_output', return_value=value):
+            writes = []
+            with patch('jujupy.client.until_timeout', autospec=True,
+                       return_value=[2, 1]):
+                with patch.object(GroupReporter, '_write', autospec=True,
+                                  side_effect=lambda _, s: writes.append(s)):
+                    with self.assertRaisesRegexp(
+                            StatusNotMet,
+                            'Timed out waiting for voting to be enabled.'):
+                        client.wait_for_ha()
+            self.assertEqual(writes[:2], ['no-vote: 0, 1, 2', ' .'])
+            self.assertEqual(writes[2:-1], ['.'] * (len(writes) - 3))
+            self.assertEqual(writes[-1:], ['\n'])
+
+    def test_wait_for_ha_timeout(self):
+        value = yaml.safe_dump({
+            'machines': {
+                '0': {'state-server-member-status': 'has-vote'},
+                '1': {'state-server-member-status': 'has-vote'},
+            },
+            'services': {},
+        })
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        status = client.status_class.from_text(value)
+        with patch('jujupy.client.until_timeout',
+                   lambda x, start=None: range(0)):
+            with patch.object(client, 'get_status', return_value=status):
+                with self.assertRaisesRegexp(
+                        StatusNotMet,
+                        'Timed out waiting for voting to be enabled.'):
+                    client.wait_for_ha()
+
+    def test_wait_for_deploy_started(self):
+        value = yaml.safe_dump({
+            'machines': {
+                '0': {'agent-state': 'started'},
+            },
+            'services': {
+                'jenkins': {
+                    'units': {
+                        'jenkins/1': {'baz': 'qux'}
+                    }
+                }
+            }
+        })
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        with patch.object(client, 'get_juju_output', return_value=value):
+            client.wait_for_deploy_started()
+
+    def test_wait_for_deploy_started_timeout(self):
+        value = yaml.safe_dump({
+            'machines': {
+                '0': {'agent-state': 'started'},
+            },
+            'services': {},
+        })
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        with patch('jujupy.client.until_timeout', lambda x: range(0)):
+            with patch.object(client, 'get_juju_output', return_value=value):
+                with self.assertRaisesRegexp(
+                        StatusNotMet,
+                        'Timed out waiting for applications to start.'):
+                    client.wait_for_deploy_started()
+
+    def test_wait_for_version(self):
+        value = self.make_status_yaml('agent-version', '1.17.2', '1.17.2')
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        with patch.object(client, 'get_juju_output', return_value=value):
+            client.wait_for_version('1.17.2')
+
+    def test_wait_for_version_timeout(self):
+        value = self.make_status_yaml('agent-version', '1.17.2', '1.17.1')
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        writes = []
+        with patch('jujupy.client.until_timeout',
+                   lambda x, start=None: [x]):
+            with patch.object(client, 'get_juju_output', return_value=value):
+                with patch.object(GroupReporter, '_write', autospec=True,
+                                  side_effect=lambda _, s: writes.append(s)):
+                    with self.assertRaisesRegexp(
+                            StatusNotMet, 'Some versions did not update'):
+                        client.wait_for_version('1.17.2')
+        self.assertEqual(writes, ['1.17.1: jenkins/0', ' .', '\n'])
+
+    def test_wait_for_version_handles_connection_error(self):
+        err = subprocess.CalledProcessError(2, 'foo')
+        err.stderr = 'Unable to connect to environment'
+        err = CannotConnectEnv(err)
+        status = self.make_status_yaml('agent-version', '1.17.2', '1.17.2')
+        actions = [err, status]
+
+        def get_juju_output_fake(*args, **kwargs):
+            action = actions.pop(0)
+            if isinstance(action, Exception):
+                raise action
+            else:
+                return action
+
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        with patch.object(client, 'get_juju_output', get_juju_output_fake):
+            client.wait_for_version('1.17.2')
+
+    def test_wait_for_version_raises_non_connection_error(self):
+        err = Exception('foo')
+        status = self.make_status_yaml('agent-version', '1.17.2', '1.17.2')
+        actions = [err, status]
+
+        def get_juju_output_fake(*args, **kwargs):
+            action = actions.pop(0)
+            if isinstance(action, Exception):
+                raise action
+            else:
+                return action
+
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        with patch.object(client, 'get_juju_output', get_juju_output_fake):
+            with self.assertRaisesRegexp(Exception, 'foo'):
+                client.wait_for_version('1.17.2')
+
+    def test_wait_just_machine_0(self):
+        value = yaml.safe_dump({
+            'machines': {
+                '0': {'agent-state': 'started'},
+            },
+        })
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        with patch.object(client, 'get_juju_output', return_value=value):
+            client.wait_for(WaitMachineNotPresent('1'), quiet=True)
+
+    def test_wait_just_machine_0_timeout(self):
+        value = yaml.safe_dump({
+            'machines': {
+                '0': {'agent-state': 'started'},
+                '1': {'agent-state': 'started'},
+            },
+        })
+        client = EnvJujuClient1X(SimpleEnvironment('local'), None, None)
+        with patch.object(client, 'get_juju_output', return_value=value), \
+            patch('jujupy.client.until_timeout',
+                  lambda x, start=None: range(1)), \
+            self.assertRaisesRegexp(
+                Exception,
+                'Timed out waiting for machine removal 1'):
+            client.wait_for(WaitMachineNotPresent('1'), quiet=True)
+
+    def test_set_model_constraints(self):
+        client = EnvJujuClient1X(SimpleEnvironment('bar', {}), None, '/foo')
+        with patch.object(client, 'juju') as juju_mock:
+            client.set_model_constraints({'bar': 'baz'})
+        juju_mock.assert_called_once_with('set-constraints', ('bar=baz',))
+
+    def test_get_model_config(self):
+        env = SimpleEnvironment('foo', None)
+        fake_popen = FakePopen(yaml.safe_dump({'bar': 'baz'}), None, 0)
+        client = EnvJujuClient1X(env, None, 'juju')
+        with patch('subprocess.Popen', return_value=fake_popen) as po_mock:
+            result = client.get_model_config()
+        assert_juju_call(
+            self, po_mock, client, (
+                'juju', '--show-log', 'get-env', '-e', 'foo'))
+        self.assertEqual({'bar': 'baz'}, result)
+
+    def test_get_env_option(self):
+        env = SimpleEnvironment('foo', None)
+        fake_popen = FakePopen('https://example.org/juju/tools', None, 0)
+        client = EnvJujuClient1X(env, None, 'juju')
+        with patch('subprocess.Popen', return_value=fake_popen) as mock:
+            result = client.get_env_option('tools-metadata-url')
+        self.assertEqual(
+            mock.call_args[0][0],
+            ('juju', '--show-log', 'get-env', '-e', 'foo',
+             'tools-metadata-url'))
+        self.assertEqual('https://example.org/juju/tools', result)
+
+    def test_set_env_option(self):
+        env = SimpleEnvironment('foo')
+        client = EnvJujuClient1X(env, None, 'juju')
+        with patch('subprocess.check_call') as mock:
+            client.set_env_option(
+                'tools-metadata-url', 'https://example.org/juju/tools')
+        environ = dict(os.environ)
+        environ['JUJU_HOME'] = client.env.juju_home
+        mock.assert_called_with(
+            ('juju', '--show-log', 'set-env', '-e', 'foo',
+             'tools-metadata-url=https://example.org/juju/tools'))
+
+    def test_unset_env_option(self):
+        env = SimpleEnvironment('foo')
+        client = EnvJujuClient1X(env, None, 'juju')
+        with patch('subprocess.check_call') as mock:
+            client.unset_env_option('tools-metadata-url')
+        environ = dict(os.environ)
+        environ['JUJU_HOME'] = client.env.juju_home
+        mock.assert_called_with(
+            ('juju', '--show-log', 'set-env', '-e', 'foo',
+             'tools-metadata-url='))
+
+    @contextmanager
+    def run_model_defaults_test(self, operation_name):
+        env = SimpleEnvironment('foo', {'type': 'local'})
+        client = EnvJujuClient1X(env, '1.23-series-arch', None)
+        yield client
+        self.assertEqual('INFO No model-defaults stored for client '
+                         '(attempted {}).\n'.format(operation_name),
+                         self.log_stream.getvalue())
+
+    def test_get_model_defaults(self):
+        with self.run_model_defaults_test('get') as client:
+            client.get_model_defaults('some-key')
+
+    def test_set_model_defaults(self):
+        with self.run_model_defaults_test('set') as client:
+            client.set_model_defaults('some-key', 'some-value')
+
+    def test_unset_model_defaults(self):
+        with self.run_model_defaults_test('unset') as client:
+            client.unset_model_defaults('some-key')
+
+    def test_set_testing_agent_metadata_url(self):
+        env = SimpleEnvironment(None, {'type': 'foo'})
+        client = EnvJujuClient1X(env, None, None)
+        with patch.object(client, 'get_env_option') as mock_get:
+            mock_get.return_value = 'https://example.org/juju/tools'
+            with patch.object(client, 'set_env_option') as mock_set:
+                client.set_testing_agent_metadata_url()
+        mock_get.assert_called_with('tools-metadata-url')
+        mock_set.assert_called_with(
+            'tools-metadata-url',
+            'https://example.org/juju/testing/tools')
+
+    def test_set_testing_agent_metadata_url_noop(self):
+        env = SimpleEnvironment(None, {'type': 'foo'})
+        client = EnvJujuClient1X(env, None, None)
+        with patch.object(client, 'get_env_option') as mock_get:
+            mock_get.return_value = 'https://example.org/juju/testing/tools'
+            with patch.object(client, 'set_env_option') as mock_set:
+                client.set_testing_agent_metadata_url()
+        mock_get.assert_called_with('tools-metadata-url')
+        self.assertEqual(0, mock_set.call_count)
+
+    def test_juju(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, 'juju')
+        with patch('subprocess.check_call') as mock:
+            client.juju('foo', ('bar', 'baz'))
+        environ = dict(os.environ)
+        environ['JUJU_HOME'] = client.env.juju_home
+        mock.assert_called_with(('juju', '--show-log', 'foo', '-e', 'qux',
+                                 'bar', 'baz'))
+
+    def test_juju_env(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+
+        def check_path(*args, **kwargs):
+            self.assertRegexpMatches(os.environ['PATH'], r'/foobar\:')
+        with patch('subprocess.check_call', side_effect=check_path):
+            client.juju('foo', ('bar', 'baz'))
+
+    def test_juju_no_check(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, 'juju')
+        environ = dict(os.environ)
+        environ['JUJU_HOME'] = client.env.juju_home
+        with patch('subprocess.call') as mock:
+            client.juju('foo', ('bar', 'baz'), check=False)
+        mock.assert_called_with(('juju', '--show-log', 'foo', '-e', 'qux',
+                                 'bar', 'baz'))
+
+    def test_juju_no_check_env(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+
+        def check_path(*args, **kwargs):
+            self.assertRegexpMatches(os.environ['PATH'], r'/foobar\:')
+        with patch('subprocess.call', side_effect=check_path):
+            client.juju('foo', ('bar', 'baz'), check=False)
+
+    def test_juju_timeout(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+        with patch('subprocess.check_call') as cc_mock:
+            client.juju('foo', ('bar', 'baz'), timeout=58)
+        self.assertEqual(cc_mock.call_args[0][0], (
+            sys.executable, get_timeout_path(), '58.00', '--', 'baz',
+            '--show-log', 'foo', '-e', 'qux', 'bar', 'baz'))
+
+    def test_juju_juju_home(self):
+        env = SimpleEnvironment('qux')
+        os.environ['JUJU_HOME'] = 'foo'
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+
+        def check_home(*args, **kwargs):
+            self.assertEqual(os.environ['JUJU_HOME'], 'foo')
+            yield
+            self.assertEqual(os.environ['JUJU_HOME'], 'asdf')
+            yield
+
+        with patch('subprocess.check_call', side_effect=check_home):
+            client.juju('foo', ('bar', 'baz'))
+            client.env.juju_home = 'asdf'
+            client.juju('foo', ('bar', 'baz'))
+
+    def test_juju_extra_env(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, 'juju')
+        extra_env = {'JUJU': '/juju', 'JUJU_HOME': client.env.juju_home}
+
+        def check_env(*args, **kwargs):
+            self.assertEqual('/juju', os.environ['JUJU'])
+
+        with patch('subprocess.check_call', side_effect=check_env) as mock:
+            client.juju('quickstart', ('bar', 'baz'), extra_env=extra_env)
+        mock.assert_called_with(
+            ('juju', '--show-log', 'quickstart', '-e', 'qux', 'bar', 'baz'))
+
+    def test_juju_backup_with_tgz(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+
+        def check_env(*args, **kwargs):
+            self.assertEqual(os.environ['JUJU_ENV'], 'qux')
+            return 'foojuju-backup-24.tgzz'
+        with patch('subprocess.check_output',
+                   side_effect=check_env) as co_mock:
+            backup_file = client.backup()
+        self.assertEqual(backup_file, os.path.abspath('juju-backup-24.tgz'))
+        assert_juju_call(self, co_mock, client, ['juju', 'backup'])
+
+    def test_juju_backup_with_tar_gz(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+        with patch('subprocess.check_output',
+                   return_value='foojuju-backup-123-456.tar.gzbar'):
+            backup_file = client.backup()
+        self.assertEqual(
+            backup_file, os.path.abspath('juju-backup-123-456.tar.gz'))
+
+    def test_juju_backup_no_file(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+        with patch('subprocess.check_output', return_value=''):
+            with self.assertRaisesRegexp(
+                    Exception, 'The backup file was not found in output'):
+                client.backup()
+
+    def test_juju_backup_wrong_file(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+        with patch('subprocess.check_output',
+                   return_value='mumu-backup-24.tgz'):
+            with self.assertRaisesRegexp(
+                    Exception, 'The backup file was not found in output'):
+                client.backup()
+
+    def test_juju_backup_environ(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+        environ = client._shell_environ()
+        environ['JUJU_ENV'] = client.env.environment
+
+        def side_effect(*args, **kwargs):
+            self.assertEqual(environ, os.environ)
+            return 'foojuju-backup-123-456.tar.gzbar'
+        with patch('subprocess.check_output', side_effect=side_effect):
+            client.backup()
+            self.assertNotEqual(environ, os.environ)
+
+    def test_restore_backup(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+        with patch.object(client, 'get_juju_output') as gjo_mock:
+            result = client.restore_backup('quxx')
+        gjo_mock.assert_called_once_with('restore', '--constraints',
+                                         'mem=2G', 'quxx')
+        self.assertIs(gjo_mock.return_value, result)
+
+    def test_restore_backup_async(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+        with patch.object(client, 'juju_async') as gjo_mock:
+            result = client.restore_backup_async('quxx')
+        gjo_mock.assert_called_once_with(
+            'restore', ('--constraints', 'mem=2G', 'quxx'))
+        self.assertIs(gjo_mock.return_value, result)
+
+    def test_enable_ha(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+        with patch.object(client, 'juju', autospec=True) as eha_mock:
+            client.enable_ha()
+        eha_mock.assert_called_once_with('ensure-availability', ('-n', '3'))
+
+    def test_juju_async(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+        with patch('subprocess.Popen') as popen_class_mock:
+            with client.juju_async('foo', ('bar', 'baz')) as proc:
+                assert_juju_call(self, popen_class_mock, client, (
+                    'baz', '--show-log', 'foo', '-e', 'qux', 'bar', 'baz'))
+                self.assertIs(proc, popen_class_mock.return_value)
+                self.assertEqual(proc.wait.call_count, 0)
+                proc.wait.return_value = 0
+        proc.wait.assert_called_once_with()
+
+    def test_juju_async_failure(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+        with patch('subprocess.Popen') as popen_class_mock:
+            with self.assertRaises(subprocess.CalledProcessError) as err_cxt:
+                with client.juju_async('foo', ('bar', 'baz')):
+                    proc_mock = popen_class_mock.return_value
+                    proc_mock.wait.return_value = 23
+        self.assertEqual(err_cxt.exception.returncode, 23)
+        self.assertEqual(err_cxt.exception.cmd, (
+            'baz', '--show-log', 'foo', '-e', 'qux', 'bar', 'baz'))
+
+    def test_juju_async_environ(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+        environ = client._shell_environ()
+        proc_mock = Mock()
+        with patch('subprocess.Popen') as popen_class_mock:
+
+            def check_environ(*args, **kwargs):
+                self.assertEqual(environ, os.environ)
+                return proc_mock
+            popen_class_mock.side_effect = check_environ
+            proc_mock.wait.return_value = 0
+            with client.juju_async('foo', ('bar', 'baz')):
+                pass
+            self.assertNotEqual(environ, os.environ)
+
+    def test_is_jes_enabled(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+        fake_popen = FakePopen(' %s' % SYSTEM, None, 0)
+        with patch('subprocess.Popen',
+                   return_value=fake_popen) as po_mock:
+            self.assertIsFalse(client.is_jes_enabled())
+        self.assertEqual(0, po_mock.call_count)
+
+    def test_get_jes_command(self):
+        env = SimpleEnvironment('qux')
+        client = EnvJujuClient1X(env, None, '/foobar/baz')
+        # Juju 1.24 and older do not have a JES command. It is an error
+        # to call get_jes_command when is_jes_enabled is False
+        fake_popen = FakePopen(' %s' % SYSTEM, None, 0)
+        with patch('subprocess.Popen',
+                   return_value=fake_popen) as po_mock:
+            with self.assertRaises(JESNotSupported):
+                client.get_jes_command()
+        self.assertEqual(0, po_mock.call_count)
+
+    def test_get_juju_timings(self):
+        env = SimpleEnvironment('foo')
+        client = EnvJujuClient1X(env, None, 'my/juju/bin')
+        client._backend.juju_timings = {("juju", "op1"): [1],
+                                        ("juju", "op2"): [2]}
+        flattened_timings = client.get_juju_timings()
+        expected = {"juju op1": [1], "juju op2": [2]}
+        self.assertEqual(flattened_timings, expected)
+
+    def test_deploy_bundle_1x(self):
+        client = EnvJujuClient1X(SimpleEnvironment('an_env', None),
+                                 '1.23-series-arch', None)
+        with patch.object(client, 'juju') as mock_juju:
+            client.deploy_bundle('bundle:~juju-qa/some-bundle')
+        mock_juju.assert_called_with(
+            'deployer', ('--debug', '--deploy-delay', '10', '--timeout',
+                         '3600', '--config', 'bundle:~juju-qa/some-bundle'))
+
+    def test_deploy_bundle_template(self):
+        client = EnvJujuClient1X(SimpleEnvironment('an_env', None),
+                                 '1.23-series-arch', None)
+        with patch.object(client, 'juju') as mock_juju:
+            client.deploy_bundle('bundle:~juju-qa/some-{container}-bundle')
+        mock_juju.assert_called_with(
+            'deployer', (
+                '--debug', '--deploy-delay', '10', '--timeout', '3600',
+                '--config', 'bundle:~juju-qa/some-lxc-bundle'))
+
+    def test_deployer(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', None)
+        with patch.object(EnvJujuClient1X, 'juju') as mock:
+            client.deployer('bundle:~juju-qa/some-bundle')
+        mock.assert_called_with(
+            'deployer', ('--debug', '--deploy-delay', '10', '--timeout',
+                         '3600', '--config', 'bundle:~juju-qa/some-bundle'))
+
+    def test_deployer_template(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', None)
+        with patch.object(EnvJujuClient1X, 'juju') as mock:
+            client.deployer('bundle:~juju-qa/some-{container}-bundle')
+        mock.assert_called_with(
+            'deployer', (
+                '--debug', '--deploy-delay', '10', '--timeout', '3600',
+                '--config', 'bundle:~juju-qa/some-lxc-bundle'))
+
+    def test_deployer_with_bundle_name(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', None)
+        with patch.object(EnvJujuClient1X, 'juju') as mock:
+            client.deployer('bundle:~juju-qa/some-bundle', 'name')
+        mock.assert_called_with(
+            'deployer', ('--debug', '--deploy-delay', '10', '--timeout',
+                         '3600', '--config', 'bundle:~juju-qa/some-bundle',
+                         'name'))
+
+    def test_quickstart_maas(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'maas'}),
+                                 '1.23-series-arch', '/juju')
+        with patch.object(EnvJujuClient1X, 'juju') as mock:
+            client.quickstart('bundle:~juju-qa/some-bundle')
+        mock.assert_called_with(
+            'quickstart', ('--constraints', 'mem=2G', '--no-browser',
+                           'bundle:~juju-qa/some-bundle'),
+            extra_env={'JUJU': '/juju'})
+
+    def test_quickstart_local(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', '/juju')
+        with patch.object(EnvJujuClient1X, 'juju') as mock:
+            client.quickstart('bundle:~juju-qa/some-bundle')
+        mock.assert_called_with(
+            'quickstart', ('--constraints', 'mem=2G', '--no-browser',
+                           'bundle:~juju-qa/some-bundle'),
+            extra_env={'JUJU': '/juju'})
+
+    def test_quickstart_template(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', '/juju')
+        with patch.object(EnvJujuClient1X, 'juju') as mock:
+            client.quickstart('bundle:~juju-qa/some-{container}-bundle')
+        mock.assert_called_with(
+            'quickstart', (
+                '--constraints', 'mem=2G', '--no-browser',
+                'bundle:~juju-qa/some-lxc-bundle'),
+            extra_env={'JUJU': '/juju'})
+
+    def test_list_models(self):
+        env = SimpleEnvironment('foo', {'type': 'local'})
+        client = EnvJujuClient1X(env, '1.23-series-arch', None)
+        client.list_models()
+        self.assertEqual(
+            'INFO The model is environment foo\n',
+            self.log_stream.getvalue())
+
+    def test__get_models(self):
+        data = """\
+            - name: foo
+              model-uuid: aaaa
+            - name: bar
+              model-uuid: bbbb
+        """
+        env = SimpleEnvironment('foo', {'type': 'local'})
+        client = fake_juju_client(cls=EnvJujuClient1X, env=env)
+        with patch.object(client, 'get_juju_output', return_value=data):
+            models = client._get_models()
+            self.assertEqual(
+                [{'name': 'foo', 'model-uuid': 'aaaa'},
+                 {'name': 'bar', 'model-uuid': 'bbbb'}],
+                models)
+
+    def test__get_models_exception(self):
+        env = SimpleEnvironment('foo', {'type': 'local'})
+        client = fake_juju_client(cls=EnvJujuClient1X, env=env)
+        with patch.object(client, 'get_juju_output',
+                          side_effect=subprocess.CalledProcessError('a', 'b')):
+            self.assertEqual([], client._get_models())
+
+    def test_get_models(self):
+        env = SimpleEnvironment('foo', {'type': 'local'})
+        client = EnvJujuClient1X(env, '1.23-series-arch', None)
+        self.assertEqual({}, client.get_models())
+
+    def test_iter_model_clients(self):
+        data = """\
+            - name: foo
+              model-uuid: aaaa
+              owner: admin@local
+            - name: bar
+              model-uuid: bbbb
+              owner: admin@local
+        """
+        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, None)
+        with patch.object(client, 'get_juju_output', return_value=data):
+            model_clients = list(client.iter_model_clients())
+        self.assertEqual(2, len(model_clients))
+        self.assertIs(client, model_clients[0])
+        self.assertEqual('bar', model_clients[1].env.environment)
+
+    def test_get_controller_client(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo'), {'bar': 'baz'},
+                                 'myhome')
+        controller_client = client.get_controller_client()
+        self.assertIs(client, controller_client)
+
+    def test_list_controllers(self):
+        env = SimpleEnvironment('foo', {'type': 'local'})
+        client = EnvJujuClient1X(env, '1.23-series-arch', None)
+        client.list_controllers()
+        self.assertEqual(
+            'INFO The controller is environment foo\n',
+            self.log_stream.getvalue())
+
+    def test_get_controller_model_name(self):
+        env = SimpleEnvironment('foo', {'type': 'local'})
+        client = EnvJujuClient1X(env, '1.23-series-arch', None)
+        controller_name = client.get_controller_model_name()
+        self.assertEqual('foo', controller_name)
+
+    def test_get_controller_endpoint_ipv4(self):
+        env = SimpleEnvironment('foo', {'type': 'local'})
+        client = EnvJujuClient1X(env, '1.23-series-arch', None)
+        with patch.object(client, 'get_juju_output',
+                          return_value='10.0.0.1:17070') as gjo_mock:
+            endpoint = client.get_controller_endpoint()
+        self.assertEqual(('10.0.0.1', '17070'), endpoint)
+        gjo_mock.assert_called_once_with('api-endpoints')
+
+    def test_get_controller_endpoint_ipv6(self):
+        env = SimpleEnvironment('foo', {'type': 'local'})
+        client = EnvJujuClient1X(env, '1.23-series-arch', None)
+        with patch.object(client, 'get_juju_output',
+                          return_value='[::1]:17070') as gjo_mock:
+            endpoint = client.get_controller_endpoint()
+        self.assertEqual(('::1', '17070'), endpoint)
+        gjo_mock.assert_called_once_with('api-endpoints')
+
+    def test_action_do(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', None)
+        with patch.object(EnvJujuClient1X, 'get_juju_output') as mock:
+            mock.return_value = \
+                "Action queued with id: 5a92ec93-d4be-4399-82dc-7431dbfd08f9"
+            id = client.action_do("foo/0", "myaction", "param=5")
+            self.assertEqual(id, "5a92ec93-d4be-4399-82dc-7431dbfd08f9")
+        mock.assert_called_once_with(
+            'action do', 'foo/0', 'myaction', "param=5"
+        )
+
+    def test_action_do_error(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', None)
+        with patch.object(EnvJujuClient1X, 'get_juju_output') as mock:
+            mock.return_value = "some bad text"
+            with self.assertRaisesRegexp(Exception,
+                                         "Action id not found in output"):
+                client.action_do("foo/0", "myaction", "param=5")
+
+    def test_action_fetch(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', None)
+        with patch.object(EnvJujuClient1X, 'get_juju_output') as mock:
+            ret = "status: completed\nfoo: bar"
+            mock.return_value = ret
+            out = client.action_fetch("123")
+            self.assertEqual(out, ret)
+        mock.assert_called_once_with(
+            'action fetch', '123', "--wait", "1m"
+        )
+
+    def test_action_fetch_timeout(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', None)
+        ret = "status: pending\nfoo: bar"
+        with patch.object(EnvJujuClient1X,
+                          'get_juju_output', return_value=ret):
+            with self.assertRaisesRegexp(Exception,
+                                         "timed out waiting for action"):
+                client.action_fetch("123")
+
+    def test_action_do_fetch(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', None)
+        with patch.object(EnvJujuClient1X, 'get_juju_output') as mock:
+            ret = "status: completed\nfoo: bar"
+            # setting side_effect to an iterable will return the next value
+            # from the list each time the function is called.
+            mock.side_effect = [
+                "Action queued with id: 5a92ec93-d4be-4399-82dc-7431dbfd08f9",
+                ret]
+            out = client.action_do_fetch("foo/0", "myaction", "param=5")
+            self.assertEqual(out, ret)
+
+    def test_run(self):
+        env = SimpleEnvironment('name', {}, 'foo')
+        client = fake_juju_client(cls=EnvJujuClient1X, env=env)
+        run_list = [
+            {"MachineId": "1",
+             "Stdout": "Linux\n",
+             "ReturnCode": 255,
+             "Stderr": "Permission denied (publickey,password)"}]
+        run_output = json.dumps(run_list)
+        with patch.object(client._backend, 'get_juju_output',
+                          return_value=run_output) as gjo_mock:
+            result = client.run(('wname',), applications=['foo', 'bar'])
+        self.assertEqual(run_list, result)
+        gjo_mock.assert_called_once_with(
+            'run', ('--format', 'json', '--service', 'foo,bar', 'wname'),
+            frozenset(['migration']),
+            'foo', 'name', user_name=None)
+
+    def test_list_space(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', None)
+        yaml_dict = {'foo': 'bar'}
+        output = yaml.safe_dump(yaml_dict)
+        with patch.object(client, 'get_juju_output', return_value=output,
+                          autospec=True) as gjo_mock:
+            result = client.list_space()
+        self.assertEqual(result, yaml_dict)
+        gjo_mock.assert_called_once_with('space list')
+
+    def test_add_space(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', None)
+        with patch.object(client, 'juju', autospec=True) as juju_mock:
+            client.add_space('foo-space')
+        juju_mock.assert_called_once_with('space create', ('foo-space'))
+
+    def test_add_subnet(self):
+        client = EnvJujuClient1X(SimpleEnvironment(None, {'type': 'local'}),
+                                 '1.23-series-arch', None)
+        with patch.object(client, 'juju', autospec=True) as juju_mock:
+            client.add_subnet('bar-subnet', 'foo-space')
+        juju_mock.assert_called_once_with('subnet add',
+                                          ('bar-subnet', 'foo-space'))
+
+    def test__shell_environ_uses_pathsep(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo'), None,
+                                 'foo/bar/juju')
+        with patch('os.pathsep', '!'):
+            environ = client._shell_environ()
+        self.assertRegexpMatches(environ['PATH'], r'foo/bar\!')
+
+    def test_set_config(self):
+        client = EnvJujuClient1X(SimpleEnvironment('bar', {}), None, '/foo')
+        with patch.object(client, 'juju') as juju_mock:
+            client.set_config('foo', {'bar': 'baz'})
+        juju_mock.assert_called_once_with('set', ('foo', 'bar=baz'))
+
+    def test_get_config(self):
+        def output(*args, **kwargs):
+            return yaml.safe_dump({
+                'charm': 'foo',
+                'service': 'foo',
+                'settings': {
+                    'dir': {
+                        'default': 'true',
+                        'description': 'bla bla',
+                        'type': 'string',
+                        'value': '/tmp/charm-dir',
+                    }
+                }
+            })
+        expected = yaml.safe_load(output())
+        client = EnvJujuClient1X(SimpleEnvironment('bar', {}), None, '/foo')
+        with patch.object(client, 'get_juju_output',
+                          side_effect=output) as gjo_mock:
+            results = client.get_config('foo')
+        self.assertEqual(expected, results)
+        gjo_mock.assert_called_once_with('get', 'foo')
+
+    def test_get_service_config(self):
+        def output(*args, **kwargs):
+            return yaml.safe_dump({
+                'charm': 'foo',
+                'service': 'foo',
+                'settings': {
+                    'dir': {
+                        'default': 'true',
+                        'description': 'bla bla',
+                        'type': 'string',
+                        'value': '/tmp/charm-dir',
+                    }
+                }
+            })
+        expected = yaml.safe_load(output())
+        client = EnvJujuClient1X(SimpleEnvironment('bar', {}), None, '/foo')
+        with patch.object(client, 'get_juju_output', side_effect=output):
+            results = client.get_service_config('foo')
+        self.assertEqual(expected, results)
+
+    def test_get_service_config_timesout(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, '/foo')
+        with patch('jujupy.client.until_timeout', return_value=range(0)):
+            with self.assertRaisesRegexp(
+                    Exception, 'Timed out waiting for juju get'):
+                client.get_service_config('foo')
+
+    def test_ssh_keys(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, None)
+        given_output = 'ssh keys output'
+        with patch.object(client, 'get_juju_output', autospec=True,
+                          return_value=given_output) as mock:
+            output = client.ssh_keys()
+        self.assertEqual(output, given_output)
+        mock.assert_called_once_with('authorized-keys list')
+
+    def test_ssh_keys_full(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, None)
+        given_output = 'ssh keys full output'
+        with patch.object(client, 'get_juju_output', autospec=True,
+                          return_value=given_output) as mock:
+            output = client.ssh_keys(full=True)
+        self.assertEqual(output, given_output)
+        mock.assert_called_once_with('authorized-keys list', '--full')
+
+    def test_add_ssh_key(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, None)
+        with patch.object(client, 'get_juju_output', autospec=True,
+                          return_value='') as mock:
+            output = client.add_ssh_key('ak', 'bk')
+        self.assertEqual(output, '')
+        mock.assert_called_once_with(
+            'authorized-keys add', 'ak', 'bk', merge_stderr=True)
+
+    def test_remove_ssh_key(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, None)
+        with patch.object(client, 'get_juju_output', autospec=True,
+                          return_value='') as mock:
+            output = client.remove_ssh_key('ak', 'bk')
+        self.assertEqual(output, '')
+        mock.assert_called_once_with(
+            'authorized-keys delete', 'ak', 'bk', merge_stderr=True)
+
+    def test_import_ssh_key(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo', {}), None, None)
+        with patch.object(client, 'get_juju_output', autospec=True,
+                          return_value='') as mock:
+            output = client.import_ssh_key('gh:au', 'lp:bu')
+        self.assertEqual(output, '')
+        mock.assert_called_once_with(
+            'authorized-keys import', 'gh:au', 'lp:bu', merge_stderr=True)
+
+    def test_disable_commands_properties(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, None)
+        self.assertEqual(
+            'destroy-environment', client.command_set_destroy_model)
+        self.assertEqual('remove-object', client.command_set_remove_object)
+        self.assertEqual('all-changes', client.command_set_all)
+
+    def test_list_disabled_commands(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, None)
+        with patch.object(client, 'get_juju_output', autospec=True,
+                          return_value=dedent("""\
+             - command-set: destroy-model
+               message: Lock Models
+             - command-set: remove-object""")) as mock:
+            output = client.list_disabled_commands()
+        self.assertEqual([{'command-set': 'destroy-model',
+                           'message': 'Lock Models'},
+                          {'command-set': 'remove-object'}], output)
+        mock.assert_called_once_with('block list',
+                                     '--format', 'yaml')
+
+    def test_disable_command(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, None)
+        with patch.object(client, 'juju', autospec=True) as mock:
+            client.disable_command('all', 'message')
+        mock.assert_called_once_with('block all', ('message', ))
+
+    def test_enable_command(self):
+        client = EnvJujuClient1X(SimpleEnvironment('foo'), None, None)
+        with patch.object(client, 'juju', autospec=True) as mock:
+            client.enable_command('all')
+        mock.assert_called_once_with('unblock', 'all')
+
+
+class TestEnvJujuClient25(ClientTest):
+
+    client_class = EnvJujuClient25
+
+    def test_enable_jes(self):
+        client = self.client_class(
+            SimpleEnvironment('baz', {}),
+            '1.25-foobar', 'path')
+        with self.assertRaises(JESNotSupported):
+            client.enable_jes()
+
+    def test_disable_jes(self):
+        client = self.client_class(
+            SimpleEnvironment('baz', {}),
+            '1.25-foobar', 'path')
+        client.feature_flags.add('jes')
+        client.disable_jes()
+        self.assertNotIn('jes', client.feature_flags)
+
+    def test_clone_unchanged(self):
+        client1 = self.client_class(
+            SimpleEnvironment('foo'), '1.27', 'full/path', debug=True)
+        client2 = client1.clone()
+        self.assertIsNot(client1, client2)
+        self.assertIs(type(client1), type(client2))
+        self.assertIs(client1.env, client2.env)
+        self.assertEqual(client1.version, client2.version)
+        self.assertEqual(client1.full_path, client2.full_path)
+        self.assertIs(client1.debug, client2.debug)
+        self.assertEqual(client1._backend, client2._backend)
+
+    def test_clone_changed(self):
+        client1 = self.client_class(
+            SimpleEnvironment('foo'), '1.27', 'full/path', debug=True)
+        env2 = SimpleEnvironment('bar')
+        client2 = client1.clone(env2, '1.28', 'other/path', debug=False,
+                                cls=EnvJujuClient1X)
+        self.assertIs(EnvJujuClient1X, type(client2))
+        self.assertIs(env2, client2.env)
+        self.assertEqual('1.28', client2.version)
+        self.assertEqual('other/path', client2.full_path)
+        self.assertIs(False, client2.debug)
+
+    def test_clone_defaults(self):
+        client1 = self.client_class(
+            SimpleEnvironment('foo'), '1.27', 'full/path', debug=True)
+        client2 = client1.clone()
+        self.assertIsNot(client1, client2)
+        self.assertIs(self.client_class, type(client2))
+        self.assertEqual(set(), client2.feature_flags)
+
+
+class TestEnvJujuClient22(ClientTest):
+
+    client_class = EnvJujuClient22
+
+    def test__shell_environ(self):
+        client = self.client_class(
+            SimpleEnvironment('baz', {'type': 'ec2'}), '1.22-foobar', 'path')
+        env = client._shell_environ()
+        self.assertEqual(env.get(JUJU_DEV_FEATURE_FLAGS), 'actions')
+
+    def test__shell_environ_juju_home(self):
+        client = self.client_class(
+            SimpleEnvironment('baz', {'type': 'ec2'}), '1.22-foobar', 'path',
+            'asdf')
+        env = client._shell_environ()
+        self.assertEqual(env['JUJU_HOME'], 'asdf')
+
+
+class TestEnvJujuClient24(ClientTest):
+
+    client_class = EnvJujuClient24
+
+    def test_no_jes(self):
+        client = self.client_class(
+            SimpleEnvironment('baz', {}),
+            '1.25-foobar', 'path')
+        with self.assertRaises(JESNotSupported):
+            client.enable_jes()
+        client._use_jes = True
+        env = client._shell_environ()
+        self.assertNotIn('jes', env.get(JUJU_DEV_FEATURE_FLAGS, '').split(","))
+
+    def test_add_ssh_machines(self):
+        client = self.client_class(SimpleEnvironment('foo', {}), None, 'juju')
+        with patch('subprocess.check_call', autospec=True) as cc_mock:
+            client.add_ssh_machines(['m-foo', 'm-bar', 'm-baz'])
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'), 0)
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-bar'), 1)
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-baz'), 2)
+        self.assertEqual(cc_mock.call_count, 3)
+
+    def test_add_ssh_machines_no_retry(self):
+        client = self.client_class(SimpleEnvironment('foo', {}), None, 'juju')
+        with patch('subprocess.check_call', autospec=True,
+                   side_effect=[subprocess.CalledProcessError(None, None),
+                                None, None, None]) as cc_mock:
+            with self.assertRaises(subprocess.CalledProcessError):
+                client.add_ssh_machines(['m-foo', 'm-bar', 'm-baz'])
+        assert_juju_call(self, cc_mock, client, (
+            'juju', '--show-log', 'add-machine', '-e', 'foo', 'ssh:m-foo'))
+
+
+class TestStatus1X(FakeHomeTestCase):
+
+    def test_model_name(self):
+        status = Status1X({'environment': 'bar'}, '')
+        self.assertEqual('bar', status.model_name)
+
+    def test_get_applications_gets_services(self):
+        status = Status1X({
+            'services': {'service': {}},
+            'applications': {'application': {}},
+            }, '')
+        self.assertEqual({'service': {}}, status.get_applications())
+
+    def test_condense_status(self):
+        status = Status1X({}, '')
+        self.assertEqual(status.condense_status(
+                             {'agent-state': 'started',
+                              'agent-state-info': 'all good',
+                              'agent-version': '1.25.1'}),
+                         {'current': 'started', 'message': 'all good',
+                          'version': '1.25.1'})
+
+    def test_condense_status_no_info(self):
+        status = Status1X({}, '')
+        self.assertEqual(status.condense_status(
+                             {'agent-state': 'started',
+                              'agent-version': '1.25.1'}),
+                         {'current': 'started', 'version': '1.25.1'})
+
+    @staticmethod
+    def run_iter_status():
+        status = Status1X({
+            'environment': 'fake-unit-test',
+            'machines': {
+                '0': {
+                    'agent-state': 'started',
+                    'agent-state-info': 'all good',
+                    'agent-version': '1.25.1',
+                    },
+                },
+            'services': {
+                'dummy-sink': {
+                    'units': {
+                        'dummy-sink/0': {
+                            'agent-state': 'started',
+                            'agent-version': '1.25.1',
+                            },
+                        'dummy-sink/1': {
+                            'workload-status': {
+                                'current': 'active',
+                                },
+                            'agent-status': {
+                                'current': 'executing',
+                                },
+                            'agent-state': 'started',
+                            'agent-version': '1.25.1',
+                            },
+                        }
+                    },
+                'dummy-source': {
+                    'service-status': {
+                        'current': 'active',
+                        },
+                    'units': {
+                        'dummy-source/0': {
+                            'agent-state': 'started',
+                            'agent-version': '1.25.1',
+                            }
+                        }
+                    },
+                },
+            }, '')
+        for sub_status in status.iter_status():
+            yield sub_status
+
+    def test_iter_status_range(self):
+        status_set = set([(status_item.item_name, status_item.status_name,
+                           status_item.current)
+                          for status_item in self.run_iter_status()])
+        APP = StatusItem.APPLICATION
+        WORK = StatusItem.WORKLOAD
+        JUJU = StatusItem.JUJU
+        self.assertEqual({
+            ('0', JUJU, 'started'), ('dummy-sink/0', JUJU, 'started'),
+            ('dummy-sink/1', JUJU, 'executing'),
+            ('dummy-sink/1', WORK, 'active'), ('dummy-source', APP, 'active'),
+            ('dummy-source/0', JUJU, 'started'),
+            }, status_set)
+
+    def test_iter_status_data(self):
+        iterator = self.run_iter_status()
+        self.assertEqual(iterator.next().status,
+                         dict(current='started', message='all good',
+                              version='1.25.1'))
+
+
+def fast_timeout(count):
+    if False:
+        yield

=== added file 'jujupy/version_client.py'
--- jujupy/version_client.py	1970-01-01 00:00:00 +0000
+++ jujupy/version_client.py	2017-01-20 20:03:12 +0000
@@ -0,0 +1,705 @@
+from contextlib import contextmanager
+import json
+import logging
+import os
+import re
+import subprocess
+
+import yaml
+
+from jujupy.client import (
+    Controller,
+    _DEFAULT_BUNDLE_TIMEOUT,
+    get_cache_path,
+    get_jenv_path,
+    get_teardown_timeout,
+    JESNotSupported,
+    _jes_cmds,
+    Juju2Backend,
+    JujuData,
+    KVM_MACHINE,
+    LXC_MACHINE,
+    make_safe_config,
+    ModelClient,
+    SimpleEnvironment,
+    Status,
+    StatusItem,
+    SYSTEM,
+    unqualified_model_name,
+    UpgradeMongoNotSupported,
+    )
+from utility import (
+    ensure_deleted,
+    scoped_environ,
+    split_address_port,
+    )
+
+
+log = logging.getLogger("jujupy.version_client")
+
+
+class BootstrapMismatch(Exception):
+
+    def __init__(self, arg_name, arg_val, env_name, env_val):
+        super(BootstrapMismatch, self).__init__(
+            '--{} {} does not match {}: {}'.format(
+                arg_name, arg_val, env_name, env_val))
+
+
+class IncompatibleConfigClass(Exception):
+    """Raised when a client is initialised with the wrong config class."""
+
+
+class VersionNotTestedError(Exception):
+
+    def __init__(self, version):
+        super(VersionNotTestedError, self).__init__(
+            'Tests for juju {} are no longer supported.'.format(version))
+
+
+class Juju1XBackend(Juju2Backend):
+    """Backend for Juju 1.x versions.
+
+    Uses -e to specify models ("environments", uses JUJU_HOME to specify home
+    directory.
+    """
+
+    _model_flag = '-e'
+
+    def shell_environ(self, used_feature_flags, juju_home):
+        """Generate a suitable shell environment.
+
+        For 2.0-alpha1 and earlier set only JUJU_HOME and not JUJU_DATA.
+        """
+        env = super(Juju1XBackend, self).shell_environ(used_feature_flags,
+                                                       juju_home)
+        env['JUJU_HOME'] = juju_home
+        del env['JUJU_DATA']
+        return env
+
+
+class Status1X(Status):
+
+    @property
+    def model_name(self):
+        return self.status['environment']
+
+    def get_applications(self):
+        return self.status.get('services', {})
+
+    def condense_status(self, item_value):
+        """Condense the scattered agent-* fields into a status dict."""
+        def shift_field(dest_dict, dest_name, src_dict, src_name):
+            if src_name in src_dict:
+                dest_dict[dest_name] = src_dict[src_name]
+        condensed = {}
+        shift_field(condensed, 'current', item_value, 'agent-state')
+        shift_field(condensed, 'version', item_value, 'agent-version')
+        shift_field(condensed, 'message', item_value, 'agent-state-info')
+        return condensed
+
+    def iter_status(self):
+        SERVICE = 'service-status'
+        AGENT = 'agent-status'
+        for machine_name, machine_value in self.get_machines({}).items():
+            yield StatusItem(StatusItem.JUJU, machine_name,
+                             self.condense_status(machine_value))
+        for app_name, app_value in self.get_applications().items():
+            if SERVICE in app_value:
+                yield StatusItem(
+                    StatusItem.APPLICATION, app_name,
+                    {StatusItem.APPLICATION: app_value[SERVICE]})
+            for unit_name, unit_value in app_value.get('units', {}).items():
+                if StatusItem.WORKLOAD in unit_value:
+                    yield StatusItem(StatusItem.WORKLOAD,
+                                     unit_name, unit_value)
+                if AGENT in unit_value:
+                    yield StatusItem(
+                        StatusItem.JUJU, unit_name,
+                        {StatusItem.JUJU: unit_value[AGENT]})
+                else:
+                    yield StatusItem(StatusItem.JUJU, unit_name,
+                                     self.condense_status(unit_value))
+
+
+class ModelClient2_1(ModelClient):
+    """Client for Juju 2.1"""
+
+    REGION_ENDPOINT_PROMPT = 'Enter the API endpoint url for the region:'
+
+
+class ModelClientRC(ModelClient2_1):
+
+    def get_bootstrap_args(
+            self, upload_tools, config_filename, bootstrap_series=None,
+            credential=None, auto_upgrade=False, metadata_source=None,
+            to=None, no_gui=False, agent_version=None):
+        """Return the bootstrap arguments for the substrate."""
+        if self.env.joyent:
+            # Only accept kvm packages by requiring >1 cpu core, see lp:1446264
+            constraints = 'mem=2G cpu-cores=1'
+        else:
+            constraints = 'mem=2G'
+        cloud_region = self.get_cloud_region(self.env.get_cloud(),
+                                             self.env.get_region())
+        # Note controller name before cloud_region
+        args = ['--constraints', constraints,
+                self.env.environment,
+                cloud_region,
+                '--config', config_filename,
+                '--default-model', self.env.environment]
+        if upload_tools:
+            if agent_version is not None:
+                raise ValueError(
+                    'agent-version may not be given with upload-tools.')
+            args.insert(0, '--upload-tools')
+        else:
+            if agent_version is None:
+                agent_version = self.get_matching_agent_version()
+            args.extend(['--agent-version', agent_version])
+        if bootstrap_series is not None:
+            args.extend(['--bootstrap-series', bootstrap_series])
+        if credential is not None:
+            args.extend(['--credential', credential])
+        if metadata_source is not None:
+            args.extend(['--metadata-source', metadata_source])
+        if auto_upgrade:
+            args.append('--auto-upgrade')
+        if to is not None:
+            args.extend(['--to', to])
+        if no_gui:
+            args.append('--no-gui')
+        return tuple(args)
+
+
+class EnvJujuClient1X(ModelClientRC):
+    """Base for all 1.x client drivers."""
+
+    default_backend = Juju1XBackend
+
+    config_class = SimpleEnvironment
+
+    status_class = Status1X
+
+    # The environments.yaml options that are replaced by bootstrap options.
+    # For Juju 1.x, no bootstrap options are used.
+    bootstrap_replaces = frozenset()
+
+    destroy_model_command = 'destroy-environment'
+
+    supported_container_types = frozenset([KVM_MACHINE, LXC_MACHINE])
+
+    agent_metadata_url = 'tools-metadata-url'
+
+    _show_status = 'status'
+
+    command_set_destroy_model = 'destroy-environment'
+
+    command_set_all = 'all-changes'
+
+    @classmethod
+    def _get_env(cls, env):
+        if isinstance(env, JujuData):
+            raise IncompatibleConfigClass(
+                'JujuData cannot be used with {}'.format(cls.__name__))
+        return env
+
+    def get_cache_path(self):
+        return get_cache_path(self.env.juju_home, models=False)
+
+    def remove_service(self, service):
+        self.juju('destroy-service', (service,))
+
+    def backup(self):
+        environ = self._shell_environ()
+        # juju-backup does not support the -e flag.
+        environ['JUJU_ENV'] = self.env.environment
+        try:
+            # Mutate os.environ instead of supplying env parameter so Windows
+            # can search env['PATH']
+            with scoped_environ(environ):
+                args = ['juju', 'backup']
+                log.info(' '.join(args))
+                output = subprocess.check_output(args)
+        except subprocess.CalledProcessError as e:
+            log.info(e.output)
+            raise
+        log.info(output)
+        backup_file_pattern = re.compile('(juju-backup-[0-9-]+\.(t|tar.)gz)')
+        match = backup_file_pattern.search(output)
+        if match is None:
+            raise Exception("The backup file was not found in output: %s" %
+                            output)
+        backup_file_name = match.group(1)
+        backup_file_path = os.path.abspath(backup_file_name)
+        log.info("State-Server backup at %s", backup_file_path)
+        return backup_file_path
+
+    def restore_backup(self, backup_file):
+        return self.get_juju_output('restore', '--constraints', 'mem=2G',
+                                    backup_file)
+
+    def restore_backup_async(self, backup_file):
+        return self.juju_async('restore', ('--constraints', 'mem=2G',
+                                           backup_file))
+
+    def enable_ha(self):
+        self.juju('ensure-availability', ('-n', '3'))
+
+    def list_models(self):
+        """List the models registered with the current controller."""
+        log.info('The model is environment {}'.format(self.env.environment))
+
+    def list_clouds(self, format='json'):
+        """List all the available clouds."""
+        return {}
+
+    def show_controller(self, format='json'):
+        """Show controller's status."""
+        return {}
+
+    def get_models(self):
+        """return a models dict with a 'models': [] key-value pair."""
+        return {}
+
+    def list_controllers(self):
+        """List the controllers."""
+        log.info(
+            'The controller is environment {}'.format(self.env.environment))
+
+    @staticmethod
+    def get_controller_member_status(info_dict):
+        return info_dict.get('state-server-member-status')
+
+    def action_fetch(self, id, action=None, timeout="1m"):
+        """Fetches the results of the action with the given id.
+
+        Will wait for up to 1 minute for the action results.
+        The action name here is just used for an more informational error in
+        cases where it's available.
+        Returns the yaml output of the fetched action.
+        """
+        # the command has to be "action fetch" so that the -e <env> args are
+        # placed after "fetch", since that's where action requires them to be.
+        out = self.get_juju_output("action fetch", id, "--wait", timeout)
+        status = yaml.safe_load(out)["status"]
+        if status != "completed":
+            name = ""
+            if action is not None:
+                name = " " + action
+            raise Exception(
+                "timed out waiting for action%s to complete during fetch" %
+                name)
+        return out
+
+    def action_do(self, unit, action, *args):
+        """Performs the given action on the given unit.
+
+        Action params should be given as args in the form foo=bar.
+        Returns the id of the queued action.
+        """
+        args = (unit, action) + args
+
+        # the command has to be "action do" so that the -e <env> args are
+        # placed after "do", since that's where action requires them to be.
+        output = self.get_juju_output("action do", *args)
+        action_id_pattern = re.compile(
+            'Action queued with id: ([a-f0-9\-]{36})')
+        match = action_id_pattern.search(output)
+        if match is None:
+            raise Exception("Action id not found in output: %s" %
+                            output)
+        return match.group(1)
+
+    def run(self, commands, applications):
+        responses = self.get_juju_output(
+            'run', '--format', 'json', '--service', ','.join(applications),
+            *commands)
+        return json.loads(responses)
+
+    def list_space(self):
+        return yaml.safe_load(self.get_juju_output('space list'))
+
+    def add_space(self, space):
+        self.juju('space create', (space),)
+
+    def add_subnet(self, subnet, space):
+        self.juju('subnet add', (subnet, space))
+
+    def add_user_perms(self, username, models=None, permissions='read'):
+        raise JESNotSupported()
+
+    def grant(self, user_name, permission, model=None):
+        raise JESNotSupported()
+
+    def revoke(self, username, models=None, permissions='read'):
+        raise JESNotSupported()
+
+    def set_model_constraints(self, constraints):
+        constraint_strings = self._dict_as_option_strings(constraints)
+        return self.juju('set-constraints', constraint_strings)
+
+    def set_config(self, service, options):
+        option_strings = ['{}={}'.format(*item) for item in options.items()]
+        self.juju('set', (service,) + tuple(option_strings))
+
+    def get_config(self, service):
+        return yaml.safe_load(self.get_juju_output('get', service))
+
+    def get_model_config(self):
+        """Return the value of the environment's configured option."""
+        return yaml.safe_load(self.get_juju_output('get-env'))
+
+    def get_env_option(self, option):
+        """Return the value of the environment's configured option."""
+        return self.get_juju_output('get-env', option)
+
+    def set_env_option(self, option, value):
+        """Set the value of the option in the environment."""
+        option_value = "%s=%s" % (option, value)
+        return self.juju('set-env', (option_value,))
+
+    def unset_env_option(self, option):
+        """Unset the value of the option in the environment."""
+        return self.juju('set-env', ('{}='.format(option),))
+
+    def get_model_defaults(self, model_key, cloud=None, region=None):
+        log.info('No model-defaults stored for client (attempted get).')
+
+    def set_model_defaults(self, model_key, value, cloud=None, region=None):
+        log.info('No model-defaults stored for client (attempted set).')
+
+    def unset_model_defaults(self, model_key, cloud=None, region=None):
+        log.info('No model-defaults stored for client (attempted unset).')
+
+    def _cmd_model(self, include_e, controller):
+        if controller:
+            return self.get_controller_model_name()
+        elif self.env is None or not include_e:
+            return None
+        else:
+            return unqualified_model_name(self.model_name)
+
+    def update_user_name(self):
+        return
+
+    def _get_substrate_constraints(self):
+        if self.env.joyent:
+            # Only accept kvm packages by requiring >1 cpu core, see lp:1446264
+            return 'mem=2G cpu-cores=1'
+        else:
+            return 'mem=2G'
+
+    def get_bootstrap_args(self, upload_tools, bootstrap_series=None,
+                           credential=None):
+        """Return the bootstrap arguments for the substrate."""
+        if credential is not None:
+            raise ValueError(
+                '--credential is not supported by this juju version.')
+        constraints = self._get_substrate_constraints()
+        args = ('--constraints', constraints)
+        if upload_tools:
+            args = ('--upload-tools',) + args
+        if bootstrap_series is not None:
+            env_val = self.env.get_option('default-series')
+            if bootstrap_series != env_val:
+                raise BootstrapMismatch(
+                    'bootstrap-series', bootstrap_series, 'default-series',
+                    env_val)
+        return args
+
+    def bootstrap(self, upload_tools=False, bootstrap_series=None):
+        """Bootstrap a controller."""
+        self._check_bootstrap()
+        args = self.get_bootstrap_args(upload_tools, bootstrap_series)
+        self.juju('bootstrap', args)
+
+    @contextmanager
+    def bootstrap_async(self, upload_tools=False):
+        self._check_bootstrap()
+        args = self.get_bootstrap_args(upload_tools)
+        with self.juju_async('bootstrap', args):
+            yield
+            log.info('Waiting for bootstrap of {}.'.format(
+                self.env.environment))
+
+    def get_jes_command(self):
+        raise JESNotSupported()
+
+    def enable_jes(self):
+        raise JESNotSupported()
+
+    def upgrade_juju(self, force_version=True):
+        args = ()
+        if force_version:
+            version = self.get_matching_agent_version(no_build=True)
+            args += ('--version', version)
+        if self.env.local:
+            args += ('--upload-tools',)
+        self._upgrade_juju(args)
+
+    def make_model_config(self):
+        config_dict = make_safe_config(self)
+        # Strip unneeded variables.
+        return config_dict
+
+    def _add_model(self, model_name, config_file):
+        seen_cmd = self.get_jes_command()
+        if seen_cmd == SYSTEM:
+            controller_option = ('-s', self.env.environment)
+        else:
+            controller_option = ('-c', self.env.environment)
+        self.juju(_jes_cmds[seen_cmd]['create'], controller_option + (
+            model_name, '--config', config_file), include_e=False)
+
+    def destroy_model(self):
+        """With JES enabled, destroy-environment destroys the model."""
+        return self.destroy_environment(force=False)
+
+    def kill_controller(self, check=False):
+        """Destroy the environment, with force. Hard kill option.
+
+        :return: Subprocess's exit code."""
+        return self.juju(
+            'destroy-environment', (self.env.environment, '--force', '-y'),
+            check=check, include_e=False, timeout=get_teardown_timeout(self))
+
+    def destroy_controller(self, all_models=False):
+        """Destroy the environment, with force. Soft kill option.
+
+        :param all_models: Ignored.
+        :raises: subprocess.CalledProcessError if the operation fails."""
+        return self.juju(
+            'destroy-environment', (self.env.environment, '-y'),
+            include_e=False, timeout=get_teardown_timeout(self))
+
+    def destroy_environment(self, force=True, delete_jenv=False):
+        if force:
+            force_arg = ('--force',)
+        else:
+            force_arg = ()
+        exit_status = self.juju(
+            'destroy-environment',
+            (self.env.environment,) + force_arg + ('-y',),
+            check=False, include_e=False,
+            timeout=get_teardown_timeout(self))
+        if delete_jenv:
+            jenv_path = get_jenv_path(self.env.juju_home, self.env.environment)
+            ensure_deleted(jenv_path)
+        return exit_status
+
+    def _get_models(self):
+        """return a list of model dicts."""
+        try:
+            return yaml.safe_load(self.get_juju_output(
+                'environments', '-s', self.env.environment, '--format', 'yaml',
+                include_e=False))
+        except subprocess.CalledProcessError:
+            # This *private* method attempts to use a 1.25 JES feature.
+            # The JES design is dead. The private method is not used to
+            # directly test juju cli; the failure is not a contract violation.
+            log.info('Call to JES juju environments failed, falling back.')
+            return []
+
+    def get_model_uuid(self):
+        raise JESNotSupported()
+
+    def deploy_bundle(self, bundle, timeout=_DEFAULT_BUNDLE_TIMEOUT):
+        """Deploy bundle using deployer for Juju 1.X version."""
+        self.deployer(bundle, timeout=timeout)
+
+    def deployer(self, bundle_template, name=None, deploy_delay=10,
+                 timeout=3600):
+        """Deploy a bundle using deployer."""
+        bundle = self.format_bundle(bundle_template)
+        args = (
+            '--debug',
+            '--deploy-delay', str(deploy_delay),
+            '--timeout', str(timeout),
+            '--config', bundle,
+        )
+        if name:
+            args += (name,)
+        self.juju('deployer', args)
+
+    def deploy(self, charm, repository=None, to=None, series=None,
+               service=None, force=False, storage=None, constraints=None):
+        args = [charm]
+        if repository is not None:
+            args.extend(['--repository', repository])
+        if to is not None:
+            args.extend(['--to', to])
+        if service is not None:
+            args.extend([service])
+        if storage is not None:
+            args.extend(['--storage', storage])
+        if constraints is not None:
+            args.extend(['--constraints', constraints])
+        return self.juju('deploy', tuple(args))
+
+    def upgrade_charm(self, service, charm_path=None):
+        args = (service,)
+        if charm_path is not None:
+            repository = os.path.dirname(os.path.dirname(charm_path))
+            args = args + ('--repository', repository)
+        self.juju('upgrade-charm', args)
+
+    def get_controller_client(self):
+        """Return a client for the controller model.  May return self."""
+        return self
+
+    def get_controller_model_name(self):
+        """Return the name of the 'controller' model.
+
+        Return the name of the 1.x environment."""
+        return self.env.environment
+
+    def get_controller_endpoint(self):
+        """Return the host and port of the state-server leader."""
+        endpoint = self.get_juju_output('api-endpoints')
+        return split_address_port(endpoint)
+
+    def upgrade_mongo(self):
+        raise UpgradeMongoNotSupported()
+
+    def create_cloned_environment(
+            self, cloned_juju_home, controller_name, user_name=None):
+        """Create a cloned environment.
+
+        `user_name` is unused in this version of juju.
+        """
+        user_client = self.clone(env=self.env.clone())
+        user_client.env.juju_home = cloned_juju_home
+        # New user names the controller.
+        user_client.env.controller = Controller(controller_name)
+        return user_client
+
+    def add_storage(self, unit, storage_type, amount="1"):
+        """Add storage instances to service.
+
+        Only type 'disk' is able to add instances.
+        """
+        self.juju('storage add', (unit, storage_type + "=" + amount))
+
+    def list_storage(self):
+        """Return the storage list."""
+        return self.get_juju_output('storage list', '--format', 'json')
+
+    def list_storage_pool(self):
+        """Return the list of storage pool."""
+        return self.get_juju_output('storage pool list', '--format', 'json')
+
+    def create_storage_pool(self, name, provider, size):
+        """Create storage pool."""
+        self.juju('storage pool create',
+                  (name, provider,
+                   'size={}'.format(size)))
+
+    def ssh_keys(self, full=False):
+        """Give the ssh keys registered for the current model."""
+        args = []
+        if full:
+            args.append('--full')
+        return self.get_juju_output('authorized-keys list', *args)
+
+    def add_ssh_key(self, *keys):
+        """Add one or more ssh keys to the current model."""
+        return self.get_juju_output('authorized-keys add', *keys,
+                                    merge_stderr=True)
+
+    def remove_ssh_key(self, *keys):
+        """Remove one or more ssh keys from the current model."""
+        return self.get_juju_output('authorized-keys delete', *keys,
+                                    merge_stderr=True)
+
+    def import_ssh_key(self, *keys):
+        """Import ssh keys from one or more identities to the current model."""
+        return self.get_juju_output('authorized-keys import', *keys,
+                                    merge_stderr=True)
+
+    def list_disabled_commands(self):
+        """List all the commands disabled on the model."""
+        raw = self.get_juju_output('block list', '--format', 'yaml')
+        return yaml.safe_load(raw)
+
+    def disable_command(self, command_set, message=''):
+        """Disable a command-set."""
+        return self.juju('block {}'.format(command_set), (message, ))
+
+    def enable_command(self, args):
+        """Enable a command-set."""
+        return self.juju('unblock', args)
+
+
+class EnvJujuClient22(EnvJujuClient1X):
+
+    used_feature_flags = frozenset(['actions'])
+
+    def __init__(self, *args, **kwargs):
+        super(EnvJujuClient22, self).__init__(*args, **kwargs)
+        self.feature_flags.add('actions')
+
+
+class EnvJujuClient25(EnvJujuClient1X):
+    """Drives Juju 2.5-series clients."""
+
+    used_feature_flags = frozenset()
+
+    def disable_jes(self):
+        self.feature_flags.discard('jes')
+
+
+class EnvJujuClient24(EnvJujuClient25):
+    """Similar to EnvJujuClient25."""
+
+    def add_ssh_machines(self, machines):
+        for machine in machines:
+            self.juju('add-machine', ('ssh:' + machine,))
+
+
+def get_client_class(version):
+    if version.startswith('1.16'):
+        raise VersionNotTestedError(version)
+    elif re.match('^1\.22[.-]', version):
+        client_class = EnvJujuClient22
+    elif re.match('^1\.24[.-]', version):
+        client_class = EnvJujuClient24
+    elif re.match('^1\.25[.-]', version):
+        client_class = EnvJujuClient25
+    elif re.match('^1\.26[.-]', version):
+        raise VersionNotTestedError(version)
+    elif re.match('^1\.', version):
+        client_class = EnvJujuClient1X
+    elif re.match('^2\.0-(alpha|beta)', version):
+        raise VersionNotTestedError(version)
+    elif re.match('^2\.0-rc[1-3]', version):
+        client_class = ModelClientRC
+    elif re.match('^2\.[0-1][.-]', version):
+        client_class = ModelClient2_1
+    else:
+        client_class = ModelClient
+    return client_class
+
+
+def client_from_config(config, juju_path, debug=False, soft_deadline=None):
+    """Create a client from an environment's configuration.
+
+    :param config: Name of the environment to use the config from.
+    :param juju_path: Path to juju binary the client should wrap.
+    :param debug=False: The debug flag for the client, False by default.
+    :param soft_deadline: A datetime representing the deadline by which
+        normal operations should complete.  If None, no deadline is
+        enforced.
+    """
+    version = ModelClient.get_version(juju_path)
+    client_class = get_client_class(str(version))
+    if config is None:
+        env = client_class.config_class('', {})
+    else:
+        env = client_class.config_class.from_config(config)
+    if juju_path is None:
+        full_path = ModelClient.get_full_path()
+    else:
+        full_path = os.path.abspath(juju_path)
+    return client_class(env, version, full_path, debug=debug,
+                        soft_deadline=soft_deadline)

=== added file 'maas/add-existing-nodes-to-maas.sh'
--- maas/add-existing-nodes-to-maas.sh	1970-01-01 00:00:00 +0000
+++ maas/add-existing-nodes-to-maas.sh	2017-01-03 21:57:52 +0000
@@ -0,0 +1,31 @@
+#!/bin/bash -x
+
+ENV=$1
+BASE_NODE_NUM=$2
+ADD_COUNT=$3
+base_name='maas-node'
+power_address='qemu+ssh://jenkins@10.0.200.1/system'
+pool_path='/images/maaspool1'
+
+if [[ ! -d $HOME/juju-ci-tools/maas ]]; then
+    echo "\$HOME/juju-ci-tools/maas is required."
+    exit 1;
+fi
+
+next_num=${BASE_NODE_NUM}
+last_num=$((next_num + ADD_COUNT))
+while (($next_num <= $last_num)); do
+    mac_to_add=$(
+        virsh dumpxml ${base_name}-${next_num} |
+        grep "mac address" |
+        head -2 |
+        tail -1 |
+        awk -F\' '{print $2}')
+    maas ${ENV} machines create \
+        architecture="amd64/generic" mac_addresses="${mac_to_add}" \
+        hostname="${base_name}-${next_num}" power_type="virsh" \
+        power_parameters_power_address="${power_address}" \
+        power_parameters_power_id="${base_name}-${next_num}" \
+        autodetect_nodegroup=1
+    next_num=$((next_num+1))
+done

=== added file 'maas/reset-fabrics.bash'
--- maas/reset-fabrics.bash	1970-01-01 00:00:00 +0000
+++ maas/reset-fabrics.bash	2017-01-05 14:55:38 +0000
@@ -0,0 +1,44 @@
+#!/bin/bash
+# Reset maas fabrics and vlans to 0, the defaults.
+set -eux
+
+
+ENV=$1
+
+# Learn all the system_id and hostnames to lookup later.
+ALL_SYSTEM_HOSTS=$(
+    maas $ENV machines read |
+    grep -P 'system_id|hostname' |
+    sed -r ' N;s/\n/ /; s/( +"system_id": .[^,]+,)(.*)/\2 \1/; s, +"(system_id|hostname)": ,,g; s/"//g; s/,/@/; s/,//')
+# Find the unwanted fabric, which is most likely the one with a number
+# greater than 9.
+FABRIC=$(
+    maas $ENV fabrics read |
+    sed -r '/\/fabrics\/[0-9][0-9]+/!d; s,.*/fabrics/([0-9][0-9]+)/.*,\1,')
+if [[ $FABRIC == "" ]]; then
+    exit 0
+fi
+# Find the vlan id of the default fabric.
+VLAN=$(
+    maas $ENV fabric read 0 |
+    sed -r '/vlans/,/}/!d; /"id"/!d; s,[^0-9],,g' |
+    head -1)
+# Learn the misconfigured interfaces by attempting to delete
+# the unwanted fabric.
+INTERFACES=$(
+    maas $ENV fabric delete $FABRIC 2>&1 |
+    sed -r 's,(^.*:|\([^\)]*\)) ,,g; s, on ,@,g; s/,//g' ||
+    true)
+# Reset the fabric and vlan of each machine on the unwanted fabric.
+for iface_machine in $INTERFACES; do
+    iface=$(echo $iface_machine | cut -d @ -f1)
+    machine=$(echo $iface_machine | cut -d @ -f2)
+    system_id=$(
+        echo "$ALL_SYSTEM_HOSTS" |
+        grep $machine@ | cut -d @ -f2)
+    maas $ENV machine release $system_id
+    sleep 5
+    maas $ENV interface update $system_id $iface fabric=0 vlan=$VLAN
+done
+# Delete the unwanted fabric.
+maas $ENV fabric delete $FABRIC

=== added file 'perfscale_controller_stress.py'
--- perfscale_controller_stress.py	1970-01-01 00:00:00 +0000
+++ perfscale_controller_stress.py	2017-01-18 20:56:26 +0000
@@ -0,0 +1,88 @@
+#!/usr/bin/env python
+"""Perfscale test that uses the observable-swarm charm collection to exercise
+the controller(s).
+
+Steps taken in this test:
+  - Bootstraps a new controller (installing monitoring software on it).
+  - Iterate a user defined amount:
+    - Add new model
+    - Deploy observable-swarm to this new model
+"""
+
+import argparse
+from datetime import datetime
+import sys
+
+from deploy_stack import (
+    BootstrapManager,
+)
+from generate_perfscale_results import (
+    DeployDetails,
+    TimingData,
+    add_basic_perfscale_arguments,
+    run_perfscale_test,
+)
+from utility import (
+    configure_logging,
+)
+
+__metaclass__ = type
+
+
+def assess_controller_stress(client, args):
+    """Deploy observable-swarm charm many times to stress test controller
+    machines.
+    """
+    test_start = datetime.utcnow()
+
+    deploy_details = dict()
+    for model_number in range(0, args.deploy_amount):
+        model_name = 'swarm-model-{}'.format(model_number)
+        deploy_time = deploy_swarm_to_new_model(client, model_name)
+        deploy_details[model_name] = '{} Seconds'.format(deploy_time)
+
+    test_end = datetime.utcnow()
+    deploy_timing = TimingData(test_start, test_end)
+
+    return DeployDetails('Controller Stress.', deploy_details, deploy_timing)
+
+
+def deploy_swarm_to_new_model(client, model_name):
+    before_add = datetime.utcnow()
+
+    new_client = client.add_model(model_name)
+    new_client.deploy(get_charm_url())
+    new_client.wait_for_started()
+    new_client.wait_for_workloads()
+
+    after_add = datetime.utcnow()
+    return int((after_add - before_add).total_seconds())
+
+
+def get_charm_url():
+    return 'cs:~containers/observable-swarm'
+
+
+def parse_args(argv):
+    """Parse all arguments."""
+    parser = argparse.ArgumentParser(
+        description='Perfscale Controller Stress test.')
+    add_basic_perfscale_arguments(parser)
+    parser.add_argument(
+        '--deploy-amount',
+        help='The amount of deploys to do.',
+        type=int,
+        default=1)
+    return parser.parse_args(argv)
+
+
+def main(argv=None):
+    args = parse_args(argv)
+    configure_logging(args.verbose)
+    bs_manager = BootstrapManager.from_args(args)
+    run_perfscale_test(assess_controller_stress, bs_manager, args)
+
+    return 0
+
+if __name__ == '__main__':
+    sys.exit(main())

=== modified file 'perfscale_longrunning.py'
--- perfscale_longrunning.py	2016-10-25 06:05:20 +0000
+++ perfscale_longrunning.py	2016-12-16 15:43:19 +0000
@@ -63,7 +63,7 @@
 
 def action_create(client, series='trusty'):
     start = datetime.utcnow()
-    new_model = client.add_model(client.env.clone('newmodel'))
+    new_model = client.add_model('newmodel')
     deploy_stack(new_model, series)
     end = datetime.utcnow()
     log.info('Create action took: {}'.format(
@@ -83,8 +83,7 @@
     new_models = []
     for i in range(0, 20):
         total_new_models += 1
-        new_model_name = 'model{}'.format(total_new_models)
-        new_model = client.add_model(client.env.clone(new_model_name))
+        new_model = client.add_model('model{}'.format(total_new_models))
         new_model.wait_for_started()
         log.info('Added model number {}'.format(total_new_models))
         new_models.append(new_model)

=== modified file 'perfscale_mass_model_destruction.py'
--- perfscale_mass_model_destruction.py	2016-10-30 23:17:23 +0000
+++ perfscale_mass_model_destruction.py	2016-12-16 15:43:19 +0000
@@ -40,7 +40,7 @@
     for item in xrange(0, model_count):
         model_name = 'model{}'.format(item)
         log.info('Creating model: {}'.format(model_name))
-        new_model = client.add_model(client.env.clone(model_name))
+        new_model = client.add_model(model_name)
         all_models.append(new_model)
 
     # Workaround for bug: https://bugs.launchpad.net/juju/+bug/1635052

=== modified file 'pipdeps.py'
--- pipdeps.py	2016-06-23 02:50:43 +0000
+++ pipdeps.py	2017-01-12 15:20:16 +0000
@@ -4,6 +4,7 @@
 from __future__ import print_function
 
 import argparse
+from distutils.version import LooseVersion
 import os
 import platform
 import subprocess
@@ -19,16 +20,23 @@
 PREFIX = "juju-ci-tools/"
 REQUIREMENTS = os.path.join(os.path.realpath(os.path.dirname(__file__)),
                             "requirements.txt")
+REQUIREMENTS_PY3 = os.path.join(os.path.realpath(os.path.dirname(__file__)),
+                                "requirements_py3.txt")
 MAC_WIN_REQS = os.path.join(os.path.realpath(os.path.dirname(__file__)),
                             "mac-win-requirements.txt")
 OBSOLETE = os.path.join(os.path.realpath(os.path.dirname(__file__)),
                         "obsolete-requirements.txt")
 
 
-def get_requirements():
+def get_requirements(python3=False):
+    """Return requirements file path."""
     if platform.dist()[0] in ('Ubuntu', 'debian'):
+        if python3:
+            return REQUIREMENTS_PY3
         return REQUIREMENTS
     else:
+        if python3:
+            return None
         return MAC_WIN_REQS
 
 
@@ -50,9 +58,35 @@
     return boto.s3.connection.S3Connection(access_key, secret_key)
 
 
-def run_pip_install(extra_args, requirements, verbose=False):
+def is_py3_supported():
+    """Determine if Python3 and pip3 are installed."""
+    try:
+        version = subprocess.check_output(
+            ['python3', '--version'], stderr=subprocess.STDOUT)
+    except OSError:
+        return False
+    version = version.strip().split()[-1]
+    # Python on yakkety reports its version as "3.5.2+", so use LooseVersion.
+    if LooseVersion(version) < LooseVersion('3.5'):
+        return False
+    try:
+        subprocess.check_call(['pip3', '--version'])
+    except OSError:
+        return False
+    return True
+
+
+def run_pip3_install(extra_args, requirements, verbose=False):
+    if requirements and is_py3_supported():
+        run_pip_install(extra_args, requirements, verbose=verbose,
+                        cmd=['pip3'])
+    else:
+        print('Python 3 or pip 3 is not installed.')
+
+
+def run_pip_install(extra_args, requirements, verbose=False, cmd=None):
     """Run pip install in a subprocess with given additional arguments."""
-    cmd = ["pip"]
+    cmd = cmd or ['pip']
     if not verbose:
         cmd.append("-q")
     cmd.extend(["install", "-r", requirements])
@@ -85,15 +119,29 @@
         subprocess.check_call(uninstall_cmd)
 
 
-def command_install(bucket, requirements, verbose=False):
+def get_pip_args(archives_url):
+    args = ["--no-index", "--find-links", archives_url]
+    # --user option is invalid when running inside virtualenv. Check
+    # sys.real_prefix to determine if it is executing inside virtualenv.
+    # Inside virtualenv, the sys.prefix points to the virtualenv directory
+    # and sys.real_prefix points to the real prefix.
+    # Running outside a virtualenv, sys should not have real_prefix attribute.
+    if not hasattr(sys, 'real_prefix'):
+        args.append("--user")
+    return args
+
+
+def command_install(bucket, requirements, verbose=False,
+                    requirements_py3=None):
     with utility.temp_dir() as archives_dir:
         for key in bucket.list(prefix=PREFIX):
             archive = key.name[len(PREFIX):]
             key.get_contents_to_filename(os.path.join(archives_dir, archive))
         archives_url = "file://" + archives_dir
+        pip_args = get_pip_args(archives_url)
         run_pip_uninstall(OBSOLETE)
-        run_pip_install(["--user", "--no-index", "--find-links", archives_url],
-                        requirements, verbose=verbose)
+        run_pip_install(pip_args, requirements, verbose=verbose)
+        run_pip3_install(pip_args, requirements_py3, verbose=verbose)
 
 
 def command_update(s3, requirements, verbose=False):
@@ -136,6 +184,10 @@
     parser.add_argument(
         "--requirements", default=get_requirements(), type=os.path.expanduser,
         help="Location requirements file to use.")
+    parser.add_argument(
+        "--requirements_py3", default=get_requirements(python3=True),
+        type=os.path.expanduser,
+        help="Location requirements file to use for Python 3.")
     subparsers = parser.add_subparsers(dest="command")
     subparsers.add_parser("install", help="Download deps from S3 and install.")
     subparsers.add_parser(
@@ -157,7 +209,8 @@
     else:
         bucket = s3.get_bucket(BUCKET)
         if args.command == "install":
-            command_install(bucket, args.requirements, args.verbose)
+            command_install(
+                bucket, args.requirements, args.verbose, args.requirements_py3)
         elif args.command == "list":
             command_list(bucket, args.verbose)
         elif args.command == "delete":

=== added file 'prepare-testing.bash'
--- prepare-testing.bash	1970-01-01 00:00:00 +0000
+++ prepare-testing.bash	2017-01-16 16:11:20 +0000
@@ -0,0 +1,47 @@
+#!/bin/bash
+source $HOME/cloud-city/juju-qa.jujuci
+set -eux
+: ${SCRIPTS=$(readlink -f $(dirname $0))}
+
+# The maas 1.8 on finfolk is shutdown.
+#ssh -i $HOME/cloud-city/staging-juju-rsa maas-1-8 "~/clean_leaked_leases.bash"
+
+# Release all allocated machines on maas
+
+ssh -i $HOME/cloud-city/staging-juju-rsa jenkins@munna \
+  'JUJU_HOME=~/cloud-city' juju-ci-tools/clean_maas.py parallel-munna-vmaas \
+  --hours=2
+
+ssh -i $HOME/cloud-city/staging-juju-rsa jenkins@finfolk \
+  'JUJU_HOME=~/cloud-city' juju-ci-tools/clean_maas.py parallel-finfolk-vmaas \
+  --hours=2
+
+ssh -i $HOME/cloud-city/staging-juju-rsa jenkins@silcoon \
+  'JUJU_HOME=~/cloud-city' juju-ci-tools/clean_maas.py parallel-silcoon-vmaas \
+  --hours=2
+
+# Delete all lxd containers left behind on several machines.
+
+ssh -i $HOME/cloud-city/staging-juju-rsa jenkins@silcoon \
+    juju-ci-tools/clean_lxd.py
+
+ssh -i $HOME/cloud-city/staging-juju-rsa jenkins@feature-slave.vapour.ws \
+    juju-ci-tools/clean_lxd.py
+
+ssh -i $HOME/cloud-city/staging-juju-rsa jenkins@lxd-slave-a.vapour.ws \
+    juju-ci-tools/clean_lxd.py
+
+ssh -i $HOME/cloud-city/staging-juju-rsa jenkins@lxd-slave-b.vapour.ws \
+    juju-ci-tools/clean_lxd.py
+
+ssh -i $HOME/cloud-city/staging-juju-rsa jenkins@arm64-slave \
+    juju-ci-tools/clean_lxd.py
+
+ssh -i $HOME/cloud-city/staging-juju-rsa jenkins@s390x-slave \
+    juju-ci-tools/clean_lxd.py
+
+ssh -i $HOME/cloud-city/staging-juju-rsa jenkins@xenial-slave.vapour.ws \
+    juju-ci-tools/clean_lxd.py
+
+ssh -i $HOME/cloud-city/staging-juju-rsa jenkins@yakkety-slave.vapour.ws \
+    juju-ci-tools/clean_lxd.py

=== modified file 'remote.py'
--- remote.py	2016-02-16 17:33:12 +0000
+++ remote.py	2016-12-15 21:25:55 +0000
@@ -3,6 +3,7 @@
 import logging
 import os
 import subprocess
+import sys
 import zlib
 
 import winrm
@@ -108,6 +109,22 @@
         self.address = unit['public-address']
 
 
+def _default_is_command_error(err):
+    """
+    Whether to treat error as issue with remote command rather than ssh.
+
+    This is a conservative default, remote commands may return a variety of
+    other return codes. However, as the fallback to local ssh binary will
+    repeat the command, those problems will be exposed later anyway.
+    """
+    return err.returncode == 1
+
+
+def _no_platform_ssh():
+    """True if no openssh binary is available on this platform."""
+    return sys.platform == "win32"
+
+
 class SSHRemote(_Remote):
     """SSHRemote represents a juju machine to access using ssh."""
 
@@ -121,22 +138,46 @@
     # Limit each operation over SSH to 2 minutes by default
     timeout = 120
 
-    def run(self, command):
-        """Run a command on the remote machine."""
+    def run(self, command_args, is_command_error=_default_is_command_error):
+        """
+        Run a command on the remote machine.
+
+        If the remote instance has a juju unit run will default to using the
+        juju ssh command. Otherwise, or if that fails, it will fall back to
+        using ssh directly.
+
+        The command_args param is a string or list of arguments to be invoked
+        on the remote machine. A string must be given if special shell
+        characters are used.
+
+        The is_command_error param is a function that takes an instance of
+        CalledProcessError and returns whether that error comes from the
+        command being run rather than ssh itself. This can be used to skip the
+        fallback to native ssh behaviour when running commands that may fail.
+        """
+        if not isinstance(command_args, (list, tuple)):
+            command_args = [command_args]
         if self.use_juju_ssh:
             logging.debug('juju ssh {}'.format(self.unit))
             try:
-                return self.client.get_juju_output("ssh", self.unit, command,
-                                                   timeout=self.timeout)
+                return self.client.get_juju_output(
+                    "ssh", self.unit, *command_args, timeout=self.timeout)
             except subprocess.CalledProcessError as e:
-                logging.warning("juju ssh to {!r} failed, returncode: {}"
-                                " output: {!r}".format(self.unit, e.returncode,
-                                                       e.output))
+                logging.warning(
+                    "juju ssh to {!r} failed, returncode: {} output: {!r}"
+                    " stderr: {!r}".format(
+                        self.unit, e.returncode, e.output,
+                        getattr(e, "stderr", None)))
+                # Don't fallback to calling ssh directly if command really
+                # failed or if there is likely to be no usable ssh client.
+                if is_command_error(e) or _no_platform_ssh():
+                    raise
                 self.use_juju_ssh = False
             self._ensure_address()
         args = ["ssh"]
         args.extend(self._ssh_opts)
-        args.extend([self.address, command])
+        args.append(self.address)
+        args.extend(command_args)
         logging.debug(' '.join(utility.quote(i) for i in args))
         return self._run_subprocess(args)
 
@@ -156,7 +197,7 @@
 
         Tildes and environment variables in the form $TMP will be expanded.
         """
-        return self.run("cat " + utility.quote(filename))
+        return self.run(["cat", filename])
 
     def _run_subprocess(self, command):
         if self.timeout:
@@ -229,6 +270,7 @@
     def __init__(self, *args, **kwargs):
         super(WinRmRemote, self).__init__(*args, **kwargs)
         self._ensure_address()
+        self.use_juju_ssh = False
         self.certs = utility.get_winrm_certs()
         self.session = _SSLSession(self.address, self.certs)
 

=== added file 'requirements_py3.txt'
--- requirements_py3.txt	1970-01-01 00:00:00 +0000
+++ requirements_py3.txt	2017-01-09 20:19:16 +0000
@@ -0,0 +1,2 @@
+juju>=0.1.2
+

=== modified file 'run-osx-client-remote.bash'
--- run-osx-client-remote.bash	2016-01-22 20:53:38 +0000
+++ run-osx-client-remote.bash	2016-12-15 13:19:58 +0000
@@ -11,6 +11,7 @@
 tar -xf $TARFILE -C $WORKSPACE
 mkdir artifacts
 deploy_job.py parallel-osx-client-base $WORKSPACE/juju-bin/juju artifacts $env \
-    --series trusty --agent-stream=revision-build-$revision_build
+    --series xenial --use-charmstore \
+    --agent-stream=revision-build-$revision_build
 # The host experiences connection issues with AWS, retry destroy just in case.
 ~/Bin/juju destroy-environment --force -y $env || true

=== modified file 'run-uitests.bash'
--- run-uitests.bash	2016-11-30 16:11:18 +0000
+++ run-uitests.bash	2016-12-15 17:28:18 +0000
@@ -52,28 +52,10 @@
 # Print a summary for the jenkins job description.
 $SCRIPTS/s3ci.py get-summary $REVISION_BUILD $JOB_NAME
 
-# Get the pristine tarfile and unpack it to build a tainted testing juju.
-JUJU_ARCHIVE=$($SCRIPTS/s3ci.py get --config $S3_CONFIG $REVISION_BUILD build-revision juju-core_.*.tar.gz)
-tar -xf $JUJU_ARCHIVE
-JUJU_DIR=$(basename $JUJU_ARCHIVE .tar.gz)
-export GOPATH=$WORKSPACE/$JUJU_DIR
-JUJU_PACKAGE=$GOPATH/src/github.com/juju/juju
-CHARM_PACKAGE=$GOPATH/src/github.com/juju/charm
-# Add the charm package to the tree.
-go get github.com/juju/charm/...
-
-# If tree must be tampered with, then let it be us who do it to ensure
-# we understand what is not being tested.
-CSCLIENT="$GOPATH/src/gopkg.in/juju/charmrepo.v2-unstable/csclient/csclient.go"
-API_URL=$(echo $STORE_URL | sed -e "s,//,//api.,")
-SERVER_PATTERN="s,https://api.jujucharms.com,$API_URL,"
-sed -i -e "$SERVER_PATTERN" $CSCLIENT
-for PACKAGE in $JUJU_PACKAGE $CHARM_PACKAGE; do
-    cd $PACKAGE
-    go install ./...
-done
+# Get the built version of juju that we want to test with.
+JUJU_SOURCE=$($SCRIPTS/s3ci.py get-juju-bin --config $S3_CONFIG $REVISION_BUILD ./)
+JUJU_SOURCE_BINARY=$(readlink -f $JUJU_SOURCE)
 cd $WORKSPACE
-ls $GOPATH/bin
 
 # Get the releases Juju GUI.
 GUI_URL=$(sstream-query http://streams.canonical.com/juju/gui/streams/v1/index.json --output-format="%(item_url)s" | head -1)
@@ -105,19 +87,19 @@
 
 # Do not reveal credentials.
 echo "devenv/bin/uitest --driver phantom \
-    -c lxd \
+    -c google \
     --gui-archive $WORKSPACE/$GUI_ARCHIVE \
-    --gopath $GOPATH \
     --credentials <SECRET> \
     --admin <SECRET> \
     --url $STORE_URL" \
+    --juju $JUJU_SOURCE_BINARY \
     "$SUITE"
 set +x
 devenv/bin/uitest --driver phantom \
     -c google \
     --gui-archive $WORKSPACE/$GUI_ARCHIVE \
-    --gopath $GOPATH \
     --credentials $STORE_CREDENTIALS \
     --admin $STORE_ADMIN \
     --url $STORE_URL \
+    --juju $JUJU_SOURCE_BINARY \
     "$SUITE"

=== modified file 'streams-from-local.bash'
--- streams-from-local.bash	2016-08-15 18:33:26 +0000
+++ streams-from-local.bash	2017-01-03 21:57:52 +0000
@@ -22,8 +22,8 @@
 fi
 
 full_version=$($JUJUD version)
-version=$(echo $full_version | cut -d '-' -f 1)
-arch=$(echo $full_version | cut -d '-' -f 3)
+version=$(echo $full_version | sed -r 's,(.*)-[^-]+-[^-]+,\1,')
+arch=$(echo $full_version | sed -r 's,.*-([^-]+),\1,')
 
 # Juju wont permit devel versions to be in released streams.
 if [[ $version =~ (alpha|beta) ]]; then

=== modified file 'substrate.py'
--- substrate.py	2016-11-28 22:04:46 +0000
+++ substrate.py	2017-01-20 20:42:55 +0000
@@ -1,28 +1,32 @@
 from contextlib import (
     contextmanager,
-)
+    )
 import json
 import logging
 import os
 import subprocess
 from time import sleep
-import urlparse
-
+try:
+    import urlparse
+except ImportError:
+    import urllib.parse as urlparse
 from boto import ec2
 from boto.exception import EC2ResponseError
 
+from dateutil import parser as date_parser
+
 import gce
 from jujuconfig import (
     get_euca_env,
     translate_to_env,
-)
+    )
 from jujupy import (
     EnvJujuClient1X
-)
+    )
 from utility import (
     temp_dir,
     until_timeout,
-)
+    )
 import winazurearm
 
 
@@ -279,7 +283,7 @@
 
     See: https://bugs.launchpad.net/juju-core/+bug/1586089
 
-    :param client: An EnvJujuClient instance.
+    :param client: A ModelClient instance.
     :param instance_ids: a list of Juju machine instance-ids
     :return: A list of ARM VM instance ids.
     """
@@ -484,6 +488,12 @@
 
     SUBNET_CONNECTION_MODES = frozenset(('AUTO', 'DHCP', 'STATIC', 'LINK_UP'))
 
+    ACQUIRING = 'User acquiring node'
+
+    CREATED = 'created'
+
+    NODE = 'node'
+
     def __init__(self, profile, url, oauth):
         self.profile = profile
         self.url = urlparse.urljoin(url, self._API_PATH)
@@ -524,6 +534,17 @@
         allocated = {node['hostname']: node for node in nodes}
         return allocated
 
+    def get_acquire_date(self, node):
+        events = self._maas(
+            self.profile, 'events', 'query', 'id={}'.format(node))
+        for event in events['events']:
+            if node != event[self.NODE]:
+                raise ValueError(
+                    'Node "{}" was not "{}".'.format(event[self.NODE], node))
+            if event['type'] == self.ACQUIRING:
+                return date_parser.parse(event[self.CREATED])
+        raise LookupError('Unable to find acquire date for "{}".'.format(node))
+
     def get_allocated_ips(self):
         """Return a dict of allocated ips with the hostname as keys.
 

=== modified file 'test-restricted-network'
--- test-restricted-network	2016-06-14 00:46:29 +0000
+++ test-restricted-network	2017-01-25 02:01:02 +0000
@@ -41,7 +41,7 @@
 export PATH="$SCRIPTS:$PATH"
 source $JUJU_HOME/ec2rc
 mkdir -p bash
-source $($SCRIPTS/s3ci.py get $REVISION_BUILD build-revision buildvars.bash bash)
+source $(s3ci.py get $REVISION_BUILD build-revision buildvars.bash)
 ec2-terminate-job-instances
 set -x
 export INSTANCE_TYPE=m1.large

=== added file 'tested-public-clouds/fpc-2016-12-14.yaml'
--- tested-public-clouds/fpc-2016-12-14.yaml	1970-01-01 00:00:00 +0000
+++ tested-public-clouds/fpc-2016-12-14.yaml	2017-01-05 14:15:47 +0000
@@ -0,0 +1,202 @@
+# DO NOT EDIT, will be overwritten, use "juju update-clouds" to refresh.
+clouds:
+  aws:
+    type: ec2
+    description: Amazon Web Services
+    auth-types: [ access-key ]
+    regions:
+      us-east-1:
+        endpoint: https://ec2.us-east-1.amazonaws.com
+      us-east-2:
+        endpoint: https://ec2.us-east-2.amazonaws.com
+      us-west-1:
+        endpoint: https://ec2.us-west-1.amazonaws.com
+      us-west-2:
+        endpoint: https://ec2.us-west-2.amazonaws.com
+      ca-central-1:
+        endpoint: https://ec2.ca-central-1.amazonaws.com
+      eu-west-1:
+        endpoint: https://ec2.eu-west-1.amazonaws.com
+      eu-west-2:
+        endpoint: https://ec2.eu-west-2.amazonaws.com
+      eu-central-1:
+        endpoint: https://ec2.eu-central-1.amazonaws.com
+      ap-south-1:
+        endpoint: https://ec2.ap-south-1.amazonaws.com
+      ap-southeast-1:
+        endpoint: https://ec2.ap-southeast-1.amazonaws.com
+      ap-southeast-2:
+        endpoint: https://ec2.ap-southeast-2.amazonaws.com
+      ap-northeast-1:
+        endpoint: https://ec2.ap-northeast-1.amazonaws.com
+      ap-northeast-2:
+        endpoint: https://ec2.ap-northeast-2.amazonaws.com
+      sa-east-1:
+        endpoint: https://ec2.sa-east-1.amazonaws.com
+  aws-china:
+    type: ec2
+    description: Amazon China
+    auth-types: [ access-key ]
+    regions:
+      cn-north-1:
+        endpoint: https://ec2.cn-north-1.amazonaws.com.cn
+  aws-gov:
+    type: ec2
+    description: Amazon (USA Government)
+    auth-types: [ access-key ]
+    regions:
+      us-gov-west-1:
+        endpoint: https://ec2.us-gov-west-1.amazonaws.com
+  google:
+    type: gce
+    description: Google Cloud Platform
+    auth-types: [ jsonfile, oauth2 ]
+    regions:
+      us-east1:
+        endpoint: https://www.googleapis.com
+      us-central1:
+        endpoint: https://www.googleapis.com
+      europe-west1:
+        endpoint: https://www.googleapis.com
+      asia-east1:
+        endpoint: https://www.googleapis.com
+  azure:
+    type: azure
+    description: Microsoft Azure
+    auth-types: [ interactive, service-principal-secret ]
+    regions:
+      centralus:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      eastus:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      eastus2:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      northcentralus:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      southcentralus:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      westus:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      northeurope:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      westeurope:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      eastasia:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      southeastasia:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      japaneast:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      japanwest:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      brazilsouth:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      australiaeast:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      australiasoutheast:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      centralindia:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      southindia:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+      westindia:
+        endpoint: https://management.azure.com
+        storage-endpoint: https://core.windows.net
+        identity-endpoint: https://graph.windows.net
+  azure-china:
+    type: azure
+    description: Microsoft Azure China
+    auth-types: [ interactive, service-principal-secret ]
+    regions:
+      chinaeast:
+        endpoint: https://management.chinacloudapi.cn
+        storage-endpoint: https://core.chinacloudapi.cn
+        identity-endpoint: https://graph.chinacloudapi.cn
+      chinanorth:
+        endpoint: https://management.chinacloudapi.cn
+        storage-endpoint: https://core.chinacloudapi.cn
+        identity-endpoint: https://graph.chinacloudapi.cn
+  rackspace:
+    type: rackspace
+    description: Rackspace Cloud
+    auth-types: [ userpass ]
+    endpoint: https://identity.api.rackspacecloud.com/v2.0
+    regions:
+      dfw:
+        endpoint: https://identity.api.rackspacecloud.com/v2.0
+      ord:
+        endpoint: https://identity.api.rackspacecloud.com/v2.0
+      iad:
+        endpoint: https://identity.api.rackspacecloud.com/v2.0
+      lon:
+        endpoint: https://lon.identity.api.rackspacecloud.com/v2.0
+      syd:
+        endpoint: https://identity.api.rackspacecloud.com/v2.0
+      hkg:
+        endpoint: https://identity.api.rackspacecloud.com/v2.0
+  joyent:
+    type: joyent
+    description: Joyent Cloud
+    auth-types: [ userpass ]
+    regions:
+      eu-ams-1:
+        endpoint: https://eu-ams-1.api.joyentcloud.com
+      us-sw-1:
+        endpoint: https://us-sw-1.api.joyentcloud.com
+      us-east-1:
+        endpoint: https://us-east-1.api.joyentcloud.com
+      us-east-2:
+        endpoint: https://us-east-2.api.joyentcloud.com
+      us-east-3:
+        endpoint: https://us-east-3.api.joyentcloud.com
+      us-west-1:
+        endpoint: https://us-west-1.api.joyentcloud.com
+  cloudsigma:
+    type: cloudsigma
+    description: CloudSigma Cloud
+    auth-types: [ userpass ]
+    regions:
+      hnl:
+        endpoint: https://hnl.cloudsigma.com/api/2.0/
+      mia:
+        endpoint: https://mia.cloudsigma.com/api/2.0/
+      sjc:
+        endpoint: https://sjc.cloudsigma.com/api/2.0/
+      wdc:
+        endpoint: https://wdc.cloudsigma.com/api/2.0/
+      zrh:
+        endpoint: https://zrh.cloudsigma.com/api/2.0/

=== modified file 'tests/__init__.py'
--- tests/__init__.py	2016-11-21 16:30:01 +0000
+++ tests/__init__.py	2017-01-08 04:36:05 +0000
@@ -2,15 +2,22 @@
 
 from contextlib import contextmanager
 import datetime
-import errno
 import logging
 import os
-import StringIO
+import io
+try:
+    from StringIO import StringIO
+except ImportError:
+    from io import StringIO
 import subprocess
+import sys
 from tempfile import NamedTemporaryFile
 import unittest
 
-from mock import patch
+try:
+    from mock import patch
+except ImportError:
+    from unittest.mock import patch
 import yaml
 
 import utility
@@ -18,7 +25,10 @@
 
 @contextmanager
 def stdout_guard():
-    stdout = StringIO.StringIO()
+    if isinstance(sys.stdout, io.TextIOWrapper):
+        stdout = io.StringIO()
+    else:
+        stdout = io.BytesIO()
     with patch('sys.stdout', stdout):
         yield
     if stdout.getvalue() != '':
@@ -90,7 +100,7 @@
         :param data_dict: A dictionary of data, which is used to overwrite
             the data in public-clouds.yaml, or None, in which case the file
             is removed."""
-        dest_file = os.path.join(self.home_dir, '.juju/public-clouds.yaml')
+        dest_file = os.path.join(self.juju_home, 'public-clouds.yaml')
         if data_dict is None:
             with utility.skip_on_missing_file():
                 os.remove(dest_file)
@@ -103,7 +113,7 @@
     log = logging.getLogger()
     testcase.addCleanup(setattr, log, 'handlers', log.handlers)
     log.handlers = []
-    testcase.log_stream = StringIO.StringIO()
+    testcase.log_stream = StringIO()
     handler = logging.StreamHandler(testcase.log_stream)
     handler.setFormatter(logging.Formatter("%(levelname)s %(message)s"))
     log.addHandler(handler)
@@ -118,7 +128,10 @@
 
 @contextmanager
 def parse_error(test_case):
-    stderr = StringIO.StringIO()
+    if isinstance(sys.stdout, io.TextIOWrapper):
+        stderr = io.StringIO()
+    else:
+        stderr = io.BytesIO()
     with test_case.assertRaises(SystemExit):
         with patch('sys.stderr', stderr):
             yield stderr
@@ -174,17 +187,14 @@
     temporary_file = NamedTemporaryFile(delete=False)
     try:
         with temporary_file as temp_file:
-            with patch('jujupy.NamedTemporaryFile',
+            with patch('utility.NamedTemporaryFile',
                        return_value=temp_file):
                 with patch.object(temp_file, '__exit__'):
                     yield temp_file
     finally:
-        try:
+        # File may have already been deleted, e.g. by temp_yaml_file.
+        with utility.skip_on_missing_file():
             os.unlink(temporary_file.name)
-        except OSError as e:
-            # File may have already been deleted, e.g. by temp_yaml_file.
-            if e.errno != errno.ENOENT:
-                raise
 
 
 @contextmanager

=== modified file 'tests/test_assess_add_cloud.py'
--- tests/test_assess_add_cloud.py	2016-11-17 15:54:38 +0000
+++ tests/test_assess_add_cloud.py	2017-01-17 18:04:08 +0000
@@ -1,43 +1,58 @@
 from contextlib import contextmanager
 from copy import deepcopy
+from StringIO import StringIO
 from textwrap import dedent
-from StringIO import StringIO
+from unittest import TestCase
 
 from mock import (
     call,
     patch,
     )
 
-from fakejuju import fake_juju_client
 from jujupy import (
     AuthNotAccepted,
-    JujuData,
+    fake_juju_client,
+    InvalidEndpoint,
     NameNotAccepted,
     TypeNotAccepted,
     )
 from assess_add_cloud import (
     assess_all_clouds,
     assess_cloud,
+    CloudMismatch,
     CloudSpec,
+    cloud_spec,
     iter_clouds,
+    NameMismatch,
     write_status,
+    xfail,
     )
 from tests import FakeHomeTestCase
 from utility import JujuAssertionError
 
 
-class TestCase(FakeHomeTestCase):
-
-    def make_fake_juju_client(self):
-        env = JujuData('foo', juju_home=self.juju_home)
-        return fake_juju_client(env=env)
-
-
-class TestAssessCloud(TestCase):
+class TestCloudSpec(TestCase):
+
+    def test_cloud_spec(self):
+        self.assertEqual(
+                CloudSpec('label1', 'name1', {'config': '1'}, None, None),
+                cloud_spec('label1', 'name1', {'config': '1'}))
+
+
+class TestXFail(TestCase):
+
+    def test_xfail(self):
+        spec = CloudSpec('label', 'name', {'config': 'value'}, 'foo', 'bar')
+        fail_spec = xfail(spec, 'baz', 'qux')
+        self.assertEqual(fail_spec, CloudSpec(
+            'label', 'name', {'config': 'value'}, 'qux', 'baz'))
+
+
+class TestAssessCloud(FakeHomeTestCase):
 
     @contextmanager
     def cloud_client(self, clouds):
-        client = self.make_fake_juju_client()
+        client = fake_juju_client(juju_home=self.juju_home)
         client.env.load_yaml()
 
         def dump(cloud_name, cloud):
@@ -83,31 +98,50 @@
             """), stderr.getvalue())
 
 
-def make_long_endpoint(spec, regions=False):
+def make_long_endpoint(spec, endpoint_validation, regions=False):
     config = deepcopy(spec.config)
     config['endpoint'] = 'A' * 4096
     if regions:
         for region in config['regions'].values():
             region['endpoint'] = 'A' * 4096
-    return CloudSpec('long-endpoint-{}'.format(spec.name), spec.name, config,
-                     exception=None)
-
-
-class TestIterClouds(TestCase):
-
-    bogus_type = CloudSpec('bogus-type', 'bogus-type', {'type': 'bogus'},
-                           exception=TypeNotAccepted)
+    spec = cloud_spec('long-endpoint-{}'.format(spec.name), spec.name, config,
+                      InvalidEndpoint)
+    if not endpoint_validation:
+        spec = xfail(spec, 1641970, CloudMismatch)
+    return spec
+
+
+class TestIterClouds(FakeHomeTestCase):
+
+    bogus_type = cloud_spec('bogus-type', 'bogus-type', {'type': 'bogus'},
+                            exception=TypeNotAccepted)
 
     def test_manual(self):
-        cloud = {'type': 'manual', 'endpoint': 'http://example.com'}
-        spec = CloudSpec('foo', 'foo', cloud, exception=None)
+        self.maxDiff = None
+        cloud = {'type': 'manual', 'endpoint': 'http://example.com'}
+        spec = cloud_spec('foo', 'foo', cloud)
+        self.assertItemsEqual([
+            self.bogus_type, xfail(spec, 1649721, InvalidEndpoint),
+            xfail(xfail(cloud_spec('long-name-foo', 'A' * 4096, cloud),
+                        1641970, NameMismatch), 1649721, InvalidEndpoint),
+            xfail(xfail(cloud_spec('invalid-name-foo', 'invalid/name', cloud,
+                        exception=NameNotAccepted), 1641981, None),
+                  1649721, InvalidEndpoint),
+            make_long_endpoint(spec, endpoint_validation=True)],
+                iter_clouds({'foo': cloud}, endpoint_validation=True))
+
+    def test_manual_no_validation(self):
+        self.maxDiff = None
+        cloud = {'type': 'manual', 'endpoint': 'http://example.com'}
+        spec = cloud_spec('foo', 'foo', cloud)
         self.assertItemsEqual([
             self.bogus_type, spec,
-            CloudSpec('long-name-foo', 'A' * 4096, cloud, exception=None),
-            CloudSpec('invalid-name-foo', 'invalid/name', cloud,
-                      exception=NameNotAccepted),
-            make_long_endpoint(spec),
-            ], iter_clouds({'foo': cloud}))
+            xfail(cloud_spec('long-name-foo', 'A' * 4096, cloud),
+                  1641970, NameMismatch),
+            xfail(cloud_spec('invalid-name-foo', 'invalid/name', cloud,
+                             exception=NameNotAccepted), 1641981, None),
+            make_long_endpoint(spec, endpoint_validation=False)],
+                iter_clouds({'foo': cloud}, endpoint_validation=False))
 
     def test_vsphere(self):
         cloud = {
@@ -115,60 +149,107 @@
             'endpoint': '1.2.3.4',
             'regions': {'q': {'endpoint': '1.2.3.4'}},
             }
-        spec = CloudSpec('foo', 'foo', cloud, exception=None)
+        spec = cloud_spec('foo', 'foo', cloud, exception=None)
         self.assertItemsEqual([
             self.bogus_type, spec,
-            CloudSpec('invalid-name-foo', 'invalid/name', cloud,
-                      exception=NameNotAccepted),
-            CloudSpec('long-name-foo', 'A' * 4096, cloud, exception=None),
-            make_long_endpoint(spec, regions=True),
-            ], iter_clouds({'foo': cloud}))
+            xfail(cloud_spec('invalid-name-foo', 'invalid/name', cloud,
+                             exception=NameNotAccepted), 1641981, None),
+            xfail(cloud_spec('long-name-foo', 'A' * 4096, cloud,
+                             exception=None), 1641970, NameMismatch),
+            xfail(make_long_endpoint(spec, regions=True,
+                                     endpoint_validation=True),
+                  1641970, CloudMismatch),
+            ], iter_clouds({'foo': cloud}, endpoint_validation=True))
 
     def test_maas(self):
         cloud = {
             'type': 'maas',
             'endpoint': 'http://example.com',
             }
-        spec = CloudSpec('foo', 'foo', cloud, exception=None)
-        self.assertItemsEqual([
-            self.bogus_type, spec,
-            CloudSpec('invalid-name-foo', 'invalid/name', cloud,
-                      exception=NameNotAccepted),
-            CloudSpec('long-name-foo', 'A' * 4096, cloud, exception=None),
-            make_long_endpoint(spec),
-            ], iter_clouds({'foo': cloud}))
+        spec = cloud_spec('foo', 'foo', cloud, exception=None)
+        self.assertItemsEqual([
+            self.bogus_type, spec,
+            xfail(cloud_spec('invalid-name-foo', 'invalid/name', cloud,
+                             exception=NameNotAccepted), 1641981, None),
+            xfail(cloud_spec('long-name-foo', 'A' * 4096, cloud,
+                             exception=None), 1641970, NameMismatch),
+            make_long_endpoint(spec, endpoint_validation=True),
+            ], iter_clouds({'foo': cloud}, endpoint_validation=True))
+
+    def test_maas_no_validation(self):
+        cloud = {
+            'type': 'maas',
+            'endpoint': 'http://example.com',
+            }
+        spec = cloud_spec('foo', 'foo', cloud, exception=None)
+        self.assertItemsEqual([
+            self.bogus_type, spec,
+            xfail(cloud_spec('invalid-name-foo', 'invalid/name', cloud,
+                             exception=NameNotAccepted), 1641981, None),
+            xfail(cloud_spec('long-name-foo', 'A' * 4096, cloud,
+                             exception=None), 1641970, NameMismatch),
+            make_long_endpoint(spec, endpoint_validation=False),
+            ], iter_clouds({'foo': cloud}, endpoint_validation=False))
 
     def test_openstack(self):
         config = {'type': 'openstack', 'endpoint': 'http://example.com',
                   'regions': {'bar': {'endpoint': 'http://baz.example.com'}}}
-        spec = CloudSpec('foo', 'foo', config, exception=None)
-        invalid_name = CloudSpec('invalid-name-foo', 'invalid/name', config,
-                                 exception=NameNotAccepted)
-        long_name = CloudSpec('long-name-foo', 'A' * 4096, config,
-                              exception=None)
-        long_region = CloudSpec('long-endpoint-foo-bar', 'foo',
-                                deepcopy(config), exception=None)
-        long_region.config['regions']['bar']['endpoint'] = 'A' * 4096
-        bogus_auth = CloudSpec('bogus-auth-foo', 'foo',
-                               deepcopy(config), exception=AuthNotAccepted)
-        bogus_auth.config['auth-types'] = ['asdf']
-        self.assertItemsEqual([
-            self.bogus_type, spec, invalid_name, long_name, long_region,
-            bogus_auth, make_long_endpoint(spec),
-            ], iter_clouds({'foo': config}))
-
-
-class TestAssessAllClouds(TestCase):
+        spec = cloud_spec('foo', 'foo', config, exception=None)
+        invalid_name = xfail(
+            cloud_spec('invalid-name-foo', 'invalid/name', config,
+                       exception=NameNotAccepted), 1641981, None)
+        long_name = xfail(
+            cloud_spec('long-name-foo', 'A' * 4096, config, exception=None),
+            1641970, NameMismatch)
+        long_region = cloud_spec(
+            'long-endpoint-foo-bar', 'foo', deepcopy(config), InvalidEndpoint)
+        long_region.config['regions']['bar']['endpoint'] = 'A' * 4096
+        bogus_auth = cloud_spec('bogus-auth-foo', 'foo',
+                                deepcopy(config), exception=AuthNotAccepted)
+        bogus_auth.config['auth-types'] = ['asdf']
+        self.assertItemsEqual([
+            self.bogus_type, spec, invalid_name, long_name, long_region,
+            bogus_auth,
+            make_long_endpoint(spec, endpoint_validation=True),
+            ], iter_clouds({'foo': config}, endpoint_validation=True))
+
+    def test_openstack_no_validation(self):
+        config = {'type': 'openstack', 'endpoint': 'http://example.com',
+                  'regions': {'bar': {'endpoint': 'http://baz.example.com'}}}
+        spec = cloud_spec('foo', 'foo', config, exception=None)
+        invalid_name = xfail(
+            cloud_spec('invalid-name-foo', 'invalid/name', config,
+                       exception=NameNotAccepted), 1641981, None)
+        long_name = xfail(
+            cloud_spec('long-name-foo', 'A' * 4096, config, exception=None),
+            1641970, NameMismatch)
+        long_region = xfail(cloud_spec(
+            'long-endpoint-foo-bar', 'foo', deepcopy(config),
+            InvalidEndpoint), 1641970, CloudMismatch)
+        long_region.config['regions']['bar']['endpoint'] = 'A' * 4096
+        bogus_auth = cloud_spec('bogus-auth-foo', 'foo',
+                                deepcopy(config), exception=AuthNotAccepted)
+        bogus_auth.config['auth-types'] = ['asdf']
+        self.assertItemsEqual([
+            self.bogus_type, spec, invalid_name, long_name, long_region,
+            bogus_auth,
+            make_long_endpoint(spec, endpoint_validation=False),
+            ], iter_clouds({'foo': config}, endpoint_validation=False))
+
+
+class TestAssessAllClouds(FakeHomeTestCase):
 
     def test_assess_all_clouds(self):
-        client = self.make_fake_juju_client()
+        client = fake_juju_client(juju_home=self.juju_home)
         clouds = {'a': {'type': 'foo'}, 'b': {'type': 'bar'}}
+        cloud_specs = iter_clouds(clouds, endpoint_validation=True)
         exception = Exception()
         with patch('assess_add_cloud.assess_cloud',
                    side_effect=[TypeNotAccepted(), None] + [exception] * 7):
             with patch('sys.stdout'):
                 with patch('logging.exception') as exception_mock:
-                    succeeded, failed = assess_all_clouds(client, clouds)
+                    succeeded, xfail, failed = assess_all_clouds(client,
+                                                                 cloud_specs)
         self.assertEqual({'bogus-type', 'a'}, succeeded)
         self.assertEqual({
             'b', 'bogus-auth-a', 'bogus-auth-b', 'invalid-name-a',
@@ -176,8 +257,22 @@
             failed)
         self.assertEqual(exception_mock.mock_calls, [call(exception)] * 7)
 
-
-class TestWriteStatus(TestCase):
+    def test_xfail(self):
+        cloud_specs = [xfail(cloud_spec('label1', 'name1', {'config': '1'}),
+                             27, TypeNotAccepted)]
+        client = fake_juju_client(juju_home=self.juju_home)
+        with patch('assess_add_cloud.assess_cloud',
+                   side_effect=TypeNotAccepted):
+            with patch('logging.exception') as exception_mock:
+                with patch('sys.stdout'):
+                    succeeded, xfailed, failed = assess_all_clouds(client,
+                                                                   cloud_specs)
+        self.assertEqual(set(), failed)
+        self.assertEqual({27: {'label1'}}, xfailed)
+        self.assertEqual(0, exception_mock.call_count)
+
+
+class TestWriteStatus(FakeHomeTestCase):
 
     def do_write(self, status, items):
         stdout = StringIO()

=== modified file 'tests/test_assess_autoload_credentials.py'
--- tests/test_assess_autoload_credentials.py	2016-11-25 16:23:45 +0000
+++ tests/test_assess_autoload_credentials.py	2017-01-17 18:04:08 +0000
@@ -12,7 +12,7 @@
 
 import assess_autoload_credentials as aac
 from deploy_stack import BootstrapManager
-from fakejuju import fake_juju_client
+from jujupy import fake_juju_client
 from tests import (
     TestCase,
     parse_error,

=== modified file 'tests/test_assess_block.py'
--- tests/test_assess_block.py	2016-11-14 20:10:26 +0000
+++ tests/test_assess_block.py	2017-01-20 20:49:41 +0000
@@ -12,7 +12,7 @@
     parse_args,
     )
 from jujupy import (
-    EnvJujuClient,
+    ModelClient,
     EnvJujuClient1X,
     SimpleEnvironment
     )
@@ -79,12 +79,12 @@
         mock_client.list_disabled_commands.side_effect = [
             make_block_list(mock_client, []),
             make_block_list(
-                mock_client, [EnvJujuClient.command_set_destroy_model]),
+                mock_client, [ModelClient.command_set_destroy_model]),
             make_block_list(mock_client, []),
             make_block_list(
-                mock_client, [EnvJujuClient.command_set_remove_object]),
+                mock_client, [ModelClient.command_set_remove_object]),
             make_block_list(mock_client, []),
-            make_block_list(mock_client, [EnvJujuClient.command_set_all]),
+            make_block_list(mock_client, [ModelClient.command_set_all]),
             make_block_list(mock_client, []),
             ]
         mock_client.env.environment = 'foo'

=== modified file 'tests/test_assess_bootstrap.py'
--- tests/test_assess_bootstrap.py	2016-11-25 15:12:11 +0000
+++ tests/test_assess_bootstrap.py	2017-01-20 21:15:26 +0000
@@ -21,10 +21,8 @@
 from deploy_stack import (
     BootstrapManager,
     )
-from fakejuju import (
+from jujupy import (
     fake_juju_client,
-    )
-from jujupy import (
     _temp_env as temp_env,
     Status,
     )
@@ -160,11 +158,11 @@
         juju_version = '1.25.5'
     call_cxt = patch('subprocess.call')
     cc_cxt = patch('subprocess.check_call')
-    gv_cxt = patch('jujupy.EnvJujuClient.get_version',
+    gv_cxt = patch('jujupy.ModelClient.get_version',
                    side_effect=lambda cls: juju_version)
-    gjo_cxt = patch('jujupy.EnvJujuClient.get_juju_output', autospec=True,
+    gjo_cxt = patch('jujupy.ModelClient.get_juju_output', autospec=True,
                     return_value='')
-    imc_cxt = patch('jujupy.EnvJujuClient.iter_model_clients',
+    imc_cxt = patch('jujupy.ModelClient.iter_model_clients',
                     autospec=True, return_value=[])
     env_cxt = temp_env({'environments': {'bar': {'type': 'foo'}}})
     with call_cxt, cc_cxt, gv_cxt, gjo_cxt, env_cxt, imc_cxt:
@@ -233,7 +231,7 @@
             self.assertEqual(myself.env.config,
                              {'name': 'bar', 'type': 'foo'})
         with extended_bootstrap_cxt():
-            with patch('jujupy.EnvJujuClient.bootstrap', side_effect=check,
+            with patch('jujupy.ModelClient.bootstrap', side_effect=check,
                        autospec=True):
                 assess_bootstrap(parse_args(['base', 'bar', '/foo']))
         self.assertRegexpMatches(
@@ -247,7 +245,7 @@
                     'name': 'qux', 'type': 'foo', 'region': 'baz'})
             self.assertEqual(myself.env.environment, 'qux')
         with extended_bootstrap_cxt():
-            with patch('jujupy.EnvJujuClient.bootstrap', side_effect=check,
+            with patch('jujupy.ModelClient.bootstrap', side_effect=check,
                        autospec=True):
                 args = parse_args(['base', 'bar', '/foo'])
                 args.region = 'baz'
@@ -273,7 +271,7 @@
             self.assertEqual(self.target_dict, myself.env._config)
             self.assertIsNotNone(metadata_source)
         with extended_bootstrap_cxt('2.0.0'):
-            with patch('jujupy.EnvJujuClient.bootstrap', side_effect=check,
+            with patch('jujupy.ModelClient.bootstrap', side_effect=check,
                        autospec=True):
                 args = parse_args(['metadata', 'bar', '/foo'])
                 args.temp_env_name = 'qux'
@@ -288,7 +286,7 @@
             self.assertEqual(self.target_dict, myself.env._config)
             self.assertEqual('agents', metadata_source)
         with extended_bootstrap_cxt('2.0.0'):
-            with patch('jujupy.EnvJujuClient.bootstrap', side_effect=check,
+            with patch('jujupy.ModelClient.bootstrap', side_effect=check,
                        autospec=True):
                 args = parse_args(['metadata', 'bar', '/foo'])
                 args.temp_env_name = 'qux'
@@ -300,7 +298,7 @@
 
     def test_assess_metadata_valid_url(self):
         with extended_bootstrap_cxt('2.0.0'):
-            with patch('jujupy.EnvJujuClient.bootstrap', autospec=True):
+            with patch('jujupy.ModelClient.bootstrap', autospec=True):
                 args = parse_args(['metadata', 'bar', '/foo'])
                 args.temp_env_name = 'qux'
                 bs_manager = BootstrapManager.from_args(args)
@@ -317,7 +315,7 @@
     def test_get_controller_address(self):
         status = Status({'machines': {"0": {'dns-name': '255.1.1.0'}}}, '')
         client = fake_juju_client()
-        with patch('jujupy.EnvJujuClient.status_until', return_value=[status],
+        with patch('jujupy.ModelClient.status_until', return_value=[status],
                    autospec=True):
             self.assertEqual('255.1.1.0', get_controller_address(client))
 
@@ -341,7 +339,7 @@
                              myself.env._config)
             self.assertEqual(DEST, to)
         with extended_bootstrap_cxt('2.0.0'):
-            with patch('jujupy.EnvJujuClient.bootstrap', side_effect=check,
+            with patch('jujupy.ModelClient.bootstrap', side_effect=check,
                        autospec=True):
                 with patch('assess_bootstrap.get_controller_hostname',
                            return_value=DEST, autospec=True):
@@ -357,7 +355,7 @@
 
     def test_assess_to_fails(self):
         with extended_bootstrap_cxt('2.0.0'):
-            with patch('jujupy.EnvJujuClient.bootstrap', autospec=True):
+            with patch('jujupy.ModelClient.bootstrap', autospec=True):
                 with patch('assess_bootstrap.get_controller_address',
                            return_value='255.1.1.0', autospec=True):
                     args = parse_args(['to', 'bar', '/foo',

=== modified file 'tests/test_assess_cloud.py'
--- tests/test_assess_cloud.py	2016-11-23 17:52:57 +0000
+++ tests/test_assess_cloud.py	2017-01-20 20:36:36 +0000
@@ -9,25 +9,27 @@
 from assess_cloud import (
     assess_cloud_combined,
     assess_cloud_kill_controller,
+    assess_cloud_provisioning,
     client_from_args,
     parse_args,
     )
 from deploy_stack import BootstrapManager
-from fakejuju import (
+from jujupy import (
+    ModelClient,
     FakeBackend,
     fake_juju_client,
-    )
-from jujupy import (
-    EnvJujuClient,
     Juju2Backend,
-    temp_yaml_file,
     )
+from jujupy.version_client import ModelClient2_1
 from tests import (
     FakeHomeTestCase,
     observable_temp_file,
     TestCase,
     )
-from utility import temp_dir
+from utility import (
+    temp_dir,
+    temp_yaml_file,
+    )
 
 
 def backend_call(client, cmd, args, model=None, check=True, timeout=None,
@@ -92,21 +94,47 @@
             ], any_order=True)
 
 
+class TestAssessCloudProvisioning(FakeHomeTestCase):
+
+    def test_assess_cloud_provisioning(self):
+        with mocked_bs_manager(self.juju_home) as (bs_manager, config_file):
+            assess_cloud_provisioning(bs_manager)
+            client = bs_manager.client
+            juju_wrapper = client._backend.juju
+        juju_wrapper.assert_has_calls([
+            backend_call(
+                client, 'bootstrap', (
+                    '--constraints', 'mem=2G', 'foo/bar', 'foo', '--config',
+                    config_file.name, '--default-model', 'foo',
+                    '--agent-version', client.version)),
+            backend_call(client, 'add-machine', ('--series', 'win2012r2'),
+                         'foo:foo'),
+            backend_call(client, 'add-machine', ('--series', 'trusty'),
+                         'foo:foo'),
+            backend_call(client, 'remove-machine', ('0',), 'foo:foo'),
+            backend_call(client, 'remove-machine', ('1',), 'foo:foo'),
+            backend_call(
+                client, 'destroy-controller',
+                ('foo', '-y', '--destroy-all-models'), timeout=600),
+            ], any_order=True)
+
+
 class TestClientFromArgs(FakeHomeTestCase):
 
     def test_client_from_args(self):
         with temp_yaml_file({}) as clouds_file:
             args = Namespace(
                 juju_bin='/usr/bin/juju', clouds_file=clouds_file,
-                cloud='mycloud', region=None, debug=False, deadline=None)
-            with patch.object(EnvJujuClient.config_class,
+                cloud='mycloud', region=None, debug=False, deadline=None,
+                config=None)
+            with patch.object(ModelClient.config_class,
                               'from_cloud_region') as fcr_mock:
-                with patch.object(EnvJujuClient, 'get_version',
+                with patch.object(ModelClient, 'get_version',
                                   return_value='2.0.x'):
                     client = client_from_args(args)
         fcr_mock.assert_called_once_with('mycloud', None, {}, {},
                                          self.juju_home)
-        self.assertIs(type(client), EnvJujuClient)
+        self.assertIs(type(client), ModelClient2_1)
         self.assertIs(type(client._backend), Juju2Backend)
         self.assertEqual(client.version, '2.0.x')
         self.assertIs(client.env, fcr_mock.return_value)
@@ -115,17 +143,32 @@
         with temp_yaml_file({}) as clouds_file:
             args = Namespace(
                 juju_bin='FAKE', clouds_file=clouds_file, cloud='mycloud',
-                region=None, debug=False, deadline=None)
-            with patch.object(EnvJujuClient.config_class,
+                region=None, debug=False, deadline=None, config=None)
+            with patch.object(ModelClient.config_class,
                               'from_cloud_region') as fcr_mock:
                 client = client_from_args(args)
         fcr_mock.assert_called_once_with('mycloud', None, {}, {},
                                          self.juju_home)
-        self.assertIs(type(client), EnvJujuClient)
+        self.assertIs(type(client), ModelClient)
         self.assertIs(type(client._backend), FakeBackend)
         self.assertEqual(client.version, '2.0.0')
         self.assertIs(client.env, fcr_mock.return_value)
 
+    def test_config(self):
+        with temp_yaml_file({}) as clouds_file:
+            with temp_yaml_file({'foo': 'bar'}) as config_file:
+                args = Namespace(
+                    juju_bin='/usr/bin/juju', clouds_file=clouds_file,
+                    cloud='mycloud', region=None, debug=False, deadline=None,
+                    config=config_file)
+                with patch.object(ModelClient.config_class,
+                                  'from_cloud_region') as fcr_mock:
+                    with patch.object(ModelClient, 'get_version',
+                                      return_value='2.0.x'):
+                        client_from_args(args)
+        fcr_mock.assert_called_once_with('mycloud', None, {'foo': 'bar'}, {},
+                                         self.juju_home)
+
 
 class TestParseArgs(TestCase):
 
@@ -138,9 +181,15 @@
             cloud='bar', clouds_file='foo', deadline=None, debug=False,
             juju_bin='baz', keep_env=False, logs=log_dir, machine=[],
             region=None, series=None, temp_env_name='qux', upload_tools=False,
-            verbose=logging.INFO, test='combined'
+            verbose=logging.INFO, test='combined', config=None,
             ))
 
+    def test_parse_args_combined_config(self):
+        with temp_dir() as log_dir:
+            args = parse_args(['combined', 'foo', 'bar', 'baz', log_dir,
+                               'qux', '--config', 'quxx'])
+        self.assertEqual('quxx', args.config)
+
     def test_parse_args_kill_controller(self):
         with temp_dir() as log_dir:
             args = parse_args(['kill-controller', 'foo', 'bar', 'baz', log_dir,
@@ -150,5 +199,29 @@
             cloud='bar', clouds_file='foo', deadline=None, debug=False,
             juju_bin='baz', keep_env=False, logs=log_dir, machine=[],
             region=None, series=None, temp_env_name='qux', upload_tools=False,
-            verbose=logging.INFO, test='kill-controller'
-            ))
+            verbose=logging.INFO, test='kill-controller', config=None,
+            ))
+
+    def test_parse_args_kill_controller_config(self):
+        with temp_dir() as log_dir:
+            args = parse_args(['kill-controller', 'foo', 'bar', 'baz', log_dir,
+                               'qux', '--config', 'quxx'])
+        self.assertEqual('quxx', args.config)
+
+    def test_parse_args_provisioning(self):
+        with temp_dir() as log_dir:
+            args = parse_args(['provisioning', 'foo', 'bar', 'baz', log_dir,
+                               'qux'])
+        self.assertEqual(args, Namespace(
+            agent_stream=None, agent_url=None, bootstrap_host=None,
+            cloud='bar', clouds_file='foo', deadline=None, debug=False,
+            juju_bin='baz', keep_env=False, logs=log_dir, machine=[],
+            region=None, series=None, temp_env_name='qux', upload_tools=False,
+            verbose=logging.INFO, test='provisioning', config=None,
+            ))
+
+    def test_parse_args_provisioning_config(self):
+        with temp_dir() as log_dir:
+            args = parse_args(['provisioning', 'foo', 'bar', 'baz', log_dir,
+                               'qux', '--config', 'quxx'])
+        self.assertEqual('quxx', args.config)

=== modified file 'tests/test_assess_cloud_display.py'
--- tests/test_assess_cloud_display.py	2016-11-30 17:07:15 +0000
+++ tests/test_assess_cloud_display.py	2017-01-17 18:04:08 +0000
@@ -6,7 +6,7 @@
     )
 import yaml
 
-from fakejuju import fake_juju_client
+from jujupy import fake_juju_client
 from assess_cloud_display import (
     assess_clouds,
     assess_show_cloud,

=== modified file 'tests/test_assess_constraints.py'
--- tests/test_assess_constraints.py	2016-10-22 02:27:20 +0000
+++ tests/test_assess_constraints.py	2017-01-20 21:41:05 +0000
@@ -25,8 +25,8 @@
     mem_to_int,
     INSTANCE_TYPES,
     )
-from fakejuju import fake_juju_client
 from jujupy import (
+    fake_juju_client,
     Status,
     )
 from tests import (
@@ -193,7 +193,7 @@
         # weird.
         fake_client.env = env
         fake_client.bootstrap()
-        with patch('jujupy.EnvJujuClient.deploy',
+        with patch('jujupy.ModelClient.deploy',
                    autospec=True) as deploy_mock:
             yield fake_client, deploy_mock
 

=== modified file 'tests/test_assess_container_networking.py'
--- tests/test_assess_container_networking.py	2016-12-06 11:19:11 +0000
+++ tests/test_assess_container_networking.py	2017-01-20 20:58:41 +0000
@@ -9,7 +9,8 @@
     )
 
 from jujupy import (
-    EnvJujuClient,
+    ModelClient,
+    fake_juju_client,
     JujuData,
     KVM_MACHINE,
     LXC_MACHINE,
@@ -18,7 +19,6 @@
     )
 
 import assess_container_networking as jcnet
-from fakejuju import fake_juju_client
 from tests import (
     FakeHomeTestCase,
     parse_error,
@@ -29,7 +29,7 @@
 __metaclass__ = type
 
 
-class JujuMock(EnvJujuClient):
+class JujuMock(ModelClient):
     """A mock of the parts of the Juju command that the tests hit."""
 
     def __init__(self, *args, **kwargs):
@@ -103,7 +103,7 @@
 
 class TestContainerNetworking(TestCase):
     def setUp(self):
-        self.client = EnvJujuClient(
+        self.client = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
 
         self.juju_mock = fake_juju_client(cls=JujuMock)

=== modified file 'tests/test_assess_controller_permissions.py'
--- tests/test_assess_controller_permissions.py	2016-09-20 14:19:15 +0000
+++ tests/test_assess_controller_permissions.py	2017-01-20 21:37:19 +0000
@@ -75,7 +75,7 @@
         sc = patch('assess_controller_permissions.assert_superuser_permission',
                    autospec=True)
         with lc as lc_mock, ac as ac_mock, sc as sc_mock:
-            with patch("jujupy.EnvJujuClient.expect",
+            with patch("jujupy.ModelClient.expect",
                        autospec=True) as expect_mock:
                 expect_mock.return_value.isalive.return_value = False
                 assess_controller_permissions(fake_client)

=== modified file 'tests/test_assess_cs_staging.py'
--- tests/test_assess_cs_staging.py	2016-09-20 14:19:15 +0000
+++ tests/test_assess_cs_staging.py	2017-01-17 18:04:08 +0000
@@ -11,7 +11,7 @@
     main,
     _set_charm_store_ip,
 )
-from fakejuju import (
+from jujupy import (
     fake_juju_client,
     fake_juju_client_optional_jes,
     )

=== modified file 'tests/test_assess_heterogeneous_control.py'
--- tests/test_assess_heterogeneous_control.py	2016-11-25 16:23:45 +0000
+++ tests/test_assess_heterogeneous_control.py	2017-01-17 18:04:08 +0000
@@ -14,11 +14,9 @@
     parse_args,
     test_control_heterogeneous,
 )
-from fakejuju import (
+from jujupy import (
     fake_juju_client,
     fake_juju_client_optional_jes,
-)
-from jujupy import (
     _temp_env,
 )
 from tests.test_deploy_stack import FakeBootstrapManager

=== modified file 'tests/test_assess_log_rotation.py'
--- tests/test_assess_log_rotation.py	2016-11-03 14:19:28 +0000
+++ tests/test_assess_log_rotation.py	2017-01-20 21:30:58 +0000
@@ -17,12 +17,10 @@
     assess_debug_log,
     assess_machine_rotation,
     )
-from fakejuju import (
+from jujupy import (
+    ModelClient,
     fake_juju_client,
     fake_juju_client_optional_jes,
-    )
-from jujupy import (
-    EnvJujuClient,
     JujuData,
     _temp_env as temp_env,
     )
@@ -222,11 +220,11 @@
     def make_client_cxt(self):
         with temp_env({'environments': {'foo': {}}}):
             with patch('subprocess.check_output', return_value=''):
-                with patch('jujupy.EnvJujuClient.get_jes_command',
+                with patch('jujupy.ModelClient.get_jes_command',
                            autospec=True, return_value='controller'):
-                    with patch('jujupy.EnvJujuClient.juju',
+                    with patch('jujupy.ModelClient.juju',
                                autospec=True, return_value=''):
-                        with patch('jujupy.EnvJujuClient.kill_controller',
+                        with patch('jujupy.ModelClient.kill_controller',
                                    autospec=True) as kill_func:
                             with patch.object(JujuData, 'load_yaml'):
                                 yield kill_func
@@ -238,6 +236,6 @@
                 agent_url=None, agent_stream=None, series=None, region=None,
                 bootstrap_host=None, machine=[]
                 ))
-        self.assertIsInstance(client, EnvJujuClient)
+        self.assertIsInstance(client, ModelClient)
         self.assertIn('/jes-homes/bar', client.env.juju_home)
         kill_func.assert_called_once_with(client)

=== modified file 'tests/test_assess_mixed_images.py'
--- tests/test_assess_mixed_images.py	2016-10-14 19:07:39 +0000
+++ tests/test_assess_mixed_images.py	2017-01-17 18:04:08 +0000
@@ -13,7 +13,7 @@
     parse_args,
     main,
     )
-from fakejuju import fake_juju_client
+from jujupy import fake_juju_client
 from tests import (
     parse_error,
     TestCase,
@@ -85,6 +85,7 @@
         mock_client.bootstrap()
         assess_mixed_images(mock_client)
         expected = {
+            'model': {'name': 'name'},
             'machines': {
                 '0': {
                     'dns-name': '0.example.com',

=== modified file 'tests/test_assess_model_config_tree.py'
--- tests/test_assess_model_config_tree.py	2016-10-14 21:10:15 +0000
+++ tests/test_assess_model_config_tree.py	2017-01-17 18:04:08 +0000
@@ -6,7 +6,7 @@
 import StringIO
 
 import assess_model_config_tree as amct
-from fakejuju import fake_juju_client
+from jujupy import fake_juju_client
 from tests import (
     parse_error,
     TestCase,

=== added file 'tests/test_assess_model_defaults.py'
--- tests/test_assess_model_defaults.py	1970-01-01 00:00:00 +0000
+++ tests/test_assess_model_defaults.py	2016-12-14 20:03:27 +0000
@@ -0,0 +1,222 @@
+"""Tests for assess_model_defaults module."""
+
+from contextlib import contextmanager
+from copy import deepcopy
+import logging
+import StringIO
+
+from mock import (
+    call,
+    Mock,
+    patch,
+    )
+
+from assess_model_defaults import (
+    assemble_model_default,
+    assess_model_defaults,
+    assess_model_defaults_no_region,
+    assess_model_defaults_region,
+    get_new_model_config,
+    juju_assert_equal,
+    main,
+    parse_args,
+    )
+from tests import (
+    parse_error,
+    TestCase,
+    )
+from utility import (
+    JujuAssertionError,
+    )
+
+
+class FakeEnvModelDefaults:
+
+    def __init__(self, region=None):
+        self._region = region
+
+    def get_region(self):
+        return self._region
+
+
+class FakeJujuModelDefaults:
+
+    def __init__(self, defaults=None, region=None):
+        self.model_defaults = {}
+        if defaults is not None:
+            for (key, value) in defaults.items():
+                self.model_defaults[key] = {'default': value}
+        self.env = FakeEnvModelDefaults(region)
+
+    def get_model_defaults(self, model_key, cloud=None, region=None):
+        return {model_key: deepcopy(self.model_defaults[model_key])}
+
+    def set_model_defaults(self, model_key, value, cloud=None, region=None):
+        if cloud is None and region is None:
+            self.model_defaults[model_key]['controller'] = value
+        else:
+            self.model_defaults[model_key].setdefault('regions', []).append({
+                'name': region, 'value': value})
+
+    def unset_model_defaults(self, model_key, cloud=None, region=None):
+        key_defaults = self.model_defaults[model_key]
+        if cloud is None and region is None:
+            del key_defaults['controller']
+        else:
+            for index, element in enumerate(key_defaults['regions']):
+                if element['name'] == region:
+                    key_defaults['regions'].pop(index)
+                    break
+            else:
+                raise KeyError(region)
+            if not key_defaults['regions']:
+                del key_defaults['regions']
+
+
+class TestParseArgs(TestCase):
+
+    def test_common_args(self):
+        args = parse_args(["an-env", "/bin/juju", "/tmp/logs", "an-env-mod"])
+        self.assertEqual("an-env", args.env)
+        self.assertEqual("/bin/juju", args.juju_bin)
+        self.assertEqual("/tmp/logs", args.logs)
+        self.assertEqual("an-env-mod", args.temp_env_name)
+        self.assertEqual(False, args.debug)
+
+    def test_help(self):
+        fake_stdout = StringIO.StringIO()
+        with parse_error(self) as fake_stderr:
+            with patch("sys.stdout", fake_stdout):
+                parse_args(["--help"])
+        self.assertEqual("", fake_stderr.getvalue())
+        self.assertNotIn("TODO", fake_stdout.getvalue())
+
+
+class TestMain(TestCase):
+
+    def test_main(self):
+        argv = ["an-env", "/bin/juju", "/tmp/logs", "an-env-mod", "--verbose"]
+        client = Mock(spec=["is_jes_enabled"])
+        with patch("assess_model_defaults.configure_logging",
+                   autospec=True) as mock_cl:
+            with patch("assess_model_defaults.BootstrapManager.booted_context",
+                       autospec=True) as mock_bc:
+                with patch('deploy_stack.client_from_config',
+                           return_value=client) as mock_cfc:
+                    with patch("assess_model_defaults.assess_model_defaults",
+                               autospec=True) as mock_assess:
+                        main(argv)
+        mock_cl.assert_called_once_with(logging.DEBUG)
+        mock_cfc.assert_called_once_with('an-env', "/bin/juju", debug=False,
+                                         soft_deadline=None)
+        self.assertEqual(mock_bc.call_count, 1)
+        mock_assess.assert_called_once_with(client, None)
+
+
+class TestAssembleModelDefault(TestCase):
+
+    def test_assemble_model_default(self):
+        self.assertEqual({'test-mode': {'default': False}},
+                         assemble_model_default('test-mode', False))
+        self.assertEqual(
+            {'test-mode': {'default': False, 'controller': True}},
+            assemble_model_default('test-mode', False, True))
+        self.assertEqual(
+            {'test-mode': {'default': False, 'regions': [
+                {'name': 'fakeregion', 'value': True},
+                {'name': 'localhost', 'value': True}]}},
+            assemble_model_default('test-mode', False, None,
+                                   {'localhost': True, 'fakeregion': True}))
+        self.assertEqual(
+            {'automatically-retry-hooks':
+                {'default': True, 'controller': False}},
+            assemble_model_default('automatically-retry-hooks', True, False))
+
+
+class TestAssert(TestCase):
+
+    def test_juju_assert_equal(self):
+        juju_assert_equal(0, 0, 'should-pass')
+        with self.assertRaises(JujuAssertionError) as context:
+            juju_assert_equal(0, 1, 'should-fail')
+        self.assertEqual(context.exception.args, ('should-fail', 0, 1))
+
+
+class TestGetNewModelConfig(TestCase):
+
+    @contextmanager
+    def wrap_new_model_test(self):
+        (mock_client, mock_env, mock_model) = (Mock(), Mock(), Mock())
+        mock_client.env.attach_mock(Mock(return_value=mock_env), 'clone')
+        mock_client.attach_mock(Mock(return_value=mock_model), 'add_model')
+        mock_model.attach_mock(Mock(return_value='<model_config>'),
+                               'get_model_config')
+        yield (mock_client, mock_env, mock_model)
+        mock_client.add_model.assert_called_once_with(mock_env)
+        mock_model.get_model_config.assert_called_once_with()
+        mock_model.destroy_model.assert_called_once_with()
+
+    def test_get_new_model_config_defaults(self):
+        with self.wrap_new_model_test() as (client, env, model):
+            config = get_new_model_config(client)
+            client.env.clone.assert_called_once_with('temp-model')
+            self.assertFalse(env.set_region.called)
+            self.assertEqual('<model_config>', config)
+
+    def test_get_new_model_config_region(self):
+        with self.wrap_new_model_test() as (client, env, model):
+            config = get_new_model_config(client, 'region')
+            client.env.clone.assert_called_once_with('temp-model')
+            env.set_region.assert_called_once_with('region')
+            self.assertEqual('<model_config>', config)
+
+    def test_get_new_model_config_model_name(self):
+        with self.wrap_new_model_test() as (client, env, model):
+            config = get_new_model_config(client, model_name='diff-name')
+            client.env.clone.assert_called_once_with('diff-name')
+            self.assertFalse(env.set_region.called)
+            self.assertEqual('<model_config>', config)
+
+
+class TestAssessModelDefaults(TestCase):
+
+    def test_model_defaults_no_region(self):
+        client = Mock(wraps=FakeJujuModelDefaults({'some-key': 'black'}))
+        with patch('assess_model_defaults.get_new_model_config',
+                   return_value={'some-key': {'value': 'yellow'}},
+                   autospec=True) as get_config_mock:
+            assess_model_defaults_no_region(client, 'some-key', 'yellow')
+        get_config_mock.assert_called_once_with(client, None)
+        self.assertEqual(4, client.get_model_defaults.call_count)
+        client.get_model_defaults.assert_has_calls([call('some-key')] * 3)
+        client.set_model_defaults.assert_called_once_with(
+            'some-key', 'yellow', None, None)
+        client.unset_model_defaults.assert_called_once_with(
+            'some-key', None, None)
+
+    def test_model_defaults_region(self):
+        client = Mock(wraps=FakeJujuModelDefaults({'some-key': 'black'}))
+        with patch('assess_model_defaults.get_new_model_config',
+                   return_value={'some-key': {'value': 'yellow'}},
+                   autospec=True) as get_config_mock:
+            assess_model_defaults_region(
+                client, 'some-key', 'yellow', 'localhost', 'localhost')
+        get_config_mock.assert_called_once_with(client, 'localhost')
+        self.assertEqual(4, client.get_model_defaults.call_count)
+        client.get_model_defaults.assert_has_calls([call('some-key')] * 3)
+        client.set_model_defaults.assert_called_once_with(
+            'some-key', 'yellow', 'localhost', 'localhost')
+        client.unset_model_defaults.assert_called_once_with(
+            'some-key', 'localhost', 'localhost')
+
+    def test_model_defaults(self):
+        fake_client = FakeJujuModelDefaults(region='localhost')
+        with patch('assess_model_defaults.assess_model_defaults_no_region',
+                   autospec=True) as assess_no_region_mock:
+            with patch('assess_model_defaults.assess_model_defaults_region',
+                       autospec=True) as assess_region_mock:
+                assess_model_defaults(fake_client, None)
+        assess_no_region_mock.assert_called_once_with(
+            fake_client, 'automatically-retry-hooks', False)
+        assess_region_mock.assert_called_once_with(
+            fake_client, 'default-series', 'trusty', region='localhost')

=== modified file 'tests/test_assess_model_migration.py'
--- tests/test_assess_model_migration.py	2016-12-02 22:51:18 +0000
+++ tests/test_assess_model_migration.py	2017-01-20 20:30:41 +0000
@@ -3,21 +3,32 @@
 import argparse
 from contextlib import contextmanager
 import logging
-from mock import call, Mock, patch
 import os
 import StringIO
 from subprocess import CalledProcessError
 from textwrap import dedent
 
+from mock import (
+    call,
+    Mock,
+    patch,
+    )
+import yaml
+
 import assess_model_migration as amm
+from assess_model_migration import (
+    assert_data_file_lists_correct_controller_for_model,
+    assert_model_has_correct_controller_uuid,
+    ensure_api_login_redirects,
+    )
 from assess_user_grant_revoke import User
 from deploy_stack import (
     BootstrapManager,
     get_random_string,
-)
-from fakejuju import fake_juju_client
+    )
 from jujupy import (
-    EnvJujuClient,
+    ModelClient,
+    fake_juju_client,
     JujuData,
     SoftDeadlineExceeded,
     )
@@ -26,13 +37,13 @@
     FakeHomeTestCase,
     parse_error,
     TestCase,
-)
+    )
 from utility import (
     JujuAssertionError,
     noop_context,
     temp_dir,
     until_timeout,
-)
+    )
 
 
 class TestParseArgs(TestCase):
@@ -215,16 +226,33 @@
 
 class TestAssertModelMigratedSuccessfully(TestCase):
 
-    def test_assert_model_migrated_successfully(self):
+    def test_assert_model_migrated_successfully_default_values(self):
         client = Mock()
         with patch.object(
-                amm, 'test_deployed_mongo_is_up', autospec=True) as m_tdmiu:
+                amm,
+                'assert_deployed_charm_is_responding',
+                autospec=True) as m_tdmiu:
             with patch.object(
                     amm, 'ensure_model_is_functional',
                     autospec=True) as m_emif:
                 amm.assert_model_migrated_successfully(client, 'test')
         client.wait_for_workloads.assert_called_once_with()
-        m_tdmiu.assert_called_once_with(client)
+        m_tdmiu.assert_called_once_with(client, None)
+        m_emif.assert_called_once_with(client, 'test')
+
+    def test_assert_model_migrated_successfully_explicit_values(self):
+        client = Mock()
+        with patch.object(
+                amm,
+                'assert_deployed_charm_is_responding',
+                autospec=True) as m_tdmiu:
+            with patch.object(
+                    amm, 'ensure_model_is_functional',
+                    autospec=True) as m_emif:
+                amm.assert_model_migrated_successfully(
+                    client, 'test', 'abc123')
+        client.wait_for_workloads.assert_called_once_with()
+        m_tdmiu.assert_called_once_with(client, 'abc123')
         m_emif.assert_called_once_with(client, 'test')
 
 
@@ -341,7 +369,7 @@
             status:
                 current: available
                 since: 4 minutes ago
-    """)
+        """)
 
     migration_status_yaml = dedent("""\
         name:
@@ -422,7 +450,7 @@
             mock_sleep.assert_called_once_with(1)
 
     def test_suppresses_deadline(self):
-        client = EnvJujuClient(JujuData('local', juju_home=''), None, None)
+        client = ModelClient(JujuData('local', juju_home=''), None, None)
         with client_past_deadline(client):
 
             real_check_timeouts = client.check_timeouts
@@ -441,7 +469,7 @@
                     amm.wait_for_model(client, 'TestModelName')
 
     def test_checks_deadline(self):
-        client = EnvJujuClient(JujuData('local', juju_home=''), None, None)
+        client = ModelClient(JujuData('local', juju_home=''), None, None)
         with client_past_deadline(client):
             def get_models():
                 return {'models': [{'name': 'TestModelName'}]}
@@ -456,25 +484,79 @@
                     amm.wait_for_model(client, 'TestModelName')
 
 
-class TestDeployMongodbToNewModel(TestCase):
-
-    def test_deploys_mongodb_to_new_model(self):
-        new_model = Mock()
-        source_client = Mock()
-        source_client.add_model.return_value = new_model
-
-        with patch.object(
-                amm, 'test_deployed_mongo_is_up', autospec=True) as m_tdmiu:
-            self.assertEqual(
-                new_model,
-                amm.deploy_mongodb_to_new_model(source_client, 'test'))
-
-        source_client.add_model.assert_called_once_with(
-            source_client.env.clone.return_value)
-        new_model.juju.assert_called_once_with('deploy', ('mongodb'))
-        new_model.wait_for_started.assert_called_once_with()
-        new_model.wait_for_workloads.assert_called_once_with()
-        m_tdmiu.assert_called_once_with(new_model)
+class TestEnsureApiLoginRedirects(FakeHomeTestCase):
+
+    def test_ensure_api_login_redirects(self):
+        def named_fake_juju_client(name):
+            env = JujuData(name, {'name': name, 'type': 'foo'})
+            return fake_juju_client(env=env)
+
+        patch_assert_data = patch.object(
+            amm, 'assert_data_file_lists_correct_controller_for_model')
+        patch_assert_uuid = patch.object(
+            amm, 'assert_model_has_correct_controller_uuid')
+        client1 = named_fake_juju_client('source')
+        client2 = named_fake_juju_client('target')
+        client3 = named_fake_juju_client('dest')
+        model_details_list = [
+            {client.env.environment: {'controller-uuid': uuid}}
+            for (client, uuid)
+            in [(client1, '12345'), (client3, 'ABCDE')]]
+        with patch('logging.Logger.info', autospec=True):
+            with patch('jujupy.ModelClient.show_model', autospec=True,
+                       side_effect=model_details_list) as show_model_mock:
+                with patch.object(amm, 'migrate_model_to_controller',
+                                  return_value=client3):
+                    with patch_assert_data, patch_assert_uuid:
+                        ensure_api_login_redirects(client1, client2)
+        show_model_mock.assert_has_calls([call(client1), call(client3)])
+
+    def create_models_yaml(self, client, controllers_to_models):
+        data = dict([(controller, {'models': models})
+                     for (controller, models)
+                     in controllers_to_models.items()])
+        with open(os.path.join(client.env.juju_home, 'models.yaml'), 'w',
+                  ) as models_yaml:
+            yaml.safe_dump({'controllers': data}, models_yaml)
+
+    def test_assert_data_file_lists_correct_controller_for_model(self):
+        client = fake_juju_client(juju_home=self.juju_home)
+        self.create_models_yaml(client, {'testing': [client.env.environment]})
+        assert_data_file_lists_correct_controller_for_model(client, 'testing')
+
+    def test_assert_data_file_lists_correct_controller_for_model_fails(self):
+        client = fake_juju_client(juju_home=self.juju_home)
+        self.create_models_yaml(client, {'testing': ['not-the-client-name']})
+        with self.assertRaises(JujuAssertionError):
+            assert_data_file_lists_correct_controller_for_model(
+                client, 'testing')
+
+    def patch_uuid_client(self, client, uuid_model, uuid_controller):
+        def get_juju_output(command, *args, **kwargs):
+            m_name = client.env.environment
+            c_name = client.env.controller.name
+            if 'show-model' == command:
+                return yaml.safe_dump({
+                    m_name: {'controller-uuid': uuid_model}})
+            elif 'show-controller' == command:
+                return yaml.safe_dump({
+                    c_name: {'details': {'uuid': uuid_controller}}})
+            else:
+                raise Exception(
+                    'get_juju_output mock does not handle ' + command)
+
+        return patch.object(client, 'get_juju_output', get_juju_output)
+
+    def test_assert_model_has_correct_controller_uuid(self):
+        client = fake_juju_client()
+        with self.patch_uuid_client(client, '12345', '12345'):
+            assert_model_has_correct_controller_uuid(client)
+
+    def test_assert_model_has_correct_controller_uuid_fails(self):
+        client = fake_juju_client()
+        with self.patch_uuid_client(client, '12345', 'ABCDE'):
+            with self.assertRaises(JujuAssertionError):
+                assert_model_has_correct_controller_uuid(client)
 
 
 class TestMigrateModelToController(FakeHomeTestCase):
@@ -628,6 +710,10 @@
     yield '/tmp/dir'
 
 
+def patch_amm(target):
+    return patch.object(amm, target, autospec=True)
+
+
 class TestAssessModelMigration(TestCase):
 
     def test_runs_develop_tests_when_requested(self):
@@ -639,43 +725,36 @@
         bs1.booted_context.return_value = noop_context()
         bs2.existing_booted_context.return_value = noop_context()
 
-        def patch_amm(target):
-            return patch.object(amm, target, autospec=True)
-
+        patch_between = patch_amm(
+            'ensure_migration_including_resources_succeeds')
         patch_user = patch_amm(
             'ensure_migrating_with_insufficient_user_permissions_fails')
         patch_super = patch_amm(
             'ensure_migrating_with_superuser_user_permissions_succeeds')
-        patch_between = patch_amm(
-            'ensure_able_to_migrate_model_between_controllers')
         patch_rollback = patch_amm('ensure_migration_rolls_back_on_failure')
-        patch_resource = patch_amm('ensure_migration_of_resources_succeeds')
-        patch_back = patch_amm(
-            'ensure_migrating_to_target_and_back_to_source_succeeds')
         patch_logs = patch_amm('ensure_model_logs_are_migrated')
         patch_superother = patch_amm(
             'ensure_superuser_can_migrate_other_user_models')
         patch_redirects = patch_amm('ensure_api_login_redirects')
+        patch_deploy_simple = patch_amm('deploy_simple_server_to_new_model')
 
-        with patch_user as m_user, patch_super as m_super:
-            with patch_between as m_between, patch_rollback as m_rollback:
-                with patch_resource as m_resource, patch_back as m_back:
-                    with patch_logs as m_logs, patch_redirects as m_redirects:
+        with patch_between as m_between:
+            with patch_user as m_user, patch_super as m_super:
+                with patch_rollback as m_rollback, patch_logs as m_logs:
+                    with patch_redirects as m_redirects:
                         with patch_superother as m_superother:
                             with patch.object(
                                     amm, 'temp_dir',
                                     autospec=True,
                                     return_value=tmp_ctx()):
-                                amm.assess_model_migration(
-                                    bs1, bs2, args)
+                                with patch_deploy_simple:
+                                    amm.assess_model_migration(bs1, bs2, args)
         source_client = bs1.client
         dest_client = bs2.client
+        m_between.assert_called_once_with(source_client, dest_client)
         m_user.assert_called_once_with(source_client, dest_client, '/tmp/dir')
         m_super.assert_called_once_with(source_client, dest_client, '/tmp/dir')
-        m_between.assert_called_once_with(source_client, dest_client)
         m_rollback.assert_called_once_with(source_client, dest_client)
-        m_resource.assert_called_once_with(source_client, dest_client)
-        m_back.assert_called_once_with(source_client, dest_client)
         m_logs.assert_called_once_with(source_client, dest_client)
         m_superother.assert_called_once_with(
             source_client, dest_client, '/tmp/dir')
@@ -690,30 +769,29 @@
         bs1.booted_context.return_value = noop_context()
         bs2.existing_booted_context.return_value = noop_context()
 
-        with patch.object(
-                amm,
-                'ensure_migrating_with_insufficient_user_permissions_fails',
-                autospec=True) as m_user:
-            with patch.object(
-                    amm,
-                    'ensure_migrating_with_superuser_user_permissions_succeeds',  # NOQA
-                    autospec=True) as m_super:
-                with patch.object(
-                        amm,
-                        'ensure_able_to_migrate_model_between_controllers',
-                        autospec=True) as m_between:
-                    with patch.object(
-                            amm,
-                            'ensure_migration_rolls_back_on_failure',
-                            autospec=True) as m_rollback:
-                        with patch.object(
-                                amm, 'temp_dir',
-                                autospec=True, return_value=tmp_ctx()):
-                            amm.assess_model_migration(bs1, bs2, args)
+        patch_user = patch_amm(
+            'ensure_migrating_with_insufficient_user_permissions_fails')
+        patch_super = patch_amm(
+            'ensure_migrating_with_superuser_user_permissions_succeeds')
+        patch_between = patch_amm(
+            'ensure_migration_including_resources_succeeds')
+        patch_rollback = patch_amm('ensure_migration_rolls_back_on_failure')
+        patch_logs = patch_amm('ensure_model_logs_are_migrated')
+
+        with patch_user as m_user:
+            with patch_super as m_super:
+                with patch_between as m_between:
+                    with patch_rollback as m_rollback:
+                        with patch_logs as m_logs:
+                            with patch.object(
+                                    amm, 'temp_dir',
+                                    autospec=True, return_value=tmp_ctx()):
+                                amm.assess_model_migration(bs1, bs2, args)
         source_client = bs1.client
         dest_client = bs2.client
         m_user.assert_called_once_with(source_client, dest_client, '/tmp/dir')
         m_super.assert_called_once_with(source_client, dest_client, '/tmp/dir')
         m_between.assert_called_once_with(source_client, dest_client)
+        m_logs.assert_called_once_with(source_client, dest_client)
 
         self.assertEqual(m_rollback.call_count, 0)

=== renamed file 'tests/test_assess_jes_deploy.py' => 'tests/test_assess_multimodel.py'
--- tests/test_assess_jes_deploy.py	2016-11-09 14:25:42 +0000
+++ tests/test_assess_multimodel.py	2017-01-20 21:34:12 +0000
@@ -1,29 +1,39 @@
+#!/usr/bin/env python
+
 from argparse import Namespace
+from contextlib import contextmanager
+import os
+
 from mock import (
+    Mock,
     patch,
     )
-import os
 
-from assess_jes_deploy import (
+from assess_multimodel import (
+    assess_destroy_current,
     check_services,
     env_token,
     hosted_environment,
-    jes_setup,
+    multimodel_setup,
     )
-from fakejuju import fake_juju_client
 from jujupy import (
     EnvJujuClient25,
+    fake_juju_client,
     JUJU_DEV_FEATURE_FLAGS,
     SimpleEnvironment,
     )
 import tests
 
 
-class TestJES(tests.FakeHomeTestCase):
+def patch_local(target, **kwargs):
+    return patch('assess_multimodel.' + target, **kwargs)
+
+
+class TestMultiModel(tests.FakeHomeTestCase):
 
     client_class = EnvJujuClient25
 
-    @patch('assess_jes_deploy.get_random_string', autospec=True)
+    @patch_local('get_random_string', autospec=True)
     @patch('jujupy.SimpleEnvironment.from_config')
     def mock_client(self, from_config_func, get_random_string_func):
         from_config_func.return_value = SimpleEnvironment('baz', {})
@@ -33,20 +43,19 @@
             '1.25-foobar', 'path')
         return client
 
-    @patch('assess_jes_deploy.get_random_string', autospec=True)
-    def test_env_token(self, get_random_string_func):
-        get_random_string_func.return_value = 'fakeran'
-        self.assertEqual(env_token('env1'), 'env1fakeran')
+    def test_env_token(self):
+        with patch_local('get_random_string', return_value='fakeran'):
+            self.assertEqual(env_token('env1'), 'env1fakeran')
 
     def test_set_jes_flag(self):
         client = self.mock_client()
         env = client._shell_environ()
         self.assertNotIn('jes', env.get(JUJU_DEV_FEATURE_FLAGS, '').split(","))
 
-    @patch('assess_jes_deploy.print_now', autospec=True)
-    @patch('assess_jes_deploy.get_random_string', autospec=True)
-    @patch('jujupy.EnvJujuClient.juju', autospec=True)
-    @patch('assess_jes_deploy.check_token', autospec=True)
+    @patch_local('print_now', autospec=True)
+    @patch_local('get_random_string', autospec=True)
+    @patch('jujupy.ModelClient.juju', autospec=True)
+    @patch_local('check_token', autospec=True)
     def test_check_services(
             self,
             check_token_func,
@@ -64,12 +73,12 @@
         check_token_func.assert_called_once_with(client, 'tokenfakeran')
         print_now_func.assert_called_once_with('checking services in token')
 
-    @patch('jujupy.EnvJujuClient.get_full_path')
-    @patch('jujupy.EnvJujuClient.add_ssh_machines', autospec=True)
-    @patch('assess_jes_deploy.boot_context', autospec=True)
-    @patch('assess_jes_deploy.configure_logging', autospec=True)
-    @patch('assess_jes_deploy.client_from_config')
-    def test_jes_setup(
+    @patch('jujupy.ModelClient.get_full_path', autospec=True)
+    @patch('jujupy.ModelClient.add_ssh_machines', autospec=True)
+    @patch_local('boot_context', autospec=True)
+    @patch_local('configure_logging', autospec=True)
+    @patch_local('client_from_config', autospec=True)
+    def test_multimodel_setup(
             self,
             by_version_func,
             configure_logging_func,
@@ -93,7 +102,8 @@
         with patch.object(expected_client, 'enable_jes'):
             with patch.object(expected_client, 'is_jes_enabled',
                               return_value=True):
-                with jes_setup(setup_args) as (client, charm_series, base_env):
+                with multimodel_setup(setup_args) as (
+                        client, charm_series, base_env):
                     self.assertEqual(1, client.is_jes_enabled.call_count)
                     self.assertEqual(0, client.enable_jes.call_count)
 
@@ -116,11 +126,29 @@
         with patch.object(expected_client, 'enable_jes'):
             with patch.object(expected_client, 'is_jes_enabled',
                               return_value=False):
-                with jes_setup(setup_args) as (client, charm_previx, base_env):
+                with multimodel_setup(setup_args) as (
+                        client, charm_previx, base_env):
                     self.assertEqual(1, client.is_jes_enabled.call_count)
                     self.assertEqual(1, client.enable_jes.call_count)
 
 
+class TestDestroyCurrent(tests.TestCase):
+
+    @contextmanager
+    def prepare(self):
+        (mock_client, mock_model) = (Mock(), Mock())
+        mock_client.attach_mock(Mock(return_value=mock_model), 'add_model')
+        yield (mock_client, mock_model)
+
+    def test_destroy_current(self):
+        with self.prepare() as (client, model):
+            assess_destroy_current(client)
+        client.add_model.assert_called_once_with('delete-me')
+        model.switch.assert_called_once_with('delete-me')
+        model.destroy_model.assert_called_once_with()
+        model.show_controller.assert_called_once_with()
+
+
 class TestHostedEnvironment(tests.FakeHomeTestCase):
 
     def test_hosted_environment(self):

=== modified file 'tests/test_assess_proxy.py'
--- tests/test_assess_proxy.py	2016-10-20 18:48:20 +0000
+++ tests/test_assess_proxy.py	2017-01-17 18:04:08 +0000
@@ -12,7 +12,7 @@
 import subprocess
 
 import assess_proxy
-from fakejuju import fake_juju_client
+from jujupy import fake_juju_client
 from tests import (
     parse_error,
     TestCase,

=== modified file 'tests/test_assess_public_clouds.py'
--- tests/test_assess_public_clouds.py	2016-11-22 21:58:33 +0000
+++ tests/test_assess_public_clouds.py	2017-01-17 18:04:08 +0000
@@ -17,12 +17,13 @@
     default_log_dir,
     iter_cloud_regions,
     main,
+    make_bootstrap_manager,
     make_logging_dir,
     parse_args,
-    prepare_cloud,
     yaml_file_load,
     )
-from fakejuju import (
+from deploy_stack import BootstrapManager
+from jujupy import (
     fake_juju_client,
     )
 from tests import (
@@ -127,23 +128,30 @@
         self.assertFalse(clean_dir_mock.called)
 
 
-class TestPrepareCloud(FakeHomeTestCase):
+class TestMakeBootstrapManager(FakeHomeTestCase):
 
-    def test_prepare_cloud(self):
+    def test_make_bootstrap_manager(self):
         client = fake_juju_client()
         bs_manager = Mock()
         with patch_local('make_logging_dir', autospec=True,
                          side_effect=os.path.join):
-            with patch_local('BootstrapManager', autospec=True,
-                             return_value=bs_manager) as bsm_mock:
-                with patch_local('assess_cloud_combined', autospec=True):
-                    prepare_cloud('config', 'region', client, 'log_dir')
-        self.assertEqual(1, bsm_mock.call_count)
-        bsm_mock.assert_called_once_with(
-            'boot-cpc-foo-region', client, client, bootstrap_host=None,
-            machines=[], series=None, agent_url=None, agent_stream=None,
-            region='region', log_dir='log_dir/config/region', keep_env=False,
-            permanent=True, jes_enabled=True)
+            with patch_local('assess_cloud_combined', autospec=True):
+                bs_manager = make_bootstrap_manager('config', 'region', client,
+                                                    'log_dir')
+        self.assertEqual(bs_manager.temp_env_name, 'boot-cpc-foo-region')
+        self.assertIs(bs_manager.client, client)
+        self.assertIs(bs_manager.tear_down_client, client)
+        self.assertIs(bs_manager.bootstrap_host, None)
+        self.assertEqual(bs_manager.machines, [])
+        self.assertIs(bs_manager.series, None)
+        self.assertIs(bs_manager.agent_url, None)
+        self.assertIs(bs_manager.agent_stream, None)
+        self.assertEqual(bs_manager.region, 'region')
+        self.assertEqual(bs_manager.log_dir, 'log_dir/config/region')
+        self.assertIs(bs_manager.keep_env, False)
+        self.assertIs(bs_manager.permanent, True)
+        self.assertIs(bs_manager.jes_enabled, True)
+        self.assertIs(bs_manager.logged_exception_exit, False)
 
 
 class TestIterCloudRegions(TestCase):
@@ -174,12 +182,14 @@
         """Handles all the patching for testing bootstrap_cloud_regions."""
         with patch_local('iter_cloud_regions', autospec=True,
                          return_value=cloud_regions) as iter_mock:
-            with patch_local('prepare_cloud', autospec=True,
+            with patch_local('assess_cloud_combined', autospec=True,
                              side_effect=error) as bootstrap_mock:
                 with patch_local('client_from_config', autospec=True,
                                  return_value=client):
                     with patch('logging.info', autospec=True) as info_mock:
-                        yield (iter_mock, bootstrap_mock, info_mock)
+                        with patch_local('make_logging_dir', autospec=True,
+                                         side_effect=lambda x, y, z: y):
+                            yield (iter_mock, bootstrap_mock, info_mock)
 
     def run_test_bootstrap_cloud_regions(self, start=0, error=None):
         pc_key = 'public_clouds'
@@ -188,14 +198,21 @@
         args = Namespace(start=start, debug=True, deadline=None,
                          juju_bin='juju', logs='/tmp/log')
         fake_client = fake_juju_client()
-        expect_calls = [call('foo.config', 'foo', fake_client, '/tmp/log'),
-                        call('bar.config', 'bar', fake_client, '/tmp/log')]
+        expect_values = [('foo.config', 'foo'),
+                         ('bar.config', 'bar')]
         with self.patch_for_test(cloud_regions, fake_client, error) as (
                 iter_mock, bootstrap_mock, info_mock):
             errors = list(bootstrap_cloud_regions(pc_key, cred_key, args))
 
         iter_mock.assert_called_once_with(pc_key, cred_key)
-        bootstrap_mock.assert_has_calls(expect_calls[start:])
+        for num, (config, region) in enumerate(expect_values[start:]):
+            acc_call = bootstrap_mock.mock_calls[num]
+            name, args, kwargs = acc_call
+            self.assertEqual({}, kwargs)
+            (bs_manager,) = args
+            self.assertIsInstance(bs_manager, BootstrapManager)
+            self.assertEqual(bs_manager.region, region)
+            self.assertEqual(bs_manager.log_dir, config)
         self.assertEqual(len(cloud_regions) - start, info_mock.call_count)
         if error is None:
             self.assertEqual([], errors)

=== modified file 'tests/test_assess_recovery.py'
--- tests/test_assess_recovery.py	2016-12-05 16:26:19 +0000
+++ tests/test_assess_recovery.py	2017-01-17 18:04:08 +0000
@@ -20,8 +20,8 @@
     restore_missing_state_server,
     )
 from deploy_stack import BootstrapManager
-from fakejuju import fake_juju_client
 from jujupy import (
+    fake_juju_client,
     Machine,
     )
 from subprocess import CalledProcessError

=== modified file 'tests/test_assess_spaces_subnets.py'
--- tests/test_assess_spaces_subnets.py	2016-11-21 20:29:25 +0000
+++ tests/test_assess_spaces_subnets.py	2017-01-20 20:38:11 +0000
@@ -11,7 +11,7 @@
 
 import assess_spaces_subnets as jss
 from jujupy import (
-    EnvJujuClient,
+    ModelClient,
     JujuData,
     Status,
     )
@@ -209,7 +209,7 @@
 
 class JujuMockTestCase(TestCase):
     def setUp(self):
-        self.client = EnvJujuClient(
+        self.client = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
 
         def nil_func():

=== modified file 'tests/test_assess_ssh_keys.py'
--- tests/test_assess_ssh_keys.py	2016-10-14 19:07:39 +0000
+++ tests/test_assess_ssh_keys.py	2017-01-17 18:04:08 +0000
@@ -10,7 +10,7 @@
     parse_args,
     parse_ssh_keys_output,
     )
-from fakejuju import fake_juju_client
+from jujupy import fake_juju_client
 from tests import (
     parse_error,
     TestCase,

=== modified file 'tests/test_assess_unregister.py'
--- tests/test_assess_unregister.py	2016-10-14 19:07:39 +0000
+++ tests/test_assess_unregister.py	2017-01-20 20:48:06 +0000
@@ -11,7 +11,7 @@
 from textwrap import dedent
 
 import assess_unregister as a_unreg
-from fakejuju import fake_juju_client
+from jujupy import fake_juju_client
 from tests import (
     parse_error,
     TestCase,
@@ -64,7 +64,7 @@
 
     def test_unregister(self):
         fake_user = Mock()
-        with patch('jujupy.EnvJujuClient.register_user',
+        with patch('jujupy.ModelClient.register_user',
                    return_value=fake_user):
             with patch.object(
                     a_unreg, 'assert_controller_list',

=== modified file 'tests/test_assess_user_grant_revoke.py'
--- tests/test_assess_user_grant_revoke.py	2016-09-20 14:19:15 +0000
+++ tests/test_assess_user_grant_revoke.py	2017-01-20 21:35:27 +0000
@@ -19,11 +19,11 @@
     main,
     parse_args,
 )
-from fakejuju import (
+from jujupy import (
     fake_juju_client,
     FakeBackend,
+    JUJU_DEV_FEATURE_FLAGS,
 )
-from jujupy import JUJU_DEV_FEATURE_FLAGS
 from tests import (
     parse_error,
     TestCase,
@@ -123,7 +123,7 @@
             fake_admin_client = fake_juju_client()
             ac = patch("assess_user_grant_revoke.assert_admin_model",
                        return_value=True)
-            with patch("jujupy.EnvJujuClient.revoke", return_value=True):
+            with patch("jujupy.ModelClient.revoke", return_value=True):
                 with patch("assess_user_grant_revoke.assert_read_model",
                            return_value=True) as read_mock:
                     with patch("assess_user_grant_revoke.assert_write_model",
@@ -185,7 +185,7 @@
                      autospec=True)
         log = patch("assess_user_grant_revoke.assert_logout_login",
                     autospec=True)
-        expect = patch("jujupy.EnvJujuClient.expect",
+        expect = patch("jujupy.ModelClient.expect",
                        autospec=True)
         read = patch("assess_user_grant_revoke.assert_read_model",
                      autospec=True)

=== modified file 'tests/test_chaos.py'
--- tests/test_chaos.py	2016-11-02 20:10:13 +0000
+++ tests/test_chaos.py	2017-01-20 21:09:25 +0000
@@ -18,7 +18,7 @@
     MonkeyRunner,
     )
 from jujupy import (
-    EnvJujuClient,
+    ModelClient,
     JujuData,
     )
 from remote import SSHRemote
@@ -31,7 +31,7 @@
 class TestBackgroundChaos(FakeHomeTestCase):
 
     def test_background_chaos(self):
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo/juju')
+        client = ModelClient(JujuData('foo', {}), None, '/foo/juju')
         with patch('chaos.MonkeyRunner.deploy_chaos_monkey',
                    autospec=True) as d_mock:
             with patch('chaos.MonkeyRunner.unleash_once',
@@ -55,7 +55,7 @@
                          w_mock.mock_calls[1][2])
 
     def test_background_chaos_exits(self):
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo/juju')
+        client = ModelClient(JujuData('foo', {}), None, '/foo/juju')
         with patch('chaos.MonkeyRunner.deploy_chaos_monkey',
                    autospec=True):
             with patch('chaos.MonkeyRunner.unleash_once',
@@ -102,12 +102,13 @@
                 ('show-status', '--format', 'yaml'): status,
                 }
             return output[args]
-        client = EnvJujuClient(JujuData('foo', {}), '1.25.0', '/foo/juju')
+        client = ModelClient(JujuData('foo', {}), '1.25.0', '/foo/juju')
         with patch.object(client, 'get_juju_output', side_effect=output,
                           autospec=True) as gjo_mock:
             with patch('subprocess.check_call', autospec=True) as cc_mock:
                 monkey_runner = MonkeyRunner('foo', client, service='ser1')
-                with patch('jujupy.GroupReporter._write', autospec=True):
+                with patch('jujupy.client.GroupReporter._write',
+                           autospec=True):
                     monkey_runner.deploy_chaos_monkey()
         assert_juju_call(
             self,
@@ -150,7 +151,7 @@
                 ('show-status', '--format', 'yaml'): status,
                 }
             return output[args]
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo/juju')
+        client = ModelClient(JujuData('foo', {}), None, '/foo/juju')
         runner = MonkeyRunner('foo', client, service='jenkins')
         with patch.object(client, 'get_juju_output', side_effect=output,
                           autospec=True):
@@ -203,7 +204,7 @@
                 ('config', 'chaos-monkey'): charm_config,
                 }
             return output[args]
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo')
+        client = ModelClient(JujuData('foo', {}), None, '/foo')
         monkey_runner = MonkeyRunner('foo', client, service='jenkins')
         monkey_runner.monkey_ids = {
             'chaos-monkey/0': 'workspace0',
@@ -315,7 +316,7 @@
                      'abcdabcdabcdabcdabcdabcdabcdabcdabcd'),
                 }
             return output[args]
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo/juju')
+        client = ModelClient(JujuData('foo', {}), None, '/foo/juju')
         monkey_runner = MonkeyRunner('foo', client, service='jenkins')
         with patch.object(client, 'get_juju_output', side_effect=output,
                           autospec=True) as gjo_mock:
@@ -384,7 +385,7 @@
                  ): 'Action fail',
                 }
             return output[args]
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo/juju')
+        client = ModelClient(JujuData('foo', {}), None, '/foo/juju')
         monkey_runner = MonkeyRunner('foo', client, service='jenkins')
         with patch.object(client, 'get_juju_output', side_effect=output,
                           autospec=True):
@@ -397,7 +398,7 @@
 
     def test_is_healthy(self):
         SCRIPT = """#!/bin/bash\necho -n 'PASS'\nexit 0"""
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo/juju')
+        client = ModelClient(JujuData('foo', {}), None, '/foo/juju')
         with NamedTemporaryFile(delete=False) as health_script:
             health_script.write(SCRIPT)
             os.fchmod(health_script.fileno(), stat.S_IEXEC | stat.S_IREAD)
@@ -413,7 +414,7 @@
 
     def test_is_healthy_fail(self):
         SCRIPT = """#!/bin/bash\necho -n 'FAIL'\nexit 1"""
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo/juju')
+        client = ModelClient(JujuData('foo', {}), None, '/foo/juju')
         with NamedTemporaryFile(delete=False) as health_script:
             health_script.write(SCRIPT)
             os.fchmod(health_script.fileno(), stat.S_IEXEC | stat.S_IREAD)
@@ -428,7 +429,7 @@
 
     def test_is_healthy_with_no_execute_perms(self):
         SCRIPT = """#!/bin/bash\nexit 0"""
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo/juju')
+        client = ModelClient(JujuData('foo', {}), None, '/foo/juju')
         with NamedTemporaryFile(delete=False) as health_script:
             health_script.write(SCRIPT)
             os.fchmod(health_script.fileno(), stat.S_IREAD)
@@ -447,7 +448,7 @@
 class TestWaitForChaos(FakeHomeTestCase):
 
     def test_wait_for_chaos_complete(self):
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo')
+        client = ModelClient(JujuData('foo', {}), None, '/foo')
         runner = MonkeyRunner('foo', client)
         units = [('blib', 'blab')]
         with patch.object(runner, 'iter_chaos_monkey_units', autospec=True,
@@ -460,14 +461,14 @@
         self.assertEqual(us_mock.call_count, 1)
 
     def test_wait_for_chaos_complete_timesout(self):
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo')
+        client = ModelClient(JujuData('foo', {}), None, '/foo')
         runner = MonkeyRunner('foo', client)
         with self.assertRaisesRegexp(
                 Exception, 'Chaos operations did not complete.'):
             runner.wait_for_chaos(timeout=0)
 
     def test_wait_for_chaos_started(self):
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo')
+        client = ModelClient(JujuData('foo', {}), None, '/foo')
         runner = MonkeyRunner('foo', client)
         units = [('blib', 'blab')]
         with patch.object(runner, 'iter_chaos_monkey_units', autospec=True,
@@ -481,7 +482,7 @@
         self.assertEqual(us_mock.call_count, 1)
 
     def test_wait_for_chaos_unexpected_state(self):
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo')
+        client = ModelClient(JujuData('foo', {}), None, '/foo')
         runner = MonkeyRunner('foo', client)
         with self.assertRaisesRegexp(
                 Exception, 'Unexpected state value: foo'):

=== modified file 'tests/test_deploy_stack.py'
--- tests/test_deploy_stack.py	2016-12-01 16:08:23 +0000
+++ tests/test_deploy_stack.py	2017-01-20 20:20:55 +0000
@@ -44,7 +44,6 @@
     _get_clients_to_upgrade,
     iter_remote_machines,
     get_remote_machines,
-    GET_TOKEN_SCRIPT,
     make_controller_strategy,
     PublicController,
     safe_print_status,
@@ -52,23 +51,21 @@
     update_env,
     wait_for_state_server_to_shutdown,
     )
-from fakejuju import (
-    fake_juju_client,
-    fake_juju_client_optional_jes,
-    )
 from jujuconfig import (
     get_environments_path,
     get_jenv_path,
     get_juju_home,
     )
 from jujupy import (
-    EnvJujuClient,
     EnvJujuClient1X,
     EnvJujuClient25,
+    fake_juju_client,
+    fake_juju_client_optional_jes,
     get_cache_path,
     get_timeout_prefix,
     JujuData,
     KILL_CONTROLLER,
+    ModelClient,
     SimpleEnvironment,
     SoftDeadlineExceeded,
     Status,
@@ -104,6 +101,8 @@
 
 class DeployStackTestCase(FakeHomeTestCase):
 
+    log_level = logging.DEBUG
+
     def test_destroy_environment(self):
         client = EnvJujuClient1X(
             SimpleEnvironment('foo', {'type': 'local'}), '1.234-76', None)
@@ -141,7 +140,7 @@
 
     def test_assess_juju_run(self):
         env = JujuData('foo', {'type': 'nonlocal'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         response_ok = json.dumps(
             [{"MachineId": "1", "Stdout": "Linux\n"},
              {"MachineId": "2", "Stdout": "Linux\n"}])
@@ -171,7 +170,7 @@
 
     def test_safe_print_status(self):
         env = JujuData('foo', {'type': 'nonlocal'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         error = subprocess.CalledProcessError(1, 'status', 'status error')
         with patch.object(client, 'juju', autospec=True,
                           side_effect=[error]) as mock:
@@ -221,7 +220,7 @@
 
     def test_dump_juju_timings(self):
         env = JujuData('foo', {'type': 'bar'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         client._backend.juju_timings = {("juju", "op1"): [1],
                                         ("juju", "op2"): [2]}
         expected = {"juju op1": [1], "juju op2": [2]}
@@ -234,7 +233,7 @@
 
     def test_check_token(self):
         env = JujuData('foo', {'type': 'local'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         status = Status.from_text("""\
             applications:
               dummy-sink:
@@ -248,12 +247,12 @@
         remote = SSHRemote(client, 'unit', None, series='xenial')
         with patch('deploy_stack.remote_from_unit', autospec=True,
                    return_value=remote):
-            with patch.object(remote, 'run', autospec=True,
-                              return_value='token') as rr_mock:
+            with patch.object(remote, 'cat', autospec=True,
+                              return_value='token') as rc_mock:
                 with patch.object(client, 'get_status', autospec=True,
                                   return_value=status):
                     check_token(client, 'token', timeout=0)
-        rr_mock.assert_called_once_with(GET_TOKEN_SCRIPT)
+        rc_mock.assert_called_once_with('/var/run/dummy-sink/token')
         self.assertTrue(remote.use_juju_ssh)
         self.assertEqual(
             ['INFO Waiting for applications to reach ready.',
@@ -263,7 +262,7 @@
 
     def test_check_token_not_found(self):
         env = JujuData('foo', {'type': 'local'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         status = Status.from_text("""\
             applications:
               dummy-sink:
@@ -275,10 +274,11 @@
 
             """)
         remote = SSHRemote(client, 'unit', None, series='xenial')
+        error = subprocess.CalledProcessError(1, 'ssh', '')
         with patch('deploy_stack.remote_from_unit', autospec=True,
                    return_value=remote):
-            with patch.object(remote, 'run', autospec=True,
-                              return_value='') as rr_mock:
+            with patch.object(remote, 'cat', autospec=True,
+                              side_effect=error) as rc_mock:
                 with patch.object(remote, 'get_address',
                                   autospec=True) as ga_mock:
                     with patch.object(client, 'get_status', autospec=True,
@@ -286,8 +286,7 @@
                         with self.assertRaisesRegexp(ValueError,
                                                      "Token is ''"):
                             check_token(client, 'token', timeout=0)
-        self.assertEqual(2, rr_mock.call_count)
-        rr_mock.assert_called_with(GET_TOKEN_SCRIPT)
+        self.assertEqual(2, rc_mock.call_count)
         ga_mock.assert_called_once_with()
         self.assertFalse(remote.use_juju_ssh)
         self.assertEqual(
@@ -297,7 +296,7 @@
 
     def test_check_token_not_found_juju_ssh_broken(self):
         env = JujuData('foo', {'type': 'local'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         status = Status.from_text("""\
             applications:
               dummy-sink:
@@ -309,10 +308,11 @@
 
             """)
         remote = SSHRemote(client, 'unit', None, series='xenial')
+        error = subprocess.CalledProcessError(1, 'ssh', '')
         with patch('deploy_stack.remote_from_unit', autospec=True,
                    return_value=remote):
-            with patch.object(remote, 'run', autospec=True,
-                              side_effect=['', 'token']) as rr_mock:
+            with patch.object(remote, 'cat', autospec=True,
+                              side_effect=[error, 'token']) as rc_mock:
                 with patch.object(remote, 'get_address',
                                   autospec=True) as ga_mock:
                     with patch.object(client, 'get_status', autospec=True,
@@ -320,8 +320,8 @@
                         with self.assertRaisesRegexp(ValueError,
                                                      "Token is 'token'"):
                             check_token(client, 'token', timeout=0)
-        self.assertEqual(2, rr_mock.call_count)
-        rr_mock.assert_called_with(GET_TOKEN_SCRIPT)
+        self.assertEqual(2, rc_mock.call_count)
+        rc_mock.assert_called_with('/var/run/dummy-sink/token')
         ga_mock.assert_called_once_with()
         self.assertFalse(remote.use_juju_ssh)
         self.assertEqual(
@@ -333,7 +333,7 @@
 
     def test_check_token_win_status(self):
         env = JujuData('foo', {'type': 'azure'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         remote = MagicMock(spec=['cat', 'is_windows'])
         remote.is_windows.return_value = True
         status = Status.from_text("""\
@@ -361,7 +361,7 @@
 
     def test_check_token_win_remote(self):
         env = JujuData('foo', {'type': 'azure'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         remote = MagicMock(spec=['cat', 'is_windows'])
         remote.is_windows.return_value = True
         remote.cat.return_value = 'token'
@@ -388,11 +388,11 @@
 
     def test_check_token_win_remote_failure(self):
         env = JujuData('foo', {'type': 'azure'})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         remote = MagicMock(spec=['cat', 'is_windows'])
         remote.is_windows.return_value = True
-        remote.cat.side_effect = winrm.exceptions.WinRMTransportError(
-            'a', 'oops')
+        error = winrm.exceptions.WinRMTransportError('a', 'oops')
+        remote.cat.side_effect = error
         status = Status.from_text("""\
             applications:
               dummy-sink:
@@ -405,18 +405,15 @@
                    return_value=remote):
             with patch.object(client, 'get_status', autospec=True,
                               return_value=status):
-                check_token(client, 'token', timeout=0)
-        # application-status did not have the token, winrm did.
+                with self.assertRaises(type(error)) as ctx:
+                    check_token(client, 'token', timeout=0)
+        self.assertIs(ctx.exception, error)
         remote.cat.assert_called_once_with('%ProgramData%\\dummy-sink\\token')
         self.assertEqual(
             ['INFO Waiting for applications to reach ready.',
-             'INFO Retrieving token.',
-             'WARNING Skipping token check because of: '
-                '500 WinRMTransport. oops'],
+             'INFO Retrieving token.'],
             self.log_stream.getvalue().splitlines())
 
-    log_level = logging.DEBUG
-
 
 class DumpEnvLogsTestCase(FakeHomeTestCase):
 
@@ -443,7 +440,7 @@
                         with patch('deploy_stack.archive_logs',
                                    autospec=True) as al_mock:
                             env = JujuData('foo', {'type': 'nonlocal'})
-                            client = EnvJujuClient(env, '1.234-76', None)
+                            client = ModelClient(env, '1.234-76', None)
                             dump_env_logs(client, '10.10.0.1', artifacts_dir)
             al_mock.assert_called_once_with(artifacts_dir)
             self.assertEqual(
@@ -472,7 +469,7 @@
                         with patch('deploy_stack.archive_logs',
                                    autospec=True) as al_mock:
                             env = JujuData('foo', {'type': 'nonlocal'})
-                            client = EnvJujuClient(env, '1.234-76', None)
+                            client = ModelClient(env, '1.234-76', None)
                             dump_env_logs(client, '10.10.0.1', artifacts_dir)
             al_mock.assert_called_once_with(artifacts_dir)
             self.assertEqual(
@@ -490,7 +487,7 @@
 
     def test_dump_env_logs_local_env(self):
         env = JujuData('foo', {'type': 'local'})
-        client = EnvJujuClient(env, '1.234-76', None)
+        client = ModelClient(env, '1.234-76', None)
         with temp_dir() as artifacts_dir:
             with patch('deploy_stack.get_remote_machines',
                        autospec=True) as grm_mock:
@@ -708,7 +705,7 @@
             self.log_stream.getvalue().splitlines())
 
     def test_get_machines_for_logs(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('cloud', {'type': 'ec2'}), '1.23.4', None)
         status = Status.from_text("""\
             machines:
@@ -724,7 +721,7 @@
             {'0': '10.11.12.13', '1': '10.11.12.14'}, machines)
 
     def test_get_machines_for_logs_with_boostrap_host(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('cloud', {'type': 'ec2'}), '1.23.4', None)
         status = Status.from_text("""\
             machines:
@@ -737,7 +734,7 @@
         self.assert_machines({'0': '10.11.111.222'}, machines)
 
     def test_get_machines_for_logs_with_no_addresses(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('cloud', {'type': 'ec2'}), '1.23.4', None)
         with patch.object(client, 'get_status', autospec=True,
                           side_effect=Exception):
@@ -760,7 +757,7 @@
         juju_data.credentials = {'credentials': {cloud_name: {'credentials': {
             'maas-oauth': 'baz',
             }}}}
-        client = EnvJujuClient(juju_data, '1.23.4', None)
+        client = ModelClient(juju_data, '1.23.4', None)
         status = Status.from_text("""\
             machines:
               "0":
@@ -781,7 +778,7 @@
             {'0': '10.11.12.13', '1': '10.11.12.14'}, machines)
 
     def test_iter_remote_machines(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('cloud', {'type': 'ec2'}), '1.23.4', None)
         status = Status.from_text("""\
             machines:
@@ -798,7 +795,7 @@
             [('0', '10.11.12.13'), ('1', '10.11.12.14')], machines)
 
     def test_iter_remote_machines_with_series(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('cloud', {'type': 'ec2'}), '1.23.4', None)
         status = Status.from_text("""\
             machines:
@@ -838,9 +835,9 @@
 
     def test_deploy_dummy_stack_sets_centos_constraints(self):
         env = JujuData('foo', {'type': 'maas'})
-        client = EnvJujuClient(env, '2.0.0', '/foo/juju')
+        client = ModelClient(env, '2.0.0', '/foo/juju')
         with patch('subprocess.check_call', autospec=True) as cc_mock:
-            with patch.object(EnvJujuClient, 'wait_for_started'):
+            with patch.object(ModelClient, 'wait_for_started'):
                 with patch('deploy_stack.get_random_string',
                            return_value='fake-token', autospec=True):
                     deploy_dummy_stack(client, 'centos')
@@ -850,7 +847,7 @@
 
     def test_assess_juju_relations(self):
         env = JujuData('foo', {'type': 'nonlocal'})
-        client = EnvJujuClient(env, None, '/foo/juju')
+        client = ModelClient(env, None, '/foo/juju')
         with patch.object(client, 'get_juju_output', side_effect='fake-token',
                           autospec=True):
             with patch('subprocess.check_call', autospec=True) as cc_mock:
@@ -886,9 +883,19 @@
             call('/tmp/repo/charms-win/dummy-sink', series='win2012hvr2')]
         self.assertEqual(dp_mock.mock_calls, calls)
 
+    def test_deploy_dummy_stack_charmstore(self):
+        client = fake_juju_client()
+        client.bootstrap()
+        with patch.object(client, 'deploy', autospec=True) as dp_mock:
+            deploy_dummy_stack(client, 'xenial', use_charmstore=True)
+        calls = [
+            call('cs:~juju-qa/dummy-source', series='xenial'),
+            call('cs:~juju-qa/dummy-sink', series='xenial')]
+        self.assertEqual(dp_mock.mock_calls, calls)
+
     def test_deploy_dummy_stack(self):
         env = JujuData('foo', {'type': 'nonlocal'})
-        client = EnvJujuClient(env, '2.0.0', '/foo/juju')
+        client = ModelClient(env, '2.0.0', '/foo/juju')
         status = yaml.safe_dump({
             'machines': {'0': {'agent-state': 'started'}},
             'services': {
@@ -899,9 +906,10 @@
         })
 
         def output(*args, **kwargs):
+            token_file = '/var/run/dummy-sink/token'
             output = {
                 ('show-status', '--format', 'yaml'): status,
-                ('ssh', 'dummy-sink/0', GET_TOKEN_SCRIPT): 'fake-token',
+                ('ssh', 'dummy-sink/0', 'cat', token_file): 'fake-token',
             }
             return output[args]
 
@@ -952,8 +960,8 @@
     return SimpleEnvironment(name, {})
 
 
-def fake_EnvJujuClient(env, path=None, debug=None):
-    return EnvJujuClient(env=env, version='1.2.3.4', full_path=path)
+def fake_ModelClient(env, path=None, debug=None):
+    return ModelClient(env=env, version='1.2.3.4', full_path=path)
 
 
 class FakeBootstrapManager:
@@ -1026,7 +1034,7 @@
     @contextmanager
     def ds_cxt(self):
         env = JujuData('foo', {})
-        client = fake_EnvJujuClient(env)
+        client = fake_ModelClient(env)
         bc_cxt = patch('deploy_stack.client_from_config',
                        return_value=client)
         fc_cxt = patch('jujupy.SimpleEnvironment.from_config',
@@ -1034,7 +1042,7 @@
         mgr = MagicMock()
         bm_cxt = patch('deploy_stack.BootstrapManager', autospec=True,
                        return_value=mgr)
-        juju_cxt = patch('jujupy.EnvJujuClient.juju', autospec=True)
+        juju_cxt = patch('jujupy.ModelClient.juju', autospec=True)
         ajr_cxt = patch('deploy_stack.assess_juju_run', autospec=True)
         dds_cxt = patch('deploy_stack.deploy_dummy_stack', autospec=True)
         with bc_cxt, fc_cxt, bm_cxt as bm_mock, juju_cxt, ajr_cxt, dds_cxt:
@@ -1049,7 +1057,7 @@
             series='trusty', debug=False, agent_url=None, agent_stream=None,
             keep_env=False, upload_tools=False, with_chaos=1, jes=False,
             region=None, verbose=False, upgrade=False, deadline=None,
-            controller_host=None,
+            controller_host=None, use_charmstore=False,
         )
         with self.ds_cxt():
             with patch('deploy_stack.background_chaos',
@@ -1073,7 +1081,7 @@
             series='trusty', debug=False, agent_url=None, agent_stream=None,
             keep_env=False, upload_tools=False, with_chaos=0, jes=False,
             region=None, verbose=False, upgrade=False, deadline=None,
-            controller_host=None,
+            controller_host=None, use_charmstore=False,
         )
         with self.ds_cxt():
             with patch('deploy_stack.background_chaos',
@@ -1092,7 +1100,7 @@
             series='trusty', debug=False, agent_url=None, agent_stream=None,
             keep_env=False, upload_tools=False, with_chaos=0, jes=False,
             region='region-foo', verbose=False, upgrade=False, deadline=None,
-            controller_host=None,
+            controller_host=None, use_charmstore=False,
         )
         with self.ds_cxt() as (client, bm_mock):
             with patch('deploy_stack.assess_juju_relations',
@@ -1115,7 +1123,7 @@
             charm_prefix=None, bootstrap_host=None, machine=None, logs=None,
             debug=None, juju_bin=None, agent_url=None, agent_stream=None,
             keep_env=None, upload_tools=None, with_chaos=None, jes=None,
-            region=None, verbose=None)
+            region=None, verbose=None, use_charmstore=False)
         with patch('deploy_stack.deploy_job_parse_args', return_value=args,
                    autospec=True):
             with patch('deploy_stack._deploy_job', autospec=True) as ds_mock:
@@ -1128,7 +1136,7 @@
             charm_prefix=None, bootstrap_host=None, machine=None, logs=None,
             debug=None, juju_bin=None, agent_url=None, agent_stream=None,
             keep_env=None, upload_tools=None, with_chaos=None, jes=None,
-            region=None, verbose=None)
+            region=None, verbose=None, use_charmstore=False)
         with patch('deploy_stack.deploy_job_parse_args', return_value=args,
                    autospec=True):
             with patch('deploy_stack._deploy_job', autospec=True) as ds_mock:
@@ -1188,20 +1196,20 @@
                 with patch('deploy_stack.check_token', autospec=True):
                     with patch('deploy_stack.get_random_string',
                                return_value="FAKETOKEN", autospec=True):
-                        with patch('jujupy.EnvJujuClient.get_version',
+                        with patch('jujupy.ModelClient.get_version',
                                    side_effect=lambda cls:
                                    '2.0-rc2-arch-series'):
                             with patch(
-                                    'jujupy.get_timeout_prefix',
+                                    'jujupy.client.get_timeout_prefix',
                                     autospec=True, return_value=()):
                                 yield (co_mock, cc_mock)
 
     def test_assess_upgrade(self):
         env = JujuData('foo', {'type': 'foo'})
-        old_client = EnvJujuClient(env, None, '/foo/juju')
+        old_client = ModelClient(env, None, '/foo/juju')
         with self.upgrade_mocks() as (co_mock, cc_mock):
             assess_upgrade(old_client, '/bar/juju')
-        new_client = EnvJujuClient(env, None, '/bar/juju')
+        new_client = ModelClient(env, None, '/bar/juju')
         # Needs to upgrade the controller first.
         assert_juju_call(self, cc_mock, new_client, (
             'juju', '--show-log', 'upgrade-juju', '-m', 'foo:controller',
@@ -1228,7 +1236,7 @@
         env = SimpleEnvironment('foo', {'type': 'foo'})
         old_client = fake_juju_client(
             env, '/foo/juju', version='1.25', cls=EnvJujuClient25)
-        with patch('jujupy.EnvJujuClient.get_version',
+        with patch('jujupy.ModelClient.get_version',
                    return_value='1.25-arch-series'):
             with patch('jujupy.EnvJujuClient25._get_models', return_value=[]):
                 [new_client] = _get_clients_to_upgrade(
@@ -1240,7 +1248,7 @@
         old_client = fake_juju_client()
         old_client.bootstrap()
 
-        with patch('jujupy.EnvJujuClient.get_version',
+        with patch('jujupy.ModelClient.get_version',
                    return_value='2.0-rc2-arch-series'):
             new_clients = _get_clients_to_upgrade(
                 old_client, '/foo/newer/juju')
@@ -1251,13 +1259,13 @@
 
     def test_mass_timeout(self):
         config = {'type': 'foo'}
-        old_client = EnvJujuClient(JujuData('foo', config), None, '/foo/juju')
+        old_client = ModelClient(JujuData('foo', config), None, '/foo/juju')
         with self.upgrade_mocks():
-            with patch.object(EnvJujuClient, 'wait_for_version') as wfv_mock:
+            with patch.object(ModelClient, 'wait_for_version') as wfv_mock:
                 assess_upgrade(old_client, '/bar/juju')
             wfv_mock.assert_has_calls([call('2.0-rc2', 600)] * 2)
             config['type'] = 'maas'
-            with patch.object(EnvJujuClient, 'wait_for_version') as wfv_mock:
+            with patch.object(ModelClient, 'wait_for_version') as wfv_mock:
                 assess_upgrade(old_client, '/bar/juju')
         wfv_mock.assert_has_calls([call('2.0-rc2', 1200)] * 2)
 
@@ -1555,7 +1563,7 @@
             self.assertNotEqual(initial_home, client.env.juju_home)
 
         ije_cxt = patch.object(client, 'is_jes_enabled')
-        with patch('jujupy.EnvJujuClient.kill_controller',
+        with patch('jujupy.ModelClient.kill_controller',
                    side_effect=check_config) as kill_mock, ije_cxt:
             with bs_manager.bootstrap_context([]):
                 pass
@@ -1694,10 +1702,20 @@
                 self.assertIsTrue(bs_manager.has_controller)
         self.assertIsTrue(bs_manager.has_controller)
 
+    def test_existing_bootstrap_context_sets_has_controller(self):
+        client = self.make_client()
+        bs_manager = BootstrapManager(
+            'foobar', client, client, None, [], None, None, None, None,
+            None, False, False, False)
+        with patch.object(client, 'kill_controller'):
+            with bs_manager.existing_bootstrap_context([]):
+                self.assertIsTrue(bs_manager.has_controller)
+        self.assertIsTrue(bs_manager.has_controller)
+
     def test_handle_bootstrap_exceptions_ignores_soft_deadline(self):
         env = JujuData('foo', {'type': 'nonlocal'})
-        client = EnvJujuClient(env, None, None)
-        tear_down_client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
+        tear_down_client = ModelClient(env, None, None)
         soft_deadline = datetime(2015, 1, 2, 3, 4, 6)
         now = soft_deadline + timedelta(seconds=1)
         client.env.juju_home = use_context(self, temp_dir())
@@ -1973,23 +1991,36 @@
                 with bs_manager.runtime_context([]):
                     raise ValueError
 
+    @contextmanager
+    def logged_exception_bs_manager(self):
+        client = fake_juju_client()
+        with temp_dir() as root:
+            log_dir = os.path.join(root, 'log-dir')
+            os.mkdir(log_dir)
+            bs_manager = BootstrapManager(
+                'foobar', client, client,
+                None, [], None, None, None, None, log_dir, False,
+                True, True)
+            juju_home = os.path.join(root, 'juju-home')
+            os.mkdir(juju_home)
+            client.env.juju_home = juju_home
+            yield bs_manager
+
     def test_booted_context_handles_logged_exception(self):
-        client = fake_juju_client()
-        with temp_dir() as root:
-            log_dir = os.path.join(root, 'log-dir')
-            os.mkdir(log_dir)
-            bs_manager = BootstrapManager(
-                'foobar', client, client,
-                None, [], None, None, None, None, log_dir, False,
-                True, True)
-            juju_home = os.path.join(root, 'juju-home')
-            os.mkdir(juju_home)
-            client.env.juju_home = juju_home
+        with self.logged_exception_bs_manager() as bs_manager:
             with self.assertRaises(SystemExit):
                 with patch.object(bs_manager, 'dump_all_logs'):
                     with bs_manager.booted_context(False):
                         raise LoggedException()
 
+    def test_booted_context_raises_logged_exception(self):
+        with self.logged_exception_bs_manager() as bs_manager:
+            bs_manager.logged_exception_exit = False
+            with self.assertRaises(LoggedException):
+                with patch.object(bs_manager, 'dump_all_logs'):
+                    with bs_manager.booted_context(False):
+                        raise LoggedException()
+
     def test_booted_context_omits_supported(self):
         client = fake_juju_client()
         client.env.juju_home = use_context(self, temp_dir())
@@ -2053,8 +2084,8 @@
         env = JujuData('foo', {'type': 'nonlocal'})
         soft_deadline = datetime(2015, 1, 2, 3, 4, 6)
         now = soft_deadline + timedelta(seconds=1)
-        client = EnvJujuClient(env, None, None)
-        tear_down_client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
+        tear_down_client = ModelClient(env, None, None)
 
         def do_check_client(*args, **kwargs):
             with client.check_timeouts():
@@ -2156,7 +2187,7 @@
 
     def test_bootstrap_context(self):
         cc_mock = self.addContext(patch('subprocess.check_call'))
-        client = EnvJujuClient(JujuData(
+        client = ModelClient(JujuData(
             'foo', {'type': 'paas', 'region': 'qux'}), '1.23', 'path')
         with self.bc_context(client, 'log_dir', jes='kill-controller'):
             with observable_temp_file() as config_file:
@@ -2196,7 +2227,7 @@
 
     def test_keep_env(self):
         cc_mock = self.addContext(patch('subprocess.check_call'))
-        client = EnvJujuClient(JujuData(
+        client = ModelClient(JujuData(
             'foo', {'type': 'paas', 'region': 'qux'}), '1.23', 'path')
         with self.bc_context(client, keep_env=True, jes='kill-controller'):
             with observable_temp_file() as config_file:
@@ -2235,7 +2266,7 @@
 
     def test_upload_tools(self):
         cc_mock = self.addContext(patch('subprocess.check_call'))
-        client = EnvJujuClient(JujuData(
+        client = ModelClient(JujuData(
             'foo', {'type': 'paas', 'region': 'qux'}), '1.23', 'path')
         with self.bc_context(client, jes='kill-controller'):
             with observable_temp_file() as config_file:
@@ -2261,7 +2292,7 @@
 
     def test_calls_update_env_2(self):
         cc_mock = self.addContext(patch('subprocess.check_call'))
-        client = EnvJujuClient(JujuData(
+        client = ModelClient(JujuData(
             'foo', {'type': 'paas', 'region': 'qux'}), '1.23', 'path')
         ue_mock = self.addContext(
             patch('deploy_stack.update_env', wraps=update_env))
@@ -2319,7 +2350,7 @@
         class FakeException(Exception):
             """A sentry exception to be raised by bootstrap."""
 
-        client = EnvJujuClient(JujuData(
+        client = ModelClient(JujuData(
             'foo', {'type': 'paas'}), '1.23', 'path')
         self.addContext(patch('deploy_stack.get_machine_dns_name',
                               return_value='foo'))
@@ -2327,7 +2358,7 @@
         tear_down_mock = self.addContext(
             patch('deploy_stack.BootstrapManager.tear_down', autospec=True))
         kill_mock = self.addContext(
-            patch('jujupy.EnvJujuClient.kill_controller', autospec=True))
+            patch('jujupy.ModelClient.kill_controller', autospec=True))
         po_mock = self.addContext(patch(
             'subprocess.Popen', autospec=True,
             return_value=FakePopen('kill-controller', '', 0)))
@@ -2393,7 +2424,7 @@
 
     def test_jes(self):
         self.addContext(patch('subprocess.check_call', autospec=True))
-        client = EnvJujuClient(JujuData(
+        client = ModelClient(JujuData(
             'foo', {'type': 'paas', 'region': 'qux'}), '1.26', 'path')
         with self.bc_context(client, 'log_dir', jes=KILL_CONTROLLER):
             with boot_context('bar', client, None, [], None, None, None,
@@ -2402,7 +2433,7 @@
 
     def test_region(self):
         self.addContext(patch('subprocess.check_call', autospec=True))
-        client = EnvJujuClient(JujuData(
+        client = ModelClient(JujuData(
             'foo', {'type': 'paas'}), '1.23', 'path')
         with self.bc_context(client, 'log_dir', jes='kill-controller'):
             with boot_context('bar', client, None, [], None, None, None,
@@ -2428,7 +2459,7 @@
         effects = [None, None, None, None, None, None, error]
         cc_mock = self.addContext(patch('subprocess.check_call', autospec=True,
                                         side_effect=effects))
-        client = EnvJujuClient(JujuData(
+        client = ModelClient(JujuData(
             'foo', {'type': 'paas', 'region': 'qux'}), '1.23', 'path')
         with self.bc_context(client, 'log_dir', jes='kill-controller'):
             with observable_temp_file() as config_file:
@@ -2478,6 +2509,7 @@
             region=None,
             deadline=None,
             controller_host=None,
+            use_charmstore=False,
         ))
 
     def test_upload_tools(self):
@@ -2495,6 +2527,11 @@
             ['foo', 'bar/juju', 'baz', 'qux', '--jes'])
         self.assertIs(args.jes, True)
 
+    def test_use_charmstore(self):
+        args = deploy_job_parse_args(
+            ['foo', 'bar/juju', 'baz', 'qux', '--use-charmstore'])
+        self.assertIs(args.use_charmstore, True)
+
 
 class TestWaitForStateServerToShutdown(FakeHomeTestCase):
 

=== modified file 'tests/test_generate_perfscale_results.py'
--- tests/test_generate_perfscale_results.py	2016-10-27 05:30:29 +0000
+++ tests/test_generate_perfscale_results.py	2017-01-25 01:12:56 +0000
@@ -12,7 +12,7 @@
 from textwrap import dedent
 import json
 
-from fakejuju import fake_juju_client
+from jujupy import fake_juju_client
 import generate_perfscale_results as gpr
 import perf_graphing
 from test_quickstart_deploy import make_bootstrap_manager
@@ -181,9 +181,11 @@
         static_config_path = '/baz/bang/config'
 
         expected_calls = [
-            call('scp', (static_config_path, '2:/tmp/collectd.config')),
-            call('scp', (static_setup_path, '2:/tmp/installer.sh')),
-            call('ssh', ('2', 'chmod +x /tmp/installer.sh')),
+            call(
+                'scp',
+                ('--proxy', static_config_path, '2:/tmp/collectd.config')),
+            call('scp', ('--proxy', static_setup_path, '2:/tmp/installer.sh')),
+            call('ssh', ('--proxy', '2', 'chmod +x /tmp/installer.sh')),
         ]
 
         with patch.object(gpr, 'PATHS', autospec=True) as m_sp:
@@ -203,10 +205,12 @@
         expected_calls = [
             call(
                 'ssh',
-                ('3',
+                ('--proxy',
+                 '3',
                  '/tmp/installer.sh /tmp/collectd.config /tmp/runner.sh')),
             call(
-                'ssh', ('3', '--', 'daemon --respawn /tmp/runner.sh'))]
+                'ssh',
+                ('--proxy', '3', '--', 'daemon --respawn /tmp/runner.sh'))]
         self.assertListEqual(
             admin_client.juju.call_args_list,
             expected_calls
@@ -229,11 +233,11 @@
         expected_calls = [
             call(
                 'scp',
-                ('--', '-r',
+                ('--proxy', '--', '-r',
                  '0:/var/lib/collectd/rrd/localhost/*', expected_dir)),
             call(
                 'scp',
-                ('0:/tmp/mongodb-stats.log', expected_dir)
+                ('--proxy', '0:/tmp/mongodb-stats.log', expected_dir)
                 )]
         self.assertEqual(
             client.juju.call_args_list,
@@ -263,13 +267,14 @@
             expected_calls.append(
                 call(
                     'scp',
-                    ('--', '-r',
+                    ('--proxy', '--', '-r',
                      '{}:/var/lib/collectd/rrd/localhost/*'.format(m_id),
                      expected_dir)))
             expected_calls.append(
                 call(
                     'scp',
-                    ('{}:/tmp/mongodb-stats.log'.format(m_id), expected_dir)))
+                    ('--proxy',
+                     '{}:/tmp/mongodb-stats.log'.format(m_id), expected_dir)))
 
         self.assertEqual(
             client.juju.call_args_list,

=== modified file 'tests/test_gotestwin.py'
--- tests/test_gotestwin.py	2016-10-15 18:31:14 +0000
+++ tests/test_gotestwin.py	2017-01-25 02:32:29 +0000
@@ -12,7 +12,8 @@
 
 
 S3_CI_PATH = os.path.join(gotestwin.SCRIPTS, 's3ci.py')
-JUJU_HOME = os.path.join(gotestwin.SCRIPTS, '..', 'cloud-city')
+JUJU_HOME = os.path.normpath(os.path.join(
+    gotestwin.SCRIPTS, '..', 'cloud-city'))
 
 
 @contextlib.contextmanager
@@ -42,11 +43,12 @@
             data['command'])
         co_mock.assert_called_once_with(
             [S3_CI_PATH, 'get', '1234', 'build-revision',
-             'juju_core.*.tar.gz', './'])
+             '.*.tar.gz', './'])
         tarfile_call = call(
             [S3_CI_PATH, 'get-summary', '1234', 'GoTestWin'])
         gotest_call = call(
-            ['workspace-run', '-v', '-i', 'cloud-city/staging-juju-rsa',
+            ['workspace-run', '-v', '-i',
+             '{}/staging-juju-rsa'.format(JUJU_HOME),
              'temp-config.yaml', 'Administrator@host'])
         self.assertEqual([tarfile_call, gotest_call], cc_mock.call_args_list)
 
@@ -64,5 +66,6 @@
              'github.com/juju/juju/cmd', '--remove', 'ci/bar.tar.gz'],
             data['command'])
         cc_mock.assert_called_once_with(
-            ['workspace-run', '-v', '-i', 'cloud-city/staging-juju-rsa',
-             'temp-config.yaml', 'Administrator@host'])
+            ['workspace-run', '-v', '-i',
+             '{}/staging-juju-rsa'.format(JUJU_HOME), 'temp-config.yaml',
+             'Administrator@host'])

=== modified file 'tests/test_industrial_test.py'
--- tests/test_industrial_test.py	2016-11-25 15:10:43 +0000
+++ tests/test_industrial_test.py	2017-01-20 20:58:04 +0000
@@ -21,7 +21,7 @@
     )
 import yaml
 
-from fakejuju import (
+from jujupy import (
     fake_juju_client,
     )
 from industrial_test import (
@@ -51,7 +51,7 @@
     )
 from jujuconfig import get_euca_env
 from jujupy import (
-    EnvJujuClient,
+    ModelClient,
     EnvJujuClient1X,
     get_timeout_prefix,
     JujuData,
@@ -101,15 +101,15 @@
 
     def setUp(self):
         super(JujuPyTestCase, self).setUp()
-        patcher = patch('jujupy.pause')
-        self.addCleanup(patcher.stop)
-        patcher.start()
-
-        patcher = patch('jujupy.GroupReporter._write')
-        self.addCleanup(patcher.stop)
-        patcher.start()
-
-        patcher = patch('jujupy.until_timeout', return_value=[2, 1])
+        patcher = patch('jujupy.client.pause')
+        self.addCleanup(patcher.stop)
+        patcher.start()
+
+        patcher = patch('jujupy.client.GroupReporter._write')
+        self.addCleanup(patcher.stop)
+        patcher.start()
+
+        patcher = patch('jujupy.client.until_timeout', return_value=[2, 1])
         self.addCleanup(patcher.stop)
         patcher.start()
 
@@ -121,7 +121,7 @@
 
     :param test: A unittest.TestCase
     :param stage: The SteppedStageAttempt to test.
-    :param client: The EnvJujuClient to use for iter_steps.
+    :param client: The ModelClient to use for iter_steps.
     """
     step_iter = stage.iter_steps(client)
     test_ids = stage.get_test_info().keys()
@@ -139,7 +139,7 @@
 
 def patch_status(client, *statuses):
     """
-    Replace calls to EnvJujuClient.get_status with preformed output.
+    Replace calls to ModelClient.get_status with preformed output.
 
     If client is None, all clients will have the given status for the duration
     of the patch, otherwise only the given client is modified.
@@ -157,7 +157,7 @@
         kwargs['return_value'] = Status.from_text(json.dumps(statuses[0]))
     if client is not None:
         return patch.object(client, 'get_status', autospec=True, **kwargs)
-    return patch('jujupy.EnvJujuClient.get_status', autospec=True, **kwargs)
+    return patch('jujupy.ModelClient.get_status', autospec=True, **kwargs)
 
 
 class TestParseArgs(TestCase):
@@ -528,7 +528,7 @@
                    side_effect=by_version):
             with patch('jujupy.SimpleEnvironment.from_config',
                        side_effect=lambda x: SimpleEnvironment(x, {})):
-                with patch.object(EnvJujuClient, 'get_full_path',
+                with patch.object(ModelClient, 'get_full_path',
                                   side_effect=lambda: 'juju'):
                     yield
 
@@ -883,8 +883,8 @@
                 self.assertEqual(industrial.new_client.debug, True)
 
     def test_run_stages(self):
-        old_client = FakeEnvJujuClient('old')
-        new_client = FakeEnvJujuClient('new')
+        old_client = FakeModelClient('old')
+        new_client = FakeModelClient('new')
         industrial = IndustrialTest(old_client, new_client, [
             FakeStepAttempt.from_result(True, True),
             FakeStepAttempt.from_result(True, True)])
@@ -970,8 +970,8 @@
             ('foo', True, False), ('bar', True, True), ('foo-id', True, True)])
 
     def test_run_stages_failure_in_last_step(self):
-        old_client = FakeEnvJujuClient('old')
-        new_client = FakeEnvJujuClient('new')
+        old_client = FakeModelClient('old')
+        new_client = FakeModelClient('new')
         fsa = FakeStepAttempt([('foo', True, True), ('bar', False, True)])
         industrial = IndustrialTest(old_client, new_client, [
             fsa, FakeStepAttempt.from_result(True, True)])
@@ -981,8 +981,8 @@
                     ('foo', True, True), ('bar', False, True)])
 
     def test_run_stages_raises_cannot_upgrade_to_old_client(self):
-        old = FakeEnvJujuClient()
-        new = FakeEnvJujuClient()
+        old = FakeModelClient()
+        new = FakeModelClient()
         industrial = IndustrialTest(old, new, [PrepareUpgradeJujuAttempt({})])
         with self.assertRaises(CannotUpgradeToOldClient):
             list(industrial.run_stages())
@@ -1136,8 +1136,8 @@
             [('foo-id', True, False), ('bar-id', False, False)])
 
     def test_iter_test_results(self):
-        old = FakeEnvJujuClient()
-        new = FakeEnvJujuClient()
+        old = FakeModelClient()
+        new = FakeModelClient()
         error = ValueError('asdf')
 
         class StubSA(SteppedStageAttempt):
@@ -1187,10 +1187,10 @@
             ('bar-id', {'title': 'Bar title', 'report_on': False})]))
 
 
-def FakeEnvJujuClient(name='steve', version='1.2', full_path='/jbin/juju'):
+def FakeModelClient(name='steve', version='1.2', full_path='/jbin/juju'):
     juju_data = JujuData(name, {'type': 'fake', 'region': 'regionx'})
     juju_data.credentials = {'credentials': {'fake': {'creds': {}}}}
-    return EnvJujuClient(
+    return ModelClient(
         juju_data, version, full_path)
 
 
@@ -1204,7 +1204,7 @@
 class TestBootstrapAttempt(JujuPyTestCase):
 
     def test_iter_steps(self):
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         bootstrap = BootstrapAttempt()
         boot_iter = iter_steps_validate_info(self, bootstrap, client)
         self.assertEqual(boot_iter.next(), {'test_id': 'bootstrap'})
@@ -1232,7 +1232,7 @@
 class TestDestroyEnvironmentAttempt(JujuPyTestCase):
 
     def test_iter_steps(self):
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         destroy_env = DestroyEnvironmentAttempt()
         iterator = iter_steps_validate_info(self, destroy_env, client)
         self.assertEqual({'test_id': 'destroy-env'}, iterator.next())
@@ -1271,7 +1271,7 @@
         csg_mock.assert_called_once_with(client, gsg_mock.return_value)
 
     def test_iter_test_results(self):
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         destroy_env = DestroyEnvironmentAttempt()
         with patch('subprocess.call'):
             with patch.object(client, 'get_jes_command',
@@ -1281,7 +1281,7 @@
             ('destroy-env', True, True), ('substrate-clean', True, True)])
 
     def test_iter_steps_failure(self):
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         destroy_env = DestroyEnvironmentAttempt()
         iterator = iter_steps_validate_info(self, destroy_env, client)
         self.assertEqual({'test_id': 'destroy-env'}, iterator.next())
@@ -1331,13 +1331,13 @@
 
     @staticmethod
     def get_aws_client():
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         client.env = get_aws_env()
         return client
 
     @staticmethod
     def get_openstack_client():
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         client.env.clouds = {'clouds': {'quxxx': {
             'type': 'openstack', 'endpoint': 'qux',
             }}}
@@ -1400,7 +1400,7 @@
         os_client.servers.list.assert_called_once_with()
 
     def test_get_security_groups_non_aws(self):
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         destroy_env = DestroyEnvironmentAttempt()
         self.assertIs(destroy_env.get_security_groups(client), None)
 
@@ -1450,7 +1450,7 @@
             aws_secret_access_key='secret-skeleton-key')
 
     def test_check_security_groups_non_aws(self):
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         destroy_env = DestroyEnvironmentAttempt()
         with patch('subprocess.check_output') as co_mock:
                 destroy_env.check_security_groups(
@@ -1461,7 +1461,7 @@
 class TestEnsureAvailabilityAttempt(JujuPyTestCase):
 
     def test_iter_steps(self):
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         controller_client = client.get_controller_client()
         ensure_av = EnsureAvailabilityAttempt()
         ensure_iter = iter_steps_validate_info(self, ensure_av, client)
@@ -1490,7 +1490,7 @@
         gs_mock.assert_called_once_with()
 
     def test_iter_steps_failure(self):
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         ensure_av = EnsureAvailabilityAttempt()
         ensure_iter = iter_steps_validate_info(self, ensure_av, client)
         ensure_iter.next()
@@ -1532,14 +1532,14 @@
     def test_iter_steps(self):
         machine_started = {'juju-status': {'current': 'idle'}}
         unit_started = {'agent-status': {'current': 'idle'}}
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         client.env.update_config({'default-series': 'angsty'})
         self.do_iter_steps(client, LXD_MACHINE, machine_started, unit_started)
 
     def test_iter_steps_1x(self):
         started_state = {'agent-state': 'started'}
-        client = FakeEnvJujuClient()
-        with patch.object(EnvJujuClient, 'supported_container_types',
+        client = FakeModelClient()
+        with patch.object(ModelClient, 'supported_container_types',
                           frozenset([KVM_MACHINE, LXC_MACHINE])):
             client.env.update_config({'default-series': 'angsty'})
             self.do_iter_steps(client, LXC_MACHINE, started_state,
@@ -1650,7 +1650,7 @@
 
     def test_iter_step_failure(self):
         deploy_many = DeployManyAttempt()
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         client.env.update_config({'default-series': 'angsty'})
         deploy_iter = iter_steps_validate_info(self, deploy_many, client)
         self.assertEqual(deploy_iter.next(), {'test_id': 'add-machine-many'})
@@ -1702,7 +1702,7 @@
 
     def test_iter_step_add_machine_failure(self):
         deploy_many = DeployManyAttempt()
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         client.env.update_config({'default-series': 'angsty'})
         deploy_iter = iter_steps_validate_info(self, deploy_many, client)
         self.assertEqual(deploy_iter.next(), {'test_id': 'add-machine-many'})
@@ -1806,7 +1806,7 @@
 
     def test_iter_steps(self):
         br_attempt = BackupRestoreAttempt()
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         aws_env = get_aws_env()
         client.env.environment = aws_env.environment
         client.env._config = aws_env._config
@@ -1879,7 +1879,7 @@
         test_id = {'test_id': 'back-up-restore'}
         br_attempt = BackupRestoreAttempt()
         azure_env = SimpleEnvironment('steve', get_azure_config())
-        client = FakeEnvJujuClient()
+        client = FakeModelClient()
         client.env.environment = azure_env.environment
         client.env._config = azure_env._config
         client.env.credentials = {'credentials': {'azure': {'creds': {}}}}
@@ -1951,8 +1951,8 @@
         self.assertEqual('b', bootstrap_client.full_path)
 
     def test_iter_steps(self):
-        future_client = FakeEnvJujuClient(full_path='/future/juju')
-        present_client = FakeEnvJujuClient(full_path='/present/juju')
+        future_client = FakeModelClient(full_path='/future/juju')
+        present_client = FakeModelClient(full_path='/present/juju')
         puj_attempt = PrepareUpgradeJujuAttempt(
             {future_client.full_path: present_client.full_path})
         puj_iterator = iter_steps_validate_info(self, puj_attempt,
@@ -1984,7 +1984,7 @@
 
     def test_iter_steps_no_previous_client(self):
         uj_attempt = PrepareUpgradeJujuAttempt({})
-        client = FakeEnvJujuClient(full_path='/present/juju')
+        client = FakeModelClient(full_path='/present/juju')
         uj_iterator = uj_attempt.iter_steps(client)
         with self.assertRaises(CannotUpgradeToClient) as exc_context:
             uj_iterator.next()
@@ -1994,7 +1994,7 @@
 class TestUpgradeJujuAttempt(JujuPyTestCase):
 
     def test_iter_steps(self):
-        future_client = FakeEnvJujuClient(full_path='/future/juju')
+        future_client = FakeModelClient(full_path='/future/juju')
         uj_attempt = UpgradeJujuAttempt()
         uj_iterator = iter_steps_validate_info(self, uj_attempt, future_client)
         self.assertEqual(uj_iterator.next(), {'test_id': 'upgrade-juju'})
@@ -2026,7 +2026,7 @@
         self.assertEqual(0o755, mode & 0o777)
 
     def test_iter_steps(self):
-        client = FakeEnvJujuClient(version='2.0.0', full_path='/future/juju')
+        client = FakeModelClient(version='2.0.0', full_path='/future/juju')
         self._iter_steps(client)
 
     def test_iter_steps_juju_1x(self):
@@ -2140,20 +2140,20 @@
 class TestMakeSubstrate(JujuPyTestCase):
 
     def test_make_substrate_manager_no_support(self):
-        client = EnvJujuClient(JujuData('foo', {'type': 'foo'}),
-                               '', '')
+        client = ModelClient(JujuData('foo', {'type': 'foo'}),
+                             '', '')
         client.env.credentials = {'credentials': {'foo': {'creds': {}}}}
         with make_substrate_manager(client, []) as substrate:
             self.assertIs(substrate, None)
 
     def test_make_substrate_no_requirements(self):
-        client = EnvJujuClient(get_aws_juju_data(), '', '')
+        client = ModelClient(get_aws_juju_data(), '', '')
         client.env.credentials = {'credentials': {'aws': {'creds': {}}}}
         with make_substrate_manager(client, []) as substrate:
             self.assertIs(type(substrate), AWSAccount)
 
     def test_make_substrate_manager_unsatisifed_requirements(self):
-        client = EnvJujuClient(get_aws_juju_data(), '', '')
+        client = ModelClient(get_aws_juju_data(), '', '')
         client.env.credentials = {'credentials': {'aws': {'creds': {}}}}
         with make_substrate_manager(client, ['foo']) as substrate:
             self.assertIs(substrate, None)
@@ -2162,7 +2162,7 @@
             self.assertIs(substrate, None)
 
     def test_make_substrate_satisfied_requirements(self):
-        client = EnvJujuClient(get_aws_juju_data(), '', '')
+        client = ModelClient(get_aws_juju_data(), '', '')
         client.env.credentials = {'credentials': {'aws': {'creds': {}}}}
         with make_substrate_manager(
                 client, ['iter_security_groups']) as substrate:

=== modified file 'tests/test_jujucharm.py'
--- tests/test_jujucharm.py	2016-09-22 18:23:05 +0000
+++ tests/test_jujucharm.py	2017-01-12 22:58:16 +0000
@@ -170,16 +170,9 @@
             sane_charm_store_api_url(None),
             CharmCommand.default_api_url)
 
-    def test_replaces_places_api_characters(self):
+    def test_appends_charmstore(self):
         api = 'https://example.com'
-        expected = 'https://api.example.com/charmstore'
-        self.assertEqual(
-            sane_charm_store_api_url(api),
-            expected)
-
-    def test_replaces_www_characters(self):
-        api = 'https://www.example.com'
-        expected = 'https://api.example.com/charmstore'
+        expected = 'https://example.com/charmstore'
         self.assertEqual(
             sane_charm_store_api_url(api),
             expected)

=== modified file 'tests/test_jujuci.py'
--- tests/test_jujuci.py	2016-10-15 18:31:14 +0000
+++ tests/test_jujuci.py	2017-01-25 02:32:29 +0000
@@ -355,9 +355,9 @@
         config = {'environments': {'local': {'type': 'local'}}}
         with jujupy._temp_env(config, set_home=True):
             with patch('jujuci.destroy_environment', autospec=True) as mock_de:
-                with patch.object(jujupy.EnvJujuClient, 'get_version'):
+                with patch.object(jujupy.ModelClient, 'get_version'):
                     with patch('jujupy.get_client_class') as gcc_mock:
-                        gcc_mock.return_value = jujupy.EnvJujuClient
+                        gcc_mock.return_value = jujupy.ModelClient
                         dirty = clean_environment('foo', verbose=False)
         self.assertFalse(dirty)
         self.assertEqual(0, mock_de.call_count)
@@ -366,19 +366,19 @@
         config = {'environments': {'foo': {'type': 'local'}}}
         with jujupy._temp_env(config, set_home=True):
             with patch('jujuci.destroy_environment', autospec=True) as mock_de:
-                with patch.object(jujupy.EnvJujuClient, 'get_version'):
+                with patch.object(jujupy.ModelClient, 'get_version'):
                     with patch('jujupy.get_client_class') as gcc_mock:
                         factory = gcc_mock.return_value
-                        factory.return_value = jujupy.EnvJujuClient(
+                        factory.return_value = jujupy.ModelClient(
                             None, None, None)
-                        with patch.object(jujupy.EnvJujuClient,
+                        with patch.object(jujupy.ModelClient,
                                           'get_full_path'):
                             with patch.object(jujupy.JujuData, 'load_yaml'):
                                 dirty = clean_environment('foo', verbose=False)
         self.assertTrue(dirty)
         self.assertEqual(1, mock_de.call_count)
         args, kwargs = mock_de.call_args
-        self.assertIsInstance(args[0], jujupy.EnvJujuClient)
+        self.assertIsInstance(args[0], jujupy.ModelClient)
         self.assertEqual('foo', args[1])
 
     def test_add_artifacts_simple(self):

=== added file 'tests/test_perfscale_controller_stress.py'
--- tests/test_perfscale_controller_stress.py	1970-01-01 00:00:00 +0000
+++ tests/test_perfscale_controller_stress.py	2017-01-25 01:16:50 +0000
@@ -0,0 +1,147 @@
+"""Tests for perfscale_controller_stress module."""
+
+from datetime import datetime, timedelta
+from mock import patch, Mock
+import StringIO
+
+import perfscale_controller_stress as pcs
+from tests import (
+    parse_error,
+    TestCase,
+)
+from generate_perfscale_results import (
+    DeployDetails,
+    TimingData,
+)
+from test_generate_perfscale_results import (
+    get_default_args,
+)
+from utility import temp_dir
+
+
+def _get_default_args(**kwargs):
+    # Wrap default args for this test.
+    amount = kwargs.pop('deploy_amount', 1)
+    return get_default_args(deploy_amount=amount, **kwargs)
+
+
+class TestAssessControllerStress(TestCase):
+
+    def test_returns_DeployDetails(self):
+        args = _get_default_args(deploy_amount=0)
+        client = Mock()
+
+        deploy_details = pcs.assess_controller_stress(client, args)
+
+        self.assertIs(type(deploy_details), DeployDetails)
+        self.assertEqual(deploy_details.name, 'Controller Stress.')
+
+    def test_deploys_and_adds_units(self):
+        args = _get_default_args(deploy_amount=1)
+        client = Mock()
+
+        with patch.object(
+                pcs, 'deploy_swarm_to_new_model', autospec=True) as m_dsnm:
+            pcs.assess_controller_stress(client, args)
+        m_dsnm.assert_called_once_with(client, 'swarm-model-0')
+
+    def test_stores_deploytimes(self):
+        start = datetime(2016, 11, 1, 2, 17, 48)
+        end = datetime(2016, 11, 1, 2, 17, 50)
+        times = [start, end]
+        comparison_timing = TimingData(start, end)
+        args = _get_default_args(deploy_amount=0)
+        client = Mock()
+
+        with patch.object(pcs, 'datetime', autospec=True) as m_dt:
+            m_dt.utcnow.side_effect = times
+            deploy_details = pcs.assess_controller_stress(client, args)
+        deploy_timing = deploy_details.timings
+        self.assertEqual(deploy_timing.start, comparison_timing.start)
+        self.assertEqual(deploy_timing.end, comparison_timing.end)
+
+
+class DeploySwarmToNewModel(TestCase):
+
+    def test_returns_elapsed_seconds(self):
+        client = Mock()
+        start = datetime.utcnow()
+        end = start + timedelta(seconds=10)
+        with patch.object(pcs, 'datetime', autospec=True) as m_dt:
+            m_dt.utcnow.side_effect = [start, end]
+            elapsed = pcs.deploy_swarm_to_new_model(client, 'testing')
+        self.assertEqual(elapsed, int((end - start).total_seconds()))
+
+    def test_deploys_charm_to_new_model(self):
+        client = Mock()
+        new_client = Mock()
+        client.add_model.return_value = new_client
+
+        pcs.deploy_swarm_to_new_model(client, 'testing')
+
+        client.add_model.assert_called_once_with('testing')
+        new_client.deploy.assert_called_once_with(
+            'cs:~containers/observable-swarm')
+        self.assertEqual(new_client.wait_for_started.call_count, 1)
+        self.assertEqual(new_client.wait_for_workloads.call_count, 1)
+
+
+class TestGetCharmUrl(TestCase):
+
+    def test_returns_correct_url(self):
+        self.assertEqual(
+            pcs.get_charm_url(), 'cs:~containers/observable-swarm')
+
+
+class TestParseArgs(TestCase):
+
+    def test_common_args(self):
+        args = pcs.parse_args(
+            [
+                'an-env',
+                '/bin/juju',
+                '/tmp/logs',
+                'an-env-mod'])
+        self.assertEqual(
+            args,
+            _get_default_args()
+        )
+
+    def test_default_deploy_amount(self):
+        args = pcs.parse_args(
+            ['an-env', '/bin/juju', '/tmp/logs', 'an-env-mod'])
+        self.assertEqual(
+            args.deploy_amount, 1)
+
+    def test_passing_deploy_amount(self):
+        args = pcs.parse_args(
+            [
+                'an-env',
+                '/bin/juju',
+                '/tmp/logs',
+                'an-env-mod',
+                '--deploy-amount', '42'])
+        self.assertEqual(args.deploy_amount, 42)
+
+    def test_help(self):
+        fake_stdout = StringIO.StringIO()
+        with parse_error(self) as fake_stderr:
+            with patch('sys.stdout', fake_stdout):
+                pcs.parse_args(['--help'])
+        self.assertEqual('', fake_stderr.getvalue())
+        self.assertIn(
+            'Perfscale Controller Stress test.', fake_stdout.getvalue())
+
+
+class TestMain(TestCase):
+    with temp_dir() as log_dir:
+        argv = ['an-env', '/bin/juju', log_dir, 'an-env-mod']
+        bs_manager = Mock()
+        with patch.object(pcs, 'run_perfscale_test') as mock_run_pt:
+            with patch.object(pcs.BootstrapManager, 'from_args',
+                              return_value=bs_manager):
+                pcs.main(argv)
+        mock_run_pt.assert_called_once_with(
+            pcs.assess_controller_stress,
+            bs_manager,
+            _get_default_args(logs=log_dir))

=== modified file 'tests/test_perfscale_deployment.py'
--- tests/test_perfscale_deployment.py	2016-10-26 04:29:40 +0000
+++ tests/test_perfscale_deployment.py	2017-01-17 18:04:08 +0000
@@ -3,7 +3,7 @@
 from mock import patch, Mock
 import StringIO
 
-from fakejuju import fake_juju_client
+from jujupy import fake_juju_client
 import perfscale_deployment as pd
 from tests import (
     parse_error,

=== modified file 'tests/test_perfscale_longrunning.py'
--- tests/test_perfscale_longrunning.py	2016-10-25 04:39:43 +0000
+++ tests/test_perfscale_longrunning.py	2017-01-17 18:04:08 +0000
@@ -4,7 +4,7 @@
 from mock import patch, Mock, call
 import StringIO
 
-from fakejuju import fake_juju_client
+from jujupy import fake_juju_client
 import perfscale_longrunning as pl
 from tests import (
     parse_error,
@@ -123,13 +123,12 @@
         client = Mock()
         new_model = Mock()
         client.add_model.return_value = new_model
-        client.env.clone.return_value = 'foo'
 
         with patch.object(pl, 'deploy_stack', autospec=True) as m_deploy_stack:
             pl.action_create(client, series='foo')
 
         m_deploy_stack.assert_called_once_with(new_model, 'foo')
-        client.add_model.assert_called_once_with('foo')
+        client.add_model.assert_called_once_with('newmodel')
 
 
 class TestPerfscaleLongrunPerf(TestCase):

=== modified file 'tests/test_perfscale_mass_model_destruction.py'
--- tests/test_perfscale_mass_model_destruction.py	2016-10-26 05:06:40 +0000
+++ tests/test_perfscale_mass_model_destruction.py	2017-01-17 18:04:08 +0000
@@ -6,7 +6,7 @@
 import perfscale_mass_model_destruction as pmmd
 from generate_perfscale_results import DeployDetails
 from test_generate_perfscale_results import get_default_args
-from fakejuju import fake_juju_client
+from jujupy import fake_juju_client
 from tests import (
     TestCase
 )

=== modified file 'tests/test_pipdeps.py'
--- tests/test_pipdeps.py	2016-06-23 02:50:43 +0000
+++ tests/test_pipdeps.py	2017-01-09 20:19:16 +0000
@@ -2,10 +2,12 @@
 
 import mock
 import os
+import subprocess
 import unittest
 
 
 import pipdeps
+import tests
 import utility
 
 
@@ -69,6 +71,10 @@
         os.path.realpath(os.path.dirname(pipdeps.__file__)),
         "requirements.txt")
 
+    req_path_py3 = os.path.join(
+        os.path.realpath(os.path.dirname(pipdeps.__file__)),
+        "requirements_py3.txt")
+
     def test_added_args(self):
         with mock.patch("subprocess.check_call", autospec=True) as cc_mock:
             pipdeps.run_pip_install(["--user"], self.req_path)
@@ -82,6 +88,18 @@
         cc_mock.assert_called_once_with([
             "pip", "install", "-r", self.req_path, "--download", "/tmp/pip"])
 
+    def test_pip3_install(self):
+        with mock.patch("subprocess.check_output", autospec=True,
+                        return_value='3.5.0') as co_mock:
+            with mock.patch("subprocess.check_call", autospec=True) as cc_mock:
+                pipdeps.run_pip3_install(["--user"], self.req_path_py3)
+        co_mock.assert_called_once_with(
+            ['python3', '--version'], stderr=subprocess.STDOUT)
+        calls = [mock.call(['pip3', '--version']),
+                 mock.call(['pip3', '-q', 'install', '-r', self.req_path_py3,
+                            '--user'])]
+        self.assertEqual(cc_mock.call_args_list, calls)
+
 
 class TestRunPipUninstall(unittest.TestCase):
 
@@ -117,3 +135,43 @@
             self.assertEqual(pipdeps.MAC_WIN_REQS, pipdeps.get_requirements())
 
             self.assertEqual(pipdeps.MAC_WIN_REQS, pipdeps.get_requirements())
+
+
+class TestIsPy3Supported(tests.TestCase):
+
+    def test_is_py3_supported(self):
+        with mock.patch("subprocess.check_output", autospec=True,
+                        return_value='3.5.0') as co_mock:
+            with mock.patch("subprocess.check_call", autospec=True) as cc_mock:
+                self.assertTrue(pipdeps.is_py3_supported())
+        co_mock.assert_called_once_with(
+            ['python3', '--version'], stderr=subprocess.STDOUT)
+        cc_mock.assert_called_once_with(['pip3', '--version'])
+
+    def test_is_py3_supported_older_python3_version(self):
+        with mock.patch("subprocess.check_output", autospec=True,
+                        return_value='3.4') as co_mock:
+            with mock.patch("subprocess.check_call", autospec=True) as cc_mock:
+                self.assertFalse(pipdeps.is_py3_supported())
+        co_mock.assert_called_once_with(
+            ['python3', '--version'], stderr=subprocess.STDOUT)
+        self.assertEqual(cc_mock.call_count, 0)
+
+    def test_is_pyt3_supported_python3_not_installed(self):
+        with mock.patch("subprocess.check_output", autospec=True,
+                        side_effect=OSError(2, 'No such file')) as co_mock:
+            with mock.patch("subprocess.check_call", autospec=True) as cc_mock:
+                self.assertFalse(pipdeps.is_py3_supported())
+        co_mock.assert_called_once_with(
+            ['python3', '--version'], stderr=subprocess.STDOUT)
+        self.assertEqual(cc_mock.call_count, 0)
+
+    def test_is_pyt3_supported_pip3_not_installed(self):
+        with mock.patch("subprocess.check_output", autospec=True,
+                        return_value='3.5.2') as co_mock:
+            with mock.patch("subprocess.check_call", autospec=True,
+                            side_effect=OSError(2, 'No such file')) as cc_mock:
+                self.assertFalse(pipdeps.is_py3_supported())
+        co_mock.assert_called_once_with(
+            ['python3', '--version'], stderr=subprocess.STDOUT)
+        cc_mock.assert_called_once_with(['pip3', '--version'])

=== modified file 'tests/test_quickstart_deploy.py'
--- tests/test_quickstart_deploy.py	2016-10-27 18:49:11 +0000
+++ tests/test_quickstart_deploy.py	2017-01-20 21:26:33 +0000
@@ -5,9 +5,9 @@
 )
 
 from deploy_stack import BootstrapManager
-from fakejuju import fake_juju_client
 from jujupy import (
-    EnvJujuClient,
+    ModelClient,
+    fake_juju_client,
     JujuData,
     )
 from quickstart_deploy import QuickstartTest
@@ -37,7 +37,7 @@
             finally:
                 do_finally()
 
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         bs_manager = make_bootstrap_manager(client)
         quickstart = QuickstartTest(bs_manager, '/tmp/bundle.yaml', 2)
@@ -55,7 +55,7 @@
             except:
                 tear_down()
 
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         bs_manager = make_bootstrap_manager(client)
         quickstart = QuickstartTest(bs_manager, '/tmp/bundle.yaml', 2)
@@ -68,14 +68,15 @@
 
     def test_iter_steps(self):
         log_dir = use_context(self, temp_dir())
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         bs_manager = make_bootstrap_manager(client, log_dir=log_dir)
         quickstart = QuickstartTest(bs_manager, '/tmp/bundle.yaml', 2)
         steps = quickstart.iter_steps()
         with patch.object(client, 'quickstart') as qs_mock:
             # Test first yield
-            with patch('jujupy.check_free_disk_space', autospec=True):
+            with patch('jujupy.client.check_free_disk_space',
+                       autospec=True):
                 with patch.object(client, 'kill_controller',
                                   autospec=True) as kill_mock:
                     step = steps.next()
@@ -103,7 +104,7 @@
         with patch('deploy_stack.safe_print_status'):
             with patch('deploy_stack.BootstrapManager.tear_down'):
                 with patch('quickstart_deploy.BootstrapManager.dump_all_logs'):
-                    with patch('jujupy.EnvJujuClient.iter_model_clients',
+                    with patch('jujupy.ModelClient.iter_model_clients',
                                return_value=[]):
                         steps.close()
 
@@ -143,7 +144,7 @@
         self.assertIs(True, bs_manager.exited_top)
 
     def test_iter_steps_quickstart_fail(self):
-        client = EnvJujuClient(
+        client = ModelClient(
             JujuData('foo', {'type': 'local'}), '1.234-76', None)
         bs_manager = FakeBootstrapManager(client)
         quickstart = QuickstartTest(bs_manager, '/tmp/bundle.yaml', 2)

=== modified file 'tests/test_remote.py'
--- tests/test_remote.py	2016-05-25 16:12:07 +0000
+++ tests/test_remote.py	2017-01-20 21:13:11 +0000
@@ -8,8 +8,7 @@
 import winrm
 
 from jujupy import (
-    EnvJujuClient,
-    get_timeout_path,
+    ModelClient,
     JujuData,
     Status,
 )
@@ -20,6 +19,7 @@
 )
 import tests
 from utility import (
+    get_timeout_path,
     temp_dir,
 )
 
@@ -52,7 +52,7 @@
 
     def test_remote_from_unit(self):
         env = JujuData("an-env", {"type": "nonlocal"})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         unit = "a-application/0"
         with patch.object(client, "get_status", autospec=True) as st:
             st.return_value = Status.from_text(self.precise_status_output)
@@ -64,7 +64,7 @@
 
     def test_remote_from_unit_with_series(self):
         env = JujuData("an-env", {"type": "nonlocal"})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         unit = "a-application/0"
         remote = remote_from_unit(client, unit, series="trusty")
         self.assertEqual(
@@ -74,7 +74,7 @@
 
     def test_remote_from_unit_with_status(self):
         env = JujuData("an-env", {"type": "nonlocal"})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         unit = "a-application/0"
         status = Status.from_text(self.win2012hvr2_status_output)
         remote = remote_from_unit(client, unit, status=status)
@@ -101,7 +101,7 @@
 
     def test_run_with_unit(self):
         env = JujuData("an-env", {"type": "nonlocal"})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         unit = "a-application/0"
         remote = remote_from_unit(client, unit, series="trusty")
         with patch.object(client, "get_juju_output") as mock_cmd:
@@ -113,14 +113,14 @@
 
     def test_run_with_unit_fallback(self):
         env = JujuData("an-env", {"type": "nonlocal"})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         unit = "a-application/0"
         with patch.object(client, "get_status") as st:
             st.return_value = Status.from_text(self.precise_status_output)
             remote = remote_from_unit(client, unit)
             with patch.object(client, "get_juju_output") as mock_gjo:
-                mock_gjo.side_effect = subprocess.CalledProcessError(1, "ssh",
-                                                                     output="")
+                mock_gjo.side_effect = subprocess.CalledProcessError(
+                    255, "ssh", output="")
                 with patch.object(remote, "_run_subprocess") as mock_run:
                     mock_run.return_value = "contents of /a/file"
                     output = remote.run("cat /a/file")
@@ -140,6 +140,47 @@
             self.log_stream.getvalue(),
             "(?m)^WARNING juju ssh to 'a-application/0' failed, .*")
 
+    def test_run_default_command_error_fallback(self):
+        env = JujuData("an-env", {"type": "nonlocal"})
+        client = ModelClient(env, None, None)
+        unit = "a-application/0"
+        error = subprocess.CalledProcessError(1, "ssh", output="bad command")
+        with patch.object(client, "get_status") as st:
+            st.return_value = Status.from_text(self.precise_status_output)
+            remote = remote_from_unit(client, unit)
+            with patch.object(client, "get_juju_output") as mock_gjo:
+                mock_gjo.side_effect = error
+                with self.assertRaises(subprocess.CalledProcessError) as c:
+                    remote.run("cat /a/file")
+        self.assertIs(c.exception, error)
+        mock_gjo.assert_called_once_with("ssh", unit, "cat /a/file",
+                                         timeout=120)
+        self.assertRegexpMatches(
+            self.log_stream.getvalue(),
+            "(?m)^WARNING juju ssh to 'a-application/0' failed, .*")
+
+    def test_run_no_platform_fallback(self):
+        env = JujuData("an-env", {"type": "nonlocal"})
+        client = ModelClient(env, None, None)
+        unit = "a-application/0"
+        error = subprocess.CalledProcessError(255, "ssh", output="")
+        with patch.object(client, "get_status") as st:
+            st.return_value = Status.from_text(self.precise_status_output)
+            remote = remote_from_unit(client, unit)
+            with patch.object(client, "get_juju_output") as mock_gjo:
+                mock_gjo.side_effect = error
+                with patch("remote._no_platform_ssh", autospec=True,
+                           return_value=True) as mock_nps:
+                    with self.assertRaises(subprocess.CalledProcessError) as c:
+                        remote.run("cat /a/file")
+        self.assertIs(c.exception, error)
+        mock_gjo.assert_called_once_with("ssh", unit, "cat /a/file",
+                                         timeout=120)
+        mock_nps.assert_called_once_with()
+        self.assertRegexpMatches(
+            self.log_stream.getvalue(),
+            "(?m)^WARNING juju ssh to 'a-application/0' failed, .*")
+
     def test_run_with_address(self):
         remote = remote_from_address("10.55.60.1")
         with patch.object(remote, "_run_subprocess") as mock_run:
@@ -167,12 +208,12 @@
             "-o", "StrictHostKeyChecking no",
             "-o", "PasswordAuthentication no",
             "10.55.60.1",
-            "cat /a/file",
+            "cat", "/a/file",
         ])
 
     def test_cat_on_windows(self):
         env = JujuData("an-env", {"type": "nonlocal"})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         unit = "a-application/0"
         with patch.object(client, "get_status", autospec=True) as st:
             st.return_value = Status.from_text(self.win2012hvr2_status_output)
@@ -204,7 +245,7 @@
 
     def test_copy_on_windows(self):
         env = JujuData("an-env", {"type": "nonlocal"})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         unit = "a-application/0"
         dest = "/local/path"
         with patch.object(client, "get_status", autospec=True) as st:
@@ -243,7 +284,7 @@
 
     def test_run_cmd(self):
         env = JujuData("an-env", {"type": "nonlocal"})
-        client = EnvJujuClient(env, None, None)
+        client = ModelClient(env, None, None)
         unit = "a-application/0"
         with patch.object(client, "get_status", autospec=True) as st:
             st.return_value = Status.from_text(self.win2012hvr2_status_output)
@@ -274,7 +315,7 @@
             "-o", "StrictHostKeyChecking no",
             "-o", "PasswordAuthentication no",
             "10.55.60.1",
-            "cat /a/file",
+            "cat", "/a/file",
             ),
             stdin=subprocess.PIPE,
         )

=== modified file 'tests/test_run_chaos_monkey.py'
--- tests/test_run_chaos_monkey.py	2016-07-08 17:30:23 +0000
+++ tests/test_run_chaos_monkey.py	2017-01-20 21:32:42 +0000
@@ -10,7 +10,7 @@
 
 from chaos import MonkeyRunner
 from jujupy import (
-    EnvJujuClient,
+    ModelClient,
     JujuData,
     )
 from run_chaos_monkey import (
@@ -32,7 +32,7 @@
         self.assertEqual(args.health_checker, 'baz')
 
     def test_run_while_healthy_or_timeout(self):
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo')
+        client = ModelClient(JujuData('foo', {}), None, '/foo')
         runner = MonkeyRunner('foo', 'bar', 'script', client, total_timeout=60)
         runner.expire_time = (datetime.now() - timedelta(seconds=1))
         with patch.object(runner, 'is_healthy', autospec=True,
@@ -45,7 +45,7 @@
         wait_mock.assert_called_once_with()
 
     def test_run_while_healthy_or_timeout_exits_non_zero(self):
-        client = EnvJujuClient(JujuData('foo', {}), None, '/foo')
+        client = ModelClient(JujuData('foo', {}), None, '/foo')
         runner = MonkeyRunner('foo', 'bar', 'script', client, total_timeout=60)
         with patch.object(runner, 'is_healthy', autospec=True,
                           return_value=False):

=== modified file 'tests/test_run_deployer.py'
--- tests/test_run_deployer.py	2016-10-14 19:07:39 +0000
+++ tests/test_run_deployer.py	2017-01-20 21:17:45 +0000
@@ -14,10 +14,10 @@
     )
 
 from jujupy import (
-    EnvJujuClient,
+    ModelClient,
+    fake_juju_client,
     JujuData,
     )
-from fakejuju import fake_juju_client
 from run_deployer import (
     apply_condition,
     assess_deployer,
@@ -79,7 +79,7 @@
     def test_basic_args(self):
         args = ['bundles', 'an-env', '/bin/juju', 'logs', 'deployer-env']
         env = JujuData('an-env')
-        client = EnvJujuClient(env, '1.234-76', None)
+        client = ModelClient(env, '1.234-76', None)
         with patch('run_deployer.client_from_config',
                    return_value=client) as c_mock:
             with patch('run_deployer.boot_context'):
@@ -95,7 +95,7 @@
                 '--bundle-verification-script',
                 'verify_mediawiki_bundle.py']
         env = JujuData('an-env')
-        client = EnvJujuClient(env, '1.234-76', None)
+        client = ModelClient(env, '1.234-76', None)
         with patch('run_deployer.client_from_config',
                    return_value=client) as c_mock:
             with patch('run_deployer.boot_context'):
@@ -116,7 +116,7 @@
                 '--bundle-verification-script',
                 'verify_landscape_bundle.py']
         env = JujuData('an-env')
-        client = EnvJujuClient(env, '1.234-76', None)
+        client = ModelClient(env, '1.234-76', None)
         with patch('run_deployer.client_from_config',
                    return_value=client) as c_mock:
             with patch('run_deployer.boot_context'):
@@ -151,7 +151,7 @@
 
     def test_health(self):
         args = self.make_args(health_cmd='/tmp/check')
-        client_mock = Mock(spec=EnvJujuClient)
+        client_mock = Mock(spec=ModelClient)
         with patch('run_deployer.check_health', autospec=True) as ch_mock:
             assess_deployer(args, client_mock, 600, 1800)
         client_mock.deployer.assert_called_once_with('bundle.yaml', 'bu')
@@ -161,7 +161,7 @@
 
     def test_upgrade(self):
         args = self.make_args(juju_bin='new/juju', upgrade=True)
-        client_mock = Mock(spec=EnvJujuClient)
+        client_mock = Mock(spec=ModelClient)
         with patch('run_deployer.assess_upgrade', autospec=True) as au_mock:
             assess_deployer(args, client_mock, 600, 1800)
         client_mock.deployer.assert_called_once_with('bundle.yaml', 'bu')
@@ -174,7 +174,7 @@
     def test_upgrade_and_health(self):
         args = self.make_args(health_cmd='/tmp/check', juju_bin='new/juju',
                               upgrade=True)
-        client_mock = Mock(spec=EnvJujuClient)
+        client_mock = Mock(spec=ModelClient)
         with patch('run_deployer.assess_upgrade', autospec=True) as au_mock:
             with patch('run_deployer.check_health', autospec=True) as ch_mock:
                 assess_deployer(args, client_mock, 600, 1800)
@@ -208,7 +208,7 @@
 
     def test_allow_native_deploy(self):
         args = self.make_args(allow_native_deploy=True)
-        client_mock = Mock(spec=EnvJujuClient)
+        client_mock = Mock(spec=ModelClient)
         assess_deployer(args, client_mock, 600, 1800)
         client_mock.deploy_bundle.assert_called_once_with('bundle.yaml')
         client_mock.wait_for_started.assert_called_once_with(timeout=600)

=== modified file 'tests/test_run_download_juju.py'
--- tests/test_run_download_juju.py	2016-04-22 21:13:16 +0000
+++ tests/test_run_download_juju.py	2017-01-08 04:19:25 +0000
@@ -10,7 +10,7 @@
     parse_args,
     main
 )
-from jujupy import ensure_dir
+from utility import ensure_dir
 from tests import (
     TestCase,
 )

=== modified file 'tests/test_scale_out.py'
--- tests/test_scale_out.py	2016-09-01 19:52:54 +0000
+++ tests/test_scale_out.py	2017-01-20 21:11:17 +0000
@@ -8,7 +8,7 @@
 from unittest import TestCase
 
 from jujupy import (
-    EnvJujuClient,
+    ModelClient,
     JujuData,
 )
 from scale_out import (
@@ -20,8 +20,8 @@
 )
 
 
-def fake_EnvJujuClient(env, path=None, debug=None):
-    return EnvJujuClient(env=env, version='1.2.3.4', full_path=path)
+def fake_ModelClient(env, path=None, debug=None):
+    return ModelClient(env=env, version='1.2.3.4', full_path=path)
 
 
 class TestScaleOut(TestCase):
@@ -29,7 +29,7 @@
     @contextmanager
     def fake_client_cxt(self):
         env = JujuData('foo', {})
-        client = fake_EnvJujuClient(env)
+        client = fake_ModelClient(env)
         bv_cxt = patch('scale_out.client_from_config',
                        return_value=client)
         with bv_cxt as bv_mock:
@@ -59,7 +59,7 @@
         self.assertEqual(args, expected)
 
     @patch('scale_out.boot_context', autospec=True)
-    @patch('jujupy.EnvJujuClient.add_ssh_machines', autospec=True)
+    @patch('jujupy.ModelClient.add_ssh_machines', autospec=True)
     def test_scaleout_setup(
             self,
             add_ssh_machines_func,
@@ -127,7 +127,7 @@
 
         with self.fake_client_cxt():
             with patch('scale_out.boot_context', autospec=True) as bc_mock:
-                with patch('jujupy.EnvJujuClient.add_ssh_machines',
+                with patch('jujupy.ModelClient.add_ssh_machines',
                            autospec=True):
                     with scaleout_setup(args) as client:
                         pass
@@ -146,8 +146,8 @@
 
     def test_deploy_charms(self):
         with self.fake_client_cxt() as (client, env, bv_mock):
-            with patch.object(EnvJujuClient, 'deploy') as d_mock:
-                with patch.object(EnvJujuClient,
+            with patch.object(ModelClient, 'deploy') as d_mock:
+                with patch.object(ModelClient,
                                   'wait_for_started') as wfs_mock:
                     deploy_charms(client, ['ubuntu', 'mysql'])
         expected = [call('ubuntu', service='ubuntu'),
@@ -157,8 +157,8 @@
 
     def test_deploy_charms_local(self):
         with self.fake_client_cxt() as (client, env, bv_mock):
-            with patch.object(EnvJujuClient, 'deploy') as d_mock:
-                with patch.object(EnvJujuClient,
+            with patch.object(ModelClient, 'deploy') as d_mock:
+                with patch.object(ModelClient,
                                   'wait_for_started') as wfs_mock:
                     deploy_charms(client, ['local:foo', 'local:bar'])
         expected = [call('local:foo', service='foo'),
@@ -168,8 +168,8 @@
 
     def test_scale_out(self):
         with self.fake_client_cxt() as (client, env, bv_mock):
-            with patch.object(EnvJujuClient, 'juju') as j_mock:
-                with patch.object(EnvJujuClient,
+            with patch.object(ModelClient, 'juju') as j_mock:
+                with patch.object(ModelClient,
                                   'wait_for_started') as wfs_mock:
                     scale_out(client, 'ubuntu')
         j_mock.assert_called_once_with('add-unit', ('ubuntu', '-n', '5'))

=== modified file 'tests/test_substrate.py'
--- tests/test_substrate.py	2016-11-28 22:04:46 +0000
+++ tests/test_substrate.py	2017-01-18 16:57:54 +0000
@@ -1,4 +1,5 @@
 from contextlib import contextmanager
+from datetime import datetime
 import json
 import os
 from subprocess import CalledProcessError
@@ -15,13 +16,13 @@
     patch,
     )
 
-from fakejuju import fake_juju_client
 from jujuconfig import (
     get_euca_env,
     translate_to_env,
     )
 from jujupy import (
     EnvJujuClient1X,
+    fake_juju_client,
     JujuData,
     SimpleEnvironment,
     )
@@ -943,6 +944,57 @@
             ('maas', 'mas', 'machines', 'list-allocated'))
         self.assertEqual(node, allocated['maas-node-1.maas'])
 
+    def make_event(self, acquire_date, type_=MAASAccount.ACQUIRING):
+        return {
+           'type': type_,
+           MAASAccount.CREATED: acquire_date.isoformat(),
+           MAASAccount.NODE: 'asdf',
+           }
+
+    def test_get_acquire_date(self):
+        acquire_date = datetime(2016, 10, 25)
+        result = self._run_acquire_date(acquire_date)
+        self.assertEqual(acquire_date, result)
+
+    def _run_acquire_date(self, acquire_date, type_=MAASAccount.ACQUIRING,
+                          node='asdf'):
+        events = {'events': [
+            self.make_event(acquire_date, type_=type_),
+            ]}
+        return self._run_acquire_date_events(events, node)
+
+    def _run_acquire_date_events(self, events, node):
+        account = self.get_account()
+        with patch('subprocess.check_output', autospec=True,
+                   return_value=json.dumps(events)) as co_mock:
+            result = account.get_acquire_date(node)
+        co_mock.assert_called_once_with(
+            ('maas', 'mas', 'events', 'query', 'id={}'.format(node)))
+        return result
+
+    def test_get_acquire_date_not_acquiring(self):
+        acquire_date = datetime(2016, 10, 25)
+        with self.assertRaisesRegexp(
+                LookupError, 'Unable to find acquire date for "asdf".'):
+            self._run_acquire_date(acquire_date, type_='Not acquiring')
+
+    def test_get_acquire_date_uses_first_entry(self):
+        acquire_date = datetime(2016, 10, 25)
+        newer_date = datetime(2016, 10, 26)
+        older_date = datetime(2016, 10, 24)
+        events = {'events': [
+            self.make_event(acquire_date),
+            self.make_event(older_date),
+            self.make_event(newer_date),
+            ]}
+        result = self._run_acquire_date_events(events, 'asdf')
+        self.assertEqual(acquire_date, result)
+
+    def test_get_acquire_date_wrong_node(self):
+        acquire_date = datetime(2016, 10, 25)
+        with self.assertRaisesRegexp(ValueError, 'Node "asdf" was not "fasd"'):
+            self._run_acquire_date(acquire_date, node='fasd')
+
     def test_get_allocated_ips(self):
         account = self.get_account()
         node = make_maas_node('maas-node-1.maas')

=== modified file 'tests/test_verify_landscape_bundle.py'
--- tests/test_verify_landscape_bundle.py	2016-10-14 19:07:39 +0000
+++ tests/test_verify_landscape_bundle.py	2017-01-17 18:04:08 +0000
@@ -2,7 +2,7 @@
     patch,
     )
 
-from fakejuju import fake_juju_client
+from jujupy import fake_juju_client
 import tests
 from verify_landscape_bundle import(
     assess_landscape_bundle,

=== modified file 'tests/test_verify_mediawiki_bundle.py'
--- tests/test_verify_mediawiki_bundle.py	2016-07-22 19:17:54 +0000
+++ tests/test_verify_mediawiki_bundle.py	2017-01-17 18:04:08 +0000
@@ -7,7 +7,7 @@
 )
 
 from assess_min_version import JujuAssertionError
-from fakejuju import (
+from jujupy import (
     fake_juju_client,
     fake_juju_client_optional_jes,
     )

=== modified file 'utility.py'
--- utility.py	2016-11-18 20:26:28 +0000
+++ utility.py	2017-01-08 04:36:05 +0000
@@ -16,13 +16,17 @@
     sleep,
     time,
     )
-from tempfile import mkdtemp
+from tempfile import (
+    mkdtemp,
+    NamedTemporaryFile,
+    )
 import warnings
 # Export shell quoting function which has moved in newer python versions
 try:
     from shlex import quote
 except ImportError:
     from pipes import quote
+import yaml
 
 quote
 
@@ -88,6 +92,9 @@
     def now():
         return datetime.now()
 
+    def __next__(self):
+        self.next()
+
     def next(self):
         elapsed = self.now() - self.start
         remaining = self.timeout - elapsed.total_seconds()
@@ -190,7 +197,7 @@
             else:
                 continue
         conn = socket.socket(*addrinfo[0][:3])
-        conn.settimeout(max(remaining, 5))
+        conn.settimeout(max(remaining or 0, 5))
         try:
             conn.connect(sockaddr)
         except socket.timeout:
@@ -441,10 +448,26 @@
         os.unlink(path)
 
 
+@contextmanager
+def temp_yaml_file(yaml_dict, encoding="utf-8"):
+    temp_file = NamedTemporaryFile(suffix='.yaml', delete=False)
+    try:
+        with temp_file:
+            yaml.safe_dump(yaml_dict, temp_file, encoding=encoding)
+        yield temp_file.name
+    finally:
+        os.unlink(temp_file.name)
+
+
 def get_candidates_path(root_dir):
     return os.path.join(root_dir, 'candidate')
 
 
+def get_timeout_path():
+    import timeout
+    return os.path.abspath(timeout.__file__)
+
+
 # GZ 2015-10-15: Paths returned in filesystem dependent order, may want sort?
 def find_candidates(root_dir, find_all=False):
     return (path for path, buildvars in _find_candidates(root_dir, find_all))

=== modified file 'write_industrial_test_metadata.py'
--- write_industrial_test_metadata.py	2016-11-01 19:47:51 +0000
+++ write_industrial_test_metadata.py	2017-01-20 20:03:12 +0000
@@ -2,9 +2,9 @@
 from argparse import ArgumentParser
 import json
 
-from jujupy import (
+from jujupy.client import (
     describe_substrate,
-    EnvJujuClient,
+    ModelClient,
     SimpleEnvironment,
     )
 
@@ -25,7 +25,7 @@
     :param buildbars_path: Path to the buildvars.json file for the new client.
     :param env_name: Name of the environment being used.
     """
-    old_version = EnvJujuClient.get_version()
+    old_version = ModelClient.get_version()
     env = SimpleEnvironment.from_config(env_name)
     with open(buildvars_path) as buildvars_file:
         buildvars = json.load(buildvars_file)

